#include "mt.h"

#include "UsbSysInit.h"
#include "UsbLpt.h"

MT_USB_EXPORT *mt_usb_export;

extern unsigned long get_data_segment(void);
extern void set_data_segment(unsigned long new_gp);

extern void localhc_irq(void);
extern void localhc_uhci1_irq(void);
extern void localhc_uhci2_irq(void);
extern int usb_block_read(unsigned long VolumeID, unsigned long SectorNo, void *Buffer);
extern int usb_block_readex(unsigned long VolumeID, unsigned long SectorNo, void *Buffer, unsigned short max_prefetch_sectors);
extern int usb_block_write(unsigned long VolumeID, unsigned long SectorNo, void *Buffer);
extern int blk_sync_cache(void);
extern void usb_storage_query_inbound_display(void);
extern void usb_storage_inner_display(void);
extern void usb_storage_cache_display(void);

extern struct DiskParamS USBDiskParam;
extern int validPartitionCnt;
extern unsigned char PartitionType[20];
extern unsigned long validPartitionID[20];

static unsigned long data_segment_usb, data_segment_mt;

//static
//const
USB_MT_EXPORT usb_mt_export = {
	get_data_segment,

	UsbSysInit,
	localhc_irq,
	localhc_uhci1_irq,
	localhc_uhci2_irq,
	Prt_Init,
	Prt_Reset,
	ParallelTest,
	Prt_Open,
	Prt_GetId,
	Prt_Get_S_EPP_Sts,
	Prt_Check_Port_Pwr_Connect,
	Prt_Sts_Check_On_Line,
	Prt_Sts_Check_Paper_End,
	Prt_Sts_Check_Error,
	Prt_Buf_Write,
	Prt_Close,
	usb_ms_unplug,
	usb_block_read,
	usb_block_readex,
	usb_block_write,
	blk_sync_cache,
	usb_storage_query_inbound_display,
	usb_storage_inner_display,
	usb_storage_cache_display,

	&USBDiskParam,
	&validPartitionCnt,
	PartitionType,
	validPartitionID
};

void * c_entry(int p_mt_export)
{
	unsigned long	usb_gp;
	void			* rt_value;

	usb_gp = get_data_segment();
	set_data_segment(usb_gp);

	mt_usb_export = (MT_USB_EXPORT *)p_mt_export;
	data_segment_usb = usb_gp;
	data_segment_mt = (*(mt_usb_export->pf_get_data_segment))();
	rt_value = (void *)&usb_mt_export;

	set_data_segment(data_segment_mt);
	return	rt_value;
}

void randomize(void)
{
	unsigned long	old_gp, new_gp;
	void (*pf_randomize)(void);

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_randomize = mt_usb_export->pf_randomize;

	set_data_segment(new_gp);
	(*pf_randomize)();
	set_data_segment(old_gp);
}

int  random(int num)
{
	unsigned long	old_gp, new_gp;
	int  (*pf_random)(int num), rt_value;

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_random = mt_usb_export->pf_random;

	set_data_segment(new_gp);
	rt_value = (*pf_random)(num);
	set_data_segment(old_gp);

	return	rt_value;
}

void	scdtsk(void)
{
	unsigned long	old_gp, new_gp;
	void (*pf_scdtsk)(void);

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_scdtsk = mt_usb_export->pf_scdtsk;

	set_data_segment(new_gp);
	(*pf_scdtsk)();
	set_data_segment(old_gp);
}

int  runtsk(uint priority,void (fAR CODE *fadr)(void),uint stksze)
{
	unsigned long	old_gp, new_gp;
	int  (*pf_runtsk)(uint priority,void (fAR CODE *fadr)(void),uint stksze,...), rt_value;

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_runtsk = mt_usb_export->pf_runtsk;

	set_data_segment(new_gp);
	rt_value = (*pf_runtsk)(priority, fadr, stksze);
	set_data_segment(old_gp);

	return	rt_value;
};

int  runtsk_parm(uint priority,void (fAR CODE *fadr)(void),uint stksze, unsigned long parm)
{
	unsigned long	old_gp, new_gp;
	int  (*pf_runtsk)(uint priority,void (fAR CODE *fadr)(void),uint stksze,...), rt_value;

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_runtsk = mt_usb_export->pf_runtsk;

	set_data_segment(new_gp);
	rt_value = (*pf_runtsk)(priority, fadr, stksze, parm);
	set_data_segment(old_gp);

	return	rt_value;
};

int  dlytsk(TASK_ID slot,byte dlytyp,uint dlytme)
{
	unsigned long	old_gp, new_gp;
	int  (*pf_dlytsk)(TASK_ID slot,byte dlytyp,uint dlytme), rt_value;

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_dlytsk = mt_usb_export->pf_dlytsk;

	set_data_segment(new_gp);
	rt_value = (*pf_dlytsk)(slot, dlytyp, dlytme);
	set_data_segment(old_gp);

	return	rt_value;
}

tick_cnt_t get_sys_time(void)
{
	unsigned long	old_gp, new_gp;
	tick_cnt_t (*pf_get_sys_time)(void), rt_value;

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_get_sys_time = mt_usb_export->pf_get_sys_time;

	set_data_segment(new_gp);
	rt_value = (*pf_get_sys_time)();
	set_data_segment(old_gp);

	return	rt_value;
}

int  getres(uint resrc,uint todly)
{
	unsigned long	old_gp, new_gp;
	int  (*pf_getres)(uint resrc,uint todly), rt_value;

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_getres = mt_usb_export->pf_getres;

	set_data_segment(new_gp);
	rt_value = (*pf_getres)(resrc, todly);
	set_data_segment(old_gp);

	return	rt_value;
}

int  relres(uint resrc)
{
	unsigned long	old_gp, new_gp;
	int  (*pf_relres)(uint resrc), rt_value;

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_relres = mt_usb_export->pf_relres;

	set_data_segment(new_gp);
	rt_value = (*pf_relres)(resrc);
	set_data_segment(old_gp);

	return	rt_value;
}

int  chkevt(uint event)
{
	unsigned long	old_gp, new_gp;
	int  (*pf_chkevt)(uint event), rt_value;

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_chkevt = mt_usb_export->pf_chkevt;

	set_data_segment(new_gp);
	rt_value = (*pf_chkevt)(event);
	set_data_segment(old_gp);

	return	rt_value;
}

int  setevt(uint event)
{
	unsigned long	old_gp, new_gp;
	int  (*pf_setevt)(uint event), rt_value;

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_setevt = mt_usb_export->pf_setevt;

	set_data_segment(new_gp);
	rt_value = (*pf_setevt)(event);
	set_data_segment(old_gp);

	return	rt_value;
}

void SIGNAL_IR(unsigned long signo)
{
	unsigned long	old_gp, new_gp;
	void (*pf_SIGNAL_IR)(unsigned long signo);

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_SIGNAL_IR = mt_usb_export->pf_SIGNAL_IR;

	set_data_segment(new_gp);
	(*pf_SIGNAL_IR)(signo);
	set_data_segment(old_gp);
}

int  wteset(uint event,uint todly)
{
	unsigned long	old_gp, new_gp;
	int  (*pf_wteset)(uint event,uint todly), rt_value;

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_wteset = mt_usb_export->pf_wteset;

	set_data_segment(new_gp);
	rt_value = (*pf_wteset)(event, todly);
	set_data_segment(old_gp);

	return	rt_value;
}

int  wteset_dec(uint event,uint todly)
{
	unsigned long	old_gp, new_gp;
	int  (*pf_wteset_dec)(uint event,uint todly), rt_value;

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_wteset_dec = mt_usb_export->pf_wteset_dec;

	set_data_segment(new_gp);
	rt_value = (*pf_wteset_dec)(event, todly);
	set_data_segment(old_gp);

	return	rt_value;
}

int  clrevt(uint event)
{
	unsigned long	old_gp, new_gp;
	int  (*pf_clrevt)(uint event), rt_value;

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_clrevt = mt_usb_export->pf_clrevt;

	set_data_segment(new_gp);
	rt_value = (*pf_clrevt)(event);
	set_data_segment(old_gp);

	return	rt_value;
}

void	usb_isr_install(void)
{
	unsigned long	old_gp, new_gp;
	void (*pf_usb_isr_install)(void);

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_usb_isr_install = mt_usb_export->pf_usb_isr_install;

	set_data_segment(new_gp);
	(*pf_usb_isr_install)();
	set_data_segment(old_gp);
}

void	usb_uhci1_isr_install(void)
{
	unsigned long	old_gp, new_gp;
	void (*pf_usb_uhci1_isr_install)(void);

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_usb_uhci1_isr_install = mt_usb_export->pf_usb_uhci1_isr_install;

	set_data_segment(new_gp);
	(*pf_usb_uhci1_isr_install)();
	set_data_segment(old_gp);
}

void	usb_uhci2_isr_install(void)
{
	unsigned long	old_gp, new_gp;
	void (*pf_usb_uhci2_isr_install)(void);

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_usb_uhci2_isr_install = mt_usb_export->pf_usb_uhci2_isr_install;

	set_data_segment(new_gp);
	(*pf_usb_uhci2_isr_install)();
	set_data_segment(old_gp);
}

void	led_usb_on(void)
{
	unsigned long	old_gp, new_gp;
	void (*pf_led_usb_on)(void);

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_led_usb_on = mt_usb_export->pf_led_usb_on;

	set_data_segment(new_gp);
	(*pf_led_usb_on)();
	set_data_segment(old_gp);
}

void	led_usb_off(void)
{
	unsigned long	old_gp, new_gp;
	void (*pf_led_usb_off)(void);

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_led_usb_off = mt_usb_export->pf_led_usb_off;

	set_data_segment(new_gp);
	(*pf_led_usb_off)();
	set_data_segment(old_gp);
}

void led_usb_flash(char isEnable)
{
	unsigned long	old_gp, new_gp;
	void (*pf_led_usb_flash)(char isEnable);

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_led_usb_flash = mt_usb_export->pf_led_usb_flash;

	set_data_segment(new_gp);
	(*pf_led_usb_flash)(isEnable);
	set_data_segment(old_gp);
}

unsigned char PciCfgRdByte( int hDev, int offset )
{
	unsigned long	old_gp, new_gp;
	unsigned char (*pf_PciCfgRdByte)( int hDev, int offset ), rt_value;

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_PciCfgRdByte = mt_usb_export->pf_PciCfgRdByte;

	set_data_segment(new_gp);
	rt_value = (*pf_PciCfgRdByte)(hDev, offset);
	set_data_segment(old_gp);

	return	rt_value;
}

unsigned short PciCfgRdWord( int hDev, int offset )
{
	unsigned long	old_gp, new_gp;
	unsigned short (*pf_PciCfgRdWord)( int hDev, int offset ), rt_value;

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_PciCfgRdWord = mt_usb_export->pf_PciCfgRdWord;

	set_data_segment(new_gp);
	rt_value = (*pf_PciCfgRdWord)(hDev, offset);
	set_data_segment(old_gp);

	return	rt_value;
}

unsigned long PciCfgRdDword( int hDev, int offset )
{
	unsigned long	old_gp, new_gp;
	unsigned long (*pf_PciCfgRdDword)( int hDev, int offset ), rt_value;

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_PciCfgRdDword = mt_usb_export->pf_PciCfgRdDword;

	set_data_segment(new_gp);
	rt_value = (*pf_PciCfgRdDword)(hDev, offset);
	set_data_segment(old_gp);

	return	rt_value;
}

void PciCfgWrByte( int hDev, int offset, unsigned char data )
{
	unsigned long	old_gp, new_gp;
	void (*pf_PciCfgWrByte)( int hDev, int offset, unsigned char data );

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_PciCfgWrByte = mt_usb_export->pf_PciCfgWrByte;

	set_data_segment(new_gp);
	(*pf_PciCfgWrByte)(hDev, offset, data);
	set_data_segment(old_gp);
}

void PciCfgWrWord( int hDev, int offset, unsigned short data )
{
	unsigned long	old_gp, new_gp;
	void (*pf_PciCfgWrWord)( int hDev, int offset, unsigned short data );

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_PciCfgWrWord = mt_usb_export->pf_PciCfgWrWord;

	set_data_segment(new_gp);
	(*pf_PciCfgWrWord)(hDev, offset, data);
	set_data_segment(old_gp);
}

void PciCfgWrDword( int hDev, int offset, unsigned long data )
{
	unsigned long	old_gp, new_gp;
	void (*pf_PciCfgWrDword)( int hDev, int offset, unsigned long data );

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_PciCfgWrDword = mt_usb_export->pf_PciCfgWrDword;

	set_data_segment(new_gp);
	(*pf_PciCfgWrDword)(hDev, offset, data);
	set_data_segment(old_gp);
}

void Nputchr(char ch)
{
	unsigned long	old_gp, new_gp;
	void (*pf_Nputchr)(char ch);

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_Nputchr = mt_usb_export->pf_Nputchr;

	set_data_segment(new_gp);
	(*pf_Nputchr)(ch);
	set_data_segment(old_gp);
}

void PrintULONG(unsigned long num)
{
	unsigned long	old_gp, new_gp;
	void (*pf_PrintULONG)(unsigned long num);

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_PrintULONG = mt_usb_export->pf_PrintULONG;

	set_data_segment(new_gp);
	(*pf_PrintULONG)(num);
	set_data_segment(old_gp);
}

void updateDiskDriveInfo(char *vendor, char *model, int driveID)
{
	unsigned long	old_gp, new_gp;
	void (*pf_updateDiskDriveInfo)(char *vendor, char *model, int driveID);

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_updateDiskDriveInfo = mt_usb_export->pf_updateDiskDriveInfo;

	set_data_segment(new_gp);
	(*pf_updateDiskDriveInfo)(vendor, model, driveID);
	set_data_segment(old_gp);
}

void updateShareConfPlug(char resType, char mode)
{
	unsigned long	old_gp, new_gp;
	void (*pf_updateShareConfPlug)(char resType, char mode);

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_updateShareConfPlug = mt_usb_export->pf_updateShareConfPlug;

	set_data_segment(new_gp);
	(*pf_updateShareConfPlug)(resType, mode);
	set_data_segment(old_gp);
}

unsigned long diskMount_status(int disk)
{
	unsigned long	old_gp, new_gp;
	unsigned long (*pf_diskMount_status)(int disk), rt_value;

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_diskMount_status = mt_usb_export->pf_diskMount_status;

	set_data_segment(new_gp);
	rt_value = (*pf_diskMount_status)(disk);
	set_data_segment(old_gp);

	return	rt_value;
}

void diskMount_set_status(int disk, unsigned long status)
{
	unsigned long	old_gp, new_gp;
	void (*pf_diskMount_set_status)(int disk, unsigned long status);

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_diskMount_set_status = mt_usb_export->pf_diskMount_set_status;

	set_data_segment(new_gp);
	(*pf_diskMount_set_status)(disk, status);
	set_data_segment(old_gp);
}

int partition_mount(unsigned long VolumeID, unsigned char volumeType, struct DiskParamS *diskInfo)
{
	unsigned long	old_gp, new_gp;
	int (*pf_partition_mount)(unsigned long VolumeID, unsigned char volumeType, struct DiskParamS *diskInfo), rt_value;

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_partition_mount = mt_usb_export->pf_partition_mount;

	set_data_segment(new_gp);
	rt_value = (*pf_partition_mount)(VolumeID, volumeType, diskInfo);
	set_data_segment(old_gp);

	return	rt_value;
}

int partition_umount(unsigned long VolumeID, unsigned char volumeType)
{
	unsigned long	old_gp, new_gp;
	int (*pf_partition_umount)(unsigned long VolumeID, unsigned char volumeType), rt_value;

	old_gp = data_segment_usb;
	new_gp = data_segment_mt;
	pf_partition_umount = mt_usb_export->pf_partition_umount;

	set_data_segment(new_gp);
	rt_value = (*pf_partition_umount)(VolumeID, volumeType);
	set_data_segment(old_gp);

	return	rt_value;
}
