/*
 * bus.c - bus driver management
 *
 * Copyright (c) 2002-3 Patrick Mochel
 * Copyright (c) 2002-3 Open Source Development Labs
 * 
 * This file is released under the GPLv2
 *
 */

#undef DEBUG

// #ifdef _ETCPIP
#define __KERNEL__
        #ifdef _NEWLIB
          #include "string.h"
        #endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
/*
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsbMsInit.h"
*/
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP


/*
#include <linux/device.h>
*/
#include "PoDevice.h" // include/linux/device.h
#include "kdev_t.h"   // also need
// B. S. Liaw Add this
#include "base/PoDeviceSc.h" // #include/linux/device.h , 2.6.0-test7
/*
#include <linux/module.h>
#include <linux/errno.h>
*/
#include "PoErrno.h" // osport_errno.h
/*
#include <linux/init.h>
#include <linux/string.h>
*/
#include "PoString.h" // linux/string.h

#include "base.h"
/*
#include "power/power.h"
*/

// B. S. Liaw add for pr_debug()
#include "PoKernel.h"


#define to_dev(node) container_of(node,struct device,bus_list)

/*
#define to_drv(node) container_of(node,struct device_driver,kobj.entry)
*/
#define to_drv(node) container_of(node,struct device_driver,bus_drv_list)


#define to_bus_attr(_attr) container_of(_attr,struct bus_attribute,attr)
#define to_bus(obj) container_of(obj,struct bus_type,subsys.kset.kobj)

/*
 * sysfs bindings for drivers
 */

#define to_drv_attr(_attr) container_of(_attr,struct driver_attribute,attr)
#define to_driver(obj) container_of(obj, struct device_driver, kobj)

// B. S, Liaw add this
struct subsystem bus_subsys ;
/*
#ifndef NO_SYS_FS
// struct attribute on sysfs.h
static ssize_t
drv_attr_show(struct kobject * kobj, struct attribute * attr, char * buf)
{
	struct driver_attribute * drv_attr = to_drv_attr(attr);
	struct device_driver * drv = to_driver(kobj);
	ssize_t ret = 0;

	if (drv_attr->show)
		ret = drv_attr->show(drv,buf);
	return ret;
}

static ssize_t
drv_attr_store(struct kobject * kobj, struct attribute * attr, 
	       const char * buf, size_t count)
{
	struct driver_attribute * drv_attr = to_drv_attr(attr);
	struct device_driver * drv = to_driver(kobj);
	ssize_t ret = 0;

	if (drv_attr->store)
		ret = drv_attr->store(drv,buf,count);
	return ret;
}

static struct sysfs_ops driver_sysfs_ops = {
	.show	= drv_attr_show,
	.store	= drv_attr_store,
};
#endif // ndef NO_SYS_FS
*/

static void driver_release(struct kobject * kobj)
{
 struct device_driver * drv = to_driver(kobj);
 #ifndef NO_SEMPHORE
	up(&drv->unload_sem);
 #endif // ndef NO_SEMPHORE
}

static struct kobj_type ktype_driver = {
 #ifndef NO_SYS_FS
	.sysfs_ops	= &driver_sysfs_ops,
 #endif // ndef NO_SYS_FS
 .release	= driver_release,
};


/*
 * sysfs bindings for buses
 */

/*
#ifndef NO_SYS_FS
   // struct attribute on sysfs.h
   static ssize_t
   bus_attr_show(struct kobject * kobj, struct attribute * attr, char * buf)
   {
	struct bus_attribute * bus_attr = to_bus_attr(attr);
	struct bus_type * bus = to_bus(kobj);
	ssize_t ret = 0;

	if (bus_attr->show)
		ret = bus_attr->show(bus,buf);
	return ret;
   }

   // struct attribute on sysfs.h
   static ssize_t
   bus_attr_store(struct kobject * kobj, struct attribute * attr, 
	       const char * buf, size_t count)
   {
	struct bus_attribute * bus_attr = to_bus_attr(attr);
	struct bus_type * bus = to_bus(kobj);
	ssize_t ret = 0;

	if (bus_attr->store)
		ret = bus_attr->store(bus,buf,count);
	return ret;
}

static struct sysfs_ops bus_sysfs_ops = {
	.show	= bus_attr_show,
	.store	= bus_attr_store,
};
#endif // ndef NO_SYS_FS
*/
int bus_create_file(struct bus_type * bus, struct bus_attribute * attr)
{
 int error;
 if (get_bus(bus)) {
	#ifndef NO_SYS_FS
		    // struct attribute on sysfs.h
		error = sysfs_create_file(&bus->subsys.kset.kobj,&attr->attr);
	 #else
		error = 0 ;
	#endif
	put_bus(bus);
   } else
	error = -EINVAL;

 return error;
}

void bus_remove_file(struct bus_type * bus, struct bus_attribute * attr)
{
	if (get_bus(bus)) {
		#ifndef NO_SYS_FS
		    // struct attribute on sysfs.h
			sysfs_remove_file(&bus->subsys.kset.kobj,&attr->attr);
		#endif // ndef NO_SYS_FS
		put_bus(bus);
	}
}
/*
#ifndef NO_SYS_FS
static struct kobj_type ktype_bus = {
	.sysfs_ops	= &bus_sysfs_ops,

};
#endif // ndef NO_SYS_FS

#ifndef NO_SYS_FS

decl_subsys(bus,&ktype_bus,NULL);

// decl_subsys(bus,&ktype_bus);

#endif // ndef NO_SYS_FS
*/
/**
 *	bus_for_each_dev - device iterator.
 *	@bus:	bus type.
 *	@start:	device to start iterating from.
 *	@data:	data for the callback.
 *	@fn:	function to be called for each device.
 *
 *	Iterate over @bus's list of devices, and call @fn for each,
 *	passing it @data. If @start is not NULL, we use that device to
 *	begin iterating from.
 *
 *	We check the return of @fn each time. If it returns anything
 *	other than 0, we break out and return that value.
 *
 *	NOTE: The device that returns a non-zero value is not retained
 *	in any way, nor is its refcount incremented. If the caller needs
 *	to retain this data, it should do, and increment the reference 
 *	count in the supplied callback.
 */
int bus_for_each_dev(struct bus_type * bus, struct device * start, 
		     void * data, int (*fn)(struct device *, void *))
{
 struct list_head * head, * entry;
 int error = 0;

 if (!(bus = get_bus(bus)))
		return -EINVAL;

 #ifndef NO_KSET
	head = start ? &start->bus_list : &bus->devices.list;
  #else
	head = start ? &start->bus_list : bus->devices_list;
 #endif

 #ifndef NO_SEMPHORE
	down_read(&bus->subsys.rwsem);
 #endif // ndef NO_SEMPHORE

 list_for_each(entry,head) {
	struct device * dev = get_device(to_dev(entry));

	error = fn(dev,data);
		put_device(dev);
		if (error)
			break;
	}

 #ifndef NO_SEMPHORE
	up_read(&bus->subsys.rwsem);
 #endif // ndef NO_SEMPHORE

 put_bus(bus);

 return error;
}

/**
 *	bus_for_each_drv - driver iterator
 *	@bus:	bus we're dealing with.
 *	@start:	driver to start iterating on.
 *	@data:	data to pass to the callback.
 *	@fn:	function to call for each driver.
 *
 *	This is nearly identical to the device iterator above.
 *	We iterate over each driver that belongs to @bus, and call
 *	@fn for each. If @fn returns anything but 0, we break out
 *	and return it. If @start is not NULL, we use it as the head
 *	of the list.
 *
 *	NOTE: we don't return the driver that returns a non-zero 
 *	value, nor do we leave the reference count incremented for that
 *	driver. If the caller needs to know that info, it must set it
 *	in the callback. It must also be sure to increment the refcount
 *	so it doesn't disappear before returning to the caller.
 */

int bus_for_each_drv(struct bus_type * bus, struct device_driver * start,
		     void * data, int (*fn)(struct device_driver *, void *))
{
 struct list_head * head, * entry;
 int error = 0;

 if(!(bus = get_bus(bus)))
		return -EINVAL;

 #ifndef NO_KSET
	head = start ? &start->kobj.entry : &bus->drivers.list;
  #else
	head = start ? &start->kobj.entry : bus->drivers_list;
 #endif
 /*
 #ifndef NO_SEMPHORE
	down_read(&bus->subsys.rwsem);
 #endif // ndef NO_SEMPHORE
 */
 list_for_each(entry,head) {
	struct device_driver *drv = get_driver( to_drv(entry)
												);

	USB_DBG_INIT(iprintf(
			"[USB INIT] bus_for_each_drv() : bus driver %08XH .. \n",drv)) ;
	error = fn(drv,data);
	put_driver(drv);
	if(error)
			break;
   }
 /*
 #ifndef NO_SEMPHORE
	up_read(&bus->subsys.rwsem);
 #endif // ndef NO_SEMPHORE
 */
 put_bus(bus);
 return error;
}

/**
 *	device_bind_driver - bind a driver to one device.
 *	@dev:	device.
 *
 *	Allow manual attachment of a driver to a deivce.
 *	Caller must have already set @dev->driver.
 *
 *	Note that this does not modify the bus reference count 
 *	nor take the bus's rwsem. Please verify those are accounted 
 *	for before calling this. (It is ok to call with no other effort
 *	from a driver's probe() method.)
 */

void device_bind_driver(struct device * dev)
{
 USB_DBG_FLOW(iprintf("[USB FLOW] device_bind_driver() Begin .. \n"));

 #ifdef USB_DBGSTRING_ON
	pr_debug("bound device '%s' to driver '%s'\n",
		 dev->bus_id,dev->driver->name);
 #endif // def USB_DBGSTRING_ON

 USB_DBG_MEM(iprintf("[USB MEM] device_bind_driver() : dev->driver->devices=%08XH \n",
			dev->driver->devices));
 #ifdef USB_BS_ASSERT
	if ( ! (IS_VALID_ADDR(&dev->driver->devices)) ) {
		iprintf("[ASSERT][USB] device_bind_driver() : " ) ;
	    iprintf("Error : Invlid address &dev->driver->devices=%08XH \n",
				&dev->driver->devices) ;
	    iprintf("[ASSERT][USB] Continue ..\n") ;
	   }
 #endif // USB_BS_ASSERT
 list_add_tail(&dev->driver_list,&dev->driver->devices);
 #ifndef NO_SYS_FS
	sysfs_create_link(&dev->driver->kobj,&dev->kobj,
			  kobject_name(&dev->kobj));
 #endif // ndef NO_SYS_FS
 USB_DBG_FLOW(iprintf("[USB FLOW] device_bind_driver() End. \n"));
}


/**
 *	bus_match - check compatibility between device & driver.
 *	@dev:	device.
 *	@drv:	driver.
 *
 *	First, we call the bus's match function, which should compare
 *	the device IDs the driver supports with the device IDs of the 
 *	device. Note we don't do this ourselves because we don't know 
 *	the format of the ID structures, nor what is to be considered
 *	a match and what is not.
 *	
 *	If we find a match, we call @drv->probe(@dev) if it exists, and 
 *	call attach() above.
 */
static int bus_match(struct device * dev, struct device_driver * drv)
{
 int error = -ENODEV;


 USB_DBG_FLOW(iprintf("[USB FLOW] bus_match() Begin .. \n"));
 USB_DBG_PLUG(iprintf("[USB PLUG] bus_match() Begin .. \n"));

 USB_DBG_FLOW(iprintf("[USB FLOW] bus_match() : dev->bus->match() .. \n"));
 if(dev->bus->match(dev,drv)) {
	dev->driver = drv;
	if (drv->probe) {
		USB_DBG_PLUG(iprintf("[USB PLUG] bus_match() : drv->probe() .. \n"));
		USB_DBG_INIT(iprintf("[USB INIT] bus_match() : drv->probe() .. \n"));
		USB_DBG_FLOW(iprintf("[USB FLOW] bus_match() : drv->probe() .. \n"));
		if ((error = drv->probe(dev))) {
			dev->driver = NULL;
			return error;
		   }
	   }
	 else {
		USB_DBG_PLUG(iprintf("[USB PLUG] bus_match() : No drv->probe() !! \n"));
		USB_DBG_INFO(iprintf("[USB INFO] bus_match() : No drv->probe() !! \n"));
	   }
	device_bind_driver(dev);
	error = 0;
   }
 USB_DBG_FLOW(iprintf("[USB FLOW] bus_match() End. \n"));
 return error;
}


/**
 *	device_attach - try to attach device to a driver.
 *	@dev:	device.
 *
 *	Walk the list of drivers that the bus has and call bus_match() 
 *	for each pair. If a compatible pair is found, break out and return.
 */
static int device_attach(struct device * dev)
{
 struct bus_type * bus = dev->bus;
 struct list_head * entry;
 int error;


 USB_DBG_FLOW(iprintf("[USB FLOW] device_attach() Begin .. \n"));
 if (dev->driver) {
	USB_DBG_FLOW(iprintf("[USB FLOW] device_attach() : device_bind_driver() .. \n"));
	device_bind_driver(dev);
	return 1;
	}

 if (bus->match) {

	#ifndef NO_KSET
		USB_DBG_MEM(iprintf("[USB MEM] device_attach() : bus->drivers.list=%08XH \n",
			bus->drivers.list )) ;
		list_for_each(entry,&bus->drivers.list) {
	 #else
		USB_DBG_MEM(iprintf("[USB MEM] device_attach() : bus->drivers_list=%08XH \n",
			bus->drivers_list )) ;
		list_for_each(entry,bus->drivers_list) {
	#endif
  
		struct device_driver *drv = to_drv(entry);


		USB_DBG_INIT(iprintf(
			"[USB INIT] device_attach() : bus driver %08XH .. \n",drv)) ;
		USB_DBG_FLOW(iprintf("[USB FLOW] device_attach() : bus_match() .. \n"));
		error = bus_match(dev,drv);
		if (!error )  
				/* success, driver matched */
				return 1; 
		if (error != -ENODEV) 
				/* driver matched but the probe failed */
			#ifdef USB_DBGSTRING_ON
				/*
				printk(KERN_WARNING 
				*/
				dprintf("[USB] "
				    "%s: probe of %s failed with error %d\n",
				    drv->name, dev->bus_id, error);
			 #else
				dprintf("[USB] : probe of %s failed with error %d\n",
				    dev->bus_id, error);
			#endif
		}
	}

 USB_DBG_FLOW(iprintf("[USB FLOW] device_attach() End. \n"));
 return 0;
}


/**
 *	driver_attach - try to bind driver to devices.
 *	@drv:	driver.
 *
 *	Walk the list of devices that the bus has on it and try to match
 *	the driver with each one.
 *	If bus_match() returns 0 and the @dev->driver is set, we've found
 *	a compatible pair.
 *
 *	Note that we ignore the -ENODEV error from bus_match(), since it's 
 *	perfectly valid for a driver not to bind to any devices.
 */

void driver_attach(struct device_driver * drv)
{
 struct bus_type * bus = drv->bus;
 struct list_head * entry;
 int error;


 USB_DBG_FLOW(iprintf("[USB FLOW] driver_attach() Begin .. \n"));
 if (!bus->match)
   {
	USB_DBG_INFO(iprintf("[USB INFO] driver_attach() : Null bus->match !! \n"));
	USB_DBG_FLOW(iprintf("[USB FLOW] driver_attach() End with Fail !! \n"));
	return;
   }
 #ifndef NO_KSET
	list_for_each(entry,&bus->devices.list) {
  #else
	list_for_each(entry,bus->devices_list) {
 #endif
		struct device * dev = container_of(entry,struct device,bus_list);
		if (!dev->driver) {
			USB_DBG_FLOW(iprintf("[USB FLOW] driver_attach() : bus_match() .. \n"));
			error = bus_match(dev,drv);
			if (error && (error != -ENODEV))
			   {
				// driver matched but the probe failed 
				/*
				printk(KERN_WARNING 
				    "%s: probe of %s failed with error %d\n",
				    drv->name, dev->bus_id, error);
				*/
				#ifdef USB_DBGSTRING_ON
					dprintf("[USB] Error : %s : probe of %s failed with error %d\n",
					    drv->name, dev->bus_id, error);
				 #else
					dprintf("[USB] Error : probe of %s failed with error %d\n",
				     dev->bus_id, error);
				#endif
			   }
		}
	}
 USB_DBG_FLOW(iprintf("[USB FLOW] driver_attach() End. \n"));
}


/**
 *	device_release_driver - manually detach device from driver.
 *	@dev:	device.
 *
 *	Manually detach device from driver.
 *	Note that this is called without incrementing the bus
 *	reference count nor taking the bus's rwsem. Be sure that
 *	those are accounted for before calling this function.
 */

void device_release_driver(struct device * dev)
{
 struct device_driver * drv = dev->driver;

 USB_DBG_PLUG(iprintf("[USB PLUG] device_release_driver() : Begin .. \n"));
 if (drv) {
	 #ifndef NO_SYS_FS
		sysfs_remove_link(&drv->kobj,kobject_name(&dev->kobj));
	 #endif // ndef NO_SYS_FS
	USB_DBG_FLOW(iprintf("[USB FLOW] device_release_driver() : list_del_init() .. \n"));
	 list_del_init(&dev->driver_list);

	 #ifndef _ETCPIP
		device_detach_shutdown(dev);
	 #endif // ndef _ETCPIP
	if (drv->remove) {
		USB_DBG_PLUG(iprintf("[USB PLUG] device_release_driver() : drv->remove() .. \n"));
		USB_DBG_INIT(iprintf("[USB INIT] device_release_driver() : drv->remove() .. \n"));
		drv->remove(dev);
	   }
	  else {
		USB_DBG_PLUG(iprintf("[USB PLUG] device_release_driver() : No drv->remove() !! \n"));
		USB_DBG_INFO(iprintf("[USB INIT] device_release_driver() : No drv->remove() !! \n"));
	   }
	dev->driver = NULL;
   }
  else {
	USB_DBG_PLUG(iprintf("[USB PLUG] device_release_driver() : No drv !! \n"));
	USB_DBG_INFO(iprintf("[USB INIT] device_release_driver() : No drv !! \n"));
   }
}


/**
 *	driver_detach - detach driver from all devices it controls.
 *	@drv:	driver.
 */

static void driver_detach(struct device_driver * drv)
{
	struct list_head * entry, * next;
	list_for_each_safe(entry,next,&drv->devices) {
		struct device * dev = container_of(entry,struct device,driver_list);
		USB_DBG_PLUG(iprintf("[USB PLUG] driver_detach() : device_release_driver() .. \n"));
		device_release_driver(dev);
	}
	
}

/**
 *	bus_add_device - add device to bus
 *	@dev:	device being added
 *
 *	- Add the device to its bus's list of devices.
 *	- Try to attach to driver.
 *	- Create link to device's physical location.
 */
int bus_add_device(struct device * dev)
{
 /*
 struct bus_type * bus = get_bus(dev->bus);
 */
 struct bus_type * bus ;
 int error = 0;


 USB_DBG_FLOW(iprintf("[USB FLOW] bus_add_device() Begin .. \n"));

 USB_DBG_PLUG(iprintf("[USB PLUG] bus_add_device() : dev=%08XH \n",
	dev));
 USB_DBG_PLUG(iprintf("[USB PLUG] bus_add_device() : dev->bus=%08XH \n",
	dev->bus));
 USB_DBG_MEM(iprintf("[USB MEM] bus_add_device() : dev->bus=%08XH \n",
	dev->bus));

 bus = get_bus(dev->bus);

 USB_DBG_PLUG(iprintf("[USB PLUG] bus_add_device() : bus=%08XH \n",bus));
 USB_DBG_MEM(iprintf("[USB MEM] bus_add_device() : bus=%08XH \n",bus));
 if (bus) {
	#ifndef NO_SEMPHORE
		down_write(&dev->bus->subsys.rwsem);
	#endif // ndef NO_SEMPHORE

	#ifdef USB_DBGSTRING_ON
		/*
		pr_debug("bus %s: add device %s\n",bus->name,dev->bus_id);
		*/
		USB_DBG_INIT(iprintf("[USB INIT] bus [%s]: add device [%s] \n",
			bus->name,dev->bus_id));
	#endif // USB_DBGSTRING_ON

	#ifndef NO_KSET
		USB_DBG_MEM(iprintf("[USB MEM] bus_add_device() : dev->bus->devices=%08XH \n",
			dev->bus->devices));
		USB_DBG_MEM(iprintf("[USB MEM] bus_add_device() : dev->bus->devices.list=%08XH \n",
			dev->bus->devices.list));
		#ifdef USB_BS_ASSERT
			if ( ! (IS_VALID_ADDR(dev->bus->devices.list)) ) {
				iprintf("[ASSERT][USB] bus_add_device() : " ) ;
    		    iprintf("Error : Invlid address dev->bus->devices.list=%08XH \n",
						dev->bus->devices.list) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT

		list_add_tail(&dev->bus_list,&dev->bus->devices.list);
	 #else
		USB_DBG_MEM(iprintf("[USB MEM] bus_add_device() : dev->bus->devices_list=%08XH \n",
			dev->bus->devices_list));

		#ifdef USB_BS_ASSERT
			if ( ! (IS_VALID_ADDR(dev->bus->devices_list)) ) {
				iprintf("[ASSERT][USB] bus_add_device() : " ) ;
    		    iprintf("Error : Invlid address dev->bus->devices_list=%08XH \n",
						dev->bus->devices_list) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT

		list_add_tail(&dev->bus_list,dev->bus->devices_list);
	#endif


	USB_DBG_FLOW(iprintf("[USB FLOW] bus_add_device() : device_attach() .. \n"));
	USB_DBG_INIT(iprintf("[USB INIT] bus_add_device() : device_attach() .. \n"));
	device_attach(dev);

	#ifndef NO_SEMPHORE
		up_write(&dev->bus->subsys.rwsem);
	#endif // ndef NO_SEMPHORE
	#ifndef NO_SYS_FS
		sysfs_create_link(&bus->devices.kobj,&dev->kobj,dev->bus_id);
	#endif // ndef NO_SYS_FS
   }
  else
   {
	USB_DBG_INFO(iprintf(
		"[USB INFO] bus_add_device() : Error : Null bus=%08XH !! \n",
		bus));
   }
 USB_DBG_FLOW(iprintf("[USB FLOW] bus_add_device() End. \n"));
 return error;
}

/**
 *	bus_remove_device - remove device from bus
 *	@dev:	device to be removed
 *
 *	- Remove symlink from bus's directory.
 *	- Delete device from bus's list.
 *	- Detach from its driver.
 *	- Drop reference taken in bus_add_device().
 */
void bus_remove_device(struct device * dev)
{
 USB_DBG_FLOW(iprintf("[USB FLOW] bus_remove_device() Begin .. \n"));
 USB_DBG_PLUG(iprintf("[USB PLUG] bus_remove_device() Begin .. \n"));
 USB_DBG_PLUG(iprintf("[USB PLUG] bus_remove_device() : dev=%08XH \n",
	dev));
 if(dev->bus) {
	#ifndef NO_SYS_FS
		sysfs_remove_link(&dev->bus->devices.kobj,dev->bus_id);
	#endif // ndef NO_SYS_FS
	#ifndef NO_SEMPHORE
		down_write(&dev->bus->subsys.rwsem);
	#endif // ndef NO_SEMPHORE

	#ifdef USB_DBGSTRING_ON
		pr_debug("bus %s: remove device %s\n",dev->bus->name,dev->bus_id);
	#endif // USB_DBGSTRING_ON

	USB_DBG_PLUG(iprintf("[USB PLUG] bus_remove_device() : device_release_driver() .. \n"));
	device_release_driver(dev);
	USB_DBG_FLOW(iprintf("[USB FLOW] bus_remove_device() : list_del_init() .. \n"));
	list_del_init(&dev->bus_list);
	#ifndef NO_SEMPHORE
		up_write(&dev->bus->subsys.rwsem);
	#endif // ndef NO_SEMPHORE
	put_bus(dev->bus);
   }
  else {
	USB_DBG_PLUG(iprintf("[USB PLUG] bus_remove_device() : No dev->bus !! \n"));
   }
}


/**
 *	bus_add_driver - Add a driver to the bus.
 *	@drv:	driver.
 *
 */

int bus_add_driver(struct device_driver * drv)
{
 struct bus_type * bus = get_bus(drv->bus);
 int error = 0;

 USB_DBG_FLOW(iprintf("[USB FLOW] bus_add_driver() Begin .. \n")) ;

 if (bus) {
		
	#ifdef USB_DBGSTRING_ON
		USB_DBG_INIT(iprintf(
			"[USB INIT] bus_add_driver() : bus [%s] add driver [%s] %08XH .. \n",
			bus->name,drv->name,drv));
		USB_DBG_INFO(iprintf(
			"[USB INFO] bus_add_driver() : bus [%s] add driver [%s] \n",bus->name,drv->name));
	 #else
		USB_DBG_INIT(iprintf(
			"[USB INIT] bus_add_driver() : bus add driver %08XH .. \n",drv)) ;
		USB_DBG_INFO(iprintf(
			"[USB INFO] bus_add_driver() : bus add driver ..\n"));
	#endif
	
 #ifdef USB_DBGSTRING_ON
	kobject_set_name(&drv->kobj,drv->name);
 #endif // USB_DBGSTRING_ON
	#ifndef NO_KSET
		drv->kobj.kset = &bus->drivers;
		if ((error = kobject_register(&drv->kobj))) {
				put_bus(bus);
				return error;
			}
	 #else
		USB_DBG_MEM(iprintf("[USB MEM] bus_add_driver() : bus->drivers_list=%08XH \n",
			bus->drivers_list )) ;
		list_add_tail(&drv->bus_drv_list,bus->drivers_list);
	#endif
	
	#ifndef NO_SEMPHORE
		down_write(&bus->subsys.rwsem);
	#endif // ndef NO_SEMPHORE
	
	driver_attach(drv);
	
	#ifndef NO_SEMPHORE
		up_write(&bus->subsys.rwsem);
	#endif // ndef NO_SEMPHORE
	
	if (error) {
		// #ifndef _ETCPIP
			kobject_unregister(&drv->kobj);
		// #endif // ndef _ETCPIP
			put_bus(bus);
		}
	
   }
 USB_DBG_FLOW(iprintf("[USB FLOW] bus_add_driver() End. \n")) ;
 return error;
}


/**
 *	bus_remove_driver - delete driver from bus's knowledge.
 *	@drv:	driver.
 *
 *	Detach the driver from the devices it controls, and remove
 *	it from its bus's list of drivers. Finally, we drop the reference
 *	to the bus we took in bus_add_driver().
 */

void bus_remove_driver(struct device_driver * drv)
{
	if (drv->bus) {
 #ifndef NO_SEMPHORE
		down_write(&drv->bus->subsys.rwsem);
 #endif // ndef NO_SEMPHORE

 #ifdef USB_DBGSTRING_ON
		pr_debug("bus %s: remove driver %s\n",drv->bus->name,drv->name);
 #endif // USB_DBGSTRING_ON

 driver_detach(drv);

 #ifndef NO_SEMPHORE
		up_write(&drv->bus->subsys.rwsem);
 #endif // ndef NO_SEMPHORE
 #ifndef _ETCPIP
		kobject_unregister(&drv->kobj);
 #endif // ndef _ETCPIP
		put_bus(drv->bus);
	}
}


/* Helper for bus_rescan_devices's iter */
static int bus_rescan_devices_helper(struct device *dev, void *data)
{
	int *count = data;

	USB_DBG_FLOW(iprintf("[USB FLOW] bus_rescan_devices_helper() : device_attach() .. \n"));
	if (!dev->driver && device_attach(dev))
		(*count)++;

	return 0;
}


/**
 *	bus_rescan_devices - rescan devices on the bus for possible drivers
 *	@bus:	the bus to scan.
 *
 *	This function will look for devices on the bus with no driver
 *	attached and rescan it against existing drivers to see if it
 *	matches any. Calls device_attach(). Returns the number of devices
 *	that were sucessfully bound to a driver.
 */
int bus_rescan_devices(struct bus_type * bus)
{
	int count = 0;

	bus_for_each_dev(bus, NULL, &count, bus_rescan_devices_helper);

	return count;
}


struct bus_type * get_bus(struct bus_type * bus)
{
 USB_DBG_FLOW(iprintf("[USB FLOW] get_bus() Called. \n"));
 return bus ? container_of(subsys_get(&bus->subsys),struct bus_type,subsys) : NULL;
}


void put_bus(struct bus_type * bus)
{
 // #ifndef _ETCPIP
	subsys_put(&bus->subsys);
 // #endif // ndef _ETCPIP
}


/**
 *	find_bus - locate bus by name.
 *	@name:	name of bus.
 *
 *	Call kset_find_obj() to iterate over list of buses to
 *	find a bus by name. Return bus if found.
 */


struct bus_type * find_bus(char * name)
{
	struct kobject * k = kset_find_obj(&bus_subsys.kset,name);
	return k ? to_bus(k) : NULL;
}


/**
 *	bus_register - register a bus with the system.
 *	@bus:	bus.
 *
 *	Once we have that, we registered the bus with the kobject
 *	infrastructure, then register the children subsystems it has:
 *	the devices and drivers that belong to the bus. 
 */

#ifndef NO_KSET
	int bus_register(struct bus_type * bus);
 #else
	int bus_register(struct bus_type * bus,struct list_head *drivers_list,
			struct list_head *devices_list)
#endif
{
 USB_DBG_FLOW(iprintf("[USB FLOW] bus_register() Begin .. \n"));
 // #ifndef _ETCPIP
 #ifdef USB_DBGSTRING_ON
	kobject_set_name(&bus->subsys.kset.kobj,bus->name);
 #endif // USB_DBGSTRING_ON

 subsys_set_kset(bus,bus_subsys);
 
 #ifndef NO_KSET
	subsystem_register(&bus->subsys);
  #else
	bus->drivers_list = drivers_list ;
	USB_DBG_MEM(iprintf("[USB MEM] bus_register() : Set Value bus->drivers_list=%08XH \n",
		bus->drivers_list )) ;
	bus->devices_list = devices_list ;
 #endif

 #ifndef NO_KSET
	kobject_set_name(&bus->devices.kobj, "devices");
	bus->devices.subsys = &bus->subsys;
 #endif // ndef NO_KSET
 #ifndef _ETCPIP
	kset_register(&bus->devices);
 #endif // ndef NO_SYS_FS

 #ifndef NO_KSET
	kobject_set_name(&bus->drivers.kobj, "drivers");
	bus->drivers.subsys = &bus->subsys;
	bus->drivers.ktype = &ktype_driver;
 #endif // ndef NO_KSET
 #ifndef _ETCPIP
	kset_register(&bus->drivers);
 #endif // ndef NO_SYS_FS
 
 #ifdef USB_DBGSTRING_ON
	USB_DBG_INIT(iprintf(
		"[USB INIT] bus_register() : bus type [%s] registered\n",
		bus->name
		));
  #else
	USB_DBG_INIT(iprintf(
		"[USB INIT] bus_register() : bus registered \n"
		));
 #endif
 
 USB_DBG_FLOW(iprintf("[USB FLOW] bus_register() End. \n"));
 return 0;
}


/**
 *	bus_unregister - remove a bus from the system 
 *	@bus:	bus.
 *
 *	Unregister the child subsystems and the bus itself.
 *	Finally, we call put_bus() to release the refcount
 */
void bus_unregister(struct bus_type * bus)
{
 #ifdef USB_DBGSTRING_ON
	/*
	pr_debug("bus %s: unregistering\n",bus->name);
	*/
	dprintf("[USB] bus [%s] : unregistering\n",bus->name);
 #else
	dprintf("[USB] bus_unregister() : unregistering \n");
 #endif

 #ifndef _ETCPIP
	kset_unregister(&bus->drivers);
	kset_unregister(&bus->devices);
	subsystem_unregister(&bus->subsys);
 #endif // ndef _ETCPIP
}

/*
int __init buses_init(void)
*/
int buses_init(void)
{
 USB_DBG_FLOW(iprintf(
	"[USB FLOW] buses_init() Begin .. \n"
	));
 #ifndef _ETCPIP
	return subsystem_register(&bus_subsys);
 #endif // ndef _ETCPIP
}

/*
EXPORT_SYMBOL(bus_for_each_dev);
EXPORT_SYMBOL(bus_for_each_drv);

EXPORT_SYMBOL(device_bind_driver);
EXPORT_SYMBOL(device_release_driver);

EXPORT_SYMBOL(bus_add_device);
EXPORT_SYMBOL(bus_remove_device);
EXPORT_SYMBOL(bus_register);
EXPORT_SYMBOL(bus_unregister);
EXPORT_SYMBOL(bus_rescan_devices);
EXPORT_SYMBOL(get_bus);
EXPORT_SYMBOL(put_bus);
EXPORT_SYMBOL(find_bus);

EXPORT_SYMBOL(bus_create_file);
EXPORT_SYMBOL(bus_remove_file);
*/

// B. S. Liaw TEST for compiler worning !!

