
/*
 *  linux/drivers/base/map.c
 *
 * (C) Copyright Al Viro 2002,2003
 *	Released under GPL v2.
 *
 * NOTE: data structure needs to be changed.  It works, but for large dev_t
 * it will be too slow.  It is isolated, though, so these changes will be
 * local to that file.
 */

// #ifdef _ETCPIP
#define __KERNEL__
        #ifdef _NEWLIB
          #include "string.h"
        #endif
   	#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP


/*
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/kdev_t.h>
*/
#include "kdev_t.h"
/*
#include <linux/kobject.h>
*/
#include "kobject.h"
/*
#include <linux/kobj_map.h>
*/
#include "kobj_map.h"

struct kobj_map {
	struct probe {
		struct probe *next;
		dev_t dev;
		unsigned long range;
		struct module *owner;
		kobj_probe_t *get;
		int (*lock)(dev_t, void *);
		void *data;
	} *probes[255];
	struct rw_semaphore *sem;
};

int kobj_map(struct kobj_map *domain, dev_t dev, unsigned long range,
	     struct module *module, kobj_probe_t *probe,
	     int (*lock)(dev_t, void *), void *data)
{
 unsigned n = MAJOR(dev + range - 1) - MAJOR(dev) + 1;
 unsigned index = MAJOR(dev);
 unsigned i;
 struct probe *p;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] kobj_map() Begin .. \n"));
 if (n > 255)
		n = 255;

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] kobj_map() : p = UsbAllocMemC() .. \n"
	));

 /*
 p = kmalloc(sizeof(struct probe) * n, GFP_KERNEL);
 */
 p = (struct probe *) UsbAllocMemC(USB_MEM_C_S
				#ifdef USB_DBG_MEM_C_ON
					, sizeof(struct probe) * n
		        #endif
           		) ;

 if(p == NULL) {
	return -ENOMEM;
   }

 for (i = 0; i < n; i++, p++) {
		p->owner = module;
		p->get = probe;
		p->lock = lock;
		p->dev = dev;
		p->range = range;
		p->data = data;
	}
 /*
 #ifndef NO_SEMPHORE
 down_write(domain->sem);
 #endif // ndef NO_SEMPHORE
 */
 for (i = 0, p -= n; i < n; i++, p++, index++) {
	struct probe **s = &domain->probes[index % 255];

	while (*s && (*s)->range < range)
			s = &(*s)->next;
		p->next = *s;
		*s = p;
   }
 /*
 #ifndef NO_SEMPHORE
 up_write(domain->sem);
 #endif // ndef NO_SEMPHORE
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] kobj_map() End. \n"));
 return 0;
}

void kobj_unmap(struct kobj_map *domain, dev_t dev, unsigned long range)
{
	unsigned n = MAJOR(dev + range - 1) - MAJOR(dev) + 1;
	unsigned index = MAJOR(dev);
	unsigned i;
	struct probe *found = NULL;

	if (n > 255)
		n = 255;

 /*
 #ifndef NO_SEMPHORE
	down_write(domain->sem);
 #endif // ndef NO_SEMPHORE
 */
	for (i = 0; i < n; i++, index++) {
		struct probe **s;
		for (s = &domain->probes[index % 255]; *s; s = &(*s)->next) {
			struct probe *p = *s;
			if (p->dev == dev && p->range == range) {
				*s = p->next;
				if (!found)
					found = p;
				break;
			}
		}
	}

 /*
 #ifndef NO_SEMPHORE
 up_write(domain->sem);
 #endif // ndef NO_SEMPHORE
 */

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] kobj_unmap() : UsbFreeMemC(found) .. \n"
		));
 /*
 kfree(found);
 */
 UsbFreeMemC(USB_MEM_C_S,(unsigned char *) found
				#if defined(USB_DBG_MEM_C_ON)
					,sizeof(struct probe)
				#endif
				);
}

struct kobject *kobj_lookup(struct kobj_map *domain, dev_t dev, int *index)
{
 struct kobject *kobj;
 struct probe *p;
 unsigned best = ~0U;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] kobj_lookup() Begin .. \n"));

 // ---------
 retry:
 // ---------

 /*
 #ifndef NO_SEMPHORE
	down_read(domain->sem);
 #endif // ndef NO_SEMPHORE
 */
 for (p = domain->probes[MAJOR(dev) % 255]; p; p = p->next) {
		struct kobject *(*probe)(dev_t, int *, void *);
		struct module *owner;
		void *data;

		if (p->dev > dev || p->dev + p->range - 1 < dev)
			continue;
		if (p->range - 1 >= best)
			break;
		if (!try_module_get(p->owner))
			continue;
		owner = p->owner;
		data = p->data;
		probe = p->get;
		best = p->range - 1;
		*index = dev - p->dev;
		if (p->lock && p->lock(dev, data) < 0) {
			module_put(owner);
			continue;
		}
 /*
 #ifndef NO_SEMPHORE
		up_read(domain->sem);
 #endif // ndef NO_SEMPHORE
 */
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] kobj_lookup() : probe() .. \n"));
		kobj = probe(dev, index, data);
		/* Currently ->owner protects _only_ ->probe() itself. */
		module_put(owner);
		if (kobj)
			return kobj;
		goto retry;
	}
 /*
 #ifndef NO_SEMPHORE
 up_read(domain->sem);
 #endif // ndef NO_SEMPHORE
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] kobj_lookup() End. \n"));
 return NULL;
}

struct kobj_map *kobj_map_init(kobj_probe_t *base_probe,
		struct subsystem *s)
{
 /*
 struct kobj_map *p = kmalloc(sizeof(struct kobj_map), GFP_KERNEL);
 struct probe *base = kmalloc(sizeof(struct probe), GFP_KERNEL);
 */
 struct kobj_map *p ;
 struct probe *base ;
 int i;


 USBMS_DBG_FLOW(iprintf(
	"[USBMS FLOW] kobj_map_init() Begin .. \n"
	));

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] kobj_map_init() : p = UsbAllocMemC() .. \n"
	));

 p = (struct kobj_map *) UsbAllocMemC(USB_MEM_C_L // size : 1024
				#ifdef USB_DBG_MEM_C_ON
					, sizeof(struct kobj_map)
		        #endif
           		) ;

 if(!p) {
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] kobj_map_init() : Out Of Memory for p !! \n"
		));
	dprintf("[USBMS] Error : Out Of Memory !! \n") ;
   }

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] kobj_map_init() : base = UsbAllocMemC() .. \n"
	));

 base = (struct probe *) UsbAllocMemC(USB_MEM_C_S
				#ifdef USB_DBG_MEM_C_ON
					, sizeof(struct probe)
		        #endif
           		) ;

 if(!base) {
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] kobj_map_init() : Out Of Memory for base !! \n"
		));
	dprintf("[USBMS] Error : Out Of Memory !! \n") ;
   }


 memset(base, 0, sizeof(struct probe));
 base->dev = 1;
 base->range = ~0;
 base->get = base_probe;
 for (i = 0; i < 255; i++)
   {
	p->probes[i] = base;
   }
 /*
 #ifndef NO_SEMPHORE
	p->sem = &s->rwsem;
 #endif // ndef NO_SEMPHORE
 */

 USBMS_DBG_FLOW(iprintf(
	"[USBMS FLOW] kobj_map_init() End. \n"
	));

 return p;
}
