 
#ifdef _ETCPIP
   	#define __KERNEL__
        #ifdef _NEWLIB
          #include "sys_type.h"
        #endif

   	        #ifdef _NEWLIB
          #include "string.h"
        #endif
   	#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
          // for deciding whether to compile
          //  Debug Version or Release Version
   	#include "etcpip.h"

   	#include "UsbDbg.h"
   
   	#include "PoList.h" // osport_list.h
   
   	#include "UsbCfg.h" // Config and Setting
   	#include "UsOsport.h" // "usb_osport.h"
	#ifndef NO_TIMER
		#include "PoTimer.h" // osport_timer.h
	#endif // ndef NO_TIMER
   
   	#ifndef NO_WAITQUEUE
		#include  "PoWait.h" // osport_wait.h
   	#endif // ndef NO_WAITQUEUE
#endif // _ETCPIP


#ifndef _ETCPIP
   #include <linux/usb.h>
 #else
   #include "usb.h"
#endif

#ifndef _ETCPIP
	#include <linux/module.h>
	#include <linux/init.h>
	#include <linux/slab.h>
	#include <asm/byteorder.h>
#endif // ndef _ETCPIP

#ifdef USB_MEM_LIB
	#include "UsMemLib.h" // UsbMemLib.h
#endif // USB_MEM_LIB



#define USB_MAXALTSETTING		128	/* Hard limit */
#define USB_MAXENDPOINTS		30	/* Hard limit */

/* these maximums are arbitrary */
#define USB_MAXCONFIG			8
#define USB_ALTSETTINGALLOC		4
#define USB_MAXINTERFACES		32

static int usb_parse_endpoint(struct usb_host_endpoint *endpoint, unsigned char *buffer, int size)
{
	struct usb_descriptor_header *header;
	unsigned char *begin;
	int parsed = 0, len, numskipped;

	header = (struct usb_descriptor_header *)buffer;

	/* Everything should be fine being passed into here, but we sanity */
	/*  check JIC */
	if (header->bLength > size) {
		err("ran out of descriptors parsing");
		return -1;
	}
		
	if (header->bDescriptorType != USB_DT_ENDPOINT) {
		warn("unexpected descriptor 0x%X, expecting endpoint, 0x%X",
			header->bDescriptorType, USB_DT_ENDPOINT);
		return parsed;
	}

	if (header->bLength == USB_DT_ENDPOINT_AUDIO_SIZE)
		memcpy(&endpoint->desc, buffer, USB_DT_ENDPOINT_AUDIO_SIZE);
	else
		memcpy(&endpoint->desc, buffer, USB_DT_ENDPOINT_SIZE);
	
	le16_to_cpus(&endpoint->desc.wMaxPacketSize);

	buffer += header->bLength;
	size -= header->bLength;
	parsed += header->bLength;

	/* Skip over the rest of the Class Specific or Vendor Specific */
	/*  descriptors */
	begin = buffer;
	numskipped = 0;
	while (size >= sizeof(struct usb_descriptor_header)) {
		header = (struct usb_descriptor_header *)buffer;

		if (header->bLength < 2) {
			err("invalid descriptor length of %d", header->bLength);
			return -1;
		}

		/* If we find another "proper" descriptor then we're done  */
		if ((header->bDescriptorType == USB_DT_ENDPOINT) ||
		    (header->bDescriptorType == USB_DT_INTERFACE) ||
		    (header->bDescriptorType == USB_DT_CONFIG) ||
		    (header->bDescriptorType == USB_DT_DEVICE))
			break;

		dbg("skipping descriptor 0x%X",
			header->bDescriptorType);
		numskipped++;

		buffer += header->bLength;
		size -= header->bLength;
		parsed += header->bLength;
	}
	if (numskipped)
		dbg("skipped %d class/vendor specific endpoint descriptors", numskipped);

	/* Copy any unknown descriptors into a storage area for drivers */
	/*  to later parse */
	len = (int)(buffer - begin);
	if (!len) {
		endpoint->extra = NULL;
		endpoint->extralen = 0;
		return parsed;
	}

 #ifndef _ETCPIP
	endpoint->extra = kmalloc(len, GFP_KERNEL);

	if (!endpoint->extra) {
		err("[USB] Error : Out of Memory for endpoint extra descriptors");
		endpoint->extralen = 0;
		return parsed;
	}
  #else
	// endpoint->extra =  NonCache_contigmalloc(len) ;
    
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

    // BLOCK_PREEMPTION;

	endpoint->extra =  (unsigned char *)
		#ifdef USB_DBG_MEM_C_ON
			UsbAllocMemC(USB_MEM_C_L,len) ;
         #else
	    	UsbAllocMemC(USB_MEM_C_L) ;
        #endif
                 
    // UNBLOCK_PREEMPTION;

    USB_DBG_INIT(iprintf("[USB INIT] endpoint->extra = %08XH \n",
       endpoint->extra ));

    if ( endpoint->extra == NULL )
       {
		dprintf("[USB] Error : Out of Memory for endpoint extra descriptors \n");
		endpoint->extralen = 0;
		return parsed;
	   }
 #endif

	memcpy(endpoint->extra, begin, len);
	endpoint->extralen = len;

	return parsed;
}

static int usb_parse_interface(struct usb_interface *interface, 
		unsigned char *buffer,
		int size)
{
 int i, len, numskipped, retval, parsed = 0;
 struct usb_descriptor_header *header;
 struct usb_host_interface *ifp;
 unsigned char *begin;

 interface->act_altsetting = 0;
 interface->num_altsetting = 0;
 interface->max_altsetting = USB_ALTSETTINGALLOC;
 device_initialize(&interface->dev);

 #ifndef _ETCPIP
	interface->altsetting = kmalloc(sizeof(*interface->altsetting) * interface->max_altsetting,
					GFP_KERNEL);
	
	if (!interface->altsetting) {
		err("couldn't kmalloc interface->altsetting");
		return -1;
	}
  #else
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

    // BLOCK_PREEMPTION;

	interface->altsetting = (struct usb_host_interface *)
		#ifdef USB_DBG_MEM_C_ON
			UsbAllocMemC(USB_MEM_C_S,sizeof(struct usb_interface_descriptor)
                			* interface->max_altsetting) ;
         #else
			UsbAllocMemC(USB_MEM_C_S) ;
        #endif
                 
    // UNBLOCK_PREEMPTION;

    USB_DBG_MEM(iprintf("[USB MEM] interface->altsetting = %08XH \n",
       interface->altsetting ));

                
    if ( interface->altsetting == NULL )
       {
		dprintf("couldn't allocate interface->altsetting \n");
		return -1;
	   }
 #endif


	while (size > 0) {
		struct usb_interface_descriptor	*d;
	
		if (interface->num_altsetting >= interface->max_altsetting) {
			struct usb_host_interface *ptr;
			int oldmas;

			oldmas = interface->max_altsetting;
			interface->max_altsetting += USB_ALTSETTINGALLOC;
			if (interface->max_altsetting > USB_MAXALTSETTING) {
				warn("too many alternate settings (incr %d max %d)\n",
					USB_ALTSETTINGALLOC, USB_MAXALTSETTING);
				return -1;
			}

 #ifndef _ETCPIP
			ptr = kmalloc(sizeof(*ptr) * interface->max_altsetting, GFP_KERNEL);
			if (ptr == NULL) {
				err("couldn't kmalloc interface->altsetting");
				return -1;
			}
  #else
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

    // BLOCK_PREEMPTION;

	ptr =  (struct usb_host_interface *)
		#ifdef USB_DBG_MEM_C_ON
				UsbAllocMemC(USB_MEM_C_M,
                	  sizeof(struct usb_interface_descriptor)
                    * interface->max_altsetting) ;
         #else
    		UsbAllocMemC(USB_MEM_C_M) ;
        #endif
                 
    //  UNBLOCK_PREEMPTION;

    USB_DBG_MEM(iprintf("[USB MEM] ptr = %08XH \n",
        ptr ));

    if ( ptr == NULL )
       {
		dprintf("couldn't allocate ptr \n");
		return -1;
	   }
 #endif
 
 memcpy(ptr, interface->altsetting,
 		sizeof(*interface->altsetting) * oldmas);
            
 USB_DBG_MEM(iprintf(
	"[USB MEM] usb_parse_interface() : free interface->altsetting .. \n"
    ));
    
 #ifndef USB_MEM_LIB
 	kfree(interface->altsetting);
 #elif defined USB_DBG_MEM_C_ON
    UsbFreeMemC(USB_MEM_C_S,(unsigned char *)interface->altsetting,
    		sizeof(struct usb_interface_descriptor)
          * interface->max_altsetting) ;
 #else
    UsbFreeMemC(USB_MEM_C_S,(unsigned char *)
    							interface->altsetting) ;
 #endif
			interface->altsetting = ptr;
		}

		ifp = interface->altsetting + interface->num_altsetting;
		ifp->endpoint = NULL;
		ifp->extra = NULL;
		ifp->extralen = 0;
		interface->num_altsetting++;

		memcpy(ifp, buffer, USB_DT_INTERFACE_SIZE);

		/* Skip over the interface */
		buffer += ifp->desc.bLength;
		parsed += ifp->desc.bLength;
		size -= ifp->desc.bLength;

		begin = buffer;
		numskipped = 0;

		/* Skip over any interface, class or vendor descriptors */
		while (size >= sizeof(struct usb_descriptor_header)) {
			header = (struct usb_descriptor_header *)buffer;

			if (header->bLength < 2) {
				err("invalid descriptor length of %d", header->bLength);
				return -1;
			}

			/* If we find another "proper" descriptor then we're done  */
			if ((header->bDescriptorType == USB_DT_INTERFACE) ||
			    (header->bDescriptorType == USB_DT_ENDPOINT) ||
			    (header->bDescriptorType == USB_DT_CONFIG) ||
			    (header->bDescriptorType == USB_DT_DEVICE))
				break;

			numskipped++;

			buffer += header->bLength;
			parsed += header->bLength;
			size -= header->bLength;
		}

		if (numskipped)
           {
			#ifdef USB_DBG_INIT_ON
				dbg("[USB INIT] skipped %d class/vendor specific "
            		"interface descriptors", numskipped);
            #endif
			USB_DBG_INFO(iprintf(
				"[USB INFO] skipped %d class/vendor specific "
                "interface descriptors\n",
                numskipped
                ));
           }

		/* Copy any unknown descriptors into a storage area for */
		/*  drivers to later parse */
		len = (int)(buffer - begin);
		if (len) {
 #ifndef _ETCPIP
			ifp->extra = kmalloc(len, GFP_KERNEL);

			if (!ifp->extra) {
				err("[USB] Error : Out of Memory for interface extra descriptors");
				ifp->extralen = 0;
				return -1;
			}
  #else
	// ifp->extra =  NonCache_contigmalloc(len) ;
        
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

    // BLOCK_PREEMPTION;

	ifp->extra =  (unsigned char *)
		#ifdef USB_DBG_MEM_C_ON
				UsbAllocMemC(USB_MEM_C_L,len) ;
         #else
     UsbAllocMemC(USB_MEM_C_L) ;
     #endif
                 
    // UNBLOCK_PREEMPTION;

    USB_DBG_MEM(iprintf("[USB MEM] ifp->extra = %08XH \n",ifp->extra ));

    if ( ifp->extra == NULL )
       {
		dprintf("[USB] Error : Out of Memory for interface extra descriptors \n");
				ifp->extralen = 0;
		return -1;
	   }
 #endif
			memcpy(ifp->extra, begin, len);
			ifp->extralen = len;
		}

		/* Did we hit an unexpected descriptor? */
		header = (struct usb_descriptor_header *)buffer;
		if ((size >= sizeof(struct usb_descriptor_header)) &&
		    ((header->bDescriptorType == USB_DT_CONFIG) ||
		     (header->bDescriptorType == USB_DT_DEVICE)))
			return parsed;

		if (ifp->desc.bNumEndpoints > USB_MAXENDPOINTS) {
			warn("too many endpoints");
			return -1;
		}

 #ifndef _ETCPIP
		ifp->endpoint = (struct usb_host_endpoint *)
			kmalloc(ifp->desc.bNumEndpoints *
			sizeof(struct usb_host_endpoint), GFP_KERNEL);
		if (!ifp->endpoint) {
			err("out of memory");
			return -1;	
		}
  #else
	//ifp->endpoint = (struct usb_endpoint_descriptor *)
    //     NonCache_contigmalloc(ifp->bNumEndpoints *
	//		sizeof(struct usb_endpoint_descriptor)) ;
        
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

    // BLOCK_PREEMPTION;

	ifp->endpoint = (struct usb_host_endpoint *)
		#ifdef USB_DBG_MEM_C_ON
				UsbAllocMemC(USB_MEM_C_L,
                ifp->desc.bNumEndpoints *
							sizeof(struct usb_host_endpoint)
				) ;
         #else
    
	        		UsbAllocMemC(USB_MEM_C_L) ;
         #endif
                 
    // UNBLOCK_PREEMPTION;

    USB_DBG_MEM(iprintf("[USB MEM] ifp->endpoint = %08XH \n",ifp->endpoint ));

    if ( ifp->endpoint == NULL )
       {
		dprintf("out of memory \n");
		return -1;
	   }
 #endif

 #ifndef _ETCPIP
		memset(ifp->endpoint, 0, ifp->desc.bNumEndpoints *
			sizeof(struct usb_host_endpoint));
 #endif	
	
		for (i = 0; i < ifp->desc.bNumEndpoints; i++) {
			header = (struct usb_descriptor_header *)buffer;

			if (header->bLength > size) {
				err("ran out of descriptors parsing");
				return -1;
			}
		
			retval = usb_parse_endpoint(ifp->endpoint + i, buffer, size);
			if (retval < 0)
				return retval;

			buffer += retval;
			parsed += retval;
			size -= retval;
		}

		/* We check to see if it's an alternate to this one */
		d = (struct usb_interface_descriptor *)buffer;
		if (size < USB_DT_INTERFACE_SIZE
				|| d->bDescriptorType != USB_DT_INTERFACE
				|| !d->bAlternateSetting)
			return parsed;
	}

	return parsed;
}

int usb_parse_configuration(struct usb_host_config *config, char *buffer)
{
	int i, retval, size;
	struct usb_descriptor_header *header;

	memcpy(&config->desc, buffer, USB_DT_CONFIG_SIZE);
	le16_to_cpus(&config->desc.wTotalLength);
	size = config->desc.wTotalLength;

	if (config->desc.bNumInterfaces > USB_MAXINTERFACES) {
		warn("too many interfaces");
		return -1;
	}

 #ifndef _ETCPIP
	config->interface = (struct usb_interface *)
		kmalloc(config->desc.bNumInterfaces *
		sizeof(struct usb_interface), GFP_KERNEL);
	dbg("kmalloc IF %p, numif %i", config->interface, config->desc.bNumInterfaces);
	if (!config->interface) {
		err("out of memory");
		return -1;	
	}
  #else
	//config->interface = (struct usb_interface *)
    //        NonCache_contigmalloc(config->desc.bNumInterfaces *
	//	sizeof(struct usb_interface)) ;
        
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

    // BLOCK_PREEMPTION;

	config->interface = (struct usb_interface *)
		#ifdef USB_DBG_MEM_C_ON
			UsbAllocMemC(USB_MEM_C_M,config->desc.bNumInterfaces *
					sizeof(struct usb_interface)) ;
         #else
    		UsbAllocMemC(USB_MEM_C_M) ;
        #endif
    // UNBLOCK_PREEMPTION;

    USB_DBG_INIT(iprintf("[USB INIT] config->interface = %08XH \n",
        config->interface ));

    if ( config->interface == NULL )
       {
		dprintf("out of memory \n");
		return -1;
	   }
 #endif

	memset(config->interface, 0,
	       config->desc.bNumInterfaces * sizeof(struct usb_interface));

	buffer += config->desc.bLength;
	size -= config->desc.bLength;
	
	config->extra = NULL;
	config->extralen = 0;

	for (i = 0; i < config->desc.bNumInterfaces; i++) {
		int numskipped, len;
		char *begin;

		/* Skip over the rest of the Class Specific or Vendor */
		/*  Specific descriptors */
		begin = buffer;
		numskipped = 0;
		while (size >= sizeof(struct usb_descriptor_header)) {
			header = (struct usb_descriptor_header *)buffer;

			if ((header->bLength > size) || (header->bLength < 2)) {
				err("invalid descriptor length of %d", header->bLength);
				return -1;
			}

			/* If we find another "proper" descriptor then we're done  */
			if ((header->bDescriptorType == USB_DT_ENDPOINT) ||
			    (header->bDescriptorType == USB_DT_INTERFACE) ||
			    (header->bDescriptorType == USB_DT_CONFIG) ||
			    (header->bDescriptorType == USB_DT_DEVICE))
				break;

			dbg("skipping descriptor 0x%X", header->bDescriptorType);
			numskipped++;

			buffer += header->bLength;
			size -= header->bLength;
		}
		if (numskipped)
			dbg("skipped %d class/vendor specific endpoint descriptors", numskipped);

		/* Copy any unknown descriptors into a storage area for */
		/*  drivers to later parse */
		len = (int)(buffer - begin);
		if (len) {
			if (config->extralen) {
				warn("extra config descriptor");
			} else {
 #ifndef _ETCPIP
				config->extra = kmalloc(len, GFP_KERNEL);
				if (!config->extra) {
					err("[USB] Error : Out of Memory for config extra descriptors");
					config->extralen = 0;
					return -1;
				}
  #else
	// config->extra = NonCache_contigmalloc(len) ;
    
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

    // BLOCK_PREEMPTION;

	config->extra = (unsigned char *)
		#ifdef USB_DBG_MEM_C_ON
			UsbAllocMemC(USB_MEM_C_L,len) ;
         #else
     		UsbAllocMemC(USB_MEM_C_L) ;
        #endif
                 
    // UNBLOCK_PREEMPTION;

    USB_DBG_INIT(iprintf("[USB INIT] config->extra = %08XH \n",config->extra ));

    if ( config->extra == NULL )
       {
		dprintf("[USB] Error : Out of Memory for config extra descriptors \n");
					config->extralen = 0;
					return -1;
	   }
 #endif

				memcpy(config->extra, begin, len);
				config->extralen = len;
			}
		}

		retval = usb_parse_interface(config->interface + i, buffer, size);
		if (retval < 0)
			return retval;

		buffer += retval;
		size -= retval;
	}

	return size;
}

// hub-only!! ... and only exported for reset/reinit path.
// otherwise used internally on disconnect/destroy path
void usb_destroy_configuration(struct usb_device *dev)
{
 int c, i, j, k;
	
 USB_DBG_FLOW(iprintf(
	"[USB FLOW] usb_destroy_configuration() Begin ..\n"
    ));
    
 // #ifdef USB_DBG_INIT_ON
 #ifdef USB_DBG_MEM_COUNT_ON
	iprintf(
   		"[USB MEM_COUNT] usb_destroy_configuration(1) : show USB Mem Count .. \n");
	showUsbMemC() ;
	showUsbMemU() ;
 #endif

 if (!dev->config)
   {
	return;
   }

 USB_DBG_INIT(iprintf(
	"[USB INIT] usb_destroy_configuration() : "
    "rawdescriptors ..\n"
    ));
 if (dev->rawdescriptors)
   {
	for (i = 0; i < dev->descriptor.bNumConfigurations; i++)
       {
        #ifndef _ETCPIP
        	 // bigbuffer
			kfree(dev->rawdescriptors[i]);
         #else
			#ifndef USB_MEM_LIB
	       		// NonCache_contigmalloc() cannot be free now !!
			#elif defined(USB_DBG_MEM_U_ON)
				#ifdef USB_BIGBUF_M_POOL
					USB_DBG_MEM_U(iprintf(
			    		"[USB MEM_U] usb_destroy_configuration() : "
                        "bigbuffer use Medium Pool !!\n"
	            		));
            	#endif
                
				#ifndef USB_BIGBUF_M_POOL
		        	UsbFreeMemU(USB_MEM_U_L, dev->rawdescriptors[i],77) ; // bigbuffer
                #else
		        	UsbFreeMemU(USB_MEM_U_M, dev->rawdescriptors[i],77) ; // bigbuffer
                #endif
		     #else
				#ifdef USB_BIGBUF_M_POOL
					USB_DBG_MEM_U(iprintf(
			    		"[USB MEM_U] usb_destroy_configuration() : "
                        "bigbuffer use Medium Pool !!\n"
	            		));
            	#endif
                
				#ifndef USB_BIGBUF_M_POOL
		        	UsbFreeMemU(USB_MEM_U_L, dev->rawdescriptors[i]) ; // bigbuffer
                #else
		        	UsbFreeMemU(USB_MEM_U_M, dev->rawdescriptors[i]) ; // bigbuffer
                #endif
			#endif
		#endif
       } // for i
    #ifndef _ETCPIP
			kfree(dev->rawdescriptors);
     #else
		#ifdef USB_DBG_MEM_C_ON
			UsbFreeMemC(USB_MEM_C_S,(unsigned char *)
            		dev->rawdescriptors,
            		sizeof(*dev->rawdescriptors));
		 #else
			UsbFreeMemC(USB_MEM_C_S,(unsigned char *)
            					dev->rawdescriptors);
        #endif
	#endif
   } // if

 USB_DBG_INIT(iprintf(
	"[USB INIT] usb_destroy_configuration() : bNumConfigurations ..\n"
    ));
    
 for (c = 0; c < dev->descriptor.bNumConfigurations; c++)
   {
	struct usb_host_config *cf = &dev->config[c];

	if (!cf->interface)
			break;

	for (i = 0; i < cf->desc.bNumInterfaces; i++) {
			struct usb_interface *ifp =
				&cf->interface[i];
				
			if (!ifp->altsetting)
				break;

			for (j = 0; j < ifp->num_altsetting; j++) {
				struct usb_host_interface *as =
					&ifp->altsetting[j];
					
				if(as->extra) {
					kfree(as->extra);
				}

				if (!as->endpoint)
					break;
					
				for(k = 0; k < as->desc.bNumEndpoints; k++) {
					if(as->endpoint[k].extra) {
						kfree(as->endpoint[k].extra);
					}
				}	
				kfree(as->endpoint);
			}

 USB_DBG_MEM(iprintf(
	"[USB MEM] usb_destroy_configuration() : free ifp->altsetting .. \n"
    ));
 #ifndef USB_MEM_LIB
 	kfree(ifp->altsetting);
 #elif defined USB_DBG_MEM_C_ON
    UsbFreeMemC(USB_MEM_C_S,(unsigned char *)ifp->altsetting,sizeof(struct usb_interface_descriptor)
                			* ifp->max_altsetting) ;
 #else
    UsbFreeMemC(USB_MEM_C_S,(unsigned char *)ifp->altsetting) ;
 #endif
		}
 USB_DBG_MEM(iprintf(
	"[USB MEM] usb_destroy_configuration() : free cf->interface .. \n"
    ));
 #ifndef USB_MEM_LIB
 	kfree(cf->interface);
 #elif defined USB_DBG_MEM_C_ON
    UsbFreeMemC(USB_MEM_C_M,(unsigned char *)cf->interface,sizeof(struct usb_interface)) ;
 #else
    UsbFreeMemC(USB_MEM_C_M,(unsigned char *)cf->interface) ;
 #endif
	}
    
 USB_DBG_INIT(iprintf(
	"[USB INIT] usb_destroy_configuration() : dev->config ..\n"
    ));
    
 USB_DBG_MEM(iprintf(
	"[USB MEM] usb_destroy_configuration() : &(dev->config) = %08XH "
    "dev->config = %08XH \n",
	&(dev->config),dev->config
    )) ;
 /*    
 dprintf(
	"[USBLPT MEM] usb_destroy_configuration() : &(dev->config) = %08XH "
    "dev->config = %08XH \n",
	&(dev->config),dev->config
    ) ;
 */   
 if (dev->config)
   {
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_destroy_configuration() : free dev->config .. \n"
    	));
    /*    
	// FixMe : Why Already Free By Others ?? 
    USB_DBG_INFO(iprintf(
	        "[USB INFO] FixMe : dev->config Already Free ! \n"
	        )) ;
	*/
	#ifndef USB_MEM_LIB
 		kfree (dev->config);
	#else
    	UsbFreeMemC(USB_MEM_C_S,      // dev->config : size : 9
    	// UsbFreeMemC(USB_MEM_C_M,   // *dev : size : 432
        	(unsigned char *)dev->config
    		#if defined USB_DBG_MEM_C_ON
	        	,sizeof(dev->config)
		 	#endif
            ) ;
 	#endif
 
    }
   else
    {
	// FixMe : Why Already Free By Others ?? 
    USB_DBG_INFO(iprintf(
	        "[USB INFO] FixMe : dev->config Already Free ! \n"
	        )) ;
    }
    
 // #ifdef USB_DBG_INIT_ON
 #ifdef USB_DBG_MEM_COUNT_ON
	iprintf(
   		"[USB MEM_COUNT] usb_destroy_configuration(2) : show USB Mem Count .. \n");
	showUsbMemC() ;
	showUsbMemU() ;
	#ifdef USB_DBG_MEM_C_ON
		USB_DBG_MEM_C(iprintf("[USB MEM_C] usb_destroy_configuration(3) : Dump mem/pool .. \n"));
	    DumpUsbMemC(1) ;
	    // DumpUsbMemC(2) ;
	#endif // USB_DBG_MEM_C_ON
	#ifdef USB_DBG_MEM_U_ON
		USB_DBG_MEM_U(iprintf("[USB MEM_U] usb_destroy_configuration(4) : Dump mem/pool .. \n"));
		DumpUsbMemU(4) ;
		// DumpUsbMemU(5) ;
	#endif // USB_DBG_MEM_U_ON
 #endif
 
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_destroy_configuration() End.\n"));
}


// hub-only!! ... and only in reset path, or usb_new_device()
// (used by real hubs and virtual root hubs)

int usb_get_configuration(struct usb_device *dev)
{
 int result;
 unsigned int cfgno, length;
 unsigned char *buffer;
 unsigned char *bigbuffer;
 struct usb_config_descriptor *desc;

 USB_DBG_FLOW(iprintf("[USB FLOW] usb_get_configuration() Begin ..\n"));
 
 USB_DBG_INIT(iprintf("[USB INIT] usb_get_configuration() ..\n"));
 
 // #ifdef USB_DBG_INIT_ON
 #ifdef USB_DBG_MEM_COUNT_ON
	iprintf(
   		"[USB MEM_COUNT] usb_get_configuration(1) : show USB Mem Count .. \n");
	showUsbMemC() ;
	showUsbMemU() ;
 #endif
 
 if (dev->descriptor.bNumConfigurations > USB_MAXCONFIG) 
   {
    /*
	warn("too many configurations");
    */
	dprintf("[USB] too many configurations\n");

	return -EINVAL;
   }

 if (dev->descriptor.bNumConfigurations < 1) 
    {
     /*
	 warn("not enough configurations");
     */
	dprintf("[USB] not enough configurations\n");

	return -EINVAL;
   }

 #ifndef _ETCPIP
	dev->config = (struct usb_host_config *)
		kmalloc(dev->descriptor.bNumConfigurations *
		sizeof(struct usb_host_config), GFP_KERNEL);
	if (!dev->config) {
		err("out of memory");
		return -ENOMEM;	
	}
  #else
	//dev->config = (struct usb_config_descriptor *)
    //              NonCache_contigmalloc(dev->descriptor.bNumConfigurations *
	//	sizeof(struct usb_config_descriptor)) ;
        
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

    //  BLOCK_PREEMPTION;

	dev->config = (struct usb_host_config *) // 9
		#ifdef USB_DBG_MEM_C_ON
			UsbAllocMemC(USB_MEM_C_S,dev->descriptor.bNumConfigurations *
		          sizeof(struct usb_config_descriptor) ) ;
         #else
        	UsbAllocMemC(USB_MEM_C_S) ;
        #endif
                 
    // UNBLOCK_PREEMPTION;

    USB_DBG_MEM(iprintf(
    	"[USB MEM] dev->config = %08XH \n",dev->config
        ));
    /*
    dprintf(
    	"[USB MEM] dev->config = %08XH \n",dev->config
        );
	*/
    if ( dev->config == NULL )
       {
		dprintf("[USB] usb_get_configuration() : out of memory \n");
		return -ENOMEM;
	   }
 #endif
 
 #ifndef _ETCPIP
	memset(dev->config, 0, dev->descriptor.bNumConfigurations *
		sizeof(struct usb_host_config));
 #endif
 
 #ifndef _ETCPIP
	dev->rawdescriptors = (char **)kmalloc(sizeof(char *) *
		dev->descriptor.bNumConfigurations, GFP_KERNEL);
	if (!dev->rawdescriptors) {
		err("out of memory");
		return -ENOMEM;
	}
  #else
	// dev->rawdescriptors = (char **)NonCache_contigmalloc(sizeof(char *) *
	//      dev->descriptor.bNumConfigurations) ;
    
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

    // BLOCK_PREEMPTION;

	dev->rawdescriptors = (char **)
		#ifdef USB_DBG_MEM_C_ON
			UsbAllocMemC(USB_MEM_C_S, sizeof(char *) *
	      		dev->descriptor.bNumConfigurations  ) ;
         #else
         	UsbAllocMemC(USB_MEM_C_S) ;
        #endif
                 
    // UNBLOCK_PREEMPTION;

    USB_DBG_MEM(iprintf("[USB MEM] dev->rawdescriptors = %08XH \n",
                 dev->rawdescriptors ));

    if ( dev->rawdescriptors == NULL )
       {
		dprintf("out of memory \n");
		return -ENOMEM;
	   }
 #endif

 #ifndef _ETCPIP
    /*
	//buffer = kmalloc(8, GFP_KERNEL);			// 20021007 remarked by drsohn
	buffer = kmalloc(8, GFP_KERNEL|GFP_DMA);			// 20021007 modified by drsohn
	if (!buffer) {
		err("unable to allocate memory for configuration descriptors");
		return -ENOMEM;
	}
    */
  #else
	#ifndef USB_MEM_LIB
		buffer = (unsigned char *)
		    NonCache_contigmalloc(8) ;
	    // NonCache_contigmalloc() cannot be free now !!
     #else
     	// alloc_mem(8) ;
        
		USB_DBG_MEM_U(iprintf(
		    "[USB MEM_U] usb_get_configuration() buffer = UsbAllocMemU() ..\n"
            ));
            
		buffer = (unsigned char *) UsbAllocMemU(USB_MEM_U_S
				#ifdef USB_DBG_MEM_U_ON
            		,8
                #endif
                    );
    #endif
    USB_DBG_MEM(iprintf("[USB MEM] buffer = %08XH \n", buffer ));
    if ( buffer == NULL )
       {
		dprintf("[USB] unable to allocate memory for configuration descriptors\n");
		return -ENOMEM;
	   }
 #endif
	desc = (struct usb_config_descriptor *)buffer;

	for (cfgno = 0; cfgno < dev->descriptor.bNumConfigurations; cfgno++) {
		/* We grab the first 8 bytes so we know how long the whole */
		/*  configuration is */
		result = usb_get_descriptor(dev, USB_DT_CONFIG, cfgno, buffer, 8);
 #ifndef TIMEOUT_LOOSER_CHECK
	if (result < 8)
  #else
    if ( (result < 8) && (result != -EINPROGRESS) )
 #endif
       {
			if (result < 0)
			dprintf("[USB] unable to get descriptor");
		else 
          {
			dprintf("[USB] config descriptor too short (expected %i, got %i)", 8, result);
				result = -EINVAL;
			}
			goto err;
		}

  	/* Get the full buffer */
	length = le16_to_cpu(desc->wTotalLength); // __u16 , Max 64K Bytes

 	#ifndef _ETCPIP
		//bigbuffer = kmalloc(length, GFP_KERNEL);		// 20021008 remarked by drsohn
		bigbuffer = kmalloc(length, GFP_KERNEL|GFP_DMA);		// 20021008 modified by drsohn
		if (!bigbuffer) {
			err("unable to allocate memory for configuration descriptors");
			result = -ENOMEM;
			goto err;
		}
  	 #else
		#ifndef USB_MEM_LIB
		    bigbuffer = (unsigned char *)
	        NonCache_contigmalloc(length) ;
			// NonCache_contigmalloc() cannot be free now !!
	     #else
	     	// alloc_mem(length) ;
			USB_DBG_MEM_C(iprintf(
			    "[USB MEM_U] usb_get_configuration() Call UsbAllocMemU ..\n"
	            ));
            
			#ifdef USB_BIGBUF_M_POOL
				USB_DBG_MEM_U(iprintf(
			    	"[USB MEM_U] bigbuffer use Medium Pool !!\n"
	            	));
            #endif
            
			bigbuffer = (unsigned char *)
			#ifndef USB_BIGBUF_M_POOL
				#ifdef USB_DBG_MEM_U_ON
	        		UsbAllocMemU(USB_MEM_U_L,length); // __u16 , Max 64K Bytes
	    	     #else
    		    	UsbAllocMemU(USB_MEM_U_L);
	        	#endif   
	         #else
				#ifdef USB_DBG_MEM_U_ON
		        	UsbAllocMemU(USB_MEM_U_M,length); // __u16 , Max 64K Bytes
	    	     #else
	        		UsbAllocMemU(USB_MEM_U_M);
	        #endif
        #endif
    #endif
       USB_DBG_MEM(iprintf("[USB MEM] bigbuffer = %08XH \n", bigbuffer ));
       if ( bigbuffer == NULL )
       	  {
		   dprintf("[USB] unable to allocate memory for configuration descriptors\n");
		   result -ENOMEM;
           goto err ;
	      }
	#endif

		/* Now that we know the length, get the whole thing */
		result = usb_get_descriptor(dev, USB_DT_CONFIG, cfgno, bigbuffer, length);
 #ifndef TIMEOUT_LOOSER_CHECK
	if (result < 0)
  #else
    if ( (result < 0) && (result != -EINPROGRESS) )
 #endif
       {
		dprintf("[USB] couldn't get all of config descriptors\n");
        #ifndef _ETCPIP
			kfree(bigbuffer);
         #else
			#ifndef USB_MEM_LIB
	       		// NonCache_contigmalloc() cannot be free now !!
			#elif defined(USB_DBG_MEM_U_ON)
				USB_DBG_MEM_U(iprintf(
			    	"[USB MEM_U] usb_get_configuration() Call UsbFreeMemU ..\n"
	            	));
                
				#ifndef USB_BIGBUF_M_POOL
		        	UsbFreeMemU(USB_MEM_U_L, bigbuffer,77) ;
                #else
		        	UsbFreeMemU(USB_MEM_U_M, bigbuffer,77) ;
                #endif
		    #else
				USB_DBG_MEM_U(iprintf(
			    	"[USB MEM_U] usb_get_configuration() Call UsbFreeMemU ..\n"
	            	));
                
				#ifndef USB_BIGBUF_M_POOL
		        	UsbFreeMemU(USB_MEM_U_L, bigbuffer) ;
                #else
		        	UsbFreeMemU(USB_MEM_U_M, bigbuffer) ;
                #endif
			#endif
		 #endif
		goto err;
	   }	
	
 #ifndef TIMEOUT_LOOSER_CHECK
	if (result < length)
  #else
    if ( (result < length) && (result != -EINPROGRESS) )
 #endif
       {
		dprintf(
            "[USB] config descriptor too short (expected %i, got %i)\n",
        	length, result);
			result = -EINVAL;
        #ifndef _ETCPIP
			kfree(bigbuffer);
         #else
			#ifndef USB_MEM_LIB
	       		// NonCache_contigmalloc() cannot be free now !!
			#elif defined(USB_DBG_MEM_U_ON)
				USB_DBG_MEM_U(iprintf(
			    	"[USB MEM_U] usb_get_configuration() Call UsbFreeMemU ..\n"
	            	));
				#ifndef USB_BIGBUF_M_POOL
		        	UsbFreeMemU(USB_MEM_U_L, bigbuffer,77) ;
                #else
		        	UsbFreeMemU(USB_MEM_U_M, bigbuffer,77) ;
                #endif
		    #else
				USB_DBG_MEM_U(iprintf(
			    	"[USB MEM_U] usb_get_configuration() Call UsbFreeMemU ..\n"
	            	));
				#ifndef USB_BIGBUF_M_POOL
		        	UsbFreeMemU(USB_MEM_U_L, bigbuffer) ;
                #else
		        	UsbFreeMemU(USB_MEM_U_M, bigbuffer) ;
                #endif
			#endif
		#endif
		goto err;
	   }

		dev->rawdescriptors[cfgno] = bigbuffer;

		result = usb_parse_configuration(&dev->config[cfgno], bigbuffer);
		if (result > 0)
			dprintf("[USB] descriptor data left");
			#ifndef TIMEOUT_LOOSER_CHECK
				else if (result < 0) 
			  #else
			    else if ( (result < 0) && (result != -EINPROGRESS) )
			 #endif
			          {
						result = -EINVAL;
						goto err;
				      }
   } // for
 /*
 kfree (buffer);
 */
 // NonCache_contigmalloc() cannot be free now !!
 // free(hub->descriptor);
 USB_DBG_MEM_U(iprintf("[USB MEM_U] usb_get_configuration() : UsbFreeMemU(buffer) ..\n"
            ));
 UsbFreeMemU(USB_MEM_U_S, (unsigned char *) buffer
			#if defined(USB_DBG_MEM_U_ON)
			,8
			#endif
			) ;
 
 // #ifdef USB_DBG_INIT_ON
 #ifdef USB_DBG_MEM_COUNT_ON
	iprintf(
   		"[USB MEM_COUNT] usb_get_configuration(2) : show USB Mem Count .. \n");
	showUsbMemC() ;
	showUsbMemU() ;
 #endif
 
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_get_configuration() End.\n"));

 return 0;
 
 err:
 
 #ifndef _ETCPIP
	kfree (buffer);
  #else
	#ifndef USB_MEM_LIB
    // NonCache_contigmalloc() cannot be free now !!
	#elif defined(USB_DBG_MEM_U_ON)
        UsbFreeMemU(USB_MEM_U_S, buffer,8) ;
	#else
		// free(hub->descriptor);
		USB_DBG_MEM_U(iprintf(
		    "[USB MEM_U] usb_get_configuration() Call UsbFreeMemU buffer ..\n"
            ));
        UsbFreeMemU(USB_MEM_U_S, buffer) ;
	#endif
 #endif
 
 dev->descriptor.bNumConfigurations = cfgno;
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_get_configuration() End with Fail.\n"));
 return result;
}

