/*
  NOTE:

  this code was lifted straight out of drivers/pci/pci.c;
  when compiling for the SAMSUNG UBS Controller 
  usb-ohci.c driver needs these routines even when the architecture
  has no pci bus...
  From Ni PalDuk Gulga
*/

// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
     #include "sys_type.h"
#endif
#ifdef _NEWLIB
          #include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
          // for deciding whether to compile
          //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
		#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
// #endif // _ETCPIP

#ifndef _ETCPIP
   #include <linux/config.h>
   #include <linux/module.h>
   #include <linux/types.h>
   #include <linux/kernel.h>
#endif // ndef _ETCPIP
    
#ifndef _ETCPIP
	#include <linux/pci.h>
#else
    #include "PoDevice.h" // osport_device.h
	#ifdef CONFIG_PCI
	    #include "UsbPoPci.h" // usb_osport_pci.h
    #endif	// CONFIG_PCI
#endif
    
#ifndef _ETCPIP
   #include <linux/string.h>
   #include <linux/init.h>
   #include <linux/slab.h>
   #include <linux/spinlock.h>
   #include <linux/bitops.h>
   
   #include <asm/page.h>
#endif // ndef _ETCPIP

#ifdef _ETCPIP
   #include "PoList.h" // osport_list.h
#endif // _ETCPIP

#ifndef USB_NO_DMA_POOL
	#include "UsDmaPol.h" // usb_pcipool.h
#endif // ndef USB_NO_DMA_POOL

#ifdef USB_MEM_LIB
	#include "UsMemLib.h" // UsbMemLib.h
#endif // USB_MEM_LIB

#ifndef USB_NO_DMA_POOL

/*
 * Pool allocator ... wraps the pci_alloc_consistent page allocator, so
 * small blocks are easily used by drivers for bus mastering controllers.
 * This should probably be sharing the guts of the slab allocator.
 */

#ifdef CONFIG_PCI
	struct pci_pool {	/* the pool */
 #else
	struct dma_pool {	/* the pool */
#endif
 struct list_head	page_list;
 #ifndef NO_SPINLOCK
	spinlock_t		lock;
 #endif // ndef NO_SPINLOCK
 size_t			blocks_per_page;
 size_t			size;
 int			flags;
 
 #ifdef CONFIG_PCI
	struct pci_dev		*dev;
 #endif
 
 size_t			allocation;
 #ifdef USB_DBGSTRING_ON
	char			name [32];
 #endif // USB_DBGSTRING_ON
 #ifndef NO_WAITQUEUE
	wait_queue_head_t	waitq;
 #endif // ndef NO_WAITQUEUE
};

struct pci_page {	/* cacheable header for 'allocation' bytes */
	struct list_head	page_list;
	void			*vaddr;
	dma_addr_t		dma;
	unsigned long		bitmap [0];
};

#define	POOL_TIMEOUT_JIFFIES	((100 /* msec */ * HZ) / 1000)
#define	POOL_POISON_BYTE	0xa7

#endif // ndef USB_NO_DMA_POOL

// #define CONFIG_PCIPOOL_DEBUG


#ifndef USB_NO_DMA_POOL
/**
 * usb_pci_pool_create - Creates a pool of pci consistent memory blocks, 
   for dma.
 * @name: name of pool, for diagnostics
 * @pdev: pci device that will be doing the DMA
 * @size: size of the blocks in this pool.
 * @align: alignment requirement for blocks; must be a power of two
 * @allocation: returned blocks won't cross this boundary (or zero)
 * @flags: SLAB_* flags (not all are supported).
 *
 * Returns a pci allocation pool with the requested characteristics, or
 * null if one can't be created.  Given one of these pools, usb_pci_pool_alloc()
 * may be used to allocate memory.  Such memory will all have "consistent"
 * DMA mappings, accessible by the device and its driver without using
 * cache flushing primitives.  The actual size of blocks allocated may be
 * larger than requested because of alignment.
 *
 * If allocation is nonzero, objects returned from usb_pci_pool_alloc() won't
 * cross that size boundary.  This is useful for devices which have
 * addressing restrictions on individual DMA transfers, such as not crossing
 * boundaries of 4KBytes.
 */
#ifdef CONFIG_PCI
	#ifdef USB_DBGSTRING_ON
		struct pci_pool *
		usb_pci_pool_create (const char *name, struct pci_dev *pdev,
			size_t size, size_t align, size_t allocation, int flags)
	 #else
		struct pci_pool *
		usb_pci_pool_create(struct pci_dev *pdev,
			size_t size, size_t align, size_t allocation, int flags)
	#endif
 #else
	#ifdef USB_DBGSTRING_ON
		struct dma_pool *usb_dma_pool_create(const char *name,
			size_t size, size_t align, size_t allocation, int flags)
	 #else
		struct dma_pool *usb_dma_pool_create(
			size_t size, size_t align, size_t allocation, int flags)
	#endif
#endif
{
 #ifdef USB_NO_DMA_POOL
  /*
  struct pci_pool * retval ;
 
  USB_DBG_FLOW(iprintf("[USB FLOW] usb_dma_pool_create() Begin ..\n"));
  
  retval = (void *)
		#ifdef USB_DBG_MEM_U_ON
        	UsbAllocMemU(USB_MEM_U_L,size);
         #else
        	UsbAllocMemU(USB_MEM_U_L);
        #endif
        
  // handle = retval ;
 
 USB_DBG_MEM(iprintf(
	"[USB MEM] usb_dma_pool_create() : retval=%08XH \n" ,
    retval
   	)) ;
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_dma_pool_create() End.\n"));
 return retval;
 */
#else

 struct dma_pool		*retval;

 USB_DBG_FLOW(iprintf("[USB FLOW] usb_dma_pool_create() Begin ..\n"));
 
 if (align == 0)
		align = 1;
	if (size == 0)
		return 0;
	else if (size < align)
		size = align;
	else if ((size % align) != 0)
	{
		size += align + 1;
		size &= ~(align - 1);
	}

 if (allocation == 0)
   {
	if (PAGE_SIZE < size)
			allocation = size;
	else
			allocation = PAGE_SIZE;
		// FIXME: round up for less fragmentation
   }
 else if (allocation < size)
		return 0;

 #ifndef _ETCPIP
	if (!(retval = kmalloc (sizeof(*retval), flags)))
	{
		return retval;
	}
  #else
	#ifndef USB_MEM_LIB
		retval = 
	    	NonCache_contigmalloc(sizeof(*retval)) ;
	    // NonCache_contigmalloc() cannot be free now !!
     #else
     	// alloc_mem(sizeof(*retval)) ;
        
		USB_DBG_MEM_U(iprintf(
		    "[USB MEM_U] usb_dma_pool_create() Call UsbAllocMemU() ..\n"
            ));
            
	retval = UsbAllocMemU(USB_MEM_U_S
				#ifdef USB_DBG_MEM_U_ON
    	        	,sizeof(*retval)
	            #endif
                );
    #endif
    
    USB_DBG_MEM(iprintf(
    	"[USB MEM] usb_dma_pool_create() : retval = %08XH \n",
        retval ));
    if ( retval == NULL ) {
		dprintf("[USB] Error : Out of Memory \n");
		return retval;
	   }
 #endif

 #ifdef	CONFIG_PCIPOOL_DEBUG
	flags |= SLAB_POISON;
 #endif

 #ifdef USB_DBGSTRING_ON
	strncpy (retval->name, name, sizeof retval->name);
	retval->name [sizeof retval->name - 1] = 0;
 #endif // USB_DBGSTRING_ON

 #ifdef CONFIG_PCI
 	retval->dev = pdev;
 #endif
 
 INIT_LIST_HEAD (&retval->page_list);
 
 #ifndef NO_SPINLOCK
	spin_lock_init (&retval->lock);
 #endif // ndef NO_SPINLOCK
 
 retval->size = size;
 retval->flags = flags;
 retval->allocation = allocation;
 retval->blocks_per_page = allocation / size;
    
 #ifndef NO_WAITQUEUE
	init_waitqueue_head (&retval->waitq);
 #endif // ndef NO_WAITQUEUE

 #ifdef CONFIG_PCIPOOL_DEBUG
	#ifndef _ETCPIP    
		printk (KERN_DEBUG "pcipool create %s/%s size %d, %d/page "
        	"(%d alloc)\n",
			pdev ? pdev->slot_name : NULL, retval->name, size,
			retval->blocks_per_page, allocation);
	 #else
		#ifdef USB_DBGSTRING_ON
        
			#ifdef CONFIG_PCI
			USB_DBG_INFO(iprintf(
			"[USB INFO] pcipool create [%s]/[%s] size %d, %d/page "
        	"(%d alloc) \n",
			pdev ? pdev->slot_name : NULL, retval->name, size,
			retval->blocks_per_page, allocation
            ));
            #else
            #endif
         #else
			USB_DBG_INFO(iprintf(
				"[USB INFO] pcipool create PCI POOL size %d, %d/page "
        		"(%d alloc) \n",
	            size,
				retval->blocks_per_page, allocation
        	    ));
        #endif
	#endif
  #else
	// #error : Please Make Sure : Not CONFIG_PCIPOOL_DEBUG !!
 #endif

 USB_DBG_FLOW(iprintf("[USB FLOW] usb_dma_pool_create() End.\n"));
 return retval;
 #endif // ndef USB_NO_DMA_POOL
} // function usb_dma_pool_create()
#endif // ndef USB_NO_DMA_POOL


#ifndef USB_NO_DMA_POOL

static struct pci_page * pool_alloc_page(struct dma_pool *pool, 
					int mem_flags)
{
 struct pci_page	*page;
 int				mapsize;

 mapsize = pool->blocks_per_page;
 mapsize = (mapsize + BITS_PER_LONG - 1) / BITS_PER_LONG;
 mapsize *= sizeof (long);

 #ifndef _ETCPIP
	page = (struct pci_page *)
    		kmalloc (mapsize + sizeof *page, mem_flags);
	if (!page)
		return 0;
  #else
	#ifndef USB_MEM_LIB
	page = (struct pci_page *)
        NonCache_contigmalloc(mapsize + sizeof *page) ;
		// NonCache_contigmalloc() cannot be free now !!
     #else
     	// alloc_mem(mapsize + sizeof *page) ;
		USB_DBG_MEM_U(iprintf(
		    "[USB MEM_U] pool_alloc_page() Call UsbAllocMemU ..\n"
            ));
		page = (struct pci_page *)
		#ifdef USB_DBG_MEM_U_ON
        	UsbAllocMemU(USB_MEM_U_M,mapsize + sizeof *page);
         #else
        	UsbAllocMemU(USB_MEM_U_M);
        #endif   
    #endif
    
    /*
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

    BLOCK_PREEMPTION;

	page =  (char **)alloc_mem(20 * sizeof (char *)) ;
                 
    UNBLOCK_PREEMPTION;
    */
    
    USB_DBG_MEM(iprintf(
		"[USB MEM] pool_alloc_page() : page = %08XH \n" ,
        page
        ));

    if ( page == NULL )
       {
		dprintf("[USB] Error : Out of Memory !!\n");
		return 0;
	   }
 #endif
        
 USB_DBG_INIT(iprintf(
	"[USB INIT] pool_alloc_page() : alloc page->vaddr .. \n"
    ));
    
 // #ifndef USB_MEM_LIB
 #ifdef CONFIG_PCI
	page->vaddr = pci_alloc_consistent (pool->dev,
				pool->allocation, &page->dma);
  #else
	page->vaddr = mem_alloc_consistent(
				pool->allocation, &page->dma);
 #endif
 // #else
 //	// 2003/5/13 B. S. Liaw : page Address FixUp
 //	page->vaddr = pci_alloc_consistent (pool->dev,
 //				pool->allocation+ (size_t)PAGE_SIZE, &page->dma);
 //    USB_DBG_INIT(iprintf(
 //       "[USB INIT] pool_alloc_page() : page->vaddr Address FixUp \n"));
 //       
 //   (unsigned long) (page->vaddr) += (unsigned long) PAGE_SIZE ;
 //   (unsigned long) (page->vaddr) &= (unsigned long) ~(PAGE_SIZE-1) ;
    
 //   (unsigned long) (page->dma) += (unsigned long) PAGE_SIZE ;
 //   (unsigned long) (page->dma) &= (unsigned long) ~(PAGE_SIZE-1) ;
 // #endif
                
 USB_DBG_MEM(iprintf(
	"[USB MEM] pool_alloc_page() : page->dma=%08XH \n" ,
    page->dma
   	)) ;
    
 if (page->vaddr)
    {
		memset (page->bitmap, 0xff, mapsize);	// bit set == free
		if (pool->flags & SLAB_POISON)
			memset (page->vaddr, POOL_POISON_BYTE, pool->allocation);
		list_add (&page->page_list, &pool->page_list);
	}
   else
    {
     #ifndef _ETCPIP
		kfree (page);
      #else
	#ifndef USB_MEM_LIB
        // NonCache_contigmalloc() cannot be free now !!
     #else
		#ifdef USB_DBG_MEM_C_ON
			UsbFreeMemC(USB_MEM_C_L,(unsigned char *)page,
            	sizeof(*page));
		 #else
			UsbFreeMemC(USB_MEM_C_L,(unsigned char *)page);
        #endif
    #endif
     #endif
	 page = 0;
	}
 return page;
}


static inline int
is_page_busy (int blocks, unsigned long *bitmap)
{
	while (blocks > 0) {
		if (*bitmap++ != ~0UL)
			return 1;
		blocks -= BITS_PER_LONG;
	}
	return 0;
}

static void
pool_free_page (struct dma_pool *pool, struct pci_page *page)
{
 dma_addr_t	dma = page->dma;

 if (pool->flags & SLAB_POISON)
		memset (page->vaddr, POOL_POISON_BYTE, pool->allocation);
        
 #ifdef CONFIG_PCI
 	pci_free_consistent (pool->dev, pool->allocation,
    					 page->vaddr, dma);
  #else
 	mem_free_consistent(pool->allocation,
    					 page->vaddr, dma);
 #endif
 
 list_del (&page->page_list);

 #ifndef _ETCPIP
	kfree (page);
  #else
	#ifndef USB_MEM_LIB
		// NonCache_contigmalloc() cannot be free now !!
	#elif defined(USB_DBG_MEM_U_ON)
        UsbFreeMemU(USB_MEM_U_M, (unsigned char *)page,77) ;
	#else
        UsbFreeMemU(USB_MEM_U_M, (unsigned char *)page) ;
	#endif
 #endif
}

#endif // ndef USB_NO_DMA_POOL

#ifndef USB_NO_DMA_POOL
/**
 * usb_pci_pool_destroy - destroys a pool of pci memory blocks.
 * @pool: pci pool that will be destroyed
 *
 * Caller guarantees that no more memory from the pool is in use,
 * and that nothing will try to use the pool after this call.
 */
#ifdef CONFIG_PCI
	void
	usb_pci_pool_destroy (struct pci_pool *pool)
 #else
	void
	usb_dma_pool_destroy (struct dma_pool *pool)
#endif
{
 #ifdef USB_NO_DMA_POOL
     /*
	 #if defined(USB_DBG_MEM_U_ON)
        UsbFreeMemU(USB_MEM_U_L, (unsigned char *)pool,PAGE_UNIT_SIZE) ;
	 #else
		// free(hub->descriptor);
        UsbFreeMemU(USB_MEM_U_L, (unsigned char *)pool) ;
	 #endif
     */
 #else
 unsigned long		flags;

 #ifdef CONFIG_PCIPOOL_DEBUG
        #ifndef _ETCPIP    
	       printk (KERN_DEBUG "pcipool destroy %s/%s\n",
     		 pool->dev ? pool->dev->slot_name : NULL,
		     pool->name);
         #else
			#ifdef USB_DBGSTRING_ON
				#ifdef CONFIG_PCI
		       	dprintf("[USB] pcipool destroy %s/%s\n",
			     pool->dev ? pool->dev->slot_name : NULL,
			     pool->name);
                #else
		       	dprintf("[USB] pcipool destroy %s\n",
			     pool->name);
                #endif
            #else
	       		dprintf("[USB] pcipool destroy PCI POOL\n") ;
            #endif
        #endif
 #endif

 #ifndef NO_SPINLOCK
	spin_lock_irqsave (&pool->lock, flags);
 #endif // ndef NO_SPINLOCK
 while (!list_empty (&pool->page_list))
   {
	struct pci_page		*page;
	page = list_entry (pool->page_list.next,
				struct pci_page, page_list);
	if (is_page_busy (pool->blocks_per_page, page->bitmap))
        {
         #ifndef _ETCPIP
			printk (KERN_ERR "usb_dma_pool_destroy %s/%s, %p busy\n",
				pool->dev ? pool->dev->slot_name : NULL,
				pool->name, page->vaddr);
          #else
			#ifdef USB_DBGSTRING_ON
				#ifdef CONFIG_PCI
				dprintf("[USB] usb_dma_pool_destroy %s/%s, %p busy\n",
					pool->dev ? pool->dev->slot_name : NULL,
					pool->name, page->vaddr);
                #else
				dprintf("[USB] usb_dma_pool_destroy %s, %p busy\n",
					pool->name, page->vaddr);
                #endif
            #else
				dprintf("[USB] usb_dma_pool_destroy PCI POOL , %p busy\n",
					page->vaddr);
            #endif
         #endif
			/* leak the still-in-use consistent memory */
		 list_del (&page->page_list);
		 kfree (page);
		}
       else
		pool_free_page (pool, page);
   } // while
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore (&pool->lock, flags);
 #endif // ndef NO_SPINLOCK
 
 #ifndef _ETCPIP
	kfree (pool);
  #else
	#ifndef USB_MEM_LIB
		// NonCache_contigmalloc() cannot be free now !!
	#elif defined(USB_DBG_MEM_U_ON)
        UsbFreeMemU(USB_MEM_U_S, (unsigned char *)pool,77) ;
	#else
		// free(hub->descriptor);
        UsbFreeMemU(USB_MEM_U_S, (unsigned char *)pool) ;
	#endif
 #endif
 #endif // ndef USB_NO_DMA_POOL
}
#endif // ndef USB_NO_DMA_POOL


/**
 * usb_pci_pool_alloc - get a block of consistent memory
 * @pool: pci pool that will produce the block
 * @mem_flags: SLAB_KERNEL or SLAB_ATOMIC
 * @handle: pointer to dma address of block
 *
 * This returns the kernel virtual address of a currently unused block,
 * and reports its dma address through the handle.
 * If such a memory block can't be allocated, null is returned.
 */

#ifdef USB_NO_DMA_POOL
   /*
   void *
   usb_dma_pool_alloc(unsigned long *handle,int size)
   {
	void * retval ;
 
	retval = (void *)
		#ifdef USB_DBG_MEM_U_ON
        	UsbAllocMemU(USB_MEM_U_M,size);
         #else
        	UsbAllocMemU(USB_MEM_U_M);
        #endif
        
	*handle = (unsigned long *) retval ;
    
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_dma_pool_alloc() : *handle=%08XH \n" ,
    	*handle
   		)) ;
 
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_dma_pool_alloc() : retval=%08XH \n" ,
    	retval
   		)) ;
        
	return retval;
   }
   */
#else
void *
usb_dma_pool_alloc (struct dma_pool *pool, int mem_flags,
					dma_addr_t *handle)
{
 unsigned long		flags;
 struct list_head	*entry;
 struct pci_page		*page;
 int			map, block;
 size_t			offset;
 void			*retval;

 restart:
 #ifndef NO_SPINLOCK
	spin_lock_irqsave (&pool->lock, flags);
 #endif // ndef NO_SPINLOCK
 
 list_for_each (entry, &pool->page_list)
   {
	int		i;
	page = list_entry (entry, struct pci_page, page_list);
	/* only cachable accesses here ... */
	for (map = 0, i = 0;
		i < pool->blocks_per_page;
		i += BITS_PER_LONG, map++
       )
       {
		if (page->bitmap [map] == 0)
				continue;
		block = ffz (~ page->bitmap [map]);
		if ((i + block) < pool->blocks_per_page)
           {
			clear_bit (block, &page->bitmap [map]);
			offset = (BITS_PER_LONG * map) + block;
			offset *= pool->size;
			goto ready;
		   }
	   } //  for map
   }
   
 if (!(page = pool_alloc_page (pool, mem_flags))) 
   {
	if (mem_flags == SLAB_KERNEL)
       {
		#ifndef NO_WAITQUEUE
		#ifndef _ETCPIP
			DECLARE_WAITQUEUE (wait, current);
	     #else
			DECLARE_WAITQUEUE (wait,0);
	    #endif 
		#endif // ndef NO_WAITQUEUE
		/*
		#ifndef _ETCPIP
			current->state = TASK_INTERRUPTIBLE;
		#endif // ndef _ETCPIP
		*/
		#ifndef NO_WAITQUEUE
			add_wait_queue (&pool->waitq, &wait);
		#endif // ndef NO_WAITQUEUE
		#ifndef NO_SPINLOCK
			spin_unlock_irqrestore (&pool->lock, flags);
		#endif // ndef NO_SPINLOCK

		#ifndef NO_TIMER
			usb_schedule_timeout (POOL_TIMEOUT_JIFFIES);
		  #else
		    #error : Please Make Sure : NO_TIMER
		#endif // ndef NO_TIMER
 
		#ifndef _ETCPIP
        	/*
		    current->state = TASK_RUNNING;
            */
		#endif // ndef _ETCPIP
 
		#ifndef NO_WAITQUEUE
			remove_wait_queue (&pool->waitq, &wait);
		#endif // ndef NO_WAITQUEUE
			goto restart;
	   }
	retval = 0;
	goto done;
   }

 clear_bit (0, &page->bitmap [0]);
 offset = 0;
    
 ready:
 
 retval = offset + page->vaddr;
 *handle = offset + page->dma;
 USB_DBG_MEM(iprintf(
	"[USB MEM] usb_dma_pool_alloc() : offset=%08XH page->dma=%08XH "
    "*handle=%08XH\n"
    ,offset, page->dma , *handle
   	)) ;
    
 done:
 
 USB_DBG_MEM(iprintf(
	"[USB MEM] usb_dma_pool_alloc() : retval=%08XH \n" ,
    retval
   	)) ;
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore (&pool->lock, flags);
 #endif // ndef NO_SPINLOCK
 return retval;
}
#endif // ndef USB_NO_DMA_POOL


#ifndef USB_NO_DMA_POOL

static struct pci_page *
pool_find_page (struct dma_pool *pool, dma_addr_t dma)
{
	unsigned long		flags;
	struct list_head	*entry;
	struct pci_page		*page;

#ifndef NO_SPINLOCK
	spin_lock_irqsave (&pool->lock, flags);
#endif // ndef NO_SPINLOCK
	list_for_each (entry, &pool->page_list) {
		page = list_entry (entry, struct pci_page, page_list);
		if (dma < page->dma)
			continue;
		if (dma < (page->dma + pool->allocation))
			goto done;
	}
	page = 0;
done:
#ifndef NO_SPINLOCK
	spin_unlock_irqrestore (&pool->lock, flags);
#endif // ndef NO_SPINLOCK
	return page;
}
#endif // ndef USB_NO_DMA_POOL


/**
 * usb_pci_pool_free - put block back into pci pool
 * @pool: the pci pool holding the block
 * @vaddr: virtual address of block
 * @dma: dma address of block
 *
 * Caller promises neither device nor driver will again touch this block
 * unless it is first re-allocated.
 */
#ifdef USB_NO_DMA_POOL
   /*
   void usb_pci_pool_free(void *vaddr)
   {
	USB_DBG_INIT(iprintf(
		"[USB INIT] usb_dma_pool_free() : free usb_dma_pool/"
    	"usb_pci_pool .. \n"
    	));
	 #if defined(USB_DBG_MEM_U_ON)
        UsbFreeMemU(USB_MEM_U_M, (unsigned char *)vaddr,PAGE_UNIT_SIZE) ;
	 #else
		// free(hub->descriptor);
        UsbFreeMemU(USB_MEM_U_M, (unsigned char *)vaddr) ;
	#endif
   }
   */
#else
   void
   usb_dma_pool_free (struct dma_pool *pool, void *vaddr, dma_addr_t dma)
   {
	struct pci_page	*page;
	unsigned long		flags;
	int			    map, block;

	USB_DBG_INIT(iprintf(
		"[USB INIT] usb_dma_pool_free() : free usb_dma_pool/"
	    "usb_pci_pool .. \n"
    	));
    
	if ((page = pool_find_page (pool, dma)) == 0)
	   {
		#ifndef _ETCPIP
		 printk (KERN_ERR "usb_dma_pool_free %s/%s, %p/%x (bad dma)\n",
			pool->dev ? pool->dev->slot_name : NULL,
			pool->name, vaddr, dma);
       #else
			#ifdef USB_DBGSTRING_ON
				#ifdef CONFIG_PCI
		 		dprintf("[USB] Error : %s/%s, %p/%x (bad dma)\n",
					pool->dev ? pool->dev->slot_name : NULL,
					pool->name, vaddr, dma);
                #else
		 		dprintf("[USB] Error : %s, %p/%x (bad dma)\n",
					pool->name, vaddr, dma);
                #endif
            #else
		 		dprintf("[USB] Error : PCI POOL , %p/%x (bad dma)\n",
					vaddr, dma);
            #endif
    #endif       
	return;
   }
   
 #ifdef	CONFIG_PCIPOOL_DEBUG
 if (((dma - page->dma) + (void *)page->vaddr) != vaddr)
   {
	#ifndef _ETCPIP    
		printk (KERN_ERR "usb_dma_pool_free %s/%s, %p (bad vaddr)/%x\n",
			pool->dev ? pool->dev->slot_name : NULL,
			pool->name, vaddr, dma);
     #else
		#ifdef USB_DBGSTRING_ON
				#ifdef CONFIG_PCI
			dprintf("[USB] Error : %s/%s, %p (bad vaddr)/%x\n",
				pool->dev ? pool->dev->slot_name : NULL,
				pool->name, vaddr, dma);
                #else
			dprintf("[USB] Error : %s, %p (bad vaddr)/%x\n",
				pool->name, vaddr, dma);
                #endif
         #else
			dprintf("[USB] Error : USB PCI POOL, %p (bad vaddr)/%08XH\n",
				vaddr, dma);
        #endif
    #endif
	return;
   }
 #endif

 block = dma - page->dma;
 block /= pool->size;
 map = block / BITS_PER_LONG;
 block %= BITS_PER_LONG;

 #ifdef	CONFIG_PCIPOOL_DEBUG
	if (page->bitmap [map] & (1UL << block))
	   {
		#ifndef _ETCPIP    
			printk (KERN_ERR "usb_dma_pool_free %s/%s, dma %x "
            	"already free\n",
				pool->dev ? pool->dev->slot_name : NULL,
				pool->name, dma);
		 #else
			#ifdef USB_DBGSTRING_ON
				#ifdef CONFIG_PCI
				dprintf("[USB] Error : %s/%s, dma %x "
            		"already free\n",
					pool->dev ? pool->dev->slot_name : NULL,
					pool->name, dma);
                #else
				dprintf("[USB] Error : %s, dma %x "
            		"already free\n",
					pool->name, dma);
                #endif
            #else
				dprintf("[USB] Error : USB PCI POOL, dma %08XH "
            		"already free\n",
					dma);
            #endif
		#endif
		return;
	   }
 #endif
 
 if (pool->flags & SLAB_POISON)
		memset (vaddr, POOL_POISON_BYTE, pool->size);

 #ifndef NO_SPINLOCK
	spin_lock_irqsave (&pool->lock, flags);
 #endif // ndef NO_SPINLOCK
 
 set_bit (block, &page->bitmap [map]);
    
 #ifndef NO_WAITQUEUE
	if (waitqueue_active (&pool->waitq))
		wake_up (&pool->waitq);
 #endif // ndef NO_WAITQUEUE
 
 /*
	 * Resist a temptation to do
	 *    if (!is_page_busy(bpp, page->bitmap)) pool_free_page(pool, page);
	 * it is not interrupt safe. Better have empty pages hang around.
	 */
  
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore (&pool->lock, flags);
 #endif // ndef NO_SPINLOCK
}


#ifndef _ETCPIP
   EXPORT_SYMBOL (usb_dma_pool_create);
   EXPORT_SYMBOL (usb_dma_pool_destroy);
   EXPORT_SYMBOL (usb_dma_pool_alloc);
   EXPORT_SYMBOL (usb_dma_pool_free);
#endif // ndef _ETCPIP

#endif // ndef USB_NO_DMA_POOL
