#ifndef _MT_USB_H_
#define _MT_USB_H_

#define RTOS_MT      1  /* MultiTask!     */

#define	    RTOS_ID	RTOS_MT

#include    <stdarg.h>





typedef	unsigned char	byte;
typedef unsigned long	uint32;

typedef byte MTflag_t;	/* mtcore flag and counter atomic type */
typedef unsigned int	uint;		/* default unsigned integer type */

#define MASK_INTS()													\
				({													\
					asm volatile("	.set	noreorder\n"			\
						"	mfc0	$8,$12\n"						\
						"	nop\n"									\
						"	srl	$8,$8,1\n"							\
						"	sll	$8,$8,1\n"							\
						"	mtc0	$8,$12\n"						\
						"	nop\n"									\
						"	.set reorder"							\
						:::"$8"	);									\
				})
#define UNMASK_INTS()												\
				({													\
					asm volatile("	.set noreorder\n"				\
						"	mfc0	$8,$12\n"						\
						"	nop\n"									\
						"	ori	$8,$8,1\n"							\
						"	mtc0	$8,$12\n"						\
						"	.set reorder"							\
						:::"$8" );									\
				})





#ifndef fAR
#define fAR
#endif

#ifndef DIRECT
#define DIRECT
#endif

#ifndef CODE
#define CODE
#endif
#ifndef DATA
#define DATA
#endif





/* CLOCKHZ defines the number of clock interrupts per second
  {$} 9 100 {$} 5 1 1000 {$} 4 "Clock interrupt frequency in Hertz" */
#define	CLOCKHZ	1000	/* 1/100 second (10ms) clock */

/* delay types for dlytsk() */
#define DLY_TICKS      1        /* delay count in scheduling ticks */
#define DLY_SECS       2        /* delay count in seconds */





typedef uint32     tick_cnt_t;
typedef byte       TASK_ID;             /* task ID type */

/* Values that need to be defined, if not already defined by other .h files */
#ifndef TRUE                /* defined in Microsoft include files */
#define TRUE       1
#define FALSE      0
#endif





typedef union cmdargs_u { /* union of arg types for queued commands */
    struct{uint priority;void (fAR CODE *fadr)(void);uint stksze;} runargs;
    struct{uint slot;} tskargs;
    struct{uint event;} evtargs;
    struct{uint resrc;} resargs;
    struct{uint mbx;} flushargs;
    struct{uint group; uint mask;} grpargs;
    struct{uint slot; uint priority;} priargs;
    struct{uint slot; uint dlytyp; uint dlytme;} dlyargs;
    struct{uint mailbox; void *msgptr; uint msgpri;} msgargs;
    struct{void *reladr;} memargs;
    struct{uint poolid; void *bufptr;} bufargs;
#ifdef HIWARE_BUG
    struct{uint mailbox; byte pad; void *msgptr; uint msgpri;} msgargs2;
    struct{byte pad; void *reladr;} memargs2;
    struct{uint poolid; byte pad; void *bufptr;} bufargs2;
#endif
/* TronTask! specific structures below: */
#if RTOS_ID == RTOS_TT
    struct{uint priority;} rotargs;         /* Used for irot_rdq */
    struct{uint mailbox; void *msgptr;} mbxargs; /* isnd_msg */
    struct{uint mbfid; void *msgptr; uint msgsz;} mbfargs; /* isnd_mbf */
    struct{void *tmrdummy;} tmrargs;
#endif /*  if TronTask */
} QARGS;

#define SUCCESS        0         /* successful operation */
#define E_RELMEM      -19        /* invalid memory release */
#define E_INVPID      -26        /* invalid pool id */





#undef  ENOMEM

#define ENOMEM  3       /* No Memory available (COLOR0) */
#define ENOENT  5       /* no file entry exists in directory */





typedef unsigned int   	size_t;

typedef struct MT_USB_EXPORT_st {
	unsigned long (*pf_get_data_segment)(void);

	void (*pf_randomize)(void);
	int  (*pf_random)(int num);
	void (*pf_scdtsk)(void);
	int  (*pf_runtsk)(uint priority,void (fAR CODE *fadr)(void),uint stksze,...);
	int  (*pf_dlytsk)(TASK_ID slot,byte dlytyp,uint dlytme);
	tick_cnt_t (*pf_get_sys_time)(void);
	int  (*pf_getres)(uint resrc,uint todly);
	int  (*pf_relres)(uint resrc);
	int  (*pf_chkevt)(uint event);
	int  (*pf_setevt)(uint event);
	void (*pf_SIGNAL_IR)(unsigned long signo);
	int  (*pf_wteset)(uint event,uint todly);
	int  (*pf_wteset_dec)(uint event,uint todly);
	int  (*pf_clrevt)(uint event);
	void (*pf_usb_isr_install)(void);
	void (*pf_usb_uhci1_isr_install)(void);
	void (*pf_usb_uhci2_isr_install)(void);
	void (*pf_led_usb_on)(void);
	void (*pf_led_usb_off)(void);
	void (*pf_led_usb_flash)(char isEnable);
	unsigned char (*pf_PciCfgRdByte)( int hDev, int offset );
	unsigned short (*pf_PciCfgRdWord)( int hDev, int offset );
	unsigned long (*pf_PciCfgRdDword)( int hDev, int offset );
	void (*pf_PciCfgWrByte)( int hDev, int offset, unsigned char data );
	void (*pf_PciCfgWrWord)( int hDev, int offset, unsigned short data );
	void (*pf_PciCfgWrDword)( int hDev, int offset, unsigned long data );
	void (*pf_Nputchr)(char ch);
	void (*pf_PrintULONG)(unsigned long num);
	void (*pf_updateDiskDriveInfo)(char *vendor, char *model, int driveID);
	void (*pf_updateShareConfPlug)(char resType, char mode);
	unsigned long (*pf_diskMount_status)(int disk);
	void (*pf_diskMount_set_status)(int disk, unsigned long status);
	int (*pf_partition_mount)(unsigned long VolumeID, unsigned char volumeType, struct DiskParamS *diskInfo);
	int (*pf_partition_umount)(unsigned long VolumeID, unsigned char volumeType);

	char	* p_debug_act;
	char	* p_debug_rt_disp;
	volatile MTflag_t DIRECT * p_mt_busy;
	DIRECT TASK_ID * p_cur_task;
	volatile tick_cnt_t DATA * p_sys_time;
} MT_USB_EXPORT;

typedef struct USB_MT_EXPORT_st {
	unsigned long (*pf_get_data_segment)(void);

	void (*pf_UsbSysInit)(void);
	void (*pf_localhc_irq)(void);
	void (*pf_localhc_uhci1_irq)(void);
	void (*pf_localhc_uhci2_irq)(void);
	unsigned char (*pf_Prt_Init)(void);
	void (*pf_Prt_Reset)(void);
	void (*pf_ParallelTest)(void);
	void (*pf_Prt_Open)(void);
	unsigned char * (*pf_Prt_GetId)(void);
	unsigned char (*pf_Prt_Get_S_EPP_Sts)(void);
	unsigned char (*pf_Prt_Check_Port_Pwr_Connect)(void);
	unsigned char (*pf_Prt_Sts_Check_On_Line)(void);
	unsigned char (*pf_Prt_Sts_Check_Paper_End)(void);
	unsigned char (*pf_Prt_Sts_Check_Error)(void);
	unsigned char (*pf_Prt_Buf_Write)(unsigned char *LV_Ptr, unsigned long LV_Legth, unsigned long *byte_sent);
	void (*pf_Prt_Close)(void);
	int (*pf_usb_ms_unplug)(void);
	int (*pf_usb_block_read)(unsigned long VolumeID, unsigned long SectorNo, void *Buffer);
	int (*pf_usb_block_readex)(unsigned long VolumeID, unsigned long SectorNo, void *Buffer, unsigned short max_prefetch_sectors);
	int (*pf_usb_block_write)(unsigned long VolumeID, unsigned long SectorNo, void *Buffer);
	int (*pf_blk_sync_cache)(void);
	void (*pf_usb_storage_query_inbound_display)(void);
	void (*pf_usb_storage_inner_display)(void);
	void (*pf_usb_storage_cache_display)(void);

	struct DiskParamS * p_USBDiskParam;
	int * p_validPartitionCnt;
	unsigned char * p_PartitionType;
	unsigned long * p_validPartitionID;
} USB_MT_EXPORT;

extern MT_USB_EXPORT *mt_usb_export;

#define		iprintf		dprintf
extern int  runtsk(uint priority,void (fAR CODE *fadr)(void),uint stksze);
extern int  runtsk_parm(uint priority,void (fAR CODE *fadr)(void),uint stksze, unsigned long parm);
extern int dprintf(char *fmt, ...);

#define debug_act	(*(mt_usb_export->p_debug_act))
#define debug_rt_disp	(*(mt_usb_export->p_debug_rt_disp))
#define mt_busy		(*(mt_usb_export->p_mt_busy))
#define cur_task	(*(mt_usb_export->p_cur_task))
#define sys_time	(*(mt_usb_export->p_sys_time))



#endif	// _MT_USB_H_
