
// Name of File : UsbMemLib.c
// ====================================================================

#define __KERNEL__

#ifdef _NEWLIB
	#include "sys_type.h"
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
		// for deciding whether to compile
        //  Debug Version or Release Version
#include "ramAddr.h"  // for  USB_?_POOL_SIZE
                      // and for CBAC_TIME_GAP_AREA (USB_BS_ASSERT)
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_TIMER
	#include "PoTimer.h" // osport_timer.h
#endif // ndef NO_TIMER
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "kobject.h"
#include "PoDevice.h" // osport_device.h

#ifdef CONFIG_PCI
	#include "UsbPoPci.h" // usb_osport_pci.h
#endif	// CONFIG_PCI
#include "UsMemLib.h" // UsbMemLib.h
#include "UsbSysInit.h"

#undef EMBEDDED_BUF

#undef __US_MEM_LIB_DBG

// ====================================================================
#ifdef USB_DBG_MEM_C_ON
	#warning: USB_DBG_MEM_C_ON !!
	// 0 : only check size
	// 1 : show message , for debug memory leak
	// 2 : show more message
	#define USB_DBG_MEM_C_LEVEL 0
#else
	#define USB_DBG_MEM_C_LEVEL 0
#endif // USB_DBG_MEM_C_ON

// ====================================================================
#ifdef USB_DBG_MEM_U_ON
	#warning: USB_DBG_MEM_U_ON !!
	// 0 : only check size
	// 1 : show message , for debug memory leak
	#define USB_DBG_MEM_U_LEVEL 0
#else
	#define USB_DBG_MEM_U_LEVEL 0
#endif // USB_DBG_MEM_C_ON

// ====================================================================

struct s_buf_link_s { // total 28 bytes
 #ifdef EMBEDDED_BUF
 	// char buf[sizeof(struct matchRule_s)];  // 'matchRule_s' buffer pointer
	char buf[USB_MEM_U_S_SIZE] ;
  #else
	unsigned int buf_number ;
 #endif
 struct s_buf_link_s *next;       // next structure pointer
 unsigned long startAddr;        // this structure start pointer, for err check
 #ifndef EMBEDDED_BUF
	unsigned long buf_startAddr;     // this structure start pointer, for err check
 #endif // ndef EMBEDDED_BUF
 char isUsed;                        // is this buffer structure being used (TRUE/FALSE,  Yes/No)
};

struct m_buf_link_s { // total 20 bytes
 #ifdef EMBEDDED_BUF
	//  char buf[sizeof(struct timeGap_s)];      // 'timeGap_s' buffer pointer
	char buf[USB_MEM_U_M_SIZE] ;          
  #else
	unsigned int buf_number ;
 #endif
 struct m_buf_link_s *next;             // next structure pointer
 unsigned long startAddr;               // this structure start pointer, for err check
 #ifndef EMBEDDED_BUF
	unsigned long buf_startAddr;               // this structure start pointer, for err check
 #endif // ndef EMBEDDED_BUF
 char isUsed;                                   // is this buffer structure being used (TRUE/FALSE,  Yes/No)
};

struct l_buf_link_s { // total 32 bytes
 #ifdef EMBEDDED_BUF
	//       char buf[sizeof(struct hostSame_s)];            // 'hostSame_s' buffer pointer
	char buf[USB_MEM_U_L_SIZE] ;          
  #else
	unsigned int buf_number ;
 #endif
 struct l_buf_link_s *next;                       // next structure pointer
 unsigned long startAddr;                         // this structure start pointer, for err check
 #ifndef EMBEDDED_BUF
	unsigned long buf_startAddr;                         // this structure start pointer, for err check
 #endif // ndef EMBEDDED_BUF
        char isUsed;                                             // is this buffer structure being used (TRUE/FALSE,  Yes/No)
};

struct bufManage_s {                                     // [0]: match rule buf,  [1]: time gap buf
        unsigned short bufNum[3];                        // total avaliable buffer entry number
        struct s_buf_link_s *bufCur_M;           // current avaliable 'matchRule_s' buffer structure
        struct s_buf_link_s *bufLast_M;          // last avaliable 'matchRule_s' buffer structure
        struct m_buf_link_s *bufCur_T;           // current avaliable 'timeGap_s' buffer structure
        struct m_buf_link_s *bufLast_T;          // last avaliable 'timeGap_s' buffer structure
        struct l_buf_link_s *bufCur_H;           // current avaliable 'hostSame_s' buffer structure
        struct l_buf_link_s *bufLast_H;          // last avaliable 'hostSame_s' buffer structure
        unsigned long bufUp[3];                          // buffer upper boundary
        unsigned long bufLo[3];                          // buffer lower boundary
};

// ====================================================================

static struct bufManage_s bufMem;

#ifndef EMBEDDED_BUF
	char *UsbSPool = (char *) 0 ;
	char *UsbMPool = (char *) 0 ;
	char *UsbLPool = (char *) 0 ;
	#ifdef USB_MEM_STATIC_ARRAY
		static char UsbSPoolArray[USB_MEM_U_S_SIZE * ((USB_S_POOL_SIZE+USB_ED_TD_MAX)+1)] ;
		static char UsbMPoolArray[USB_MEM_U_M_SIZE * (USB_M_POOL_SIZE+1)] ;
		static char UsbLPoolArray[USB_MEM_U_L_SIZE * ((USB_L_POOL_SIZE+USB_DMAPOOL_L_MAX)+1)] ;
    	#endif
#endif // ndef EMBEDDED_BUF

#ifndef USB_MEM_STATIC_ARRAY
	//struct s_buf_link_s *mrBuf = (struct s_buf_link_s *)CBAC_MATCH_RULE_AREA;
	struct s_buf_link_s *UsbSBuf = (struct s_buf_link_s *) 0 ;
	//struct m_buf_link_s *tgBuf = (struct m_buf_link_s *)CBAC_TIME_GAP_AREA;
	struct m_buf_link_s *UsbMBuf = (struct m_buf_link_s *) 0 ;
	//struct l_buf_link_s *hsBuf = (struct l_buf_link_s *)CBAC_SAME_HOST_AREA;
	struct l_buf_link_s *UsbLBuf = (struct l_buf_link_s *) 0 ;
#else
	static struct s_buf_link_s UsbSBuf[(USB_S_POOL_SIZE+USB_ED_TD_MAX)] ;
	static struct m_buf_link_s UsbMBuf[USB_M_POOL_SIZE] ;
	static struct l_buf_link_s UsbLBuf[(USB_L_POOL_SIZE+USB_DMAPOOL_L_MAX)] ;
#endif

// ====================================================================
/*
#ifdef USB_MEM_STATIC_ARRAY
	static char conti_array[USB_NONCACHE_SIZE] ;
#endif

#ifndef USB_MEM_STATIC_ARRAY
	static unsigned long curptr;
#else
	static unsigned long curptr =(unsigned long)conti_array;
#endif

#ifdef USB_MEM_STATIC_ARRAY
   void UsbNCCntMalloc_init(void)
   {
	curptr =(unsigned long)conti_array;
	iprintf("Non-Cache Area : from %08XH to %08XH\n", conti_array,(unsigned long)&conti_array[USB_NONCACHE_SIZE]-1);
   }

   void * UsbNCCntMalloc(unsigned long usize)
   {
	unsigned long oldptr;

	curptr += usize;
 
	USB_DBG_MEM(iprintf("[MEM] UsbNCCntMalloc() : Size:%08XH\n", usize));
    
	if ((curptr > (unsigned long)&conti_array[USB_NONCACHE_SIZE]) || (curptr < (unsigned long)conti_array))
	{
		dprintf("[USB] UsbNCCntMalloc() : Error: no memory allocated!! "
        		", curpt=%08XH \n", curptr
        		);
         	return((void*)0);
        }
    
	USB_DBG_MEM(iprintf(
        "[MEM] UsbNCCntMalloc() : Addr:%08XH Size:%08XH\n",
        oldptr,usize
        ));
	return( (void *)(USB_TO_UN_CACHED(oldptr)) );
   }

   void show_usb_contigmem(void)
   {
	 dprintf("[USB] contigmem end at %08x\n", curptr);
   }

#endif // ndef _S3C2510
*/

//#if _USB2 == 1	// VIA vt6212
#if 0
void usb_hc_res_mem_lock(void) {
	if (mt_busy == 0) {
		while (getres(MT_RES_USB_MEM, 0xffffffff) != SUCCESS) {
			dlytsk(cur_task, DLY_TICKS, 1);
		}
	}
}

void usb_hc_res_mem_unlock(void) {
	if (mt_busy == 0) {
		relres(MT_RES_USB_MEM);
	}
}
#else	// _USB2
#define usb_hc_res_mem_lock()		BLOCK_PREEMPTION
#define usb_hc_res_mem_unlock()		UNBLOCK_PREEMPTION
#endif	// _USB2
//*********************************************************************
// FUNCTION:  initUsbPool : Un-Cached
// PURPOSE:   initialize fixed size memory pool
// RETURN:
//         TRUE: init success,
//         FALSE: fail (will terminal system operation)
//*********************************************************************

int initUsbPool(void)
{
 int i;
 struct s_buf_link_s *curS;
 struct m_buf_link_s *curM;
 struct l_buf_link_s *curL;

/* #ifdef USB_MEM_STATIC_ARRAY
	UsbNCCntMalloc_init() ;
    	show_usb_contigmem() ;
 #endif
 */
 
 #ifndef EMBEDDED_BUF
	#ifndef USB_MEM_STATIC_ARRAY
		UsbSPool = (char *)
			UsbNCCntMalloc(
               USB_MEM_U_S_SIZE * ((USB_S_POOL_SIZE+USB_ED_TD_MAX)+2)
               ) ;
               
	#else
		UsbSPool = (char *) UsbSPoolArray ;
    	#endif
    
	if (UsbSPool == (char *)0)
	   {
		dprintf("[USB] Error : Out of NonCache Memory for UsbSPool !!\n") ;
	   }
    
	// 2003/5/14 B. S. Liaw : page Address FixUp      
   (unsigned long) (UsbSPool) += (unsigned long) (USB_MEM_U_S_SIZE-1) ;
   (unsigned long) (UsbSPool) &= (unsigned long) ~(USB_MEM_U_S_SIZE-1) ;    
 #endif // ndef EMBEDDED_BUF
 
 #ifndef USB_MEM_STATIC_ARRAY
	UsbSBuf = (struct s_buf_link_s *)
			UsbNCCntMalloc(
               sizeof(struct s_buf_link_s) * (USB_S_POOL_SIZE+USB_ED_TD_MAX)
               ) ;
	// NonCache_contigmalloc() cannot be free now !!
 #endif               
 
 if (UsbSBuf == (struct s_buf_link_s *)0)
    {
     dprintf("[USB] Error : Out of NonCache Memory for UsbSBuf !!\n") ;
    }
            
 #ifndef EMBEDDED_BUF
	#ifndef USB_MEM_STATIC_ARRAY
		UsbMPool = (char *)
			UsbNCCntMalloc(
               USB_MEM_U_M_SIZE * (USB_M_POOL_SIZE+2)
               ) ;
		// NonCache_contigmalloc() cannot be free now !!
	#else
		UsbMPool = (char *) UsbMPoolArray ;
    	#endif

	if (UsbMPool == (char *)0)
	   {
		dprintf("[USB] Error : Out of NonCache Memory for UsbMPool !!\n") ;
	   }
    
	// 2003/5/14 B. S. Liaw : page Address FixUp     
   (unsigned long) (UsbMPool) += (unsigned long) (USB_MEM_U_M_SIZE-1) ;
   (unsigned long) (UsbMPool) &= (unsigned long) ~(USB_MEM_U_M_SIZE-1) ;
 #endif // ndef EMBEDDED_BUF
 
 #ifndef USB_MEM_STATIC_ARRAY
	UsbMBuf = (struct m_buf_link_s *)
			UsbNCCntMalloc(
               sizeof(struct m_buf_link_s) * USB_M_POOL_SIZE
               ) ;
	// NonCache_contigmalloc() cannot be free now !!
 #endif
 
 if (UsbMBuf == (struct m_buf_link_s *)0)
    {
     dprintf("[USB] Error : Out of NonCache Memory for UsbMBuf !!\n") ;
    }
            
 #ifndef EMBEDDED_BUF
	#ifndef USB_MEM_STATIC_ARRAY
		UsbLPool = (char *)
			UsbNCCntMalloc(
               USB_MEM_U_L_SIZE * ((USB_L_POOL_SIZE+USB_DMAPOOL_L_MAX)+2)
               ) ;
		// NonCache_contigmalloc() cannot be free now !!
        
	#else
		UsbLPool = (char *) UsbLPoolArray ;
    	#endif
    
	if (UsbLPool == (char *)0)
	   {
		dprintf("[USB] Error : Out of NonCache Memory for UsbLPool !!\n") ;
	   }
    
	// 2003/5/14 B. S. Liaw : page Address FixUp      
   (unsigned long) (UsbLPool) += (unsigned long) (USB_MEM_U_L_SIZE-1) ;
   (unsigned long) (UsbLPool) &= (unsigned long) ~(USB_MEM_U_L_SIZE-1) ;
 #endif // ndef EMBEDDED_BUF
 
 #ifndef USB_MEM_STATIC_ARRAY
	UsbLBuf = (struct l_buf_link_s *)
			UsbNCCntMalloc(
               sizeof(struct l_buf_link_s) * (USB_L_POOL_SIZE+USB_DMAPOOL_L_MAX)
               ) ;
	// NonCache_contigmalloc() cannot be free now !!
 #endif
 
 if (UsbLBuf == (struct l_buf_link_s *)0)
    {
     dprintf("[USB] Error : Out of NonCache Memory for UsbLBuf !!\n") ;
    }
    
 memset(&bufMem, 0, sizeof(struct bufManage_s));
 
 usb_hc_res_mem_lock();
 // init match rule buf (building free entry pool linking-list)
 for (i = 0 ; i < (USB_S_POOL_SIZE+USB_ED_TD_MAX) ; i++)
 {
	curS = &UsbSBuf[i];
    	curS->isUsed = FALSE;
    	curS->startAddr = (unsigned long)curS;
    	if (i < ((USB_S_POOL_SIZE+USB_ED_TD_MAX) - 1)) 
    		curS->next = &UsbSBuf[i + 1];
       	else
        	curS->next = NULL;
	#ifndef EMBEDDED_BUF
		curS->buf_number = i ;
		curS->buf_startAddr = (unsigned long) (UsbSPool + USB_MEM_U_S_SIZE * i) ;
	#endif // ndef EMBEDDED_BUF
 }
 bufMem.bufCur_M = &UsbSBuf[0];
 bufMem.bufLast_M = &UsbSBuf[(USB_S_POOL_SIZE+USB_ED_TD_MAX) - 1];
 bufMem.bufNum[0] = (USB_S_POOL_SIZE+USB_ED_TD_MAX);
 #ifdef EMBEDDED_BUF
	bufMem.bufUp[0] = (unsigned long)&UsbSBuf[(USB_S_POOL_SIZE+USB_ED_TD_MAX)];
 #else
    	bufMem.bufUp[0] = UsbSBuf[(USB_S_POOL_SIZE+USB_ED_TD_MAX)-1].buf_startAddr + USB_MEM_U_S_SIZE - 1;
 #endif
 
 #ifdef EMBEDDED_BUF
	bufMem.bufLo[0] = (unsigned long)&UsbSBuf[0];
 #else
    	bufMem.bufLo[0] = UsbSBuf[0].buf_startAddr;
 #endif

 // init time gap rule buf (building free entry pool linking-list)
 for (i = 0 ; i < USB_M_POOL_SIZE ; i++)
 {
	curM = &UsbMBuf[i];
    	curM->isUsed = FALSE;
    	curM->startAddr = (unsigned long)curM;
    	if (i < (USB_M_POOL_SIZE - 1)) 
		curM->next = &UsbMBuf[i + 1];
    	else
		curM->next = NULL;
	#ifndef EMBEDDED_BUF
		curM->buf_number = i ;
		curM->buf_startAddr = (unsigned long) (UsbMPool + USB_MEM_U_M_SIZE * i) ;
	#endif // ndef EMBEDDED_BUF
 }
   
 bufMem.bufCur_T = &UsbMBuf[0];
 bufMem.bufLast_T = &UsbMBuf[USB_M_POOL_SIZE - 1];
 bufMem.bufNum[1] = USB_M_POOL_SIZE;
 #ifdef EMBEDDED_BUF
	bufMem.bufUp[1] = (unsigned long)&UsbMBuf[USB_M_POOL_SIZE];
 #else
    	bufMem.bufUp[1] = UsbMBuf[USB_M_POOL_SIZE-1].buf_startAddr + USB_MEM_U_M_SIZE - 1;
 #endif
 
 #ifdef EMBEDDED_BUF
	bufMem.bufLo[1] = (unsigned long)&UsbMBuf[0];
 #else
    	bufMem.bufLo[1] = UsbMBuf[0].buf_startAddr;
 #endif

 // init same host buf (building free entry pool linking-list)
 for (i = 0 ; i < (USB_L_POOL_SIZE+USB_DMAPOOL_L_MAX) ; i++)
 {
	curL = &UsbLBuf[i];
    	curL->isUsed = FALSE;
    	curL->startAddr = (unsigned long)curL;
    	if (i < ((USB_L_POOL_SIZE+USB_DMAPOOL_L_MAX) - 1)) 
            curL->next = &UsbLBuf[i + 1];
    	else
            curL->next = NULL;
	#ifndef EMBEDDED_BUF
		curL->buf_number = i ;
		curL->buf_startAddr = (unsigned long) (UsbLPool + USB_MEM_U_L_SIZE * i) ;
	#endif // ndef EMBEDDED_BUF
 }
   
 bufMem.bufCur_H = &UsbLBuf[0];
 bufMem.bufLast_H = &UsbLBuf[(USB_L_POOL_SIZE+USB_DMAPOOL_L_MAX) - 1];
 bufMem.bufNum[2] = (USB_L_POOL_SIZE+USB_DMAPOOL_L_MAX);
 #ifdef EMBEDDED_BUF
 	bufMem.bufUp[2] = (unsigned long)&UsbLBuf[(USB_L_POOL_SIZE+USB_DMAPOOL_L_MAX)];
  #else
    	bufMem.bufUp[2] = UsbLBuf[(USB_L_POOL_SIZE+USB_DMAPOOL_L_MAX)-1].buf_startAddr + USB_MEM_U_L_SIZE - 1;
 #endif
 
 #ifdef EMBEDDED_BUF
	bufMem.bufLo[2] = (unsigned long)&UsbLBuf[0];
 #else
    	bufMem.bufLo[2] = UsbLBuf[0].buf_startAddr;
 #endif
 
 usb_hc_res_mem_unlock();
        
 #ifndef EMBEDDED_BUF
	USB_DBG_INFO(iprintf(
		"[USB INFO] UsbSPool: 0x%08x ~ 0x%08x\n", UsbSPool,
        UsbSPool + USB_MEM_U_S_SIZE*(USB_S_POOL_SIZE+USB_ED_TD_MAX) - 1
        ));
	#ifdef USB_BS_ASSERT
		#warning : USB_BS_ASSERT : ON !!

		#ifdef CBAC_TIME_GAP_AREA
    		if  ( !(  ( (unsigned long)UsbSPool + USB_MEM_U_S_SIZE
            			 *(USB_S_POOL_SIZE+USB_ED_TD_MAX)
    					 - 1
		              )
		    		< ( (unsigned long)CBAC_TIME_GAP_AREA)
		        )  )
		        {
		         iprintf("[ASSERT][USB] initUsbPool() : " ) ;
		         iprintf("Error : UsbSPool overlap TimeGap !!\n" ) ;
		         iprintf("[ASSERT][USB] Continue ....\n") ;
		        }
	     #else
			iprintf("[ASSERT][USB] initUsbPool() : " ) ;
		    iprintf("CBAC_TIME_GAP_AREA not defined !!\n" ) ;
		    iprintf("[ASSERT][USB] Continue ....\n") ;
		#endif 
	  #else
        /*
	    #error : Please Make Sure : Not USB_BS_ASSERT
        */
	    // #warning : Please Make Sure : Not USB_BS_ASSERT
	#endif
 #endif // ndef EMBEDDED_BUF

 USB_DBG_INFO(iprintf(
	"[USB INFO] init Small Un-Cached Mem , Have: %u\n",
    bufMem.bufNum[0]
    ));
 USB_DBG_INFO(iprintf(
	"[USB INFO] Small Memory Address: 0x%08x ~ 0x%08x\n",
    UsbSBuf, &UsbSBuf[(USB_S_POOL_SIZE+USB_ED_TD_MAX) + 1]
    ));
    
 #ifdef USB_BS_ASSERT
	#ifdef CBAC_TIME_GAP_AREA
    if (!(  ((unsigned long)&UsbSBuf[(USB_S_POOL_SIZE+USB_ED_TD_MAX) + 1]
	        )
	      < ((unsigned long)CBAC_TIME_GAP_AREA)
       ) )
       {
		iprintf("[ASSERT][USB] initUsbPool() : " ) ;
        iprintf("Error : UsbSBuf overlap TimeGap !!\n" ) ;
        iprintf("[ASSERT][USB] Continue ....\n") ;
       }
     #else
		iprintf("[ASSERT][USB] initUsbPool() : " ) ;
        iprintf("CBAC_TIME_GAP_AREA not defined !!\n" ) ;
        iprintf("[ASSERT][USB] Continue ....\n") ;
    #endif 
 #endif
        
 #ifndef EMBEDDED_BUF
 	iprintf("[USB] UsbSPoolArray: %08XH ~ %08XH\n", UsbSPoolArray, UsbSPoolArray + sizeof(UsbSPoolArray) - 1);
 	iprintf("[USB] UsbMPoolArray: %08XH ~ %08XH\n", UsbMPoolArray, UsbMPoolArray + sizeof(UsbMPoolArray) - 1);
 	iprintf("[USB] UsbLPoolArray: %08XH ~ %08XH\n", UsbLPoolArray, UsbLPoolArray + sizeof(UsbLPoolArray) - 1);
 	
 	iprintf("[USB] Small Un-Cached Memory(%d): %08XH ~ %08XH\n", bufMem.bufNum[0], bufMem.bufLo[0], bufMem.bufUp[0]);
 	iprintf("[USB] Medium Un-Cached Memory(%d): %08XH ~ %08XH\n", bufMem.bufNum[1], bufMem.bufLo[1], bufMem.bufUp[1]);
 	iprintf("[USB] Large Un-Cached Memory(%d): %08XH ~ %08XH\n", bufMem.bufNum[2], bufMem.bufLo[2], bufMem.bufUp[2]);
 #endif // ndef EMBEDDED_BUF       

 return TRUE;
}

//*********************************************************************
// FUNCTION:  showUsbMemU
// PURPOSE:   show current memory pool status
//*********************************************************************
#ifdef USB_DBG_MEM_COUNT_ON
   void showUsbMemU(void)
   {
        iprintf("[USB MEM_COUNT] showUsbMemU() : SmallUnCached, have: %d\n",
        	bufMem.bufNum[0]
            );
        iprintf("[USB MEM_COUNT] showUsbMemU() : MediumUnCached, have: %d\n",
        	bufMem.bufNum[1]
            );
        iprintf("[USB MEM_COUNT] showUsbMemU() : LargeUnCached,  have: %d\n",
        	bufMem.bufNum[2]
            );
   }
#endif // USB_DBG_MEM_COUNT_ON

// ********************************************************************
// FUNCTION:  UsbAllocMemU
// PURPOSE:   get block memory from memory pool
// PARAMETERS:
//                      uint pID        : pool ID
// RETURN:
//     void *: block memory start point,  NULL: get fail 
//                     (error poolID or out of buffer)
// ********************************************************************

#ifdef USB_DBG_MEM_U_ON
	void *UsbAllocMemU(unsigned int pID,unsigned long usize)
 #else
	void *UsbAllocMemU(unsigned int pID)
#endif
{
 void *retVar = NULL;
 unsigned long	usb_alloc_bytes = 0;
 #if defined(USB_DBG_INIT_ON) || defined(USB_DBG_MEM_U)
	char *poolName[3] = {"SmallUnca", "MedUnca", "LrgUnca"};
 #endif

 #ifdef USB_MEM_STATIC_ARRAY
	unsigned long 	retVar_U32 ;
 #endif
        
 #if (USB_DBG_MEM_U_LEVEL >= 2)
	USB_DBG_MEM_U(iprintf(
		"[USB MEM_U] UsbAllocMemU() : get pID=%d, Have: %u\n",
	    pID, bufMem.bufNum[pID-USB_MEM_U_S]
	    ));
 #endif // USB_DBG_MEM_U_LEVEL >= 2
 
 usb_hc_res_mem_lock();
 if (bufMem.bufNum[pID-USB_MEM_U_S])
 {
	if (pID == USB_MEM_U_S)
	{
		bufMem.bufCur_M->isUsed = TRUE;
		#ifdef EMBEDDED_BUF
	        	retVar = (void *)bufMem.bufCur_M->buf;
        	#else
         		retVar = (void *)bufMem.bufCur_M->buf_startAddr;
		#endif
        
		#ifdef USB_MEM_STATIC_ARRAY
			retVar_U32 = (unsigned long) retVar ;
			retVar_U32 = USB_TO_UN_CACHED(retVar_U32) ;
			retVar =  (void *) retVar_U32 ;
		#endif
        
        usb_alloc_bytes = USB_MEM_U_S_SIZE;
        
        	bufMem.bufCur_M = bufMem.bufCur_M->next;
        	if (bufMem.bufCur_M == NULL)
        		bufMem.bufLast_M = NULL;
		#ifdef USB_DBG_MEM_U_ON
          		if(usize > USB_MEM_U_S_SIZE)
           		{
				iprintf(
            			"[USB MEM_U] UsbAllocMemU() : Error : usize = %d > USB_MEM_U_S_SIZE !!\n"
                		,usize
                		) ;
				retVar = NULL;
           		}
		#endif // USB_DBG_MEM_U_ON
	}
    	else if (pID == USB_MEM_U_M)
       	{
        	bufMem.bufCur_T->isUsed = TRUE;
		#ifdef EMBEDDED_BUF
	        	retVar = (void *)bufMem.bufCur_T->buf;
        	#else
         		retVar = (void *)bufMem.bufCur_T->buf_startAddr;
        	#endif
        
		#ifdef USB_MEM_STATIC_ARRAY
			retVar_U32 = (unsigned long) retVar ;
			retVar_U32 = USB_TO_UN_CACHED(retVar_U32) ;
			retVar =  (void *) retVar_U32 ;
		#endif

        usb_alloc_bytes = USB_MEM_U_M_SIZE;

        	bufMem.bufCur_T = bufMem.bufCur_T->next;
        	if (bufMem.bufCur_T == NULL)
        		bufMem.bufLast_T = NULL;
		#ifdef USB_DBG_MEM_U_ON
        		if(usize > USB_MEM_U_M_SIZE)
               		{
            			iprintf(
            				"[USB MEM_U] UsbAllocMemU() : Error : usize = %d > USB_MEM_U_M_SIZE !!\n"
                			,usize
                		) ;
				retVar = NULL;
               		}
		#endif // USB_DBG_MEM_U_ON
	}
    	else if (pID == USB_MEM_U_L)
       	{
        	bufMem.bufCur_H->isUsed = TRUE;
		#ifdef EMBEDDED_BUF
        		retVar = (void *)bufMem.bufCur_H->buf;
         	#else
         		retVar = (void *)bufMem.bufCur_H->buf_startAddr;
        	#endif
        
		#ifdef USB_MEM_STATIC_ARRAY
			retVar_U32 = (unsigned long) retVar ;
			retVar_U32 = USB_TO_UN_CACHED(retVar_U32) ;
			retVar =  (void *) retVar_U32 ;
		#endif
        
		#ifdef USB_MEM_POOL_LARGE_CLEAR
			usb_alloc_bytes = USB_MEM_U_L_SIZE;
		#endif // USB_MEM_POOL_LARGE_CLEAR
        
        	bufMem.bufCur_H = bufMem.bufCur_H->next;
        	if (bufMem.bufCur_H == NULL)
        		bufMem.bufLast_H = NULL;
        
		#ifdef USB_DBG_MEM_U_ON
        		if(usize > USB_MEM_U_L_SIZE)
               		{
	            		iprintf(
            			"[USB MEM_U] UsbAllocMemU() : Error : usize = %d > USB_MEM_U_L_SIZE !!\n"
                    		,usize
                    		) ;
				retVar = NULL;
               		}
		#endif // USB_DBG_MEM_U_ON
	}
    	else
       	{
			usb_hc_res_mem_unlock();
		#ifdef USB_DBG_INIT_ON
             		dprintf("UsbAllocMemU()> Invalid Pool ID %d !!!\n", pID);
		#endif
       		return NULL;
	}
                
	bufMem.bufNum[pID-USB_MEM_U_S]--;

    if (retVar == NULL)
      iprintf("[USB MEM_U] Error: U%d couldn't get memory\n", pID-USB_MEM_U_S);

	#if (USB_DBG_MEM_U_LEVEL >= 2)
		USB_DBG_MEM_U(iprintf(
    		"[USB MEM_U] UsbAllocMemU() : pID=%d pool: %s,  remain: %u\n",
	        pID , poolName[pID-USB_MEM_U_S], bufMem.bufNum[pID-USB_MEM_U_S]
		    ));
	#endif // (USB_DBG_MEM_U_LEVEL >= 1) 
 }
 else
   iprintf("[USB MEM_U] Run out of U%d memory\n", pID-USB_MEM_U_S);
   
 usb_hc_res_mem_unlock();

 if ((retVar != NULL) && (usb_alloc_bytes > 0)) {
   memset(retVar, 0, usb_alloc_bytes);
 }
 
 #if (USB_DBG_MEM_U_LEVEL >= 1)
	USB_DBG_MEM_U(iprintf(
		"[USB MEM_U] UsbAllocMemU() : pID=%d usize=%d ret %08XH\n",
		pID,usize , retVar
	    )) ;
 #endif // (USB_DBG_MEM_U_LEVEL >= 1) 

 #ifdef USB_DBG_MEM_U_ON
 	// showUsbMemU() ;
 #endif // USB_DBG_MEM_U_ON
 
 USB_DBG_MEM(iprintf("[USB MEM] UsbAllocMemU() : ret %08XH \n"
	, retVar
    )) ;
 
 return retVar;
}

//**************************************************************************************************
// FUNCTION:  UsbFreeMemU
// PURPOSE:   release block memory to memory pool
// PARAMETERS:
//                      uint pID        : pool ID
//                      void *point     : memory block start pointer
// RETURN:
//                      SUCCESS: release memory success,  E_INVPID: invalid pool ID,  E_RELMEM: error release memory
//***************************************************************************************************
#ifdef USB_DBG_MEM_U_ON
	int UsbFreeMemU(unsigned int pID, unsigned char *data_addr,int size)

 #else
	int UsbFreeMemU(unsigned int pID, unsigned char *data_addr)
	// int UsbFreeMemU(unsigned int pID )
#endif
{
 int retVar = E_RELMEM;
 unsigned long in = (unsigned long)data_addr ;
 
 struct s_buf_link_s *tmpM;
 struct m_buf_link_s *tmpT;
 struct l_buf_link_s *tmpH;
 #ifndef EMBEDDED_BUF
	int i ;
	struct s_buf_link_s *curS;
	struct m_buf_link_s *curM;
	struct l_buf_link_s *curL;
 #endif // ndef EMBEDDED_BUF
 unsigned char *point ;

 #ifdef USB_MEM_STATIC_ARRAY
	unsigned long point_U32 ;
 #endif
 #ifdef USB_DBG_MEM_U_ON
	char *poolName[3] = {"SmallUnca", "MedUnca", "LrgUnca"};
 #endif
         
 #ifdef USB_MEM_STATIC_ARRAY
	point_U32 = (unsigned long) data_addr ;
	point_U32 = USB_TO_CACHED(point_U32) ;  
    	point = (unsigned char *) point_U32 ;
    	in = (unsigned long) point_U32 ;
 #else
 	point = data_addr ;
 #endif
        
 usb_hc_res_mem_lock();
 
 #if (USB_DBG_MEM_U_LEVEL >= 1)
	USB_DBG_MEM_U(iprintf(
		"[USB MEM_U] UsbFreeMemU() : pID=%d pool:[%s] addr %08XH \n" ,
	    // pID , poolName[pID-USB_MEM_U_S] , point
	    pID , poolName[pID-USB_MEM_U_S] , data_addr
	    ));
 #endif // (USB_DBG_MEM_U_LEVEL >= 1) 

 if (point)
 {
	// check validation
    	if ((in < bufMem.bufLo[pID-USB_MEM_U_S]) || (in > bufMem.bufUp[pID-USB_MEM_U_S]))
       	{
		#ifdef USB_DBG_MEM_U_ON
    			dprintf(
            		"[USB MEM_U] UsbFreeMemU() : Error : [%s], "
                	"Invalid address boundary : %08XH !!!\n",
                	poolName[pID-USB_MEM_U_S], point );
		#endif
    		goto END_CODE;
	}

	if (pID == USB_MEM_U_S)
	{
		#ifdef USB_DBG_MEM_U_ON
          		if(size > USB_MEM_U_S_SIZE)
           		{
				iprintf(
            			"[USB MEM_U] UsbFreeMemU() : Error : size=%d > USB_MEM_U_S!!\n"
                    		,size
                    		) ;
				goto END_CODE;
           		}
		#endif // USB_DBG_MEM_U_ON

		#ifdef EMBEDDED_BUF
			tmpM = (struct s_buf_link_s *)point;
         	#else
			for (i = 0 ; i < (USB_S_POOL_SIZE+USB_ED_TD_MAX) ; i++)
			{
				curS = &UsbSBuf[i];
				if(curS->buf_startAddr == (unsigned long) point)
                   		{
	    				tmpM = curS ;
                    			break ;
                   		}
			} // for i
			if (i == (USB_S_POOL_SIZE+USB_ED_TD_MAX))
               		{
				iprintf("[USB MEM_U] Can't free U%d memory x%08x\n", pID, point);
				USB_DBG_MEM_U(iprintf(
               				"[USB MEM_U] UsbFreeMemU() : invalid addr %08XH "
                    			"!!\n" , point ));
				goto END_CODE;
               		}
        	#endif
		if (
			#ifdef EMBEDDED_BUF
        		(tmpM->startAddr != in) || 
			#endif // EMBEDDED_BUF
        		!tmpM->isUsed)
		{
			#ifdef USB_DBG_MEM_U_ON
     	       			dprintf(
               				"[USB MEM_U] UsbFreeMemU() : Error : pool:[%s], memory is "
                    			"unused !!!\n",
               				poolName[pID-USB_MEM_U_S] ) ;
			#endif
            		// maybe some task duplicate call this function, so..... :(
            		// FIX ME in the future. pert, 2002/06/25
            		//if (!tmpM->isUsed) retVar = SUCCESS;
			goto END_CODE;
           	}
		// release entry & change flags
        	tmpM->isUsed = FALSE;
        	tmpM->next = NULL;
		// Begin 2002/12/19 Pert Bug Fix
        	if (bufMem.bufLast_M)
        		bufMem.bufLast_M->next = tmpM;
        	else
        		bufMem.bufCur_M = tmpM;
        	bufMem.bufLast_M = tmpM;
	}
    	else if (pID == USB_MEM_U_M)
       	{
		#ifdef USB_DBG_MEM_U_ON
			if(size > USB_MEM_U_M_SIZE)
           		{
				iprintf(
            			"[USB MEM_U] UsbFreeMemU() : Error : size=%d > USB_MEM_U_M!!\n"
                    		,size
                    		) ;
				goto END_CODE;
           		}
		#endif // USB_DBG_MEM_U_ON
		#ifdef EMBEDDED_BUF
            		tmpT = (struct m_buf_link_s *)point;
         	#else
			for (i = 0 ; i < USB_M_POOL_SIZE ; i++)
			{
				curM = &UsbMBuf[i];
				if(curM->buf_startAddr == (unsigned long) point)
                   		{
	    				tmpT = curM ;
                    			break ;
                   		}
			} // for i
			if (i == USB_M_POOL_SIZE)
               		{
				iprintf("[USB MEM_U] Can't free U%d memory x%08x\n", pID, point);
				USB_DBG_MEM_U(iprintf(
               				"[USB MEM_U] UsbFreeMemU() : invalid addr %08XH "
                    			"!!\n" , point ));
				goto END_CODE;
               		}
        	#endif
        	if (
			#ifdef EMBEDDED_BUF
		        (tmpT->startAddr != in)	 || 
			#endif // EMBEDDED_BUF
            		!tmpT->isUsed)
            	{
                	#ifdef USB_DBG_MEM_U_ON
        			dprintf(
            			"[USB MEM_U] UsbFreeMemU(): Error : pool:[%s], memory is "
            			"unused !!!\n",
                		poolName[pID-USB_MEM_U_S] );
			#endif
                        // maybe some task duplicate call this function, so..... :(
                        // FIX ME in the future. pert, 2002/06/05
                        //if (!tmpT->isUsed) retVar = SUCCESS;
                        goto END_CODE;
		}	                                      
        	// release entry & change flags
        	tmpT->isUsed = FALSE;
        	tmpT->next = NULL;
		// Begin 2002/12/19 Pert Bug Fix
        	if (bufMem.bufLast_T)
        		bufMem.bufLast_T->next = tmpT;
        	else
        		bufMem.bufCur_T = tmpT;
        	bufMem.bufLast_T = tmpT;
		// End 2002/12/19 Pert Bug Fix
       	}
	else if (pID == USB_MEM_U_L)
       	{
		#ifdef USB_DBG_MEM_U_ON
	 		if(size > USB_MEM_U_L_SIZE)
           		{
				iprintf(
            			"[USB MEM_U] UsbFreeMemU() : Error : size=%d > USB_MEM_U_L!!\n"
                    		,size
                    		) ;
				goto END_CODE;
           		}
		#endif // USB_DBG_MEM_U_ON
		#ifdef EMBEDDED_BUF
            		tmpH = (struct l_buf_link_s *)point;
         	#else
			for (i = 0 ; i < (USB_L_POOL_SIZE+USB_DMAPOOL_L_MAX) ; i++)
			   {
				curL = &UsbLBuf[i];
				if(curL->buf_startAddr == (unsigned long) point)
                   		{
	    				tmpH = curL ;
                    			break ;
                   		}
			   } // for i
			if (i == (USB_L_POOL_SIZE+USB_DMAPOOL_L_MAX))
               		{
				iprintf("[USB MEM_U] Can't free U%d memory x%08x\n", pID, point);
				USB_DBG_MEM_U(iprintf(
               				"[USB MEM_U] UsbFreeMemU() : invalid addr %08XH "
                    			"!!\n" , point ));
				goto END_CODE;
               		}
        	#endif
        	if (
			#ifdef EMBEDDED_BUF
		        (tmpH->startAddr != in) || 
			#endif // EMBEDDED_BUF
	        	!tmpH->isUsed)
	        {
                	#ifdef USB_DBG_MEM_U_ON
            			dprintf(
                			"[USB MEM_U] UsbFreeMemU() : Error : pool:[%s], memory is "
                    			"unused !!!\n", poolName[pID-USB_MEM_U_S] );
			#endif
                        // maybe some task duplicate call this function, so..... :(
                        // FIX ME in the future. pert, 2002/06/05
                        //if (!tmpT->isUsed) retVar = SUCCESS;
                        goto END_CODE;
		}
		                                        
        	// release entry & change flags
        	tmpH->isUsed = FALSE;
        	tmpH->next = NULL;
		// Begin 2002/12/19 Pert Bug Fix
        	if (bufMem.bufLast_H)
        		bufMem.bufLast_H->next = tmpH;
        	else
        		bufMem.bufCur_H = tmpH;
        	bufMem.bufLast_H = tmpH;
		// End 2002/12/19 Pert Bug Fix
	}
	else
       	{
		#ifdef USB_DBG_MEM_U_ON
        		dprintf(
            		"[USB MEM_U] UsbFreeMemU() : Error : Invalid Pool ID %u !!!\n",
                	pID);
		#endif
        	retVar = E_INVPID;
        	goto END_CODE;
	}

	bufMem.bufNum[pID-USB_MEM_U_S]++;
    	retVar = SUCCESS;
 }

 END_CODE:
 
 usb_hc_res_mem_unlock();
 
 #ifdef USB_DBG_MEM_U_ON
 	// showUsbMemU() ;
 #endif // USB_DBG_MEM_U_ON
 
 return retVar;
}

// ====================================================================

// #ifdef USB_CODE_TYPE_DEBUG 
#ifdef USB_DBG_MEM_U_ON
   void DumpUsbMemU(uint pID)
   {
	#ifndef EMBEDDED_BUF
		int i ;
		struct s_buf_link_s *curS;
		struct m_buf_link_s *curM;
		struct l_buf_link_s *curL;
	#endif // ndef EMBEDDED_BUF
	char *poolName[3] = {"SmallUnca", "MedUnca", "LrgUnca"};
	// #ifdef _ADM5120
	#ifdef USB_MEM_STATIC_ARRAY
		unsigned long point_U32 ;
	#endif
 
        
	usb_hc_res_mem_lock();
 
	#if (USB_DBG_MEM_U_LEVEL >= 2)
		iprintf("[USB MEM_U] DumpUsbMemU() : pID=%d pool:[%s]  \n" ,
		    pID , poolName[pID-USB_MEM_U_S] 
    		);
	#endif // USB_DBG_MEM_U_LEVEL >= 2
    
	if (pID == USB_MEM_U_S) {
		#ifdef EMBEDDED_BUF
	    	tmpM = (struct s_buf_link_s *)point;
         #else
			for (i = 0 ; i < (USB_S_POOL_SIZE+USB_ED_TD_MAX) ; i++)
			   {
				curS = &UsbSBuf[i];
        	    if(curS->isUsed)
                  {
					// #ifdef _ADM5120
					#ifdef USB_MEM_STATIC_ARRAY
						point_U32 = (unsigned long) curS->buf_startAddr ;
						/*
						point_U32 |= 0x20000000L ;
						*/
						point_U32 = USB_TO_UN_CACHED(point_U32) ;
						iprintf(
	        	       		"[USB MEM_U] DumpUsbMemU() : pID=%d pool:[%s] addr %08XH !!\n",
							pID , poolName[pID-USB_MEM_U_S], point_U32 );
					 #else
						iprintf(
    	        	   		"[USB MEM_U] DumpUsbMemU() : pID=%d pool:[%s] addr %08XH "
        		            "!!\n" , pID , poolName[pID-USB_MEM_U_S], curS->buf_startAddr );
					#endif
                  }
			   } // for i
        #endif 
       }
    else if (pID == USB_MEM_U_M) {
		#ifdef EMBEDDED_BUF
            tmpT = (struct m_buf_link_s *)point;
         #else
			for (i = 0 ; i < USB_M_POOL_SIZE ; i++)
			   {
				curM = &UsbMBuf[i];
        	    if(curM->isUsed)
                  {
					// #ifdef _ADM5120
					#ifdef USB_MEM_STATIC_ARRAY
						point_U32 = (unsigned long) curM->buf_startAddr ;
						/*
						point_U32 |= 0x20000000L ;
						*/
						point_U32 = USB_TO_UN_CACHED(point_U32) ;
						iprintf(
	        	       		"[USB MEM_U] DumpUsbMemU() : pID=%d pool:[%s] addr %08XH !!\n",
							pID , poolName[pID-USB_MEM_U_S], point_U32 );
					 #else
						iprintf(
    	           		"[USB MEM_U] DumpUsbMemU() : pID=%d pool:[%s] addr %08XH "
        	            "!!\n" , pID , poolName[pID-USB_MEM_U_S] , curM->buf_startAddr );
					#endif
                  }
			   } // for i
        #endif 
       }
	else if (pID == USB_MEM_U_L) {
		#ifdef EMBEDDED_BUF
            tmpH = (struct l_buf_link_s *)point;
         #else
			for (i = 0 ; i < (USB_L_POOL_SIZE+USB_DMAPOOL_L_MAX) ; i++)
			   {
				curL = &UsbLBuf[i];
        	    if(curL->isUsed)  {
					// #ifdef _ADM5120
					#ifdef USB_MEM_STATIC_ARRAY
						point_U32 = (unsigned long) curL->buf_startAddr ;
						/*
						point_U32 |= 0x20000000L ;
						*/
						point_U32 = USB_TO_UN_CACHED(point_U32) ;

						iprintf(
	        	       		"[USB MEM_U] DumpUsbMemU() : pID=%d pool:[%s] addr %08XH !!\n",
							pID , poolName[pID-USB_MEM_U_S], point_U32 );
					 #else
						iprintf(
	        	       		"[USB MEM_U] DumpUsbMemU() : pID=%d pool:[%s] addr %08XH !!\n",
							pID , poolName[pID-USB_MEM_U_S], curL->buf_startAddr );
					#endif
                  }
			   } // for i
        #endif 
       }
      else
       {
        iprintf(
            	"[USB MEM_U] DumpUsbMemU() : Invalid Pool ID %u !!!\n",
                pID);
        goto END_CODE;
       }


	END_CODE:
 
	usb_hc_res_mem_unlock();
 
	#ifdef USB_DBG_MEM_COUNT_ON
		showUsbMemU() ;
	#endif // USB_DBG_MEM_COUNT_ON
   }
#endif // USB_CODE_TYPE_DEBUG 

// ====================================================================
// ====================================================================

struct usb_mem_link_s {
 unsigned int mem_number ;
 struct usb_mem_link_s *next;    // next structure pointer
 unsigned long startAddr;        // this structure start pointer, for err check
 unsigned long mem_startAddr;    // this structure start pointer, for err check
 char isUsed;                                                            // is this buffer structure being used (TRUE/FALSE,  Yes/No)
};


struct memManage_s {                                                    // [0]: match rule buf,  [1]: time gap buf
        unsigned short memNum[3];                                       // total avaliable buffer entry number
        struct usb_mem_link_s *memCur_S;    // current avaliable 'matchRule_s' buffer structure
        struct usb_mem_link_s *memLast_S;   // last avaliable 'matchRule_s' buffer structure
        struct usb_mem_link_s *memCur_M;    // current avaliable 'timeGap_s' buffer structure
        struct usb_mem_link_s *memLast_M;   // last avaliable 'timeGap_s' buffer structure
        struct usb_mem_link_s *memCur_L;    // current avaliable 'hostSame_s' buffer structure
        struct usb_mem_link_s *memLast_L;   // last avaliable 'hostSame_s' buffer structure
        unsigned long memUp[3];                                         // buffer upper boundary
        unsigned long memLo[3];                                         // buffer lower boundary
};

// ====================================================================


static struct memManage_s memMem;

char UsbSMem[(USBMEM_C_S_SZ+1) * USB_MEM_C_S_SIZE];
char UsbMMem[(USBMEM_C_M_SZ+1) * USB_MEM_C_M_SIZE];
char UsbLMem[(USB_L_MEM_BLOCK+1) * USB_MEM_C_L_SIZE];

struct usb_mem_link_s UsbSMemLink[USBMEM_C_S_SZ];
struct usb_mem_link_s UsbMMemLink[USBMEM_C_M_SZ];
struct usb_mem_link_s UsbLMemLink[USB_L_MEM_BLOCK];

// ====================================================================

int UsbMem_init(void)
{
 int i;
 struct usb_mem_link_s *curS;
 struct usb_mem_link_s *curM;
 struct usb_mem_link_s *curL;
 char *ptr;
    
 iprintf("[USB] UsbSMem: %08XH ~ %08XH\n", UsbSMem, (unsigned long)UsbSMem + sizeof(UsbSMem) - 1);
 iprintf("[USB] UsbMMem: %08XH ~ %08XH\n", UsbMMem, (unsigned long)UsbMMem + sizeof(UsbMMem) - 1);  
 iprintf("[USB] UsbLMem: %08XH ~ %08XH\n", UsbLMem, (unsigned long)UsbLMem + sizeof(UsbLMem) - 1);
 
 iprintf("[USB] UsbSMemLink: %08XH ~ %08XH\n", UsbSMemLink, (unsigned long)&UsbSMemLink[USBMEM_C_S_SZ] - 1);          
 iprintf("[USB] UsbMMemLink: %08XH ~ %08XH\n", UsbMMemLink, (unsigned long)&UsbMMemLink[USBMEM_C_M_SZ] - 1);
 iprintf("[USB] UsbLMemLink: %08XH ~ %08XH\n", UsbLMemLink, (unsigned long)&UsbLMemLink[USB_L_MEM_BLOCK] - 1);
 
 memset(&memMem, 0, sizeof(struct memManage_s));
 
 usb_hc_res_mem_lock();
 // init match rule mem (building free entry pool linking-list)
 ptr = (char *)(((unsigned long)UsbSMem + USB_MEM_C_S_SIZE - 1) & ~(USB_MEM_C_S_SIZE - 1));
 for (i=0; i < USBMEM_C_S_SZ; i++)
 {
	curS = &UsbSMemLink[i];
	curS->isUsed = FALSE;
	curS->startAddr = (unsigned long)curS;
	if (i < (USBMEM_C_S_SZ - 1)) 
	{
    		curS->next = &UsbSMemLink[i + 1];
	}
	else
	{
		curS->next = NULL;
	}
	USB_DBG_MEM(iprintf("[USB MEM] UsbMem_init() : curS->next = %08XH\n",curS->next));
	curS->mem_number = i ;
	curS->mem_startAddr = (unsigned long)ptr;
	ptr += USB_MEM_C_S_SIZE;
 }
 memMem.memCur_S = &UsbSMemLink[0];
 memMem.memLast_S = &UsbSMemLink[USBMEM_C_S_SZ-1];
 memMem.memNum[0] = USBMEM_C_S_SZ;
 memMem.memUp[0] = UsbSMemLink[USBMEM_C_S_SZ-1].mem_startAddr + USB_MEM_C_S_SIZE - 1;
 memMem.memLo[0] = UsbSMemLink[0].mem_startAddr;

 // init time gap rule mem (building free entry pool linking-list)
 ptr = (char *)(((unsigned long)UsbMMem + USB_MEM_C_M_SIZE - 1) & ~(USB_MEM_C_M_SIZE - 1));
 for (i=0; i < USBMEM_C_M_SZ; i++)
 {
	curM = &UsbMMemLink[i];
	curM->isUsed = FALSE;
	curM->startAddr = (unsigned long)curM;
	if (i < (USBMEM_C_M_SZ - 1)) 
	{
		curM->next = &UsbMMemLink[i + 1];
	}
	else
	{
        	curM->next = NULL;
	}
	USB_DBG_MEM(iprintf("[USB MEM] UsbMem_init() : curM->next = %08XH\n",curM->next));
	curM->mem_number = i ;
	curM->mem_startAddr = (unsigned long)ptr;
        ptr += USB_MEM_C_M_SIZE;
 }
   
 memMem.memCur_M = &UsbMMemLink[0];
 memMem.memLast_M = &UsbMMemLink[USBMEM_C_M_SZ-1];
 memMem.memNum[1] = USBMEM_C_M_SZ;
 memMem.memUp[1] = UsbMMemLink[USBMEM_C_M_SZ-1].mem_startAddr + USB_MEM_C_M_SIZE - 1;
 memMem.memLo[1] = UsbMMemLink[0].mem_startAddr;

 // init same host mem (building free entry pool linking-list)
 ptr = (char *)(((unsigned long)UsbLMem + USB_MEM_C_L_SIZE - 1) & ~(USB_MEM_C_L_SIZE - 1));
 for (i=0; i < USB_L_MEM_BLOCK; i++)
 {
	curL = &UsbLMemLink[i];
	curL->isUsed = FALSE;
	curL->startAddr = (unsigned long)curL;
	if (i < (USB_L_MEM_BLOCK - 1)) 
	{
        	curL->next = &UsbLMemLink[i + 1];
	}
	else
	{
        	curL->next = NULL;
	}
	USB_DBG_MEM(iprintf("[USB MEM] UsbMem_init() : curL->next = %08XH\n",curL->next));
	curL->mem_number = i ;
	curL->mem_startAddr = (unsigned long)ptr;
	ptr += USB_MEM_C_L_SIZE;
 } 
 memMem.memCur_L = &UsbLMemLink[0];
 memMem.memLast_L = &UsbLMemLink[USB_L_MEM_BLOCK-1];
 memMem.memNum[2] = USB_L_MEM_BLOCK;
 memMem.memUp[2] = UsbLMemLink[USB_L_MEM_BLOCK-1].mem_startAddr + USB_MEM_C_L_SIZE - 1;
 memMem.memLo[2] = UsbLMemLink[0].mem_startAddr;

 usb_hc_res_mem_unlock();

 iprintf("[USB] Small Cache Memory (%d): %08XH ~ %08XH\n", memMem.memNum[0], memMem.memLo[0], memMem.memUp[0]);
 iprintf("[USB] Medium Cache Memory (%d): %08XH ~ %08XH\n", memMem.memNum[1], memMem.memLo[1], memMem.memUp[1]); 
 iprintf("[USB] Large Cache Memory (%d): %08XH ~ %08XH\n", memMem.memNum[2], memMem.memLo[2], memMem.memUp[2]);                

 #ifdef USB_DBG_MEM_C_ON
   // showUsbMemC() ;
 #endif

 return TRUE; 
}


//*********************************************************************
// FUNCTION:  showUsbMemU
// PURPOSE:   show current memory pool status
//*********************************************************************

// #ifdef USB_CODE_TYPE_DEBUG 
#ifdef USB_DBG_MEM_COUNT_ON
   void showUsbMemC(void)
   {
	iprintf("[USB MEM_COUNT] showUsbMemC() : SmallCached, have: %d\n",
        	memMem.memNum[0]
            );
	iprintf("[USB MEM_COUNT] showUsbMemC() : MediumCached, have: %d\n",
        	memMem.memNum[1]
            );
	iprintf("[USB MEM_COUNT] showUsbMemC() : LargeCached,  have: %d\n",
        	memMem.memNum[2]
            );
   }
#endif // USB_CODE_TYPE_DEBUG 


//**************************************************************************************************
// FUNCTION:  reqUsbMem
// PURPOSE:   get block memory from memory pool
// PARAMETERS:
//                      uint pID        : pool ID
// RETURN:
//                      void *: block memory start point,  NULL: get fail (error poolID or out of memfer)
//***************************************************************************************************

// UsbAllocMemC : Cached

#ifdef USB_DBG_MEM_C_ON
	// extern void *reqUsbMem(unsigned int pID,unsigned long usize)
   void *UsbAllocMemC(unsigned int pID,unsigned long usize)
#else
	// extern void *reqUsbMem(unsigned int pID)
   void *UsbAllocMemC(unsigned int pID)
#endif
{
 void *retVar = NULL;
 unsigned long	usb_alloc_bytes = 0;
 #if defined(USB_DBG_INIT_ON) || defined(USB_DBG_MEM_ON) || defined(USB_DBG_MEM_C_ON)
	char *memName[3] = {"SmlCach", "MedCach", "LrgCach"};
 #endif
 
 USB_DBG_MEM(iprintf("[USB MEM] UsbAllocMemC() : pID=%d, Have: %u\n", pID, memMem.memNum[pID-USB_MEM_C_S]));
    
 #if (USB_DBG_MEM_C_LEVEL >= 2)
	USB_DBG_MEM_C(iprintf("[USB MEM_C] Cached , pID=%d usize=%d\n",pID,usize));
 #endif // (USB_DBG_MEM_C_LEVEL >= 1)

 usb_hc_res_mem_lock();
 
 if (memMem.memNum[pID-USB_MEM_C_S])
 {
	if (pID == USB_MEM_C_S)
	{
		#ifdef __US_MEM_LIB_DBG
			iprintf("[USB MEM_U] mem , small\n") ;
        	#endif
        
		memMem.memCur_S->isUsed = TRUE;
        	retVar = (void *) memMem.memCur_S->mem_startAddr;
        	usb_alloc_bytes = USB_MEM_C_S_SIZE;
		memMem.memCur_S = memMem.memCur_S->next;
        	if (memMem.memCur_S == NULL)
			memMem.memLast_S = NULL;
		#ifdef USB_DBG_MEM_C_ON
          		if (usize > USB_MEM_C_S_SIZE)
           		{
				iprintf("[USB MEM_C] UsbAllocMemC() : Error : usize = %d > USB_MEM_C_S_SIZE !!\n",usize);
				retVar = NULL;
           		}
		#endif // USB_DBG_MEM_U_ON
	}
    	else if (pID == USB_MEM_C_M)
	{
		#ifdef __US_MEM_LIB_DBG
			iprintf("[USB MEM_U] mem , medium\n") ;
        	#endif
        	memMem.memCur_M->isUsed = TRUE;
        	retVar = (void *) memMem.memCur_M->mem_startAddr;
        	usb_alloc_bytes = USB_MEM_C_M_SIZE;
        	memMem.memCur_M = memMem.memCur_M->next;
        	if (memMem.memCur_M == NULL)
        		memMem.memLast_M = NULL;
		#ifdef USB_DBG_MEM_C_ON
        		if(usize > USB_MEM_C_M_SIZE)
               		{
            			iprintf("[USB MEM_C] UsbAllocMemC() : Error : usize = %d > USB_MEM_C_M_SIZE !!\n",usize);
				retVar = NULL;
			}
		#endif // USB_DBG_MEM_U_ON
	}
	else if (pID == USB_MEM_C_L)
	{
		#ifdef __US_MEM_LIB_DBG
			iprintf("[USB MEM_U] mem , large\n") ;
        	#endif
		memMem.memCur_L->isUsed = TRUE;
		retVar = (void *) memMem.memCur_L->mem_startAddr;
		#ifdef USB_MEM_POOL_LARGE_CLEAR
			usb_alloc_bytes = USB_MEM_C_L_SIZE;
		#endif // USB_MEM_POOL_LARGE_CLEAR       
		memMem.memCur_L = memMem.memCur_L->next;
        	if (memMem.memCur_L == NULL)
			memMem.memLast_L = NULL;
        
		#ifdef USB_DBG_MEM_C_ON
        		if(usize > USB_MEM_C_L_SIZE)
               		{
	            		iprintf("[USB MEM_C] UsbAllocMemC() : Error : usize = %d > USB_MEM_C_L_SIZE !!\n",usize);
				retVar = NULL;
			}
		#endif // USB_DBG_MEM_U_ON
	}
	else
	{
			usb_hc_res_mem_unlock();
		#ifdef USB_DBG_INIT_ON
			dprintf("UsbAllocMemC()> Invalid Mem ID %d !!!\n", pID);
		#endif
        	return NULL;
        }
                
	memMem.memNum[pID-USB_MEM_C_S]--;

    if (retVar == NULL)
      iprintf("[USB MEM_C] Error: C%d couldn't get memory\n", pID-USB_MEM_C_S);

	#if (USB_DBG_MEM_C_LEVEL >= 2)
		USB_DBG_MEM_C(iprintf("[USB MEM_C] UsbAllocMemC() : mem: %s,  remain: %u\n",memName[pID-USB_MEM_C_S], memMem.memNum[pID-USB_MEM_C_S]));
	#endif // (USB_DBG_MEM_C_LEVEL >= 1)
	#if (USB_DBG_MEM_C_LEVEL >= 1)
		USB_DBG_MEM_C(iprintf("[USB MEM_C] UsbAllocMemC() : [%s] remain:%03u ",memName[pID-USB_MEM_C_S], memMem.memNum[pID-USB_MEM_C_S]));
	#endif // (USB_DBG_MEM_C_LEVEL >= 1)
	USB_DBG_MEM(iprintf("[USB MEM] UsbAllocMemC() : mem: %s,  remain: %u\n",memName[pID-USB_MEM_C_S], memMem.memNum[pID-USB_MEM_C_S]));
 }
 else
   iprintf("[USB MEM_C] Run out of C%d memory\n", pID-USB_MEM_C_S);
   
 usb_hc_res_mem_unlock();

 if ((retVar != NULL) && (usb_alloc_bytes > 0)) {
   memset(retVar, 0, usb_alloc_bytes);
 }

 #ifdef USB_DBG_MEM_C_ON
 	// showUsbMemC() ;
 #endif // USB_DBG_MEM_U_ON
 
 #if (USB_DBG_MEM_C_LEVEL >= 1)
	USB_DBG_MEM_C(iprintf("usize=%d ret=%08XH \n", usize ,retVar));
 #endif // (USB_DBG_MEM_C_LEVEL >= 1)
 #if (USB_DBG_MEM_C_LEVEL >= 2)
	USB_DBG_MEM_C(iprintf("[USB MEM_C] UsbAllocMemC() : pID=%d usize=%d , ret %08XH \n",pID,usize ,retVar));
 #endif // (USB_DBG_MEM_C_LEVEL >= 1)
 USB_DBG_MEM(iprintf("[USB MEM] UsbAllocMemC() : ret %08XH \n", retVar));
 return retVar;
}

//**************************************************************************************************
// FUNCTION:  UsbFreeMemC(
// PURPOSE:   release block memory to memory mem
// PARAMETERS:
//                      uint pID        : mem ID
//                      void *point     : memory block start pointer
// RETURN:
//                      SUCCESS: release memory success,  E_INVPID: invalid mem ID,  E_RELMEM: error release memory
//***************************************************************************************************

#ifdef USB_DBG_MEM_C_ON
	int UsbFreeMemC(uint pID, unsigned char *point,int size)
	// void UsbFreeMemC(void *ptr1,int size)
#else
	// int relUsbMem(uint pID, unsigned char *point)
	int UsbFreeMemC(uint pID, unsigned char *point)
#endif
// int relUsbMem(uint pID, unsigned char *point)
{
 int i ;
 struct usb_mem_link_s *curS;
 struct usb_mem_link_s *curM;
 struct usb_mem_link_s *curL;
 
 int retVar = E_RELMEM;
 unsigned long in = (unsigned long)point;
 struct usb_mem_link_s *tmpS;
 struct usb_mem_link_s *tmpM;
 struct usb_mem_link_s *tmpL;
 #ifdef USB_DBG_MEM_C_ON
	char *memName[3] = {"SmlCach", "MedCach", "LrgCach"};
 #endif
         
 usb_hc_res_mem_lock();
 
 #if (USB_DBG_MEM_C_LEVEL >= 1)
	USB_DBG_MEM_C(iprintf("[USB MEM_C] UsbFreeMemC() :  [%s]                    addr %08XH \n" , memName[pID-USB_MEM_C_S] , point));
 #endif // (USB_DBG_MEM_C_LEVEL >= 1)
    
 if (point)
 {
	// check validation
	if ((in < memMem.memLo[pID-USB_MEM_C_S]) || (in > memMem.memUp[pID-USB_MEM_C_S]))
	{
		#ifdef USB_DBG_MEM_C_ON
    			dprintf(
            		"[USB MEM_C] relUsbMem() : Error : Address = %08XH [%s], "
                	"Invalid address boundary !!!\n"
                	, point , memName[pID-USB_MEM_C_S]);
		#endif
    		goto END_CODE;
	}

	if (pID == USB_MEM_C_S)
	{
		#ifdef USB_DBG_MEM_C_ON
			if(size > USB_MEM_C_S_SIZE)
           		{
				iprintf("[USB MEM_U] relUsbMem() : Error : size=%d > USB_MEM_C_S!!\n",size);
				goto END_CODE;
           		}
		#endif // USB_DBG_MEM_U_ON
        
		for (i=0; i < USBMEM_C_S_SZ ; i++)
		{
			curS = &UsbSMemLink[i];
			if (curS->mem_startAddr == (unsigned long) point)
                   	{
	    			tmpS = curS ;
                    		break ;
                   	}
		} // for i
		if (i == USBMEM_C_S_SZ)
		{
			iprintf("[USB MEM_C] Can't free C%d memory x%08x\n", pID, point);
			#if (USB_DBG_MEM_C_LEVEL >= 0)
				USB_DBG_MEM_C(iprintf(
               				"[USB MEM_C] relUsbMem() : invalid addr %08XH "
                    			"!!\n" , point ));
			#endif // (USB_DBG_MEM_C_LEVEL >= 0)
			goto END_CODE;
		}
		if (!tmpS->isUsed)
		{
			#ifdef USB_DBG_MEM_C_ON
     	       			dprintf("[USB MEM_C] relUsbMem() : Error : mem:[%s], memory is "
                    			"unused !!!\n",
               				memName[pID-USB_MEM_C_S]);
			#endif
            		// maybe some task duplicate call this function, so..... :(
            		// FIX ME in the future. pert, 2002/06/25
	            	//if (!tmpS->isUsed) retVar = SUCCESS;
			goto END_CODE;
           	}

		// release entry & change flags
        	tmpS->isUsed = FALSE;
        	tmpS->next = NULL;
		// Begin 2002/12/19 Pert Bug Fix
        	if (memMem.memLast_S)
			memMem.memLast_S->next = tmpS;
		else
			memMem.memCur_S = tmpS;
        	memMem.memLast_S = tmpS;
		// End 2002/12/19 Pert Bug Fix
	}
    	else if (pID == USB_MEM_C_M)
       	{
		#ifdef USB_DBG_MEM_C_ON
			if(size > USB_MEM_C_M_SIZE)
			{
				iprintf(
            				"[USB MEM_U] relUsbMem() : Error : size=%d > USB_MEM_C_M!!\n"
                    			,size
                    		) ;
				goto END_CODE;
           		}
		#endif // USB_DBG_MEM_U_ON
		for (i = 0 ; i < (USBMEM_C_M_SZ) ; i++)
		{
			curM = &UsbMMemLink[i];
			if(curM->mem_startAddr == (unsigned long) point)
                   	{
	    			tmpM = curM ;
                    		break ;
                   	}
		} // for i
		if (i == (USBMEM_C_M_SZ))
		{
			iprintf("[USB MEM_C] Can't free C%d memory x%08x\n", pID, point);
			#if (USB_DBG_MEM_C_LEVEL >= 0)
				USB_DBG_MEM_C(iprintf(
               			"[USB MEM_C] relUsbMem() : invalid addr %08XH "
                    		"!!\n" , point ));
			#endif // (USB_DBG_MEM_C_LEVEL >= 0)
			goto END_CODE;
		}
		if (!tmpM->isUsed)
		{
                	#ifdef USB_DBG_MEM_C_ON
        			dprintf(
            				"[USB MEM_C] relUsbMem(): Error : mem:[%s], memory is "
            				"unused !!!\n",
                		memName[pID-USB_MEM_C_S]);
			#endif
                        // maybe some task duplicate call this function, so..... :(
                        // FIX ME in the future. pert, 2002/06/05
                        //if (!tmpM->isUsed) retVar = SUCCESS;
                        goto END_CODE;
		}
                                        
		// release entry & change flags
        	tmpM->isUsed = FALSE;
        	tmpM->next = NULL;
		// Begin 2002/12/19 Pert Bug Fix
        	if (memMem.memLast_M)
        		memMem.memLast_M->next = tmpM;
        	else
        		memMem.memCur_M = tmpM;
        	memMem.memLast_M = tmpM;
		// End 2002/12/19 Pert Bug Fix
	}
	else if (pID == USB_MEM_C_L)
	{
		#ifdef USB_DBG_MEMPOL_L_ON
          		if(size > USB_MEM_C_L_SIZE)
           		{
				iprintf(
            			"[USB MEM_U] relUsbMem() : Error : size=%d > USB_MEM_C_L!!\n"
                    		,size
                    		) ;
				goto END_CODE;
           		}
		#endif // USB_DBG_MEM_U_ON
		for (i = 0 ; i < USB_L_MEM_BLOCK ; i++)
		{
			curL = &UsbLMemLink[i];
			if(curL->mem_startAddr == (unsigned long) point)
                   	{
	    			tmpL = curL ;
                    		break ;
                   	}
		} // for i
		if (i == USB_L_MEM_BLOCK)
		{
			iprintf("[USB MEM_C] Can't free C%d memory x%08x\n", pID, point);
			#if (USB_DBG_MEM_C_LEVEL >= 0)
				USB_DBG_MEM_C(iprintf(
               			"[USB MEM_C] relUsbMem() : invalid addr %08XH "
                    		"!!\n" , point ));
			#endif // (USB_DBG_MEM_C_LEVEL >= 0)
			goto END_CODE;
		}
        	if (!tmpL->isUsed)
        	{              
			#ifdef USB_DBG_MEM_C_ON
            			dprintf(
                		"[USB MEM_C] relUsbMem() : Error : mem:[%s], memory is "
                    		"unused !!!\n", memName[pID-USB_MEM_C_S]);
			#endif
                        // maybe some task duplicate call this function, so..... :(
                        // FIX ME in the future. pert, 2002/06/05
                        //if (!tmpM->isUsed) retVar = SUCCESS;
                        goto END_CODE;
                }
                                        
		// release entry & change flags
        	tmpL->isUsed = FALSE;
        	tmpL->next = NULL;
		// Begin 2002/12/19 Pert Bug Fix
        	if (memMem.memLast_L)
        		memMem.memLast_L->next = tmpL;
        	else
        		memMem.memCur_L = tmpL;
        	memMem.memLast_L = tmpL;
		// End 2002/12/19 Pert Bug Fix
	}
      	else
	{
		#ifdef USB_DBG_MEM_C_ON
        		dprintf(
            		"[USB MEM_C] relUsbMem() : Invalid Mem ID %u !!!\n",
                	pID);
		#endif
        	retVar = E_INVPID;
        	goto END_CODE;
	}

	memMem.memNum[pID-USB_MEM_C_S]++;
    	retVar = SUCCESS;
 }

 END_CODE:
 
 usb_hc_res_mem_unlock();
 
 #ifdef USB_DBG_MEM_C_ON
 	// showUsbMemC() ;
 #endif // USB_DBG_MEM_U_ON
 
 return retVar;
}

#if 0
void *UsbAllocMemCs(unsigned int pID, int num)  /* Charles 20040513 */
{
 void *retVar = NULL;
    
 usb_hc_res_mem_lock();
 
 if (memMem.memNum[pID-USB_MEM_C_S] >= num)
 {
	if (pID == USB_MEM_C_S)
	{       
		memMem.memCur_S->isUsed = TRUE;
                retVar = (void *) (UsbSMem + (memMem.memCur_S->mem_number) * USB_MEM_C_S_SIZE);       
		memset(retVar, 0, USB_MEM_C_S_SIZE);        
        	memMem.memCur_S = memMem.memCur_S->next;
        	if (memMem.memCur_S == NULL)
        		memMem.memLast_S = NULL;
	}
	else if (pID == USB_MEM_C_M)
	{
		memMem.memCur_M->isUsed = TRUE;
        	retVar = (void *) (UsbMMem + (memMem.memCur_M->mem_number) * USB_MEM_C_M_SIZE);
		memset(retVar, 0, USB_MEM_C_M_SIZE);       
        	memMem.memCur_M = memMem.memCur_M->next;
        	if (memMem.memCur_M == NULL)
        		memMem.memLast_M = NULL;
	}
	else if (pID == USB_MEM_C_L)
	{           
		memMem.memCur_L->isUsed = TRUE;
		retVar = (void *) (UsbLMem + (memMem.memCur_L->mem_number) * USB_MEM_C_L_SIZE) ;
		#ifdef USB_MEM_POOL_LARGE_CLEAR
			memset(retVar, 0, USB_MEM_C_L_SIZE);
		#endif // USB_MEM_POOL_LARGE_CLEAR      
        	memMem.memCur_L = memMem.memCur_L->next;
        	if (memMem.memCur_L == NULL)
			memMem.memLast_L = NULL;        
	}
	else
	{
        	usb_hc_res_mem_unlock();
		#ifdef USB_DBG_INIT_ON
			dprintf("UsbAllocMemC()> Invalid Mem ID %d !!!\n", pID);
		#endif
		return NULL;
	}
                
	memMem.memNum[pID-USB_MEM_C_S]--;    
 }
   
 usb_hc_res_mem_unlock();

 return retVar;
}

int UsbFreeMemCs(uint pID, unsigned char *point, int num)    /* Charles 20040513 */
{
 int i ;
 int retVar = E_RELMEM;
 unsigned long in = (unsigned long)point;
 struct usb_mem_link_s *tmpS;
 struct usb_mem_link_s *tmpM;
 struct usb_mem_link_s *tmpL;
         
 usb_hc_res_mem_lock();
     
 if (point)
 {
 // check validation
	if ((in < memMem.memLo[pID-USB_MEM_C_S]) || (in > memMem.memUp[pID-USB_MEM_C_S]))
	{
    		iprintf("UsbFreeMemCs : Error : [%d] Address = %08XH\n", pID, point);
    		goto END_CODE;
	}

	if (pID == USB_MEM_C_S)
	{       
		//dprintf("match block valid boundary !!! \n");
		for (i = 0 ; i < (USBMEM_C_S_SZ) ; i++)
		{
			curS = &UsbSMemLink[i];
			if (UsbSMemLink[i].mem_startAddr == (unsigned long) point)
			{
	    			tmpS = &UsbSMemLink[i];
				break;
			}
		} // for i
		
		if (i == (USBMEM_C_S_SZ))
		{
			iprintf("UsbFreeMemCs : Invalid [%d] Address %08XH\n", pID, point);
			goto END_CODE;
		}
		
		if (!tmpS->isUsed)
		{
			iprintf("UsbFreeMemCs : [%d] memory is unused!!!\n", pID);
			goto END_CODE;
		}
                                        
		// release entry & change flags
        	tmpS->isUsed = FALSE;
        	tmpS->next = NULL;
		// Begin 2002/12/19 Pert Bug Fix
        	if (memMem.memLast_S)
			memMem.memLast_S->next = tmpS;
        	else
			memMem.memCur_S = tmpS;
        	memMem.memLast_S = tmpS;
	}
    else if (pID == USB_MEM_C_M)
       {
		#ifdef USB_DBG_MEM_C_ON
		// #ifdef ETCPIP_BS_ASSERT
          if(size > USB_MEM_C_M_SIZE)
           {
			iprintf(
            	"[USB MEM_U] relUsbMem() : Error : size=%d > USB_MEM_C_M!!\n"
                    ,size
                    ) ;
			goto END_CODE;
           }
		#endif // USB_DBG_MEM_U_ON
		//dprintf("time block valid boundary !!!\n");
			for (i = 0 ; i < (USBMEM_C_M_SZ) ; i++)
			   {
				curM = &UsbMMemLink[i];
				if(curM->mem_startAddr == (unsigned long) point)
                   {
	    			tmpM = curM ;
                    break ;
                   }
			   } // for i
			if (i == (USBMEM_C_M_SZ))
               {
				#if (USB_DBG_MEM_C_LEVEL >= 0)
				USB_DBG_MEM_C(iprintf(
               		"[USB MEM_C] relUsbMem() : invalid addr %08XH "
                    "!!\n" , point ));
				#endif // (USB_DBG_MEM_C_LEVEL >= 0)
				goto END_CODE;
               }
        if (
            !tmpM->isUsed) {
                                
		#ifdef USB_DBG_MEM_C_ON
        	dprintf(
            	"[USB MEM_C] relUsbMem(): Error : mem:[%s], memory is "
            	"unused !!!\n",
                memName[pID-USB_MEM_C_S]);
		#endif
                                // maybe some task duplicate call this function, so..... :(
                                // FIX ME in the future. pert, 2002/06/05
                                //if (!tmpM->isUsed) retVar = SUCCESS;
                                
                                goto END_CODE;
                        }
		//dprintf("time block valid checksum & use flag !!!\n");
                                        
        // release entry & change flags
        tmpM->isUsed = FALSE;
        tmpM->next = NULL;
		// Begin 2002/12/19 Pert Bug Fix
        if (memMem.memLast_M) memMem.memLast_M->next = tmpM;
        else memMem.memCur_M = tmpM;
        memMem.memLast_M = tmpM;
		// End 2002/12/19 Pert Bug Fix
       }
	else if (pID == USB_MEM_C_L)
       {
		#ifdef USB_DBG_MEMPOL_L_ON
		// #ifdef ETCPIP_BS_ASSERT
          if(size > USB_MEM_C_L_SIZE)
           {
			iprintf(
            	"[USB MEM_U] relUsbMem() : Error : size=%d > USB_MEM_C_L!!\n"
                    ,size
                    ) ;
			goto END_CODE;
           }
		#endif // USB_DBG_MEM_U_ON
		//dprintf("LargeMem block valid boundary !!!\n");
			for (i = 0 ; i < USB_L_MEM_BLOCK ; i++)
			   {
				curL = &UsbLMemLink[i];
				if(curL->mem_startAddr == (unsigned long) point)
                   {
	    			tmpL = curL ;
                    break ;
                   }
			   } // for i
			if (i == USB_L_MEM_BLOCK)
               {
				#if (USB_DBG_MEM_C_LEVEL >= 0)
				USB_DBG_MEM_C(iprintf(
               		"[USB MEM_C] relUsbMem() : invalid addr %08XH "
                    "!!\n" , point ));
				#endif // (USB_DBG_MEM_C_LEVEL >= 0)
				goto END_CODE;
               }
        if (
	        !tmpL->isUsed) {
                                
			#ifdef USB_DBG_MEM_C_ON
            	dprintf(
                	"[USB MEM_C] relUsbMem() : Error : mem:[%s], memory is "
                    "unused !!!\n", memName[pID-USB_MEM_C_S]);
			#endif
                                // maybe some task duplicate call this function, so..... :(
                                // FIX ME in the future. pert, 2002/06/05
                                //if (!tmpM->isUsed) retVar = SUCCESS;
                                
                                goto END_CODE;
                        }
		//dprintf("LargeMem block valid checksum & use flag !!!\n");
                                        
        // release entry & change flags
        tmpL->isUsed = FALSE;
        tmpL->next = NULL;
		// Begin 2002/12/19 Pert Bug Fix
        if (memMem.memLast_L) memMem.memLast_L->next = tmpL;
        else memMem.memCur_L = tmpL;
        memMem.memLast_L = tmpL;
		// End 2002/12/19 Pert Bug Fix
       }
      else
       {
		#ifdef USB_DBG_MEM_C_ON
        	dprintf(
            	"[USB MEM_C] relUsbMem() : Invalid Mem ID %u !!!\n",
                pID);
		#endif
        retVar = E_INVPID;
        goto END_CODE;
       }

	memMem.memNum[pID-USB_MEM_C_S]++;
    retVar = SUCCESS;
   }

 END_CODE:
 
 usb_hc_res_mem_unlock();
 
 #ifdef USB_DBG_MEM_C_ON
 	// showUsbMemC() ;
 #endif // USB_DBG_MEM_U_ON
 
 return retVar;
}
#endif

// #ifdef USB_CODE_TYPE_DEBUG 
#ifdef USB_DBG_MEM_C_ON
   void DumpUsbMemC(uint pID)
   {
	int i ;
	struct usb_mem_link_s *curP;
	char *memName[3] = {"SmlCach", "MedCach", "LrgCach"};
 
        
	usb_hc_res_mem_lock();
 
	iprintf("[USB MEM_C] DumpUsbMemC() : mem:[%s] \n" ,
	    memName[pID-USB_MEM_C_S]
	    );
    
	if (pID == USB_MEM_C_S)
	   {
        
		for (i = 0 ; i < (USBMEM_C_S_SZ) ; i++)
		   {
			curP = &UsbSMemLink[i];
                
	        if (curP->isUsed)
               {
				iprintf(
               		"[USB MEM_C] DumpUsbMemC() :  mem:[%s] addr %08XH "
                    "!!\n" , memName[pID-USB_MEM_C_S] , curP->mem_startAddr );
               }
		   } // for i
       }
    else if (pID == USB_MEM_C_M)
       {
		for (i = 0 ; i < (USBMEM_C_M_SZ) ; i++)
		   {
			curP = &UsbMMemLink[i];
                
	        if (curP->isUsed)
               {
				iprintf(
               		"[USB MEM_C] DumpUsbMemC() :  mem:[%s] addr %08XH "
                    "!!\n" , memName[pID-USB_MEM_C_S] , curP->mem_startAddr );
               }
		   } // for i
       }
	else if (pID == USB_MEM_C_L)
       {
		//dprintf("LargeMem block valid boundary !!!\n");
		for (i = 0 ; i < USB_L_MEM_BLOCK ; i++)
		   {
			curP = &UsbLMemLink[i];
                
	        if (curP->isUsed)
               {
				iprintf(
               		"[USB MEM_C] DumpUsbMemC() :  mem:[%s] addr %08XH "
                    "!!\n" , memName[pID-USB_MEM_C_S] , curP->mem_startAddr );
               }
		   } // for i
       }
      else
       {
        dprintf("[USB MEM_C] DumpUsbMemC() : Invalid Mem ID %u !!!\n",
                pID);
        goto END_CODE;
       }


	END_CODE:
 
	usb_hc_res_mem_unlock();
 
	#ifdef USB_DBG_MEM_COUNT_ON
		showUsbMemC() ;
	#endif // USB_DBG_MEM_COUNT_ON
   }
#endif // USB_CODE_TYPE_DEBUG 


// ====================================================================
// ====================================================================

/*
// #if (_UPNP <= 0)
#ifdef USB_MEM_LIB

   #ifdef USB_DBG_MEM_U_ON
	   static int alloc_count = 0 ;
       static int mem_usage =  0 ;
       static int mem_usage_max =  0 ;
	   static int alloc_max = 0 ;
	   static int alloc_min = 10000 ;
   #endif // USB_DBG_MEM_U_ON
   
   // =================================================================
   void *UsbAllocMemC(int size)
   {
	void * p = 0;
	int	i;
	unsigned int type;
	

	#ifdef USB_DBG_MEM_U_ON
		alloc_count ++ ;
        if (size > alloc_max)
           {
            alloc_max = size ;   
           }
        if (size < alloc_min)
           {
            alloc_min = size ;   
           }
        mem_usage +=  size ;
        if (mem_usage > mem_usage_max)
           {
            mem_usage_max = mem_usage ;   
           }
    	iprintf("[USB MEM_U] UsbAllocMemC() : alloc_count=%d "
        		"size=%d mem_usage=%d\n",
        		alloc_count,size,mem_usage) ;
    	iprintf("[USB MEM_U] alloc_max=%d alloc_min=%d "
        		"mem_usage_max=%d\n",
        		alloc_max,alloc_min,mem_usage_max ) ;
	#endif // USB_DBG_MEM_U_ON
    
	i = size % sizeof(long);
	if (i!=0) 
		size += sizeof(long)-i;		//For data alignment and remain a little space 900829    
	#if 0
	if (size <= 16) {
		p = reqmem(0);
		if (p == NULL) dprintf("*****UsbAllocMemC: upnp_b0_mem: out of memory\n");
	}
	if ( (p == NULL) && (size <= 64) ) {
		p = reqmem(1);
		if (p == NULL) dprintf("****UsbAllocMemC: upnp_b1_mem: out of memory...\n");
	}
	if (p == NULL) { 
	#endif		
     
	 //if (stateVarInit)
	 //	type = LOCAL;
	 //else
     
	 	type = GLOBAL;
	 
		//if (size > 256) dprintf("[%d]alloc_mem %d bytes\n", cur_task, size);
    		p = reqmem(type | COLOR1, size);
	//}

	if ( p == 0 )
       {
    	 USB_DBG_MEM_U(iprintf(
			"[USB MEM_U] UsbAllocMemC: Error : [task %d] out of memory, %d !!\n",
            cur_task, size
            ));
		// _chkmem(COLOR1, "alloc_mem");
        
		p = reqmem(type | COLOR0, size);
        
    	 USB_DBG_MEM_U(iprintf(
			"[USB MEM_U] UsbAllocMemC: reqmem by COLOR0 : %08XH \n",
        	p
            ));
	   }
    else if ((long)p & 3) {
		dprintf("UsbAllocMemC: %d bytes from COLOR1, return %p\n", size, p);
		_chkmem(COLOR1, "UsbAllocMemC");
		p = reqmem(type | COLOR0, size);
	} else goto alloc_OK;

	if(p) {                               // got some 
		if (((long)p & 3) != 0) {
			dprintf("UsbAllocMemC: %d bytes from COLOR0, return %p\n", size, p);
			return NULL;
		}
    // ===============
    alloc_OK:
    // ===============
    
		memset(p,0,size);               // clear memory 
		return p;
	} else 
		dprintf("allocate memory fail from COLOR0 !!!\n");
	//else                                // none left 
	//	errno = ENOMEM;
	return NULL;	
   } // function UsbAllocMemC()

   // =================================================================
   #ifdef USB_DBG_MEM_U_ON
		void UsbFreeMemC(void *ptr1,int size)
    #else
		void UsbFreeMemC(void *ptr1)
   #endif
   {
	char *ptr = (char *)ptr1;
	int retVal;
	
	#ifdef USB_DBG_MEM_U_ON
		alloc_count -- ;
        mem_usage -=  size ;
    	iprintf("[USB MEM_U] UsbFreeMemC() : alloc_count=%d size=%d mem_usage=%d\n",
        		alloc_count,size,mem_usage) ;
	#endif // USB_DBG_MEM_U_ON
    
	if (ptr == NULL) return;
	#if 0
	if ( (ptr >= upnp_b0_mem) && (ptr < upnp_b0_end) )
		relmem(0, ptr);
	else if ( (ptr >= upnp_b1_mem) && (ptr < upnp_b1_end) )
		relmem(1, ptr);
	else
	#endif
	if ((long)ptr & 3)
       {
		dprintf("UsbFreeMemC: at %p\n", ptr);
		return;
	   }
	
	retVal = relmem(ptr);
	if (retVal != 0) dprintf("UsbFreeMemC: relmem return %d\n", retVal);
	return;
   } // function UsbFreeMemC()
   
#endif // _UPNP <= 0
*/

// ====================================================================
// End of File : UsbMemLib.c
