
// include/asm-generic/bitops.h

#ifndef _ASM_GENERIC_BITOPS_H_
#define _ASM_GENERIC_BITOPS_H_


// ====================================================================

// from include/asm-mips/bitops.h

#define _MIPS_SZLONG 32

#if (_MIPS_SZLONG == 32)
	#define SZLONG_LOG 5
	#define SZLONG_MASK 31UL
	#define __LL    "ll"
	#define __SC    "sc"
#elif (_MIPS_SZLONG == 64)
	#define SZLONG_LOG 6
	#define SZLONG_MASK 63UL
	#define __LL    "lld"
	#define __SC    "scd"
#endif




/*
 * find_next_zero_bit - find the first zero bit in a memory region
 * @addr: The address to base the search on
 * @offset: The bitnumber to start searching at
 * @size: The maximum size to search
 */
/*
static inline unsigned long find_next_zero_bit(unsigned long *addr,
        unsigned long size, unsigned long offset)
{
        unsigned long *p = ((unsigned long *) addr) + (offset >> SZLONG_LOG);
        unsigned long result = offset & ~SZLONG_MASK;
        unsigned long tmp;

        if (offset >= size)
                return size;
        size -= result;
        offset &= SZLONG_MASK;
        if (offset) {
                tmp = *(p++);
                tmp |= ~0UL >> (_MIPS_SZLONG-offset);
                if (size < _MIPS_SZLONG)
                        goto found_first;
                if (~tmp)
                        goto found_middle;
                size -= _MIPS_SZLONG;
                result += _MIPS_SZLONG;
        }
        while (size & ~SZLONG_MASK) {
                if (~(tmp = *(p++)))
                        goto found_middle;
                result += _MIPS_SZLONG;
                size -= _MIPS_SZLONG;
        }
        if (!size)
                return result;
        tmp = *p;

found_first:
        tmp |= ~0UL << size;
        if (tmp == ~0UL)                // Are any bits zero? 
                return result + size;   // Nope. 
found_middle:
        return result + ffz(tmp);
}
*/

#define find_first_zero_bit(addr, size) \
        find_next_zero_bit((addr), (size), 0)

/*
 * __test_and_clear_bit - Clear a bit and return its old value
 * @nr: Bit to clear
 * @addr: Address to count from
 *
 * This operation is non-atomic and can be reordered.
 * If two examples of this operation race, one can appear to succeed
 * but actually fail.  You must protect multiple accesses with a lock.
 */
/*
static inline int __test_and_clear_bit(unsigned long nr,
        volatile unsigned long * addr)
{
 volatile unsigned long *a = addr;
 unsigned long mask;
 int retval;
 a += (nr >> SZLONG_LOG);

 mask = 1UL << (nr & SZLONG_MASK);
 retval = ((mask & *a) != 0);
 *a &= ~mask;

 return retval;
}
*/
// 2004/1/12 B. S. Liaw ..
static inline int __test_and_clear_bit(unsigned long nr, volatile void *addr)
{
 int retval;


 retval = ((
               ((unsigned char *) addr)[nr >> 3]
			 & (1U << (nr & 7))
		  ) != 0) ;

 ((unsigned char *) addr)[nr >> 3] &= ~(1U << (nr & 7));
 return retval;
}


// ====================================================================


/*
 * For the benefit of those who are trying to port Linux to another
 * architecture, here are some C-language equivalents.  You should
 * recode these in the native assembly language, if at all possible.
 * To guarantee atomicity, these routines call cli() and sti() to
 * disable interrupts while they operate.  (You have to provide inline
 * routines to cli() and sti().)
 *
 * Also note, these routines assume that you have 32 bit longs.
 * You will have to change this if you are trying to port Linux to the
 * Alpha architecture or to a Cray.  :-)
 * 
 * C language equivalents written by Theodore Ts'o, 9/26/92
 */
/* 
extern __inline__ int set_bit(int nr,long * addr)
{
	int	mask, retval;

	addr += nr >> 5;
	mask = 1 << (nr & 0x1f);
	cli();
	retval = (mask & *addr) != 0;
	*addr |= mask;
	sti();
	return retval;
}

extern __inline__ int clear_bit(int nr, long * addr)
{
	int	mask, retval;

	addr += nr >> 5;
	mask = 1 << (nr & 0x1f);
	cli();
	retval = (mask & *addr) != 0;
	*addr &= ~mask;
	sti();
	return retval;
}
*/
static inline int test_bit(int nr, long * addr)
{
	int	mask;

	addr += nr >> 5;
	mask = 1 << (nr & 0x1f);
	return ((mask & *addr) != 0);
}

/*
 * fls: find last bit set.
 */

#define fls(x) generic_fls(x)

#ifdef __KERNEL__

/*
 * ffs: find first bit set. This is defined the same way as
 * the libc and compiler builtin ffs routines, therefore
 * differs in spirit from the above ffz (man ffs).
 */

#define ffs(x) generic_ffs(x)

/*
 * hweightN: returns the hamming weight (i.e. the number
 * of bits set) of a N-bit word
 */

#define hweight32(x) generic_hweight32(x)
#define hweight16(x) generic_hweight16(x)
#define hweight8(x) generic_hweight8(x)

#endif /* __KERNEL__ */

#endif /* _ASM_GENERIC_BITOPS_H */
