
// include/linux/fs.h

#ifndef _LINUX_FS_H
#define _LINUX_FS_H

/*
 * This file has definitions for some important file table
 * structures etc.
 */
#ifdef _NEWLIB
	#include "string.h"
#endif
/*
// #ifdef _ETCPIP
#define __KERNEL__
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
		#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
// #endif // _ETCPIP
*/

/*
#include <linux/config.h>
#include <linux/linkage.h>
#include <linux/limits.h>
*/
/*
#include <linux/wait.h>
*/
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // linux/wait.h , osport_wait.h
#endif

/*
#include <linux/types.h>
#include <linux/kdev_t.h>
*/
#include "kdev_t.h"
/*
#include <linux/ioctl.h>
*/
/*
#include <linux/list.h>
*/
#include "PoList.h" // include/linux/list.h
/*
	#include <linux/dcache.h>
*/
/*
#include <linux/stat.h>
*/
#include "stat.h"

/*
#include <linux/cache.h>
#include <linux/radix-tree.h>
*/
/*
#include <linux/kobject.h>
*/
#include "kobject.h"
/*
#include <asm/atomic.h>
*/

struct iovec;
struct nameidata;
struct pipe_inode_info;
struct poll_table_struct;
struct kstatfs;
#ifndef USB_NO_VMM // vitrtul memory management
	struct vm_area_struct;
#endif // ndef USB_NO_VMM
#ifndef NO_VFS  // struct vfsmount
	struct vfsmount;
#endif // ndef NO_VFS

/*
 * It's silly to have NR_OPEN bigger than NR_FILE, but you can change
 * the file limit at runtime and only root can increase the per-process
 * nr_file rlimit, so it's safe to set up a ridiculously high absolute
 * upper limit on files-per-process.
 *
 * Some programs (notably those using select()) may have to be 
 * recompiled to take full advantage of the new limits..  
 */

/* Fixed constants first: */
#undef NR_OPEN
#define NR_OPEN (1024*1024)	/* Absolute upper limit on fd num */
#define INR_OPEN 1024		/* Initial setting for nfile rlimits */

#define BLOCK_SIZE_BITS 10
#define BLOCK_SIZE (1<<BLOCK_SIZE_BITS)

/* And dynamically-tunable limits and defaults: */
struct files_stat_struct {
	int nr_files;		/* read only */
	int nr_free_files;	/* read only */
	int max_files;		/* tunable */
};
extern struct files_stat_struct files_stat;

struct inodes_stat_t {
	int nr_inodes;
	int nr_unused;
	int dummy[5];
};
extern struct inodes_stat_t inodes_stat;

extern int leases_enable, dir_notify_enable, lease_break_time;

#define NR_FILE  8192	/* this can well be larger on a larger system */
#define NR_RESERVED_FILES 10 /* reserved for root */
#define NR_SUPER 256

#define MAY_EXEC 1
#define MAY_WRITE 2
#define MAY_READ 4
#define MAY_APPEND 8

#define FMODE_READ 1
#define FMODE_WRITE 2

#define RW_MASK		1
#define RWA_MASK	2
#define READ 0
#define WRITE 1
#define READA 2		/* read-ahead  - don't block if no resources */
#define SPECIAL 4	/* For non-blockdevice requests in request queue */

#define SEL_IN		1
#define SEL_OUT		2
#define SEL_EX		4

/* public flags for file_system_type */
#define FS_REQUIRES_DEV 1 
#define FS_REVAL_DOT	16384	/* Check the paths ".", ".." for staleness */
#define FS_ODD_RENAME	32768	/* Temporary stuff; will go away as soon
				  * as nfs_rename() will be cleaned up
				  */
/*
 * These are the fs-independent mount-flags: up to 32 flags are supported
 */
#define MS_RDONLY	 1	/* Mount read-only */
#define MS_NOSUID	 2	/* Ignore suid and sgid bits */
#define MS_NODEV	 4	/* Disallow access to device special files */
#define MS_NOEXEC	 8	/* Disallow program execution */
#define MS_SYNCHRONOUS	16	/* Writes are synced at once */
#define MS_REMOUNT	32	/* Alter flags of a mounted FS */
#define MS_MANDLOCK	64	/* Allow mandatory locks on an FS */
#define MS_DIRSYNC	128	/* Directory modifications are synchronous */
#define MS_NOATIME	1024	/* Do not update access times. */
#define MS_NODIRATIME	2048	/* Do not update directory access times */
#define MS_BIND		4096
#define MS_MOVE		8192
#define MS_REC		16384
#define MS_VERBOSE	32768
#define MS_POSIXACL	(1<<16)	/* VFS does not apply the umask */
#define MS_ONE_SECOND	(1<<17)	/* fs has 1 sec a/m/ctime resolution */
#define MS_ACTIVE	(1<<30)
#define MS_NOUSER	(1<<31)

/*
 * Superblock flags that can be altered by MS_REMOUNT
 */
#define MS_RMT_MASK	(MS_RDONLY|MS_SYNCHRONOUS|MS_MANDLOCK|MS_NOATIME|\
			 MS_NODIRATIME)

/*
 * Old magic mount flag and mask
 */
#define MS_MGC_VAL 0xC0ED0000
#define MS_MGC_MSK 0xffff0000

/* Inode flags - they have nothing to superblock flags now */

#define S_SYNC		1	/* Writes are synced at once */
#define S_NOATIME	2	/* Do not update access times */
#define S_QUOTA		4	/* Quota initialized for file */
#define S_APPEND	8	/* Append-only file */
#define S_IMMUTABLE	16	/* Immutable file */
#define S_DEAD		32	/* removed, but still open directory */
#define S_NOQUOTA	64	/* Inode is not counted to quota */
#define S_DIRSYNC	128	/* Directory modifications are synchronous */

/*
 * Note that nosuid etc flags are inode-specific: setting some file-system
 * flags just means all the inodes inherit those flags by default. It might be
 * possible to override it selectively if you really wanted to with some
 * ioctl() that is not currently implemented.
 *
 * Exception: MS_RDONLY is always applied to the entire file system.
 *
 * Unfortunately, it is possible to change a filesystems flags with it mounted
 * with files in use.  This means that all of the inodes will not have their
 * i_flags updated.  Hence, i_flags no longer inherit the superblock mount
 * flags, so these have to be checked separately. -- rmk@arm.uk.linux.org
 */
#ifndef NO_INODE
	#define __IS_FLG(inode,flg) ((inode)->i_sb->s_flags & (flg))

	#define IS_RDONLY(inode) ((inode)->i_sb->s_flags & MS_RDONLY)
#endif // ndef NO_INODE

#ifndef INODE_REDUCED  
	// for ->i_flags
	#define IS_SYNC(inode)		(__IS_FLG(inode, MS_SYNCHRONOUS) || \
					((inode)->i_flags & S_SYNC))
	#define IS_DIRSYNC(inode)	(__IS_FLG(inode, MS_SYNCHRONOUS|MS_DIRSYNC) || \
					((inode)->i_flags & (S_SYNC|S_DIRSYNC)))
#endif // ndef INODE_REDUCED  

#ifndef NO_INODE
	#define IS_MANDLOCK(inode)	__IS_FLG(inode, MS_MANDLOCK)
#endif // ndef NO_INODE

#ifndef INODE_REDUCED  
	// for ->i_flags
	#define IS_QUOTAINIT(inode)	((inode)->i_flags & S_QUOTA)
	#define IS_NOQUOTA(inode)	((inode)->i_flags & S_NOQUOTA)
	#define IS_APPEND(inode)	((inode)->i_flags & S_APPEND)
	#define IS_IMMUTABLE(inode)	((inode)->i_flags & S_IMMUTABLE)
	#define IS_NOATIME(inode)	(__IS_FLG(inode, MS_NOATIME) || ((inode)->i_flags & S_NOATIME))
#endif // ndef INODE_REDUCED  

#ifndef NO_INODE
	#define IS_NODIRATIME(inode)	__IS_FLG(inode, MS_NODIRATIME)
	#define IS_POSIXACL(inode)	__IS_FLG(inode, MS_POSIXACL)
	#define IS_ONE_SECOND(inode)	__IS_FLG(inode, MS_ONE_SECOND)
#endif // ndef NO_INODE

#ifndef INODE_REDUCED  
	#define IS_DEADDIR(inode)	((inode)->i_flags & S_DEAD)
#endif // ndef INODE_REDUCED  

/* the read-only stuff doesn't really belong here, but any other place is
   probably as bad and I don't want to create yet another include file. */

#define BLKROSET   _IO(0x12,93)	/* set device read-only (0 = read-write) */
#define BLKROGET   _IO(0x12,94)	/* get read-only status (0 = read_write) */
#define BLKRRPART  _IO(0x12,95)	/* re-read partition table */
#define BLKGETSIZE _IO(0x12,96)	/* return device size /512 (long *arg) */
#define BLKFLSBUF  _IO(0x12,97)	/* flush buffer cache */
#define BLKRASET   _IO(0x12,98)	/* set read ahead for block device */
#define BLKRAGET   _IO(0x12,99)	/* get current read ahead setting */
#define BLKFRASET  _IO(0x12,100)/* set filesystem (mm/filemap.c) read-ahead */
#define BLKFRAGET  _IO(0x12,101)/* get filesystem (mm/filemap.c) read-ahead */
#define BLKSECTSET _IO(0x12,102)/* set max sectors per request (ll_rw_blk.c) */
#define BLKSECTGET _IO(0x12,103)/* get max sectors per request (ll_rw_blk.c) */
#define BLKSSZGET  _IO(0x12,104)/* get block device sector size */
#if 0
#define BLKPG      _IO(0x12,105)/* See blkpg.h */

/* Some people are morons.  Do not use sizeof! */

#define BLKELVGET  _IOR(0x12,106,size_t)/* elevator get */
#define BLKELVSET  _IOW(0x12,107,size_t)/* elevator set */
/* This was here just to show that the number is taken -
   probably all these _IO(0x12,*) ioctls should be moved to blkpg.h. */
#endif
/* A jump here: 108-111 have been used for various private purposes. */
#define BLKBSZGET  _IOR(0x12,112,size_t)
#define BLKBSZSET  _IOW(0x12,113,size_t)
#define BLKGETSIZE64 _IOR(0x12,114,size_t)	/* return device size in bytes (u64 *arg) */

#define BMAP_IOCTL 1		/* obsolete - kept for compatibility */
#define FIBMAP	   _IO(0x00,1)	/* bmap access */
#define FIGETBSZ   _IO(0x00,2)	/* get the block size used for bmap */

#ifdef __KERNEL__

#ifndef _ETCPIP
	#include <asm/semaphore.h>
#endif // ndef _ETCPIP
/*
#include <asm/byteorder.h>
*/
#include "byteorder/byteorder-mips.h" // include/asm-mips/byteorder.h


#ifndef NO_INODE  
    // B. S. Liaw : add this to stop complier complain 
    //			    `struct inode' declared inside parameter list
	struct inode ;
#endif // ndef NO_INODE

#ifndef NO_INODE
    /* Used to be a macro which just called the function, now just a function */
	extern void update_atime (struct inode *);
#endif // ndef NO_INODE

#ifndef INODE_REDUCED
	extern void inode_init(unsigned long);
 #else
	extern void inode_init(void) ;
#endif
extern void mnt_init(unsigned long);
extern void files_init(unsigned long);

struct buffer_head;

#ifndef NO_INODE
	typedef int (get_block_t)(struct inode *inode, sector_t iblock,
			struct buffer_head *bh_result, int create);
	typedef int (get_blocks_t)(struct inode *inode, sector_t iblock,
			unsigned long max_blocks,
			struct buffer_head *bh_result, int create);
    typedef void (dio_iodone_t)(struct inode *inode, loff_t offset,
			ssize_t bytes, void *private);
#endif // ndef NO_INODE 

/*
 * Attribute flags.  These should be or-ed together to figure out what
 * has been changed!
 */
#define ATTR_MODE	1
#define ATTR_UID	2
#define ATTR_GID	4
#define ATTR_SIZE	8
#define ATTR_ATIME	16
#define ATTR_MTIME	32
#define ATTR_CTIME	64
#define ATTR_ATIME_SET	128
#define ATTR_MTIME_SET	256
#define ATTR_FORCE	512	/* Not a change, but a change it */
#define ATTR_ATTR_FLAG	1024
#define ATTR_KILL_SUID	2048
#define ATTR_KILL_SGID	4096

/*
 * This is the Inode Attributes structure, used for notify_change().  It
 * uses the above definitions as flags, to know which values have changed.
 * Also, in this manner, a Filesystem can look at only the values it cares
 * about.  Basically, these are the attributes that the VFS layer can
 * request to change from the FS layer.
 *
 * Derek Atkins <warlord@MIT.EDU> 94-10-20
 */
struct iattr {
	unsigned int	ia_valid;
	umode_t		ia_mode;
	uid_t		ia_uid;
	gid_t		ia_gid;
	loff_t		ia_size;
	/* comment out By B. S. Liaw
	struct timespec	ia_atime;
	struct timespec	ia_mtime;
	struct timespec	ia_ctime;
	*/
	unsigned int	ia_attr_flags;
};

/*
 * This is the inode attributes flag definitions
 */
#define ATTR_FLAG_SYNCRONOUS	1 	/* Syncronous write */
#define ATTR_FLAG_NOATIME	2 	/* Don't update atime */
#define ATTR_FLAG_APPEND	4 	/* Append-only file */
#define ATTR_FLAG_IMMUTABLE	8 	/* Immutable file */
#define ATTR_FLAG_NODIRATIME	16 	/* Don't update atime for directory */

/*
 * Includes for diskquotas.
 */
#ifndef _ETCPIP
	#include <linux/quota.h>
#endif // ndef _ETCPIP

/*
 * oh the beauties of C type declarations.
 */
/*
#ifndef USB_NO_VMM // vitrtul memory management
	// struct page is on include/linux/mm.h
	struct page;
#endif // ndef USB_NO_VMM
*/
struct address_space;
#ifndef NO_VFS
	struct writeback_control;
#endif // ndef NO_VFS
struct kiocb;

struct address_space_operations {
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	// struct page is on include/linux/mm.h
	int (*writepage)(struct page *page, struct writeback_control *wbc);
	int (*readpage)(struct file *, struct page *);
	int (*sync_page)(struct page *);
  #else
 */
 /*
 int (*writepage)(struct inode *inode,
				unsigned char *buf, struct writeback_control *wbc);
 */
 int (*writepage)(struct inode *inode,
				sector_t iblock , // block number on Disk Partition
				unsigned int num_sectors ,
				unsigned char *buf ) ;
 int (*readpage)(struct inode *inode,
				sector_t iblock , // block number on Disk Partition
				unsigned int num_sectors ,
				unsigned char *buf ) ;
 // int (*sync_page)(unsigned char *buf);
 /*
 #endif 
 */
 /* Write back some dirty pages from this mapping. */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	 int (*writepages)(struct address_space *, struct writeback_control *);
 #endif // ndef USB_NO_VMM
 */
 /* Set a page dirty */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	// struct page is on include/linux/mm.h
	int (*set_page_dirty)(struct page *page);
  #else
 */
	int (*set_page_dirty)(unsigned char *buf);
 /*
 #endif 
 */
 int (*readpages)(struct file *filp, struct address_space *mapping,
			struct list_head *pages, unsigned nr_pages);

 /*
  * ext3 requires that a successful prepare_write() call be followed
  * by a commit_write() call - they must be balanced
  */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	// struct page is on include/linux/mm.h
	int (*prepare_write)(struct file *, struct page *, unsigned, unsigned);
	int (*commit_write)(struct file *, struct page *, unsigned, unsigned);
  #else
	int (*prepare_write)(struct file *, unsigned char *buf, unsigned, unsigned);
	int (*commit_write)(struct file *, unsigned char *buf, unsigned, unsigned);
 #endif 
 */
 /* Unfortunately this kludge is needed for FIBMAP. Don't use it */
 sector_t (*bmap)(struct address_space *, sector_t);
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	// struct page is on include/linux/mm.h
	int (*invalidatepage) (struct page *, unsigned long);
	int (*releasepage) (struct page *, int);
  #else
 */
	int (*invalidatepage) (unsigned char *buf, unsigned long);
	int (*releasepage) (unsigned char *buf, int);
 /*
 #endif 
 */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	int (*direct_IO)(int, struct kiocb *, const struct iovec *iov,
			loff_t offset, unsigned long nr_segs);
 #endif // ndef USB_NO_VMM
 */
};

#ifndef NO_BACKDEV
	struct backing_dev_info;
#endif // ndef NO_BACKDEV

struct address_space {
 #ifndef NO_INODE
	struct inode		*host;		/* owner: inode, block_device */
 #endif // ndef NO_INODE
 #ifndef USB_NO_VMM // vitrtul memory management
	struct radix_tree_root	page_tree;	/* radix tree of all pages */
 #endif // ndef USB_NO_VMM
 #ifndef NO_SPINLOCK
	// spinlock_t : on spinlock.h
	spinlock_t		page_lock;	/* and spinlock protecting it */
 #endif // ndef NO_SPINLOCK
 #ifndef USB_NO_VMM // vitrtul memory management
	struct list_head	clean_pages;	/* list of clean pages */
	struct list_head	dirty_pages;	/* list of dirty pages */
	struct list_head	locked_pages;	/* list of locked pages */
	struct list_head	io_pages;	/* being prepared for I/O */
	unsigned long		nrpages;	/* number of total pages */
 #endif // ndef USB_NO_VMM

 struct address_space_operations *a_ops;	/* methods */

 #ifndef USB_NO_VMM // vitrtul memory management
	struct list_head	i_mmap;		/* list of private mappings */
	struct list_head	i_mmap_shared;	/* list of shared mappings */
 #endif // ndef USB_NO_VMM
 #ifndef NO_SEMPHORE
	struct semaphore	i_shared_sem;	/* protect both above lists */
 #endif // ndef NO_SEMPHORE
 #ifndef USB_NO_VMM // vitrtul memory management
	atomic_t		truncate_count;	/* Cover race condition with truncate */
	unsigned long		dirtied_when;	/* jiffies of first page dirtying */
	unsigned long		flags;		/* error bits/gfp mask */
 #endif // ndef USB_NO_VMM
 #ifndef NO_BACKDEV 
	struct backing_dev_info *backing_dev_info; /* device readahead, etc */
 #endif // ndef NO_BACKDEV 
  
 #ifndef NO_SPINLOCK
	spinlock_t		private_lock;	/* for use by the address_space */
 #endif // ndef NO_SPINLOCK
 #ifndef USB_NO_VMM // vitrtul memory management
	struct list_head	private_list;	/* ditto */
	struct address_space	*assoc_mapping;	/* ditto */
 #endif // ndef USB_NO_VMM
};

struct block_device {
 dev_t    bd_dev;  /* not a kdev_t - it's a search key */
 #ifndef NO_INODE
	struct inode *		bd_inode;	/* will die */
 #endif // ndef NO_INODE

 int			bd_openers;
 #ifndef NO_SEMPHORE
	struct semaphore	bd_sem;	/* open/close mutex */
 #endif // ndef NO_SEMPHORE
 #ifndef INODE_REDUCED  
	struct list_head	bd_inodes;
 #endif // ndef INODE_REDUCED  
 void *			bd_holder;
 int			bd_holders;
 struct block_device *	bd_contains;
 #ifndef INODE_REDUCED  
	unsigned		bd_block_size;
 #endif // ndef INODE_REDUCED  
 struct hd_struct *	bd_part;
 unsigned		bd_part_count;
 int			bd_invalidated;
 struct gendisk *	bd_disk;
 struct list_head	bd_list;
};

/*
 * Use sequence counter to get consistent i_size on 32-bit processors.
 */
#if BITS_PER_LONG==32 && defined(CONFIG_SMP)
#include <linux/seqlock.h>
#define __NEED_I_SIZE_ORDERED
#define i_size_ordered_init(inode) seqcount_init(&inode->i_size_seqcount)
#else
#define i_size_ordered_init(inode) do { } while (0)
#endif

#ifndef NO_INODE  
   struct inode {
	#ifndef INODE_REDUCED
		struct hlist_node	i_hash;
		struct list_head	i_list;
		struct list_head	i_dentry;
	#endif // ndef INODE_REDUCED

	unsigned long		i_ino;

	#ifndef INODE_REDUCED
		atomic_t		i_count;
	#endif // ndef INODE_REDUCED
	umode_t			i_mode;
	#ifndef INODE_REDUCED
		unsigned int		i_nlink;
		uid_t			i_uid;
		gid_t			i_gid;
	#endif // ndef INODE_REDUCED

	dev_t			i_rdev;
	loff_t			i_size;

	#ifndef INODE_REDUCED
		/* comment out By B. S. Liaw
		struct timespec		i_atime;
		struct timespec		i_mtime;
		struct timespec		i_ctime;
		*/
		unsigned int		i_blkbits;
		/*
		unsigned long		i_blksize;
		unsigned long		i_version;
		unsigned long		i_blocks;
		unsigned short      i_bytes;
		*/
	#endif // ndef INODE_REDUCED
	#ifndef NO_SPINLOCK
		spinlock_t		i_lock;	/* i_blocks, i_bytes, maybe i_size */
	#endif // ndef NO_SPINLOCK
	#ifndef NO_SEMPHORE
		struct semaphore	i_sem;
	#endif // ndef NO_SEMPHORE
	#ifndef INODE_REDUCED
		struct inode_operations	*i_op;
		struct file_operations	*i_fop;	/* former ->i_op->default_file_ops */
	#endif // ndef INODE_REDUCED
	struct super_block	*i_sb;
	#ifndef INODE_REDUCED
		struct file_lock	*i_flock;
	#endif // ndef INODE_REDUCED

	struct address_space	*i_mapping;
	struct address_space	i_data;

	#ifndef INODE_REDUCED
		struct dquot		*i_dquot[MAXQUOTAS];
	#endif // ndef INODE_REDUCED
	/* These three should probably be a union */
	#ifndef INODE_REDUCED
		struct list_head	i_devices;
		struct pipe_inode_info	*i_pipe;
	#endif // ndef INODE_REDUCED
	struct block_device	*i_bdev;
	#ifndef INODE_REDUCED
		struct cdev		*i_cdev;
		int			i_cindex;

		unsigned long		i_dnotify_mask; /* Directory notify events */
		struct dnotify_struct	*i_dnotify; /* for directory notifications */

	#endif // ndef INODE_REDUCED

	unsigned long		i_state;

	#ifndef INODE_REDUCED
		unsigned int		i_flags;
		unsigned char		i_sock;

		atomic_t		i_writecount;
		void			*i_security;
		__u32			i_generation;
		union {
			void		*generic_ip;
		} u;
	#endif // ndef INODE_REDUCED
	#ifdef __NEED_I_SIZE_ORDERED
		seqcount_t		i_size_seqcount;
	#endif
   };
#endif // ndef NO_INODE  
/*
 * NOTE: in a 32bit arch with a preemptable kernel and
 * an UP compile the i_size_read/write must be atomic
 * with respect to the local cpu (unlike with preempt disabled),
 * but they don't need to be atomic with respect to other cpus like in
 * true SMP (so they need either to either locally disable irq around
 * the read or for example on x86 they can be still implemented as a
 * cmpxchg8b without the need of the lock prefix). For SMP compiles
 * and 64bit archs it makes no difference if preempt is enabled or not.
 */
#ifndef NO_INODE
static inline loff_t i_size_read(struct inode *inode)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] i_size_read() Begin .. \n"));
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] i_size_read() : inode->i_size=%d (%08XH)\n",
	inode->i_size, inode->i_size ));

 #if BITS_PER_LONG==32 && defined(CONFIG_SMP)
	loff_t i_size;
	unsigned int seq;

	do {
		seq = read_seqcount_begin(&inode->i_size_seqcount);
		i_size = inode->i_size;
	} while (read_seqcount_retry(&inode->i_size_seqcount, seq));
	return i_size;
 #elif BITS_PER_LONG==32 && defined(CONFIG_PREEMPT)
	loff_t i_size;

	preempt_disable();
	i_size = inode->i_size;
	preempt_enable();
	return i_size;
 #else
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] i_size_read() End. \n"));
	return inode->i_size;
 #endif
}
#endif // ndef NO_INODE


#ifndef NO_INODE
static inline void i_size_write(struct inode *inode, loff_t i_size)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] i_size_write() Begin .. \n")) ;
 #if BITS_PER_LONG==32 && defined(CONFIG_SMP)
	write_seqcount_begin(&inode->i_size_seqcount);
	inode->i_size = i_size;
	write_seqcount_end(&inode->i_size_seqcount);
 #elif BITS_PER_LONG==32 && defined(CONFIG_PREEMPT)
	preempt_disable();
	inode->i_size = i_size;
	preempt_enable();
 #else
	inode->i_size = i_size;
 #endif
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] i_size_write() : Set Value : inode->i_size=%ld (%08XH) \n",
	inode->i_size, inode->i_size ));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] i_size_write() End. \n")) ;
}
#endif // ndef NO_INODE

#ifndef NO_INODE  
   static inline unsigned iminor(struct inode *inode)
   {
	return MINOR(inode->i_rdev);
   }

   static inline unsigned imajor(struct inode *inode)
   {
	return MAJOR(inode->i_rdev);
   }
#endif // ndef NO_INODE  

struct fown_struct {
 #ifndef NO_SPINLOCK
	// rwlock_t : on spinlock.h
	rwlock_t lock;          /* protects pid, uid, euid fields */
 #endif // ndef NO_SPINLOCK
	int pid;		/* pid or -pgrp where SIGIO should be sent */
	uid_t uid, euid;	/* uid/euid of process setting the owner */
	int signum;		/* posix.1b rt signal to be delivered on IO */
	void *security;
};

/*
 * Track a single file's readahead state
 */
struct file_ra_state {
	unsigned long start;		/* Current window */
	unsigned long size;
	unsigned long next_size;	/* Next window size */
	unsigned long prev_page;	/* Cache last read() position */
	unsigned long ahead_start;	/* Ahead window */
	unsigned long ahead_size;
	unsigned long ra_pages;		/* Maximum readahead window */
	unsigned long mmap_hit;		/* Cache hit stat for mmap accesses */
	unsigned long mmap_miss;	/* Cache miss stat for mmap accesses */
};

struct file {
	struct list_head	f_list;
	struct dentry		*f_dentry;
 #ifndef NO_VFS  // struct vfsmount
	struct vfsmount         *f_vfsmnt;
 #endif // ndef NO_VFS
	struct file_operations	*f_op;
	atomic_t		f_count;
	unsigned int 		f_flags;
	mode_t			f_mode;
	loff_t			f_pos;
	struct fown_struct	f_owner;
	unsigned int		f_uid, f_gid;
	int			f_error;
	struct file_ra_state	f_ra;

	unsigned long		f_version;
	void			*f_security;

	/* needed for tty driver, and maybe others */
	void			*private_data;

	/* Used by fs/eventpoll.c to link all the hooks to this file */
	struct list_head	f_ep_links;
 #ifndef NO_SPINLOCK
	spinlock_t		f_ep_lock;
 #endif // ndef NO_SPINLOCK
};

 #ifndef NO_SPINLOCK
	extern spinlock_t files_lock;
#endif // ndef NO_SPINLOCK

#ifndef NO_SPINLOCK
	#define file_list_lock() spin_lock(&files_lock);
	#define file_list_unlock() spin_unlock(&files_lock);
#endif // ndef NO_SPINLOCK

#define get_file(x)	atomic_inc(&(x)->f_count)
#define file_count(x)	atomic_read(&(x)->f_count)

/* Initialize and open a private file and allocate its security structure. */
extern int open_private_file(struct file *, struct dentry *, int);
/* Release a private file and free its security structure. */
extern void close_private_file(struct file *file);

#define	MAX_NON_LFS	((1UL<<31) - 1)

/* Page cache limit. The filesystems should put that into their s_maxbytes 
   limits, otherwise bad things can happen in VM. */ 
#if BITS_PER_LONG==32
 #ifndef USB_NO_VMM // no vitrtul memory management
	#define MAX_LFS_FILESIZE	(((u64)PAGE_CACHE_SIZE << (BITS_PER_LONG-1))-1) 
 #endif // ndef USB_NO_VMM
#elif BITS_PER_LONG==64
#define MAX_LFS_FILESIZE 	0x7fffffffffffffff
#endif

#define FL_POSIX	1
#define FL_FLOCK	2
#define FL_ACCESS	8	/* not trying to lock, just looking */
#define FL_LOCKD	16	/* lock held by rpc.lockd */
#define FL_LEASE	32	/* lease held on this file */
#define FL_SLEEP	128	/* A blocking lock */

/*
 * The POSIX file lock owner is determined by
 * the "struct files_struct" in the thread group
 * (or NULL for no owner - BSD locks).
 *
 * Lockd stuffs a "host" pointer into this.
 */
typedef struct files_struct *fl_owner_t;

/* that will die - we need it for nfs_lock_info */
#ifndef _ETCPIP
	#include <linux/nfs_fs_i.h>
#endif // ndef _ETCPIP

struct file_lock {
	struct file_lock *fl_next;	/* singly linked list for this inode  */
	struct list_head fl_link;	/* doubly linked list of all locks */
	struct list_head fl_block;	/* circular list of blocked processes */
	fl_owner_t fl_owner;
	unsigned int fl_pid;
 #ifndef NO_WAITQUEUE
	wait_queue_head_t fl_wait;
 #endif // ndef NO_WAITQUEUE
	struct file *fl_file;
	unsigned char fl_flags;
	unsigned char fl_type;
	loff_t fl_start;
	loff_t fl_end;

	void (*fl_notify)(struct file_lock *);	/* unblock callback */
	void (*fl_insert)(struct file_lock *);	/* lock insertion callback */
	void (*fl_remove)(struct file_lock *);	/* lock removal callback */

	struct fasync_struct *	fl_fasync; /* for lease break notifications */
	unsigned long fl_break_time;	/* for nonblocking lease breaks */
	/* comment out By B. S. Liaw
	union {
		struct nfs_lock_info	nfs_fl;
	} fl_u;
	*/
};

/* The following constant reflects the upper bound of the file/locking space */
#ifndef OFFSET_MAX
#define INT_LIMIT(x)	(~((x)1 << (sizeof(x)*8 - 1)))
#define OFFSET_MAX	INT_LIMIT(loff_t)
#define OFFT_OFFSET_MAX	INT_LIMIT(off_t)
#endif

extern struct list_head file_lock_list;

/*
#include <linux/fcntl.h>
*/
#include "fcntl.h"
/*
extern int fcntl_getlk(struct file *, struct flock __user *);
extern int fcntl_setlk(struct file *, unsigned int, struct flock __user *);
*/
extern int fcntl_getlk(struct file *, struct flock *);
extern int fcntl_setlk(struct file *, unsigned int, struct flock *);

#if BITS_PER_LONG == 32
/*
extern int fcntl_getlk64(struct file *, struct flock64 __user *);
extern int fcntl_setlk64(struct file *, unsigned int, struct flock64 __user *);
*/
extern int fcntl_getlk64(struct file *, struct flock64  *);
extern int fcntl_setlk64(struct file *, unsigned int, struct flock64  *);
#endif

extern void send_sigio(struct fown_struct *fown, int fd, int band);
extern int fcntl_setlease(unsigned int fd, struct file *filp, long arg);
extern int fcntl_getlease(struct file *filp);

/* fs/locks.c */
extern void locks_init_lock(struct file_lock *);
extern void locks_copy_lock(struct file_lock *, struct file_lock *);
extern void locks_remove_posix(struct file *, fl_owner_t);
extern void locks_remove_flock(struct file *);
extern struct file_lock *posix_test_lock(struct file *, struct file_lock *);
extern int posix_lock_file(struct file *, struct file_lock *);
extern void posix_block_lock(struct file_lock *, struct file_lock *);
extern void posix_unblock_lock(struct file *, struct file_lock *);
extern int posix_locks_deadlock(struct file_lock *, struct file_lock *);
#ifndef NO_INODE
	extern int __break_lease(struct inode *inode, unsigned int flags);
	extern void lease_get_mtime(struct inode *, struct timespec *time);
	extern int lock_may_read(struct inode *, loff_t start, unsigned long count);
	extern int lock_may_write(struct inode *, loff_t start, unsigned long count);
#endif // ndef NO_INODE

struct fasync_struct {
	int	magic;
	int	fa_fd;
	struct	fasync_struct	*fa_next; /* singly linked list */
	struct	file 		*fa_file;
};

#define FASYNC_MAGIC 0x4601

/* SMP safe fasync helpers: */
extern int fasync_helper(int, struct file *, int, struct fasync_struct **);
/* can be called from interrupts */
extern void kill_fasync(struct fasync_struct **, int, int);
/* only for net: no internal synchronization */
extern void __kill_fasync(struct fasync_struct *, int, int);

extern int f_setown(struct file *filp, unsigned long arg, int force);
extern void f_delown(struct file *filp);
extern int send_sigurg(struct fown_struct *fown);

/*
 *	Umount options
 */

#define MNT_FORCE	0x00000001	/* Attempt to forcibily umount */
#define MNT_DETACH	0x00000002	/* Just detach from the tree */

extern struct list_head super_blocks;
/*
#ifndef NO_SPINLOCK
	extern spinlock_t sb_lock;
#endif // ndef NO_SPINLOCK
*/
#define sb_entry(list)	list_entry((list), struct super_block, s_list)
#define S_BIAS (1<<30)
struct super_block {
 struct list_head	s_list;		/* Keep this first */
 dev_t			s_dev;		/* search index; _not_ kdev_t */
 unsigned long		s_blocksize;
 unsigned long		s_old_blocksize;
 #ifndef INODE_REDUCED
	unsigned char		s_blocksize_bits;
 #endif // ndef INODE_REDUCED
 unsigned char		s_dirt;
 unsigned long long	s_maxbytes;	/* Max file size */
 struct file_system_type	*s_type;
 struct super_operations	*s_op;
 struct dquot_operations	*dq_op;
 struct quotactl_ops	*s_qcop;
 struct export_operations *s_export_op;
 unsigned long		s_flags;
 unsigned long		s_magic;
 struct dentry		*s_root;
 #ifndef NO_SEMPHORE
	struct rw_semaphore	s_umount;
	struct semaphore	s_lock;
 #endif // ndef NO_SEMPHORE
 int			s_count;
 int			s_syncing;
 int			s_need_sync_fs;
 atomic_t		s_active;
 void                    *s_security;

 struct list_head	s_dirty;	/* dirty inodes */
 struct list_head	s_io;		/* parked for writeback */
 struct hlist_head	s_anon;		/* anonymous dentries for (nfs) exporting */
 struct list_head	s_files;

 struct block_device	*s_bdev;
 struct list_head	s_instances;
 /* comment out By B. S. Liaw
 struct quota_info	s_dquot;	// Diskquota specific options 
 */
 char s_id[32];				/* Informational name */

 struct kobject           kobj;          /* anchor for sysfs */
 void 			*s_fs_info;	/* Filesystem private info */

 /*
  * The next field is for VFS *only*. No filesystems have any business
  * even looking at it. You had been warned.
  */
 #ifndef NO_SEMPHORE
	struct semaphore s_vfs_rename_sem;	/* Kludge */
 #endif // ndef NO_SEMPHORE
};

/*
 * Superblock locking.
 */
static inline void lock_super(struct super_block * sb)
{
 #ifndef NO_SEMPHORE
	down(&sb->s_lock);
 #endif // ndef NO_SEMPHORE
}

static inline void unlock_super(struct super_block * sb)
{
 #ifndef NO_SEMPHORE
	up(&sb->s_lock);
 #endif // ndef NO_SEMPHORE
}

/*
 * VFS helper functions..
 */
#ifndef INODE_REDUCED  
    extern int vfs_create(struct inode *, struct dentry *, int, struct nameidata *);
	extern int vfs_mkdir(struct inode *, struct dentry *, int);
	extern int vfs_mknod(struct inode *, struct dentry *, int, dev_t);
	extern int vfs_symlink(struct inode *, struct dentry *, const char *);
	extern int vfs_link(struct dentry *, struct inode *, struct dentry *);
	extern int vfs_rmdir(struct inode *, struct dentry *);
	extern int vfs_unlink(struct inode *, struct dentry *);
	extern int vfs_rename(struct inode *, struct dentry *, struct inode *, struct dentry *);
#endif // ndef INODE_REDUCED  

/*
 * File types
 *
 * NOTE! These match bits 12..15 of stat.st_mode
 * (ie "(i_mode >> 12) & 15").
 */
#define DT_UNKNOWN	0
#define DT_FIFO		1
#define DT_CHR		2
#define DT_DIR		4
#define DT_BLK		6
#define DT_REG		8
#define DT_LNK		10
#define DT_SOCK		12
#define DT_WHT		14

/*
 * This is the "filldir" function type, used by readdir() to let
 * the kernel specify what kind of dirent layout it wants to have.
 * This allows the kernel to read directories into kernel space or
 * to have different dirent layouts depending on the binary type.
 */
typedef int (*filldir_t)(void *, const char *, int, loff_t, ino_t, unsigned);

struct block_device_operations {
 #ifndef NO_INODE
	int (*open) (struct inode *, struct file *);
	int (*release) (struct inode *, struct file *);
	#ifndef USBMS_NO_IOCTL
		int (*ioctl) (struct inode *, struct file *, unsigned, unsigned long);
	#endif // ndef USBMS_NO_IOCTL
  #else
	int (*open) (struct block_device *bdev , struct file *);
	int (*release) (struct block_device *bdev , struct file *);
	#ifndef USBMS_NO_IOCTL
		int (*ioctl) (struct block_device *bdev , struct file *, unsigned, unsigned long);
	#endif // ndef USBMS_NO_IOCTL
 #endif
 int (*media_changed) (struct gendisk *);
 int (*revalidate_disk) (struct gendisk *);
 /*
 #ifndef _ETCPIP
	struct module *owner;
 #endif // ndef _ETCPIP
 */
};

/*
 * "descriptor" for what we're up to with a read for sendfile().
 * This allows us to use the same read code yet
 * have multiple different users of the data that
 * we read from a file.
 *
 * The simplest case just copies the data to user
 * mode.
 */
typedef struct {
	size_t written;
	size_t count;
	/*
	char __user * buf;
	*/
	char * buf;
	int error;
} read_descriptor_t;

#ifndef USB_NO_VMM // vitrtul memory management
	// struct page is on include/linux/mm.h
	typedef int (*read_actor_t)(read_descriptor_t *, struct page *, unsigned long, unsigned long);
 #else
	typedef int (*read_actor_t)(read_descriptor_t *, unsigned char *buf, unsigned long, unsigned long);
#endif 

/*
 * NOTE:
 * read, write, poll, fsync, readv, writev can be called
 *   without the big kernel lock held in all filesystems.
 */
struct file_operations {
	/*
	#ifndef _ETCPIP
		struct module *owner;
	#endif // ndef _ETCPIP
	*/
	loff_t (*llseek) (struct file *, loff_t, int);
	/*
	ssize_t (*read) (struct file *, char __user *, size_t, loff_t *);
	ssize_t (*aio_read) (struct kiocb *, char __user *, size_t, loff_t);
	ssize_t (*write) (struct file *, const char __user *, size_t, loff_t *);
	ssize_t (*aio_write) (struct kiocb *, const char __user *, size_t, loff_t);
	*/
	ssize_t (*read) (struct file *, char *, size_t, loff_t *);
	ssize_t (*aio_read) (struct kiocb *, char *, size_t, loff_t);
	ssize_t (*write) (struct file *, const char *, size_t, loff_t *);
	ssize_t (*aio_write) (struct kiocb *, const char *, size_t, loff_t);
	int (*readdir) (struct file *, void *, filldir_t);
 /*
 #ifndef _ETCPIP
	unsigned int (*poll) (struct file *, struct poll_table_struct *);
 #endif // ndef _ETCPIP
 */
 #ifndef USBMS_NO_IOCTL
 #ifndef NO_INODE
	int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long);
  #else
	int (*ioctl) (struct file *, unsigned int, unsigned long);
 #endif
 #endif // ndef USBMS_NO_IOCTL

 #ifndef USB_NO_VMM // vitrtul memory management
	int (*mmap) (struct file *, struct vm_area_struct *);
 #endif // ndef USB_NO_VMM

 #ifndef NO_INODE
	int (*open) (struct inode *, struct file *);
  #else
	// int (*open) (struct gendisk *disk , struct file *);
	int (*open) (struct file *);
 #endif

 int (*flush) (struct file *);

 #ifndef NO_INODE
	int (*release) (struct inode *, struct file *);
  #else
	// int (*release) (struct gendisk *disk, struct file *);
	int (*release) (struct file *);
 #endif

 int (*fsync) (struct file *, struct dentry *, int datasync);
 int (*aio_fsync) (struct kiocb *, int datasync);
 int (*fasync) (int, struct file *, int);
 int (*lock) (struct file *, int, struct file_lock *);
 ssize_t (*readv) (struct file *, const struct iovec *, unsigned long, loff_t *);
 ssize_t (*writev) (struct file *, const struct iovec *, unsigned long, loff_t *);
 /*
 ssize_t (*sendfile) (struct file *, loff_t *, size_t, read_actor_t, void __user *);
 */
 ssize_t (*sendfile) (struct file *, loff_t *, size_t, read_actor_t, void *);

 #ifndef USB_NO_VMM // vitrtul memory management
	// struct page is on include/linux/mm.h
	ssize_t (*sendpage) (struct file *, struct page *, int, size_t, loff_t *, int);
  #else
	ssize_t (*sendpage) (struct file *, unsigned char *buf, int, size_t, loff_t *, int);
 #endif 

 unsigned long (*get_unmapped_area)(struct file *, unsigned long, unsigned long, unsigned long, unsigned long);
};

struct inode_operations {
 #ifndef INODE_REDUCED  
	int (*create) (struct inode *,struct dentry *,int, struct nameidata *);
	struct dentry * (*lookup) (struct inode *,struct dentry *, struct nameidata *);
	int (*link) (struct dentry *,struct inode *,struct dentry *);
	int (*unlink) (struct inode *,struct dentry *);
	int (*symlink) (struct inode *,struct dentry *,const char *);
	int (*mkdir) (struct inode *,struct dentry *,int);
	int (*rmdir) (struct inode *,struct dentry *);
	int (*mknod) (struct inode *,struct dentry *,int,dev_t);
	int (*rename) (struct inode *, struct dentry *,
			struct inode *, struct dentry *);
 #endif // ndef INODE_REDUCED  
    /*
	int (*readlink) (struct dentry *, char __user *,int);
	*/
	int (*readlink) (struct dentry *, char *,int);
	int (*follow_link) (struct dentry *, struct nameidata *);
 #ifndef INODE_REDUCED  
	void (*truncate) (struct inode *);
	int (*permission) (struct inode *, int, struct nameidata *);
 #endif // ndef INODE_REDUCED  
	int (*setattr) (struct dentry *, struct iattr *);
 #ifndef NO_VFS  // struct vfsmount
	int (*getattr) (struct vfsmount *mnt, struct dentry *, struct kstat *);
 #endif // ndef NO_VFS
 int (*setxattr) (struct dentry *, const char *,const void *,size_t,int);
 ssize_t (*getxattr) (struct dentry *, const char *, void *, size_t);
 ssize_t (*listxattr) (struct dentry *, char *, size_t);
 int (*removexattr) (struct dentry *, const char *);
};

struct seq_file;
/*
extern ssize_t vfs_read(struct file *, char __user *, size_t, loff_t *);
extern ssize_t vfs_write(struct file *, const char __user *, size_t, loff_t *);
extern ssize_t vfs_readv(struct file *, const struct iovec __user *,
		unsigned long, loff_t *);
extern ssize_t vfs_writev(struct file *, const struct iovec __user *,
		unsigned long, loff_t *);
		*/
extern ssize_t vfs_read(struct file *, char *, size_t, loff_t *);
extern ssize_t vfs_write(struct file *, const char *, size_t, loff_t *);
extern ssize_t vfs_readv(struct file *, const struct iovec *,
		unsigned long, loff_t *);
extern ssize_t vfs_writev(struct file *, const struct iovec *,
		unsigned long, loff_t *);

/*
 * NOTE: write_inode, delete_inode, clear_inode, put_inode can be called
 * without the big kernel lock held in all filesystems.
 */
struct super_operations {
 #ifndef NO_INODE
   	struct inode *(*alloc_inode)(struct super_block *sb);
	void (*destroy_inode)(struct inode *);

	void (*read_inode) (struct inode *);
  
   	void (*dirty_inode) (struct inode *);
	void (*write_inode) (struct inode *, int);
	void (*put_inode) (struct inode *);
	void (*drop_inode) (struct inode *);
	void (*delete_inode) (struct inode *);
 #endif // ndef NO_INODE
	void (*put_super) (struct super_block *);
	void (*write_super) (struct super_block *);
	int (*sync_fs)(struct super_block *sb, int wait);
	void (*write_super_lockfs) (struct super_block *);
	void (*unlockfs) (struct super_block *);

 #ifndef INODE_REDUCED  
	int (*statfs) (struct super_block *, struct kstatfs *);
 #endif // ndef INODE_REDUCED  

 int (*remount_fs) (struct super_block *, int *, char *);
 #ifndef NO_INODE
	void (*clear_inode) (struct inode *);
 #endif // ndef NO_INODE
 void (*umount_begin) (struct super_block *);

 #ifndef NO_VFS  // struct vfsmount
	int (*show_options)(struct seq_file *, struct vfsmount *);
 #endif // ndef NO_VFS
};

/* Inode state bits.  Protected by inode_lock. */
#define I_DIRTY_SYNC		1 /* Not dirty enough for O_DATASYNC */
#define I_DIRTY_DATASYNC	2 /* Data-related inode changes pending */
#define I_DIRTY_PAGES		4 /* Data-related inode changes pending */
#define I_LOCK			8
#define I_FREEING		16
#define I_CLEAR			32
#define I_NEW			64

#define I_DIRTY (I_DIRTY_SYNC | I_DIRTY_DATASYNC | I_DIRTY_PAGES)

#ifndef INODE_REDUCED  
	extern void __mark_inode_dirty(struct inode *, int);
	static inline void mark_inode_dirty(struct inode *inode)
	{
	__mark_inode_dirty(inode, I_DIRTY);
	}

	static inline void mark_inode_dirty_sync(struct inode *inode)
	{
	__mark_inode_dirty(inode, I_DIRTY_SYNC);
	}
#endif // ndef INODE_REDUCED


/**
 * &export_operations - for nfsd to communicate with file systems
 * decode_fh:      decode a file handle fragment and return a &struct dentry
 * encode_fh:      encode a file handle fragment from a dentry
 * get_name:       find the name for a given inode in a given directory
 * get_parent:     find the parent of a given directory
 * get_dentry:     find a dentry for the inode given a file handle sub-fragment
 *
 * Description:
 *    The export_operations structure provides a means for nfsd to communicate
 *    with a particular exported file system  - particularly enabling nfsd and
 *    the filesystem to co-operate when dealing with file handles.
 *
 *    export_operations contains two basic operation for dealing with file handles,
 *    decode_fh() and encode_fh(), and allows for some other operations to be defined
 *    which standard helper routines use to get specific information from the
 *    filesystem.
 *
 *    nfsd encodes information use to determine which filesystem a filehandle
 *    applies to in the initial part of the file handle.  The remainder, termed a
 *    file handle fragment, is controlled completely by the filesystem.
 *    The standard helper routines assume that this fragment will contain one or two
 *    sub-fragments, one which identifies the file, and one which may be used to
 *    identify the (a) directory containing the file.
 *
 *    In some situations, nfsd needs to get a dentry which is connected into a
 *    specific part of the file tree.  To allow for this, it passes the function
 *    acceptable() together with a @context which can be used to see if the dentry
 *    is acceptable.  As there can be multiple dentrys for a given file, the filesystem
 *    should check each one for acceptability before looking for the next.  As soon
 *    as an acceptable one is found, it should be returned.
 *
 * decode_fh:
 *    @decode_fh is given a &struct super_block (@sb), a file handle fragment (@fh, @fh_len)
 *    and an acceptability testing function (@acceptable, @context).  It should return
 *    a &struct dentry which refers to the same file that the file handle fragment refers
 *    to,  and which passes the acceptability test.  If it cannot, it should return
 *    a %NULL pointer if the file was found but no acceptable &dentries were available, or
 *    a %ERR_PTR error code indicating why it couldn't be found (e.g. %ENOENT or %ENOMEM).
 *
 * encode_fh:
 *    @encode_fh should store in the file handle fragment @fh (using at most @max_len bytes)
 *    information that can be used by @decode_fh to recover the file refered to by the
 *    &struct dentry @de.  If the @connectable flag is set, the encode_fh() should store
 *    sufficient information so that a good attempt can be made to find not only
 *    the file but also it's place in the filesystem.   This typically means storing
 *    a reference to de->d_parent in the filehandle fragment.
 *    encode_fh() should return the number of bytes stored or a negative error code
 *    such as %-ENOSPC
 *
 * get_name:
 *    @get_name should find a name for the given @child in the given @parent directory.
 *    The name should be stored in the @name (with the understanding that it is already
 *    pointing to a a %NAME_MAX+1 sized buffer.   get_name() should return %0 on success,
 *    a negative error code or error.
 *    @get_name will be called without @parent->i_sem held.
 *
 * get_parent:
 *    @get_parent should find the parent directory for the given @child which is also
 *    a directory.  In the event that it cannot be found, or storage space cannot be
 *    allocated, a %ERR_PTR should be returned.
 *
 * get_dentry:
 *    Given a &super_block (@sb) and a pointer to a file-system specific inode identifier,
 *    possibly an inode number, (@inump) get_dentry() should find the identified inode and
 *    return a dentry for that inode.
 *    Any suitable dentry can be returned including, if necessary, a new dentry created
 *    with d_alloc_root.  The caller can then find any other extant dentrys by following the
 *    d_alias links.  If a new dentry was created using d_alloc_root, DCACHE_NFSD_DISCONNECTED
 *    should be set, and the dentry should be d_rehash()ed.
 *
 *    If the inode cannot be found, either a %NULL pointer or an %ERR_PTR code can be returned.
 *    The @inump will be whatever was passed to nfsd_find_fh_dentry() in either the
 *    @obj or @parent parameters.
 *
 * Locking rules:
 *  get_parent is called with child->d_inode->i_sem down
 *  get_name is not (which is possibly inconsistent)
 */

struct export_operations {
	struct dentry *(*decode_fh)(struct super_block *sb, __u32 *fh, int fh_len, int fh_type,
			 int (*acceptable)(void *context, struct dentry *de),
			 void *context);
	int (*encode_fh)(struct dentry *de, __u32 *fh, int *max_len,
			 int connectable);

	/* the following are only called from the filesystem itself */
	int (*get_name)(struct dentry *parent, char *name,
			struct dentry *child);
	struct dentry * (*get_parent)(struct dentry *child);
	struct dentry * (*get_dentry)(struct super_block *sb, void *inump);

	/* This is set by the exporting module to a standard helper */
	struct dentry * (*find_exported_dentry)(
		struct super_block *sb, void *obj, void *parent,
		int (*acceptable)(void *context, struct dentry *de),
		void *context);


};


struct file_system_type {
 #ifdef USB_DBGSTRING_ON
	const char *name;
 #endif // USB_DBGSTRING_ON
 int fs_flags;
	struct super_block *(*get_sb) (struct file_system_type *, int,
				       const char *, void *);
 void (*kill_sb) (struct super_block *);
 /*
 #ifndef _ETCPIP
	struct module *owner;
 #endif // ndef _ETCPIP
 */
 struct file_system_type * next;
 struct list_head fs_supers;
};

struct super_block *get_sb_bdev(struct file_system_type *fs_type,
	int flags, const char *dev_name, void *data,
	int (*fill_super)(struct super_block *, void *, int));
struct super_block *get_sb_single(struct file_system_type *fs_type,
	int flags, void *data,
	int (*fill_super)(struct super_block *, void *, int));
struct super_block *get_sb_nodev(struct file_system_type *fs_type,
	int flags, void *data,
	int (*fill_super)(struct super_block *, void *, int));
void generic_shutdown_super(struct super_block *sb);
void kill_block_super(struct super_block *sb);
void kill_anon_super(struct super_block *sb);
void kill_litter_super(struct super_block *sb);
void deactivate_super(struct super_block *sb);
int set_anon_super(struct super_block *s, void *data);
struct super_block *sget(struct file_system_type *type,
			int (*test)(struct super_block *,void *),
			int (*set)(struct super_block *,void *),
			void *data);
struct super_block *get_sb_pseudo(struct file_system_type *, char *,
			struct super_operations *ops, unsigned long);

/* Alas, no aliases. Too much hassle with bringing module.h everywhere */
#define fops_get(fops) \
	(((fops) && try_module_get((fops)->owner) ? (fops) : NULL))
#define fops_put(fops) \
	do { if (fops) module_put((fops)->owner); } while(0)

extern int register_filesystem(struct file_system_type *);
extern int unregister_filesystem(struct file_system_type *);
#ifndef NO_VFS  // struct vfsmount
	extern struct vfsmount *kern_mount(struct file_system_type *);
	extern int may_umount(struct vfsmount *);
#endif // ndef NO_VFS
extern long do_mount(char *, char *, char *, unsigned long, void *);

extern int vfs_statfs(struct super_block *, struct kstatfs *);

/* Return value for VFS lock functions - tells locks.c to lock conventionally
 * REALLY kosha for root NFS and nfs_lock
 */ 
#define LOCK_USE_CLNT 1

#define FLOCK_VERIFY_READ  1
#define FLOCK_VERIFY_WRITE 2

#ifndef NO_INODE
	extern int locks_mandatory_locked(struct inode *);
	extern int locks_mandatory_area(int, struct inode *, struct file *, loff_t, size_t);
#endif // ndef NO_INODE

/*
 * Candidates for mandatory locking have the setgid bit set
 * but no group execute bit -  an otherwise meaningless combination.
 */
#ifndef INODE_REDUCED  
	#define MANDATORY_LOCK(inode) \
	(IS_MANDLOCK(inode) && ((inode)->i_mode & (S_ISGID | S_IXGRP)) == S_ISGID)

	static inline int locks_verify_locked(struct inode *inode)
	{
	if (MANDATORY_LOCK(inode))
		return locks_mandatory_locked(inode);
	return 0;
	}

	static inline int locks_verify_area(int read_write, struct inode *inode,
				    struct file *filp, loff_t offset,
				    size_t count)
	{
	if (inode->i_flock && MANDATORY_LOCK(inode))
		return locks_mandatory_area(read_write, inode, filp, offset, count);
	return 0;
	}

	static inline int locks_verify_truncate(struct inode *inode,
				    struct file *filp,
				    loff_t size)
	{
	if (inode->i_flock && MANDATORY_LOCK(inode))
		return locks_mandatory_area(
			FLOCK_VERIFY_WRITE, inode, filp,
			size < inode->i_size ? size : inode->i_size,
			(size < inode->i_size ? inode->i_size - size
			 : size - inode->i_size)
		);
	return 0;
	}
#endif // ndef INODE_REDUCED  

#ifndef NO_INODE
	static inline int break_lease(struct inode *inode, unsigned int mode)
	{
	#ifndef INODE_REDUCED  
		if (inode->i_flock)
			return __break_lease(inode, mode);
	#endif // ndef INODE_REDUCED  
	return 0;
	}
#endif // ndef NO_INODE

/* fs/open.c */
/*
asmlinkage long sys_open(const char __user *, int, int);
*/
long sys_open(const char *, int, int);
/*
asmlinkage long sys_close(unsigned int);	// yes, it's really unsigned 
*/
long sys_close(unsigned int);	/* yes, it's really unsigned */

extern int do_truncate(struct dentry *, loff_t start);

extern struct file *filp_open(const char *, int, int);
#ifndef NO_VFS  // struct vfsmount
	extern struct file * dentry_open(struct dentry *, struct vfsmount *, int);
#endif // ndef NO_VFS
extern int filp_close(struct file *, fl_owner_t id);
/*
extern char * getname(const char __user *);
*/
extern char * getname(const char *);

/* fs/dcache.c */
extern void vfs_caches_init(unsigned long);

/*
#define __getname()	kmem_cache_alloc(names_cachep, SLAB_KERNEL)
*/
#ifdef USB_DBG_MEM_C_ON
	#define __getname() UsbAllocMemC(USB_MEM_C_S	, sizeof(names_cachep) )
 #else
	#define __getname() UsbAllocMemC(USB_MEM_C_S)
#endif

/*
#define putname(name)	kmem_cache_free(names_cachep, (void *)(name))
*/
#if defined(USB_DBG_MEM_C_ON)
	#define putname(name) UsbFreeMemC(USB_MEM_C_S,(unsigned char *) name \
												, sizeof(name) \
												)
 #else
	#define putname(name) UsbFreeMemC(USB_MEM_C_S,(unsigned char *) name	)
#endif

enum {BDEV_FILE, BDEV_SWAP, BDEV_FS, BDEV_RAW};

#ifndef USBMS_ONLY_ONE_DISK
	extern int register_blkdev(unsigned int, const char *);
 #else
	extern int register_blkdev(const char *name) ;
#endif // USBMS_ONLY_ONE_DISK

#ifndef USBMS_ONLY_ONE_DISK
	extern int unregister_blkdev(unsigned int, const char *);
 #else
	extern int unregister_blkdev(const char *name) ;
#endif // USBMS_ONLY_ONE_DISK

extern struct block_device *bdget(dev_t);
#ifndef NO_INODE
	extern int bd_acquire(struct inode *inode);
	extern void bd_forget(struct inode *inode);
	// bdput() is on fs/block_dev.c
	extern void bdput(struct block_device *);
	extern int blkdev_open(struct inode *, struct file *);
	extern int blkdev_close(struct inode *, struct file *);
#endif // ndef NO_INODE
extern struct block_device *open_by_devnum(dev_t, unsigned, int);
extern struct file_operations def_blk_fops;
extern struct address_space_operations def_blk_aops;
extern struct file_operations def_chr_fops;
extern struct file_operations bad_sock_fops;
extern struct file_operations def_fifo_fops;
extern int ioctl_by_bdev(struct block_device *, unsigned, unsigned long);
#ifndef NO_INODE
	extern int blkdev_ioctl(struct inode *, struct file *, unsigned, unsigned long);
#endif // ndef NO_INODE  
#if _USB2 > 0	// reduce function call layer
extern int blkdev_get_block(struct inode *inode, sector_t iblock, struct buffer_head *bh, int create);
#endif	// _USB2

// blkdev_get() on fs/block_dev.c
/*
#ifndef _ETCPIP
*/
extern int blkdev_get(struct block_device *, mode_t, unsigned, int);
/*
#else
extern int blkdev_get(struct block_device *, mode_t, unsigned, int ,
						struct gendisk *disk) ;
#endif
*/

extern int blkdev_put(struct block_device *, int);
extern int bd_claim(struct block_device *, void *);
extern void bd_release(struct block_device *);
/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	extern void blk_run_queues(void) ;
 //#else
 //	extern void blk_run_queues(request_queue_t *q) ;
#endif
*/
/* fs/char_dev.c */
extern int alloc_chrdev_region(dev_t *, unsigned, unsigned, char *);
extern int register_chrdev_region(dev_t, unsigned, char *);
extern int register_chrdev(unsigned int, const char *,
			   struct file_operations *);
extern int unregister_chrdev(unsigned int, const char *);
extern void unregister_chrdev_region(dev_t, unsigned);
#ifndef INODE_REDUCED  
	extern int chrdev_open(struct inode *, struct file *);
#endif // ndef INODE_REDUCED  

/* fs/block_dev.c */
#define BDEVNAME_SIZE	32	/* Largest string for a blockdev identifier */
extern const char *__bdevname(dev_t, char *buffer);
extern const char *bdevname(struct block_device *bdev, char *buffer);
extern struct block_device *lookup_bdev(const char *);
extern struct block_device *open_bdev_excl(const char *, int, int, void *);
extern void close_bdev_excl(struct block_device *, int);

#ifndef NO_INODE
	extern void init_special_inode(struct inode *, umode_t, dev_t);

	/* Invalid inode operations -- fs/bad_inode.c */
	extern void make_bad_inode(struct inode *);
	extern int is_bad_inode(struct inode *);
#endif // ndef NO_INODE

extern struct file_operations read_fifo_fops;
extern struct file_operations write_fifo_fops;
extern struct file_operations rdwr_fifo_fops;
extern struct file_operations read_pipe_fops;
extern struct file_operations write_pipe_fops;
extern struct file_operations rdwr_pipe_fops;

extern int fs_may_remount_ro(struct super_block *);

// #ifndef USB_NO_VMM // vitrtul memory management
	/*
	 * return READ, READA, or WRITE
	 */
	#define bio_rw(bio)		((bio)->bi_rw & (RW_MASK | RWA_MASK))

	/*
	 * return data direction, READ or WRITE
	 */
	#define bio_data_dir(bio)	((bio)->bi_rw & 1)
// #endif // ndef USB_NO_VMM

extern int check_disk_change(struct block_device *);
extern int invalidate_inodes(struct super_block *);
#ifndef INODE_REDUCED  
	// __invalidate_device() is on fs/inode.c
	extern int __invalidate_device(struct block_device *, int);
#endif // ndef INODE_REDUCED  
extern int invalidate_partition(struct gendisk *, int);
/* comment out By B. S. Liaw
unsigned long invalidate_mapping_pages(struct address_space *mapping,
					pgoff_t start, pgoff_t end);
*/
unsigned long invalidate_inode_pages(struct address_space *mapping);

#ifndef NO_INODE
   static inline void invalidate_remote_inode(struct inode *inode)
   {
	if (S_ISREG(inode->i_mode) || S_ISDIR(inode->i_mode) ||
	    S_ISLNK(inode->i_mode))
	   {
		#ifndef USB_NO_VMM // vitrtul memory management
			invalidate_inode_pages(inode->i_mapping);
		#endif // ndef USB_NO_VMM
	   }
   }
#endif // ndef NO_INODE

extern void invalidate_inode_pages2(struct address_space *mapping);
#ifndef NO_INODE
	extern void write_inode_now(struct inode *, int);
#endif // ndef NO_INODE
extern int filemap_fdatawrite(struct address_space *);
extern int filemap_flush(struct address_space *);
extern int filemap_fdatawait(struct address_space *);
extern void sync_supers(void);
extern void sync_filesystems(int wait);
extern void emergency_sync(void);
extern void emergency_remount(void);
extern int do_remount_sb(struct super_block *sb, int flags,
			 void *data, int force);
#ifndef NO_INODE
	extern sector_t bmap(struct inode *, sector_t);
#endif // ndef NO_INODE
extern int setattr_mask(unsigned int);
extern int notify_change(struct dentry *, struct iattr *);
#ifndef NO_INODE
    extern int permission(struct inode *, int, struct nameidata *);
	extern int vfs_permission(struct inode *, int);
	extern int get_write_access(struct inode *);
#endif // ndef NO_INODE
extern int deny_write_access(struct file *);
#ifndef NO_INODE
   static inline void put_write_access(struct inode * inode)
   {
	#ifndef INODE_REDUCED  
		atomic_dec(&inode->i_writecount);
	#endif // ndef INODE_REDUCED  
   }
#endif // ndef NO_INODE

static inline void allow_write_access(struct file *file)
{
 #ifndef INODE_REDUCED  
	if (file)
		atomic_inc(&file->f_dentry->d_inode->i_writecount);
 #endif // ndef INODE_REDUCED  
}
extern int do_pipe(int *);

extern int open_namei(const char *, int, int, struct nameidata *);
extern int may_open(struct nameidata *, int, int);

extern int kernel_read(struct file *, unsigned long, char *, unsigned long);
extern struct file * open_exec(const char *);
 
/* fs/dcache.c -- generic fs support functions */
extern int is_subdir(struct dentry *, struct dentry *);
extern ino_t find_inode_number(struct dentry *, struct qstr *);

#ifndef _ETCPIP
	#include <linux/err.h>
#endif // ndef _ETCPIP

/* needed for stackable file system support */
extern loff_t default_llseek(struct file *file, loff_t offset, int origin);

#ifndef NO_INODE
	extern void inode_init_once(struct inode *);
	extern void iput(struct inode *);
	extern struct inode * igrab(struct inode *);
#endif // ndef NO_INODE
extern ino_t iunique(struct super_block *, ino_t);
#ifndef NO_INODE
	extern int inode_needs_sync(struct inode *inode);
	extern void generic_delete_inode(struct inode *inode);

	extern struct inode *ilookup5(struct super_block *sb, unsigned long hashval,
		int (*test)(struct inode *, void *), void *data);
	extern struct inode *ilookup(struct super_block *sb, unsigned long ino);

	extern struct inode * iget5_locked(struct super_block *, unsigned long, int (*test)(struct inode *, void *), int (*set)(struct inode *, void *), void *);
	extern struct inode * iget_locked(struct super_block *, unsigned long);
	extern void unlock_new_inode(struct inode *);

   static inline struct inode *iget(struct super_block *sb, unsigned long ino)
   {
	struct inode *inode = iget_locked(sb, ino);
	
	if (inode && (inode->i_state & I_NEW)) {
		sb->s_op->read_inode(inode);
		// #ifndef INODE_REDUCED
			unlock_new_inode(inode);
		// #endif // ndef INODE_REDUCED
	   }

	return inode;
   }

	extern void __iget(struct inode * inode);
	extern void clear_inode(struct inode *);
	extern void destroy_inode(struct inode *);
	extern struct inode *new_inode(struct super_block *);
#endif // ndef NO_INODE

extern void remove_suid(struct dentry *);

#ifndef NO_INODE
	extern void __insert_inode_hash(struct inode *, unsigned long hashval);
	extern void remove_inode_hash(struct inode *);
#endif // ndef NO_INODE
/*
#ifndef NO_INODE
   static inline void insert_inode_hash(struct inode *inode) {
	#ifndef INODE_REDUCED
		__insert_inode_hash(inode, inode->i_ino);
	#endif // ndef INODE_REDUCED
   }
#endif // ndef NO_INODE
*/
extern struct file * get_empty_filp(void);
extern void file_move(struct file *f, struct list_head *list);
extern void file_kill(struct file *f);

// #ifndef USB_NO_VMM // vitrtul memory management
	struct bio;
	extern int submit_bio(int, struct bio *);
// #endif // ndef USB_NO_VMM

extern int bdev_read_only(struct block_device *);
extern int set_blocksize(struct block_device *, int);
extern int sb_set_blocksize(struct super_block *, int);
extern int sb_min_blocksize(struct super_block *, int);

#ifndef USB_NO_VMM // vitrtul memory management
	extern int generic_file_mmap(struct file *, struct vm_area_struct *);
	extern int generic_file_readonly_mmap(struct file *, struct vm_area_struct *);
#endif // ndef USB_NO_VMM
#ifndef USB_NO_VMM // vitrtul memory management
	// struct page is on include/linux/mm.h
	extern int file_read_actor(read_descriptor_t * desc, struct page *page, unsigned long offset, unsigned long size);
	extern int file_send_actor(read_descriptor_t * desc, struct page *page, unsigned long offset, unsigned long size);
 #else
	extern int file_read_actor(read_descriptor_t * desc, unsigned char *buf, unsigned long offset, unsigned long size);
	extern int file_send_actor(read_descriptor_t * desc, unsigned char *buf, unsigned long offset, unsigned long size);
#endif 

/*
extern ssize_t generic_file_read(struct file *, char __user *, size_t, loff_t *);
*/
extern ssize_t generic_file_read(struct file *, char *, size_t, loff_t *);
#ifndef NO_INODE
	int generic_write_checks(struct inode *inode, struct file *file,
			loff_t *pos, size_t *count, int isblk);
#endif // ndef NO_INODE 
/*
extern ssize_t generic_file_write(struct file *, const char __user *, size_t, loff_t *);
extern ssize_t generic_file_aio_read(struct kiocb *, char __user *, size_t, loff_t);
extern ssize_t __generic_file_aio_read(struct kiocb *, const struct iovec *, unsigned long, loff_t *);
extern ssize_t generic_file_aio_write(struct kiocb *, const char __user *, size_t, loff_t);
*/
extern ssize_t generic_file_write(struct file *, const char *, size_t, loff_t *);
extern ssize_t generic_file_aio_read(struct kiocb *, char *, size_t, loff_t);
extern ssize_t generic_file_aio_write(struct kiocb *, const char *, size_t, loff_t);
extern ssize_t generic_file_aio_write_nolock(struct kiocb *, const struct iovec *,
				unsigned long, loff_t *);
/*
extern ssize_t do_sync_read(struct file *filp, char __user *buf, size_t len, loff_t *ppos);
extern ssize_t do_sync_write(struct file *filp, const char __user *buf, size_t len, loff_t *ppos);
*/
extern ssize_t do_sync_read(struct file *filp, char *buf, size_t len, loff_t *ppos);
extern ssize_t do_sync_write(struct file *filp, const char *buf, size_t len, loff_t *ppos);
ssize_t generic_file_write_nolock(struct file *file, const struct iovec *iov,
				unsigned long nr_segs, loff_t *ppos);
/*
extern ssize_t generic_file_sendfile(struct file *, loff_t *, size_t, read_actor_t, void __user *);
*/
extern ssize_t generic_file_sendfile(struct file *, loff_t *, size_t, read_actor_t, void *);
extern void do_generic_mapping_read(struct address_space *, struct file_ra_state *, struct file *,
				    loff_t *, read_descriptor_t *, read_actor_t);
extern void
file_ra_state_init(struct file_ra_state *ra, struct address_space *mapping);
extern ssize_t generic_file_direct_IO(int rw, struct kiocb *iocb,
	const struct iovec *iov, loff_t offset, unsigned long nr_segs);

#ifndef NO_INODE
	extern int blockdev_direct_IO(int rw, struct kiocb *iocb, struct inode *inode, 
		struct block_device *bdev, const struct iovec *iov, loff_t offset, 
		unsigned long nr_segs, get_blocks_t *get_blocks, dio_iodone_t *end_io);
#endif // ndef NO_INODE

extern ssize_t generic_file_readv(struct file *filp, const struct iovec *iov, 
	unsigned long nr_segs, loff_t *ppos);
ssize_t generic_file_writev(struct file *filp, const struct iovec *iov, 
			unsigned long nr_segs, loff_t *ppos);
extern loff_t no_llseek(struct file *file, loff_t offset, int origin);
extern loff_t generic_file_llseek(struct file *file, loff_t offset, int origin);
extern loff_t remote_llseek(struct file *file, loff_t offset, int origin);
#ifndef NO_INODE
	extern int generic_file_open(struct inode * inode, struct file * filp);
#endif // ndef NO_INODE
/*
#ifndef USB_NO_VMM // vitrtul memory management
   static inline void do_generic_file_read(struct file * filp, loff_t *ppos,
					read_descriptor_t * desc,
					read_actor_t actor)
   {
	do_generic_mapping_read(filp->f_dentry->d_inode->i_mapping,
				&filp->f_ra,
				filp,
				ppos,
				desc,
				actor);
   }
#endif // ndef USB_NO_VMM
*/
extern struct file_operations generic_ro_fops;

#define special_file(m) (S_ISCHR(m)||S_ISBLK(m)||S_ISFIFO(m)||S_ISSOCK(m))
/*
extern int vfs_readlink(struct dentry *, char __user *, int, const char *);
*/
extern int vfs_readlink(struct dentry *, char *, int, const char *);
extern int vfs_follow_link(struct nameidata *, const char *);
/*
extern int page_readlink(struct dentry *, char __user *, int);
*/
extern int page_readlink(struct dentry *, char *, int);
extern int page_follow_link(struct dentry *, struct nameidata *);
#ifndef NO_INODE
	extern int page_symlink(struct inode *inode, const char *symname, int len);
#endif // ndef NO_INODE  
extern struct inode_operations page_symlink_inode_operations;
#ifndef NO_INODE
	extern void generic_fillattr(struct inode *, struct kstat *);
#endif // ndef NO_INODE
#ifndef NO_VFS  // struct vfsmount
	extern int vfs_getattr(struct vfsmount *, struct dentry *, struct kstat *);
#endif // ndef NO_VFS
#ifndef NO_INODE
	void inode_add_bytes(struct inode *inode, loff_t bytes);
	void inode_sub_bytes(struct inode *inode, loff_t bytes);
	loff_t inode_get_bytes(struct inode *inode);
	void inode_set_bytes(struct inode *inode, loff_t bytes);
#endif // ndef NO_INODE

extern int vfs_readdir(struct file *, filldir_t, void *);
/*
extern int vfs_stat(char __user *, struct kstat *);
extern int vfs_lstat(char __user *, struct kstat *);
*/
extern int vfs_stat(char *, struct kstat *);
extern int vfs_lstat(char *, struct kstat *);
extern int vfs_fstat(unsigned int, struct kstat *);

extern struct file_system_type *get_fs_type(const char *name);
extern struct super_block *get_super(struct block_device *);
extern struct super_block *user_get_super(dev_t);
extern void drop_super(struct super_block *sb);

#ifndef NO_INODE
	extern int dcache_dir_open(struct inode *, struct file *);
	extern int dcache_dir_close(struct inode *, struct file *);
#endif // ndef NO_INODE
extern loff_t dcache_dir_lseek(struct file *, loff_t, int);
extern int dcache_readdir(struct file *, void *, filldir_t);
#ifndef NO_VFS  // struct vfsmount
	extern int simple_getattr(struct vfsmount *, struct dentry *, struct kstat *);
#endif // ndef NO_VFS
extern int simple_statfs(struct super_block *, struct kstatfs *);
#ifndef NO_INODE
	extern int simple_link(struct dentry *, struct inode *, struct dentry *);
	extern int simple_unlink(struct inode *, struct dentry *);
	extern int simple_rmdir(struct inode *, struct dentry *);
	extern int simple_rename(struct inode *, struct dentry *, struct inode *, struct dentry *);
#endif // ndef NO_INODE
extern int simple_sync_file(struct file *, struct dentry *, int);
extern int simple_empty(struct dentry *);
/*
#ifndef USB_NO_VMM // vitrtul memory management
	// struct page is on include/linux/mm.h
	extern int simple_readpage(struct file *file, struct page *page);
	extern int simple_prepare_write(struct file *file, struct page *page,
			unsigned offset, unsigned to);
	extern int simple_commit_write(struct file *file, struct page *page,
				unsigned offset, unsigned to);
 #else
*/
	extern int simple_readpage(struct file *file, unsigned char *buf);
	extern int simple_prepare_write(struct file *file, unsigned char *buf,
			unsigned offset, unsigned to);
	extern int simple_commit_write(struct file *file, unsigned char *buf,
				unsigned offset, unsigned to);
/*
#endif 
*/

#ifndef NO_INODE
extern struct dentry *simple_lookup(struct inode *, struct dentry *, struct nameidata *);
#endif // ndef NO_INODE
/*
extern ssize_t generic_read_dir(struct file *, char __user *, size_t, loff_t *);
*/
extern ssize_t generic_read_dir(struct file *, char *, size_t, loff_t *);
extern struct file_operations simple_dir_operations;
extern struct inode_operations simple_dir_inode_operations;
struct tree_descr { char *name; struct file_operations *ops; int mode; };
extern int simple_fill_super(struct super_block *, int, struct tree_descr *);
#ifndef NO_VFS  // struct vfsmount
	extern int simple_pin_fs(char *name, struct vfsmount **mount, int *count);
	extern void simple_release_fs(struct vfsmount **mount, int *count);
#endif // ndef NO_VFS

#ifndef NO_INODE
	extern int inode_change_ok(struct inode *, struct iattr *);
	extern int inode_setattr(struct inode *, struct iattr *);

	extern void inode_update_time(struct inode *inode, int ctime_too);
#endif // ndef NO_INODE

static inline ino_t parent_ino(struct dentry *dentry)
{
 ino_t res;

 #ifndef NO_SPINLOCK
	spin_lock(&dentry->d_lock);
 #endif // ndef NO_SPINLOCK

 #ifndef INODE_REDUCED 
	res = dentry->d_parent->d_inode->i_ino;
 #endif // ndef INODE_REDUCED  

 #ifndef NO_SPINLOCK
	spin_unlock(&dentry->d_lock);
 #endif // ndef NO_SPINLOCK
	return res;
}

#endif /* __KERNEL__ */
#endif /* _LINUX_FS_H */
