
// Name of File : blk_scsi_ioctl.c
// ====================================================================

// drivers/block/scsi_ioctl.c

/*
 * Copyright (C) 2001 Jens Axboe <axboe@suse.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public Licens
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-
 *
 */
// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP

/*
#include <linux/kernel.h>
*/
#include "PoKernel.h"
/*
#include <linux/errno.h>
*/
#include "PoErrno.h" // osport_errno.h
/*
#include <linux/string.h>
*/
#include "PoString.h" // linux/string.h
/*
#include <linux/module.h>
*/
/*
#include <linux/blkdev.h>
*/
#include "bitops.h" // B. S. Liaw add for test_bit()
#include "blkdev.h"
/*
#include <linux/completion.h>
*/
/*
#include <linux/cdrom.h>
*/
#include "cdrom.h"
/*
#include <linux/slab.h>
#include <linux/bio.h>
*/
#include "bio.h"
/*
#include <linux/times.h>
#include <asm/uaccess.h>
*/
/*
#include <scsi/scsi.h>
*/
#include "scsi_is.h" // include/scsi/scsi.h
/*
#include <scsi/scsi_ioctl.h>
*/
#include "scsi_ioctl.h"

/* Command group 3 is reserved and should never be used.  */
const unsigned char scsi_command_size[8] = {
 6, 10, 10, 12,
 16, 12, 10, 10
};

// ====================================================================
/*
#ifndef USBMS_NO_IOCTL

#define BLK_DEFAULT_TIMEOUT	(60 * HZ)

// defined in ../scsi/scsi.h  ... should it be included? 
#ifndef SCSI_SENSE_BUFFERSIZE
#define SCSI_SENSE_BUFFERSIZE 64
#endif

static int blk_do_rq(request_queue_t *q, struct block_device *bdev, 
		     struct request *rq)
{
 char sense[SCSI_SENSE_BUFFERSIZE];
 
 #ifndef _ETCPIP
    // struct completion in <linux/completion.h>
	DECLARE_COMPLETION(wait);
 #endif // ndef _ETCPIP
 
 int err = 0;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_do_rq() Begin .. \n")) ;
 rq->rq_disk = bdev->bd_disk;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] blk_do_rq() : Set Value : rq->rq_disk=%08XH \n",
	rq->rq_disk));

 
 // * we need an extra reference to the request, so we can look at
 // * it after io completion
  
 rq->ref_count++;

 if (!rq->sense) {
		memset(sense, 0, sizeof(sense));
		rq->sense = sense;
		rq->sense_len = 0;
	}

 rq->flags |= REQ_NOMERGE;
 
 #ifndef _ETCPIP
    // struct completion in <linux/completion.h>
	rq->waiting = &wait;
 #else
 
 #ifdef USBMS_USB_WAIT_BY_EVENT     // 2004/1/2 B. S. Liaw
	// 2003/12/02 B. S. Liaw :
	// req->waiting replaced by MT_EVT_USBMS_RQ/MT_EVT_USBMS_RQBI on kernel/local.h
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] blk_do_rq() : clrevt MT_EVT_USBMS_RQBI=%d ..\n",
		MT_EVT_USBMS_RQBI
		)) ;
    if (clrevt(MT_EVT_USBMS_RQBI) != SUCCESS) {
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] blk_do_rq() : Error : clrevt fail: Invalid MT_EVT_USBMS_RQBI=%d\n",
			MT_EVT_USBMS_RQBI
			)) ;
        dprintf("[USBMS] Error : clrevt fail: Invalid MT_EVT_USBMS_RQBI %d\n",
				MT_EVT_USBMS_RQBI);
	   }

  #else
	req->evt_blk_rq_bi = 0 ;
 #endif // ndef USBMS_USB_WAIT_BY_EVENT

 rq->waiting_evt = MT_EVT_USBMS_RQBI ;

 
 #endif
 

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	// elv_add_request() is on include/linux/elevator.h
	elv_add_request(q, rq, ELEVATOR_INSERT_BACK, 1);
  #else
		USBMS_DBG_INFO(iprintf("[USBMS INFO] blk_do_rq() : Bug on req !! \n"));

		USBMS_DBG_MEM(iprintf("[USBMS MEM] blk_do_rq() : q->req=%08XH \n",
			q->req)) ;

		#ifdef USBMS_BS_ASSERT
			if (!(   (unsigned long) q->req
			   > (unsigned long) 0x200000L
    	     )
		   )
		   {
			iprintf("[ASSERT][USBMS] blk_do_rq() : " ) ;
		    iprintf("Error : Invlid address q->req=%08XH \n",
				q->req ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
		#endif // USBMS_BS_ASSERT

		rq = q->req;
		
		//if ((rq->flags & REQ_DONTPREP) || !q->prep_rq_fn)
		//	break;
		
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_do_rq() : q->prep_rq_fn() .. \n"));
		
		//ret = q->prep_rq_fn(q, rq);
		//if (ret == BLKPREP_OK) {
		
		if (q->prep_rq_fn(q, rq) == BLKPREP_OK) {
			// break;
		   } 
		
		//  else if (ret == BLKPREP_DEFER) {
		//	req = NULL;
		//	break;
		//   }
		
		//  else if (ret == BLKPREP_KILL) {
		//	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_do_rq() : blkdev_dequeue_request() .. \n"));
		//	blkdev_dequeue_request(req);
		//	rq->flags |= REQ_QUIET;
		//	while (end_that_request_first(rq, 0, rq->nr_sectors))
		//		;
		//	end_that_request_last(req);
		//   }
		
		 else {
			printk("%s: q->prep_rq_fn() Failed !! \n", __FUNCTION__);
			// break;
		   }
		
 #endif // ndef NO_ELEVATOR 
 
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_do_rq() : generic_unplug_device() .. \n"));
 generic_unplug_device(q);
 
 //#ifndef _ETCPIP
 //   // struct completion in <linux/completion.h>
 //	wait_for_completion(&wait);
 // #else
 
 USBMS_DBG_INIT(iprintf("[USBMS INIT] blk_do_rq() *** task sleeping \n")) ;

 #ifdef USBMS_USB_WAIT_BY_EVENT     // 2004/1/2 B. S. Liaw
	// 2003/12/02 B. S. Liaw :
	// req->waiting replaced by MT_EVT_USBMS_RQ/MT_EVT_USBMS_RQBI on kernel/local.h
    (void)wteset_dec(MT_EVT_USBMS_RQBI, 0); // 0 : no timeout

    if (clrevt(MT_EVT_USBMS_RQBI) != SUCCESS) {
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] blk_do_rq() : Error : clrevt fail: Invalid MT_EVT_USBMS_RQBI=%d\n",
			MT_EVT_USBMS_RQBI
			)) ;
        dprintf("[USBMS] Error : clrevt fail: Invalid MT_EVT_USBMS_RQBI %d\n",
				MT_EVT_USBMS_RQBI);
	   }
 #else
	for(;;) {
		if (req->evt_blk_rq_bi) {
			break ;
		   }
	   }

	req->evt_blk_rq_bi = 0 ;
 #endif // ndef USBMS_USB_WAIT_BY_EVENT

 USBMS_DBG_INIT(iprintf("[USBMS INIT] blk_do_rq() *** task awakened \n")) ;
 
 //#endif
 

 if (rq->errors)
		err = -EIO;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_do_rq() End. \n")) ;
 return err;
}
#endif // ndef USBMS_NO_IOCTL
*/


#ifndef USBMS_NO_IOCTL
/*
#include <scsi/sg.h>
*/
#include "sg.h"

static int sg_get_version(int *p)
{
	static int sg_version_num = 30527;
 /*
	return put_user(sg_version_num, p);
 */
}

static int scsi_get_idlun(request_queue_t *q, int *p)
{
 /*
	return put_user(0, p);
 */
}

static int scsi_get_bus(request_queue_t *q, int *p)
{
 /*
	return put_user(0, p);
 */
}

static int sg_get_timeout(request_queue_t *q)
{
 #ifndef NO_TIMER
	return q->sg_timeout / (HZ / USER_HZ);
 #endif // ndef NO_TIMER
}

static int sg_set_timeout(request_queue_t *q, int *p)
{
 /*
 int timeout, err = get_user(timeout, p);
 */
 int timeout ;

 #ifndef NO_TIMER
	if (!err)
		q->sg_timeout = timeout * (HZ / USER_HZ);
 #endif // ndef NO_TIMER
 /*
	return err;
 */
}

static int sg_get_reserved_size(request_queue_t *q, int *p)
{
 /*
	return put_user(q->sg_reserved_size, p);
 */
}

static int sg_set_reserved_size(request_queue_t *q, int *p)
{
 /*
	int size, err = get_user(size, p);
 */
 int size ;


 /*
	if (err)
		return err;
 */
	if (size > (q->max_sectors << 9))
		return -EINVAL;

	q->sg_reserved_size = size;
	return 0;
}

/*
 * will always return that we are ATAPI even for a real SCSI drive, I'm not
 * so sure this is worth doing anything about (why would you care??)
 */
static int sg_emulated_host(request_queue_t *q, int *p)
{
 /*
	return put_user(1, p);
 */
}
#endif // ndef USBMS_NO_IOCTL


/*
#ifndef USBMS_NO_IOCTL

static int sg_io(request_queue_t *q, struct block_device *bdev,
		 struct sg_io_hdr *hdr)
{
 unsigned long start_time;
 int reading, writing;
 struct request *rq;
 // #ifndef USB_NO_VMM // vitrtul memory management
 struct bio *bio;
 // #endif // ndef USB_NO_VMM
 char sense[SCSI_SENSE_BUFFERSIZE];
 void *buffer;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sg_io() Begin .. \n"));
 if (hdr->interface_id != 'S')
		return -EINVAL;
 if (hdr->cmd_len > sizeof(rq->cmd))
		return -EINVAL;

	
 //	 * we'll do that later
	 
 if (hdr->iovec_count)
		return -EOPNOTSUPP;

 if (hdr->dxfer_len > (q->max_sectors << 9))
		return -EIO;

	reading = writing = 0;
	buffer = NULL;
 // #ifndef USB_NO_VMM // vitrtul memory management
	bio = NULL;
 // #endif // ndef USB_NO_VMM
	if (hdr->dxfer_len) {
		unsigned int bytes = (hdr->dxfer_len + 511) & ~511;

		switch (hdr->dxfer_direction) {
		default:
			return -EINVAL;
		case SG_DXFER_TO_FROM_DEV:
			reading = 1;
			// fall through 
		case SG_DXFER_TO_DEV:
			writing = 1;
			break;
		case SG_DXFER_FROM_DEV:
			reading = 1;
			break;
		}

 //		
 //		 * first try to map it into a bio. reading from device will
 //		 * be a write to vm.
 //		 
 #ifndef USB_NO_VMM // vitrtul memory management
		bio = bio_map_user(bdev, (unsigned long) hdr->dxferp,
				   hdr->dxfer_len, reading);
 #endif // ndef USB_NO_VMM

		
 //		 * if bio setup failed, fall back to slow approach
		 
 // #ifndef USB_NO_VMM // vitrtul memory management
		if (!bio) {

			// B. S. Liaw : request_buffer need uncached
			USBMS_DBG_MEM(iprintf(
				"[USBMS MEM] sg_io() : buffer = UsbAllocMemU() .. \n"
				));

			
			// buffer = kmalloc(bytes, q->bounce_gfp | GFP_USER);
			
			
			// buffer = (void *) UsbAllocMemC(USB_MEM_C_S
			
			buffer = (void *) UsbAllocMemU(USB_MEM_U_S
				#ifdef USB_DBG_MEM_U_ON
					, hdr->dxfer_len
		        #endif
           		) ;

			if (!buffer)
				return -ENOMEM;

			if (writing) {
				if (copy_from_user(buffer, hdr->dxferp,
						   hdr->dxfer_len))
					goto out_buffer;
			} else
				memset(buffer, 0, hdr->dxfer_len);
		}
 // #endif // ndef USB_NO_VMM
	}

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sg_io() : blk_get_request() .. \n"));
	rq = blk_get_request(q, writing ? WRITE : READ, __GFP_WAIT);

	
 //	 * fill in request structure
	 
	rq->cmd_len = hdr->cmd_len;
	memcpy(rq->cmd, hdr->cmdp, hdr->cmd_len);
	if (sizeof(rq->cmd) != hdr->cmd_len)
	   {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sg_io() : memset() .. \n" )) ;
		memset(rq->cmd + hdr->cmd_len, 0, sizeof(rq->cmd) - hdr->cmd_len);
	   }

	memset(sense, 0, sizeof(sense));
	rq->sense = sense;
	rq->sense_len = 0;

	rq->flags |= REQ_BLOCK_PC;

 // #ifndef USB_NO_VMM // vitrtul memory management
	rq->bio = rq->biotail = NULL;

	if (bio)
		blk_rq_bio_prep(q, rq, bio);
 // #endif // ndef USB_NO_VMM

	rq->data = buffer;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] sg_io() : Set Value : rq->data=%08XH \n",
	rq->data
	));
	rq->data_len = hdr->dxfer_len;

	rq->timeout = (hdr->timeout * HZ) / 1000;
	if (!rq->timeout)
		rq->timeout = q->sg_timeout;
	if (!rq->timeout)
		rq->timeout = BLK_DEFAULT_TIMEOUT;

 #ifndef NO_TIMER
	start_time = jiffies;
 #endif // ndef NO_TIMER

 // ignore return value. All information is passed back to caller
 // (if he doesn't check that is his problem).
 // N.B. a non-zero SCSI status is _not_ necessarily an error.
  
 blk_do_rq(q, bdev, rq);

 #ifndef USB_NO_VMM // vitrtul memory management
	if (bio)
		bio_unmap_user(bio, reading);
 #endif // ndef USB_NO_VMM

	// write to all output members 
	hdr->status = rq->errors;	
	hdr->masked_status = (hdr->status >> 1) & 0x1f;
	hdr->msg_status = 0;
	hdr->host_status = 0;
	hdr->driver_status = 0;
	hdr->info = 0;
	if (hdr->masked_status || hdr->host_status || hdr->driver_status)
		hdr->info |= SG_INFO_CHECK;
	hdr->resid = rq->data_len;
 #ifndef NO_TIMER
	hdr->duration = ((jiffies - start_time) * 1000) / HZ;
 #endif // ndef NO_TIMER
	hdr->sb_len_wr = 0;

	if (rq->sense_len && hdr->sbp) {
		int len = min((unsigned int) hdr->mx_sb_len, rq->sense_len);

		if (!copy_to_user(hdr->sbp, rq->sense, len))
			hdr->sb_len_wr = len;
	}

	blk_put_request(rq);

	if (buffer) {
		if (reading)
			if (copy_to_user(hdr->dxferp, buffer, hdr->dxfer_len))
				goto out_buffer;

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] sg_io() : UsbFreeMemC(buffer) , size=%d .. \n" ,
	hdr->dxfer_len
	));
	
	// kfree(buffer);
	
	UsbFreeMemC(USB_MEM_C_S,(unsigned char *) buffer
					#if defined(USB_DBG_MEM_C_ON)
						, hdr->dxfer_len
					#endif
					);
	}

	// may not have succeeded, but output values written to control
	// structure (struct sg_io_hdr).  
	return 0;

 out_buffer:

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] sg_io() : UsbFreeMemC(buffer) .. \n"
	));
 
 // kfree(buffer);
 
 UsbFreeMemC(USB_MEM_C_S,(unsigned char *) buffer
					#if defined(USB_DBG_MEM_C_ON)
						, hdr->dxfer_len
					#endif
					);

 return -EFAULT;
}
#endif // ndef USBMS_NO_IOCTL
*/


#ifndef USBMS_NO_IOCTL

/*
#define FORMAT_UNIT_TIMEOUT		(2 * 60 * 60 * HZ)
#define START_STOP_TIMEOUT		(60 * HZ)
#define MOVE_MEDIUM_TIMEOUT		(5 * 60 * HZ)
#define READ_ELEMENT_STATUS_TIMEOUT	(5 * 60 * HZ)
#define READ_DEFECT_DATA_TIMEOUT	(60 * HZ )
*/
#define FORMAT_UNIT_TIMEOUT		(2 * 60 * 60)
#define START_STOP_TIMEOUT		(60)
#define MOVE_MEDIUM_TIMEOUT		(5 * 60)
#define READ_ELEMENT_STATUS_TIMEOUT	(5 * 60)
#define READ_DEFECT_DATA_TIMEOUT	(60)

#define OMAX_SB_LEN 16          /* For backward compatibility */
#endif // ndef USBMS_NO_IOCTL


/*
#ifndef USBMS_NO_IOCTL

static int sg_scsi_ioctl(request_queue_t *q, struct block_device *bdev,
			 Scsi_Ioctl_Command *sic)
{
 struct request *rq;
 int err, in_len, out_len, bytes, opcode, cmdlen;
 char *buffer = NULL, sense[SCSI_SENSE_BUFFERSIZE];


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sg_scsi_ioctl() Begin .. \n"));
	
 //  * get in an out lengths, verify they don't exceed a page worth of data
	 
 //	
 //	if (get_user(in_len, &sic->inlen))
 //		return -EFAULT;
	
	
 //	if (get_user(out_len, &sic->outlen))
 //		return -EFAULT;
	
	if (in_len > PAGE_SIZE || out_len > PAGE_SIZE)
		return -EINVAL;
	
 //	if (get_user(opcode, sic->data))
 //		return -EFAULT;
	

	bytes = max(in_len, out_len);
	if (bytes) {

	// B. S. Liaw : request_buffer need uncached
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] sg_scsi_ioctl() : buffer = UsbAllocMemU() .. \n"
		));

		
		// buffer = kmalloc(bytes, q->bounce_gfp | GFP_USER);
		
		buffer = (char *) UsbAllocMemU(USB_MEM_U_S
				#ifdef USB_DBG_MEM_C_ON
					, bytes
		        #endif
           		) ;

		if (!buffer)
			return -ENOMEM;

		memset(buffer, 0, bytes);
	}

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sg_scsi_ioctl() : blk_get_request() .. \n"));
	rq = blk_get_request(q, in_len ? WRITE : READ, __GFP_WAIT);

	cmdlen = COMMAND_SIZE(opcode);

	
 //	 * get command and data to send to device, if any
	 
	err = -EFAULT;
	rq->cmd_len = cmdlen;
	if (copy_from_user(rq->cmd, sic->data, cmdlen))
		goto error;

	if (copy_from_user(buffer, sic->data + cmdlen, in_len))
		goto error;

	switch (opcode) {
		case SEND_DIAGNOSTIC:
		case FORMAT_UNIT:
			rq->timeout = FORMAT_UNIT_TIMEOUT;
			break;
		case START_STOP:
			rq->timeout = START_STOP_TIMEOUT;
			break;
		case MOVE_MEDIUM:
			rq->timeout = MOVE_MEDIUM_TIMEOUT;
			break;
		case READ_ELEMENT_STATUS:
			rq->timeout = READ_ELEMENT_STATUS_TIMEOUT;
			break;
		case READ_DEFECT_DATA:
			rq->timeout = READ_DEFECT_DATA_TIMEOUT;
			break;
		default:
			rq->timeout = BLK_DEFAULT_TIMEOUT;
			break;
	}

	memset(sense, 0, sizeof(sense));
	rq->sense = sense;
	rq->sense_len = 0;

	rq->data = buffer;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] sg_scsi_ioctl() : Set Value : rq->data=%08XH \n",
	rq->data
	));

	rq->data_len = bytes;
	rq->flags |= REQ_BLOCK_PC;

	blk_do_rq(q, bdev, rq);
	err = rq->errors & 0xff;	// only 8 bit SCSI status 
	if (err) {
		if (rq->sense_len && rq->sense) {
			bytes = (OMAX_SB_LEN > rq->sense_len) ?
				rq->sense_len : OMAX_SB_LEN;
			if (copy_to_user(sic->data, rq->sense, bytes))
				err = -EFAULT;
		}
	} else {
		if (copy_to_user(sic->data, buffer, out_len))
			err = -EFAULT;
	}
	
 error:
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] sg_scsi_ioctl) : UsbFreeMemC(buffer) .. \n"
	));
 
 // kfree(buffer);
 
 UsbFreeMemC(USB_MEM_C_S,(unsigned char *) buffer
					#if defined(USB_DBG_MEM_C_ON)
						, bytes
					#endif
					);

 blk_put_request(rq);
 return err;
}
#endif // ndef USBMS_NO_IOCTL
*/

#ifndef USBMS_NO_IOCTL
int scsi_cmd_ioctl(struct block_device *bdev, unsigned int cmd, unsigned long arg)
{
 request_queue_t *q;
 struct request *rq;
 int close = 0, err;
 // unsigned char *tmp_sense_buffer_pr ; // B. S. Liaw add this !!


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_cmd_ioctl() Begin .. \n"));

 q = bdev_get_queue(bdev);
 if (!q)
		return -ENXIO;

 if (blk_get_queue(q))
		return -ENXIO;

 switch (cmd) {
	/*
	 * new sgv3 interface
	 */
	case SG_GET_VERSION_NUM:
			err = sg_get_version((int *) arg);
			break;
	case SCSI_IOCTL_GET_IDLUN:
			err = scsi_get_idlun(q, (int *) arg);
			break;
	case SCSI_IOCTL_GET_BUS_NUMBER:
			err = scsi_get_bus(q, (int *) arg);
			break;
	case SG_SET_TIMEOUT:
			err = sg_set_timeout(q, (int *) arg);
			break;
	case SG_GET_TIMEOUT:
			err = sg_get_timeout(q);
			break;
	case SG_GET_RESERVED_SIZE:
			err = sg_get_reserved_size(q, (int *) arg);
			break;
	case SG_SET_RESERVED_SIZE:
			err = sg_set_reserved_size(q, (int *) arg);
			break;
	case SG_EMULATED_HOST:
			err = sg_emulated_host(q, (int *) arg);
			break;
	case SG_IO: {
		struct sg_io_hdr hdr;

		if (copy_from_user(&hdr, (struct sg_io_hdr *) arg, sizeof(hdr))) {
				err = -EFAULT;
				break;
			}
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_cmd_ioctl() : sg_io() .. \n"));
		err = sg_io(q, bdev, &hdr);
		if (copy_to_user((struct sg_io_hdr *) arg, &hdr, sizeof(hdr)))
				err = -EFAULT;
			break;
	   }
	case CDROM_SEND_PACKET: {
		struct cdrom_generic_command cgc;
		struct sg_io_hdr hdr;

		if (copy_from_user(&cgc, (struct cdrom_generic_command *) arg, sizeof(cgc))) {
			err = -EFAULT;
			break;
		   }
		/*
		#ifndef NO_TIMER
			cgc.timeout = clock_t_to_jiffies(cgc.timeout);
		#endif // ndef NO_TIMER
		*/
		memset(&hdr, 0, sizeof(hdr));
		hdr.interface_id = 'S';
		hdr.cmd_len = sizeof(cgc.cmd);
		hdr.dxfer_len = cgc.buflen;
		err = 0;
		switch (cgc.data_direction) {
				case CGC_DATA_UNKNOWN:
					hdr.dxfer_direction = SG_DXFER_UNKNOWN;
					break;
				case CGC_DATA_WRITE:
					hdr.dxfer_direction = SG_DXFER_TO_DEV;
					break;
				case CGC_DATA_READ:
					hdr.dxfer_direction = SG_DXFER_FROM_DEV;
					break;
				case CGC_DATA_NONE:
					hdr.dxfer_direction = SG_DXFER_NONE;
					break;
				default:
					err = -EINVAL;
			}
			if (err)
				break;

			hdr.dxferp = cgc.buffer;
			hdr.sbp = (char *) cgc.sense;
			if (hdr.sbp)
				hdr.mx_sb_len = sizeof(struct request_sense);
			hdr.timeout = cgc.timeout;
			hdr.cmdp = cgc.cmd;
			hdr.cmd_len = sizeof(cgc.cmd);
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_cmd_ioctl() : sg_io() .. \n"));
			err = sg_io(q, bdev, &hdr);

			if (hdr.status)
				err = -EIO;

			cgc.stat = err;
			cgc.buflen = hdr.resid;
			if (copy_to_user((struct cdrom_generic_command *) arg, &cgc, sizeof(cgc)))
				err = -EFAULT;

			break;
		}

	/*
	 * old junk scsi send command ioctl
	 */
	case SCSI_IOCTL_SEND_COMMAND:
		err = -EINVAL;
		if (!arg)
				break;

		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_cmd_ioctl() : sg_scsi_ioctl() .. \n"));
		err = sg_scsi_ioctl(q, bdev, (Scsi_Ioctl_Command *)arg);
		break;
	case CDROMCLOSETRAY:
		close = 1;
	case CDROMEJECT:
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_cmd_ioctl() : blk_get_request() .. \n"));
		rq = blk_get_request(q, WRITE, __GFP_WAIT);
		rq->flags |= REQ_BLOCK_PC;
		rq->data = NULL;
		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] scsi_cmd_ioctl() : Set Value : rq->data=%08XH \n",
			rq->data
			));
		rq->data_len = 0;
		rq->timeout = BLK_DEFAULT_TIMEOUT;
		// rq->cmd : unsigned char [16] blkdev.h
		// tmp_sense_buffer_pr = rq->cmd->sense_buffer ; // B. S. Liaw : backup
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_cmd_ioctl() : memset() .. \n" )) ;
		memset(rq->cmd, 0, sizeof(rq->cmd));
		// rq->cmd->sense_buffer =	tmp_sense_buffer_pr ; // B. S. Liaw : restore

		rq->cmd[0] = GPCMD_START_STOP_UNIT;
		rq->cmd[4] = 0x02 + (close != 0);
		rq->cmd_len = 6;
		err = blk_do_rq(q, bdev, rq);
		blk_put_request(rq);
		break;
	default:
		err = -ENOTTY;
   }

 blk_put_queue(q);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_cmd_ioctl() End. \n"));
 return err;
}

/*
EXPORT_SYMBOL(scsi_cmd_ioctl);
EXPORT_SYMBOL(scsi_command_size);
*/

#endif // ndef USBMS_NO_IOCTL

// #error : For Test ..

// ====================================================================
// End of File : blk_scsi_ioctl.c

