
// obj-$(CONFIG_MSDOS_PARTITION) += msdos.o

/*
 *  fs/partitions/msdos.c
 *
 *  Code extracted from drivers/block/genhd.c
 *  Copyright (C) 1991-1998  Linus Torvalds
 *
 *  Thanks to Branko Lankester, lankeste@fwi.uva.nl, who found a bug
 *  in the early extended-partition checks and added DM partitions
 *
 *  Support for DiskManager v6.0x added by Mark Lord,
 *  with information provided by OnTrack.  This now works for linux fdisk
 *  and LILO, as well as loadlin and bootln.  Note that disks other than
 *  /dev/hda *must* have a "DOS" type 0x51 partition in the first slot (hda1).
 *
 *  More flexible handling of extended partitions - aeb, 950831
 *
 *  Check partition table on IDE disks for common CHS translations
 *
 *  Re-organised Feb 1998 Russell King
 */

// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP


/*
#include <linux/config.h>
*/
#include "check.h"
#include "msdos.h"
#include "part.h"	// define in file system

#ifdef CONFIG_EFI_PARTITION
	#include "efi.h"
#endif // CONFIG_EFI_PARTITION

/*
 * Many architectures don't like unaligned accesses, while
 * the nr_sects and start_sect partition table entries are
 * at a 2 (mod 4) address.
 */
/*
#include <asm/unaligned.h>
*/
#if defined(_S3C2510)
	#include "unaligned-arm.h"
#elif defined(_AR7300)
	#include "unaligned-mips.h"
#elif defined(_ADM5120)
	#include "unaligned-mips.h"
#elif defined(_ATHEROS)
	#include "unaligned-mips.h"
#endif


#define SYS_IND(p)	(get_unaligned(&p->sys_ind))
#define NR_SECTS(p)	({ __typeof__(p->nr_sects) __a =	\
				get_unaligned(&p->nr_sects);	\
				le32_to_cpu(__a); \
			})

#define START_SECT(p)	({ __typeof__(p->start_sect) __a =	\
				get_unaligned(&p->start_sect);	\
				le32_to_cpu(__a); \
			})

/*
static inline int is_extended_partition(struct partition *p)
*/
static inline int is_extended_partition(struct hd_partition *p)
{
	return (SYS_IND(p) == DOS_EXTENDED_PARTITION ||
		SYS_IND(p) == WIN98_EXTENDED_PARTITION ||
		SYS_IND(p) == LINUX_EXTENDED_PARTITION);
}

#define MSDOS_LABEL_MAGIC1	0x55
#define MSDOS_LABEL_MAGIC2	0xAA

static inline int
msdos_magic_present(unsigned char *p)
{
 return (p[0] == MSDOS_LABEL_MAGIC1 && p[1] == MSDOS_LABEL_MAGIC2);
}

extern unsigned char PartitionType[20];

/*
 * Create devices for each logical partition in an extended partition.
 * The logical partitions form a linked list, with each entry being
 * a partition table with two entries.  The first entry
 * is the real data partition (with a start relative to the partition
 * table start).  The second is a pointer to the next logical partition
 * (with a start relative to the entire extended partition).
 * We do not create a Linux partition for the partition tables, but
 * only for the actual data partitions.
 */

// static void
static int
parse_extended(struct inode *inode,struct parsed_partitions *state, struct block_device *bdev,
			u32 first_sector, u32 first_size)
{
 /*
 struct partition *p;
 */
 struct hd_partition *p;
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	Sector sect;
  #else
 */
	unsigned char Sector_buf[USBMS_SECTOR_SIZE] ;
 /*
 #endif
 */
 unsigned char *data;
 u32 this_sector, this_size;
 int sector_size = bdev_hardsect_size(bdev) / 512;
 int loopct = 0;		/* number of links followed
				   without finding a data partition */
 int i;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] parse_extended() Begin .. \n"));
 this_sector = first_sector;
 this_size = first_size;

 while (1) {
	if (++loopct > 100)
			return;
	if (state->next == state->limit)
			return;
	
	//#ifndef USB_NO_VMM // vitrtul memory management
	//	data = read_dev_sector(bdev, this_sector, &sect);
	// #else
	
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] parse_extended(1) : read_dev_sector() .. \n"));
		
		// data = read_dev_sector(inode,bdev, this_sector, Sector_buf);
		
		
		// data = read_dev_sector(inode,bdev, this_sector,
		//  			disk_global_Sector_buf);
		
		data = disk_global_Sector_buf ;
		if ( read_dev_sector(inode,bdev, this_sector, 1, 
			 disk_global_Sector_buf) <=0 ) {
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] parse_extended(2) End with Failed !! \n"));
			return -EIO ;
		   }
	
	// #endif
	
	if (!data) {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] parse_extended(3) End with Failed !! \n"));
		return -EIO ;
	   }
	if (!msdos_magic_present(data + 510)) {
		goto done ;
	   }

	USBMS_DBG_PARTITION(iprintf(
		"[USBMS PARTITION] parse_extended(1) : Ms-Dos Partition Found \n"
		));
	// dprintf("[USBMS]        Ms-Dos Partition Found in Extended Partition \n");

	
	// p = (struct partition *) (data + 0x1be);
	
	p = (struct hd_partition *) (data + 0x1be);


		// * Usually, the first entry is the real data partition,
		// * the 2nd entry is the next extended partition, or empty,
		// * and the 3rd and 4th entries are unused.
		// * However, DRDOS sometimes has the extended partition as
		// * the first entry (when the data partition is empty),
		// * and OS/2 seems to use all four entries.
		 

		 
		// * First process the data partition(s)
	for (i=0; i<4; i++, p++) {
		u32 offs, size, next;

		if (!NR_SECTS(p) || is_extended_partition(p))
				continue;

		// Check the 3rd and 4th entries -
		//	   these sometimes contain random garbage
		
		offs = START_SECT(p)*sector_size;
		size = NR_SECTS(p)*sector_size;

		USBMS_DBG_PARTITION(iprintf(
			"[USBMS PARTITION] parse_extended(2) : i=%d \n",i
			));
		USBMS_DBG_PARTITION(iprintf(
			"[USBMS PARTITION] parse_extended(3) : offset=%08XH \n",
			this_sector + offs
			));
		USBMS_DBG_PARTITION(iprintf(
			"[USBMS PARTITION] parse_extended(4) : size=%08XH \n",size
			));
		if (!size) {
			USBMS_DBG_PARTITION(iprintf(
				"[USBMS PARTITION] parse_extended(5) : Zero Partirion !! \n"
				));
			// continue;
		   }
		dprintf("[USBMS]      Extended %d Start Sector = %08XH \n",
				i,this_sector + offs);
		dprintf("[USBMS]                         Size = %d (%08XH) Sectors \n",
				size,size);

		next = this_sector + offs;
		if (i >= 2) {
			if (offs + size > this_size)
					continue;
			if (next < first_sector)
					continue;
			if (next + size > first_sector + first_size)
					continue;
		   }

		USBMS_DBG_PARTITION(iprintf(
			"[USBMS PARTITION] parse_extended(6) : put_partition() .. \n"
			));
		USBMS_DBG_FLOW(iprintf(
			"[USBMS FLOW] parse_extended(4) : put_partition() .. \n"
			));
		PartitionType[state->next - 1] = (unsigned char) SYS_IND(p);
		put_partition(state, state->next, next, size);
		if (SYS_IND(p) == LINUX_RAID_PARTITION)
			state->parts[state->next].flags = 1;
		loopct = 0;
		if (++state->next == state->limit)
			goto done;
	   } // for i

	
	// * Next, process the (first) extended partition, if present.
	// * (So far, there seems to be no reason to make
	// *  parse_extended()  recursive and allow a tree
	// *  of extended partitions.)
	// * It should be a link to the next logical partition.
	
	p -= 4;
	for (i=0; i<4; i++, p++)
			if (NR_SECTS(p) && is_extended_partition(p))
				break;
	if (i == 4)
			goto done;	 // nothing left to do 

	this_sector = first_sector + START_SECT(p) * sector_size;
	this_size = NR_SECTS(p) * sector_size;
	
	//#ifndef USB_NO_VMM // vitrtul memory management
	//	put_dev_sector(sect);
	// #else
	
		put_dev_sector(Sector_buf);
	
	//#endif
	
   } // while 1

 // -------------
 done:
 // -------------

 //#ifndef USB_NO_VMM // vitrtul memory management
//	put_dev_sector(sect);
//  #else
 
	put_dev_sector(Sector_buf);
 
 // #endif
 
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] parse_extended() End. \n"));
 return 1 ;
}

/* james@bpgc.com: Solaris has a nasty indicator: 0x82 which also
   indicates linux swap.  Be careful before believing this is Solaris. */

static void
parse_solaris_x86(struct parsed_partitions *state, struct block_device *bdev,
			u32 offset, u32 size, int origin)
{
 #ifdef CONFIG_SOLARIS_X86_PARTITION
	// Sector sect;
	struct solaris_x86_vtoc *v;
	int i;

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] parse_solaris_x86() : read_dev_sector() .. \n"));
	/*
	v = (struct solaris_x86_vtoc *)read_dev_sector(inode,bdev, offset+1, &sect);
	*/
	/*
	v = (struct solaris_x86_vtoc *)read_dev_sector(inode,bdev, offset+1,
		disk_global_Sector_buf);
	*/
	v = (struct solaris_x86_vtoc *)	disk_global_Sector_buf ;
	if ( read_dev_sector(inode,bdev, offset+1, 1, disk_global_Sector_buf) <=0 ) {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] parse_solaris_x86() End with Failed !! \n"));
		return -1;
	   }

	if (!v)
		return;
	if (le32_to_cpu(v->v_sanity) != SOLARIS_X86_VTOC_SANE) {
		put_dev_sector(sect);
		return;
	}
	printk(" %s%d: <solaris:", state->name, origin);
	if (le32_to_cpu(v->v_version) != 1) {
		printk("  cannot handle version %d vtoc>\n",
			le32_to_cpu(v->v_version));
		put_dev_sector(sect);
		return;
	}
	for (i=0; i<SOLARIS_X86_NUMSLICE && state->next<state->limit; i++) {
		struct solaris_x86_slice *s = &v->v_slice[i];
		if (s->s_size == 0)
			continue;
		printk(" [s%d]", i);
		/* solaris partitions are relative to current MS-DOS
		 * one; must add the offset of the current partition */
		USBMS_DBG_PARTITION(iprintf("[USBMS PARTITION] parse_solaris_x86() : put_partition() .. \n"));
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] parse_solaris_x86() : put_partition() .. \n"));
		put_partition(state, state->next++,
				 le32_to_cpu(s->s_start)+offset,
				 le32_to_cpu(s->s_size));
	}
	put_dev_sector(sect);
	printk(" >\n");
#endif
}

#if defined(CONFIG_BSD_DISKLABEL) || defined(CONFIG_NEC98_PARTITION)
/* 
 * Create devices for BSD partitions listed in a disklabel, under a
 * dos-like partition. See parse_extended() for more information.
 */
void
parse_bsd(struct parsed_partitions *state, struct block_device *bdev,
		u32 offset, u32 size, int origin, char *flavour,
		int max_partitions)
{
 // Sector sect;
 struct bsd_disklabel *l;
 struct bsd_partition *p;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] parse_bsd() : read_dev_sector() .. \n"));
 /*
	l = (struct bsd_disklabel *)read_dev_sector(inode, bdev, offset+1, &sect);
 */
 /*
 l = (struct bsd_disklabel *)read_dev_sector(inode, bdev, offset+1,
		disk_global_Sector_buf);
 */
 l = (struct unixware_disklabel *) disk_global_Sector_buf ;
 if( read_dev_sector(inode,bdev, offset+1, 1, disk_global_Sector_buf) <=0 ) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] parse_bsd() End with Failed !! \n"));
	return -1;
   }

 if (!l)
		return;
	if (le32_to_cpu(l->d_magic) != BSD_DISKMAGIC) {
		put_dev_sector(sect);
		return;
	}
	printk(" %s%d: <%s:", state->name, origin, flavour);

	if (le16_to_cpu(l->d_npartitions) < max_partitions)
		max_partitions = le16_to_cpu(l->d_npartitions);
	for (p = l->d_partitions; p - l->d_partitions < max_partitions; p++) {
		u32 bsd_start, bsd_size;

		if (state->next == state->limit)
			break;
		if (p->p_fstype == BSD_FS_UNUSED) 
			continue;
		bsd_start = le32_to_cpu(p->p_offset);
		bsd_size = le32_to_cpu(p->p_size);
		if (offset == bsd_start && size == bsd_size)
			/* full parent partition, we have it already */
			continue;
		if (offset > bsd_start || offset+size < bsd_start+bsd_size) {
			printk("bad subpartition - ignored\n");
			continue;
		}
		USBMS_DBG_PARTITION(iprintf("[USBMS PARTITION] parse_bsd() : put_partition() .. \n"));
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] parse_bsd() : put_partition() .. \n"));
		put_partition(state, state->next++, bsd_start, bsd_size);
	}
	put_dev_sector(sect);
	printk(" >\n");
}
#endif

static void
parse_freebsd(struct parsed_partitions *state, struct block_device *bdev,
		u32 offset, u32 size, int origin)
{
#ifdef CONFIG_BSD_DISKLABEL
	parse_bsd(state, bdev, offset, size, origin,
			"bsd", BSD_MAXPARTITIONS);
#endif
}

static void
parse_netbsd(struct parsed_partitions *state, struct block_device *bdev,
		u32 offset, u32 size, int origin)
{
#ifdef CONFIG_BSD_DISKLABEL
	parse_bsd(state, bdev, offset, size, origin,
			"netbsd", BSD_MAXPARTITIONS);
#endif
}

static void
parse_openbsd(struct parsed_partitions *state, struct block_device *bdev,
		u32 offset, u32 size, int origin)
{
#ifdef CONFIG_BSD_DISKLABEL
	parse_bsd(state, bdev, offset, size, origin,
			"openbsd", OPENBSD_MAXPARTITIONS);
#endif
}

/*
 * Create devices for Unixware partitions listed in a disklabel, under a
 * dos-like partition. See parse_extended() for more information.
 */
static void
parse_unixware(struct parsed_partitions *state, struct block_device *bdev,
		u32 offset, u32 size, int origin)
{
 #ifdef CONFIG_UNIXWARE_DISKLABEL
	// Sector sect;
	struct unixware_disklabel *l;
	struct unixware_slice *p;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] parse_unixware() : read_dev_sector() .. \n"));
 /*
 l = (struct unixware_disklabel *)read_dev_sector(inode,bdev, offset+29, &sect);
 */
 /*
 l = (struct unixware_disklabel *)read_dev_sector(inode,bdev, offset+29,
		disk_global_Sector_buf);
 */
 l = (struct unixware_disklabel *) disk_global_Sector_buf ;
 if( read_dev_sector(inode,bdev, offset+29, 1, disk_global_Sector_buf) <=0 ) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] parse_unixware() End with Failed !! \n"));
	return -1;
   }

 if (!l)
		return;
 if (le32_to_cpu(l->d_magic) != UNIXWARE_DISKMAGIC ||
	    le32_to_cpu(l->vtoc.v_magic) != UNIXWARE_DISKMAGIC2) {
		put_dev_sector(sect);
		return;
	}
	printk(" %s%d: <unixware:", state->name, origin);
	p = &l->vtoc.v_slice[1];
	/* I omit the 0th slice as it is the same as whole disk. */
	while (p - &l->vtoc.v_slice[0] < UNIXWARE_NUMSLICE) {
		if (state->next == state->limit)
			break;

		if (p->s_label != UNIXWARE_FS_UNUSED)
		   {
			USBMS_DBG_PARTITION(iprintf("[USBMS PARTITION] parse_unixware() : put_partition() .. \n"));
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] parse_unixware() : put_partition() .. \n"));
			put_partition(state, state->next++,
						START_SECT(p), NR_SECTS(p));
		   }
		p++;
	}
	put_dev_sector(sect);
	printk(" >\n");
#endif
}

/*
 * Minix 2.0.0/2.0.2 subpartition support.
 * Anand Krishnamurthy <anandk@wiproge.med.ge.com>
 * Rajeev V. Pillai    <rajeevvp@yahoo.com>
 */
static void
parse_minix(struct parsed_partitions *state, struct block_device *bdev,
		u32 offset, u32 size, int origin)
{
 #ifdef CONFIG_MINIX_SUBPARTITION
	// Sector sect;
	unsigned char *data;
	/*
	struct partition *p;
	*/
	struct hd_partition *p;
	int i;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] parse_minix() : read_dev_sector() .. \n"));
 /*
 data = read_dev_sector(inode,bdev, offset, &sect);
 */
 // data = read_dev_sector(inode,bdev, offset, disk_global_Sector_buf);
 data = disk_global_Sector_buf ;
 if( read_dev_sector(inode,bdev, offset,
				1 ,
	disk_global_Sector_buf) <=0 ) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] parse_minix() End with Failed !! \n"));
	return -1;
   }

 if (!data)
		return;
 /*
	p = (struct partition *)(data + 0x1be);
 */
	p = (struct hd_partition *)(data + 0x1be);

	/* The first sector of a Minix partition can have either
	 * a secondary MBR describing its subpartitions, or
	 * the normal boot sector. */
	if (msdos_magic_present (data + 510) &&
	    SYS_IND(p) == MINIX_PARTITION) { /* subpartition table present */

		printk(" %s%d: <minix:", state->name, origin);
		for (i = 0; i < MINIX_NR_SUBPARTITIONS; i++, p++) {
			if (state->next == state->limit)
				break;
			/* add each partition in use */
			if (SYS_IND(p) == MINIX_PARTITION) {
				USBMS_DBG_PARTITION(iprintf("[USBMS PARTITION] parse_minix() : put_partition() .. \n"));
				USBMS_DBG_FLOW(iprintf("[USBMS FLOW] parse_minix() : put_partition() .. \n"));
				put_partition(state, state->next++,
					      START_SECT(p), NR_SECTS(p));
			   }
		}
		printk(" >\n");
	}
	put_dev_sector(sect);
#endif /* CONFIG_MINIX_SUBPARTITION */
}

static struct {
	unsigned char id;
	void (*parse)(struct parsed_partitions *, struct block_device *,
			u32, u32, int);
} subtypes[] = {
	{FREEBSD_PARTITION, parse_freebsd},
	{NETBSD_PARTITION, parse_netbsd},
	{OPENBSD_PARTITION, parse_openbsd},
	{MINIX_PARTITION, parse_minix},
	{UNIXWARE_PARTITION, parse_unixware},
	{SOLARIS_X86_PARTITION, parse_solaris_x86},
	{0, NULL},
};
 
int msdos_partition(struct inode *inode,struct parsed_partitions *state, struct block_device *bdev)
{
 /*
 int sector_size = bdev_hardsect_size(bdev) / 512;
 */
 int sector_size;
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	Sector sect;
  #else
 */
 // B. S. Liaw : Sector_buf need uncached
 // unsigned char Sector_buf[USBMS_SECTOR_SIZE] ;
 unsigned char *Sector_buf ;
 /*
 #endif
 */
 unsigned char *data;
 /*
 struct partition *p;
 */
 struct hd_partition *p;
 int slot;
#if 1  /* Charles 20040524 */
 int hardsect_size;
#endif
#ifdef _ETCPIP
 struct hd_partition *p_start;
 char partition_slot[1+4], partition_type[1+4];
 unsigned long partition_start[1+4];
 unsigned long partition_size[1+4];
#endif	// _ETCPIP

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] msdos_partition() Begin .. \n"));

 /*
 #ifdef USBMS_ONLY_ONE_DISK
	// B. S. Liaw : disk_global_Sector_buf need uncached
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] msdos_partition() : disk_global_Sector_buf = UsbAllocMemU() size=%d .. \n",
		USBMS_SECTOR_SIZE
		));

	disk_global_Sector_buf = (unsigned char *) UsbAllocMemU(USB_MEM_U_M
				#ifdef USB_DBG_MEM_U_ON
					, USBMS_SECTOR_SIZE   // size : 512
		        #endif
           		) ;
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] msdos_partition() : disk_global_Sector_buf=%08XH \n",
		disk_global_Sector_buf
		));
	 if (!disk_global_Sector_buf)
	   {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] msdos_partition(1) End with Failed !! \n"));
		return -1;
	   }
  #else
 */
	/*
	// B. S. Liaw : Sector_buf need uncached
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] msdos_partition() : Sector_buf = UsbAllocMemU() size=%d .. \n",
		USBMS_SECTOR_SIZE
		));

	// BLOCK_PREEMPTION;
	Sector_buf = (unsigned char *) UsbAllocMemU(USB_MEM_U_M
				#ifdef USB_DBG_MEM_U_ON
					, USBMS_SECTOR_SIZE   // size : 512
		        #endif
           		) ;
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] msdos_partition() : Sector_buf=%08XH \n",
		Sector_buf
		));
	disk_global_Sector_buf = Sector_buf ;
	// UNBLOCK_PREEMPTION;
	*/
	Sector_buf = disk_global_Sector_buf ;
	if (!Sector_buf) {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] msdos_partition(2) End with Failed !! \n"));
		return -1;
	   }
	
	
 /*
 #endif // nedf USBMS_ONLY_ONE_DISK
 */

 USBMS_DBG_MEM(iprintf("[USBMS MEM] msdos_partition() : bdev=%08XH \n",bdev));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] msdos_partition(3) : bdev_hardsect_size() .. \n"));
 hardsect_size = bdev_hardsect_size(bdev);
 sector_size = hardsect_size / 512;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] msdos_partition(4) : read_dev_sector() .. \n"));
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	data = read_dev_sector(bdev, 0, &sect);
  #else
 */
 /*
 #ifdef USBMS_ONLY_ONE_DISK
	data = read_dev_sector(inode,bdev, 0, disk_global_Sector_buf);
  #else
 */
	/*
	data = read_dev_sector(inode,bdev, 0, Sector_buf);
	*/
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] msdos_partition() : disk_global_Sector_buf=%08XH \n",
		disk_global_Sector_buf
		));
	data = disk_global_Sector_buf ;
	if ( read_dev_sector(inode,bdev, 0, 1 , disk_global_Sector_buf) <=0  ) {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] msdos_partition(5) End with Failed !! \n"));
		return -1;
	   }

 #ifdef USBMS_DBG_INIT_ON	   
	for (slot=0; slot < hardsect_size; slot++)
	{
		iprintf("%02X ", data[slot]);
		if ((slot & 0x0f)== 0x0f)
			iprintf("\n");
	}
 #endif
 /*
 #endif // nedf USBMS_ONLY_ONE_DISK
 */
 /*
 #endif
 */
 if(!data) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] msdos_partition(6) End with Failed !! \n"));
	return -1;
   }
 
#if 0  /* Charles 20040429 */
 if(!msdos_magic_present(data + 510)) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] msdos_partition(7) : put_dev_sector() .. \n"));
	/*
	#ifndef USB_NO_VMM // vitrtul memory management
		put_dev_sector(sect);
	 #else
	*/
	/*
	#ifdef USBMS_ONLY_ONE_DISK
		put_dev_sector(disk_global_Sector_buf);
	 #else
	*/
		put_dev_sector(Sector_buf);
	/*
	#endif // nedf USBMS_ONLY_ONE_DISK
	*/
	/*
	#endif
	*/
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] msdos_partition(8) End with Failed !! \n"));
	return 0;
   }
#endif

 USBMS_DBG_PARTITION(iprintf(
	"[USBMS PARTITION] msdos_partition(1) : Ms-Dos Partition Found \n"
	));
 // dprintf("[USBMS] Ms-Dos Partition Found \n");

 
 #ifdef USBMS_ONLY_ONE_DISK
	// 2003/12/17 B. S. Liaw add these ..
	// on fs/check.c
	
	//disk_global_inode = inode ;
	//USBMS_DBG_MEM(iprintf(
	//	"[USBMS MEM] msdos_partition() : disk_global_inode=%08XH \n",
	//	disk_global_inode
	//	));
	
	disk_global_bdev[0] = bdev ;

	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] msdos_partition() : Set Value : disk_global_bdev[0]=%08XH \n",
		disk_global_bdev[0]
		));
 #endif // USBMS_ONLY_ONE_DISK

 // for those disk with boot sector in the front (without MBR)
 // 0xEB 3C 90 for FAT16 - by Microsoft
 // 0xEB 58 90 for FAT32 - by Microsoft
 // 0xEB 52 90 for NTFS - by Microsoft
 if ( ((data[0] == 0xeb) && ((data[1] == 0x3c) || (data[1] == 0x3e) || (data[1] == 0x52) || (data[1] == 0x58)) && (data[2] == 0x90))
      || (!msdos_magic_present(data + 510)) )
// if ((data[0] == 0xeb) && ((data[1] == 0x3c) || (data[1] == 0x3e) || (data[1] == 0x58)) && (data[2] == 0x90))
 {
 	sector_t size;
 	
 	state->next = 1;
#if 1  /* Charles 20040524 */
	size = inode->i_size;
#else
 	size = inode->i_size >> 9;
#endif
	dprintf("[USBMS] Add Virtual Partition 0 Start Sector = %08XH \n", 0);
	dprintf("[USBMS]                                 Size = %d (%08XH) Sectors \n", size,size);
	PartitionType[state->next - 1] = PARTITION_ID_SINGLE;
	put_partition(state, state->next, 0, size);
	put_dev_sector(Sector_buf);
	return 1;
 }

 /*
 p = (struct partition *) (data + 0x1be);
 */
 p = (struct hd_partition *) (data + 0x1be);
 #ifdef CONFIG_EFI_PARTITION
	for (slot = 1 ; slot <= 4 ; slot++, p++) {
		/* If this is an EFI GPT disk, msdos should ignore it. */
		if (SYS_IND(p) == EFI_PMBR_OSTYPE_EFI_GPT) {
			/*
			#ifndef USB_NO_VMM // vitrtul memory management
				put_dev_sector(sect);
			 #else
			*/
			/*
			#ifdef USBMS_ONLY_ONE_DISK
				put_dev_sector(disk_global_Sector_buf);
			 #else
			*/
				put_dev_sector(Sector_buf);
			/*
			#endif // nedf USBMS_ONLY_ONE_DISK
			*/
			/*
			#endif
			*/
			return 0;
		}
	}
	/*
	p = (struct partition *) (data + 0x1be);
	*/
	p = (struct hd_partition *) (data + 0x1be);
 #endif

 /*
  * Look for partitions in two passes:
  * First find the primary and DOS-type extended partitions.
  * On the second pass look inside *BSD, Unixware and Solaris partitions.
  */

 #ifndef _ETCPIP
	state->next = 5 ;
  #else
	state->next = 0 ;
 #endif
#ifdef _ETCPIP
 {
 	int		slot_compare;
 	char	slot_swap;
 	p_start = p;
	for (slot = 1 ; slot <= 4 ; slot++, p++) {
		partition_slot[slot] = slot;
		partition_type[slot] = SYS_IND(p);
		partition_start[slot] = START_SECT(p);
		partition_size[slot] = NR_SECTS(p);
	}
	for (slot = 1; slot <= 3; slot++) {
		for (slot_compare = slot+1; slot_compare <= 4; slot_compare++) {
			slot_swap = 0;
			if (partition_type[slot_compare] != 0) {
				if (partition_type[slot] == 0)	slot_swap = 1;
				else if (partition_start[slot_compare] < partition_start[slot])	slot_swap = 1;
			}
			if (slot_swap != 0) {
				partition_slot[0] = partition_slot[slot_compare];
				partition_type[0] = partition_type[slot_compare];
				partition_start[0] = partition_start[slot_compare];
				partition_size[0] = partition_size[slot_compare];
				partition_slot[slot_compare] = partition_slot[slot];
				partition_type[slot_compare] = partition_type[slot];
				partition_start[slot_compare] = partition_start[slot];
				partition_size[slot_compare] = partition_size[slot];
				partition_slot[slot] = partition_slot[0];
				partition_type[slot] = partition_type[0];
				partition_start[slot] = partition_start[0];
				partition_size[slot] = partition_size[0];
			}
		}
	}
	if (partition_type[1] == 0) {
		partition_start[1] = 0;
		partition_size[1] = inode->i_size;
	}
 }
 for (slot = 1 ; slot <= 4 ; slot++) {
	u32 start = partition_start[slot]*sector_size;
	u32 size = partition_size[slot]*sector_size;
	p = p_start + partition_slot[slot] - 1;
#else	// _ETCPIP
 for (slot = 1 ; slot <= 4 ; slot++, p++) {
	u32 start = START_SECT(p)*sector_size;
	u32 size = NR_SECTS(p)*sector_size;
#endif	// _ETCPIP

	USBMS_DBG_PARTITION(iprintf(
		"[USBMS PARTITION] msdos_partition(2) : slot=%d \n",slot
		));
	USBMS_DBG_PARTITION(iprintf(
		"[USBMS PARTITION] msdos_partition(3) : start=%08XH \n",start
		));
	USBMS_DBG_PARTITION(iprintf(
		"[USBMS PARTITION] msdos_partition(4) : size=%08XH \n",size
		));
	if (!size) {
		USBMS_DBG_PARTITION(iprintf(
			"[USBMS PARTITION] msdos_partition(5) : Zero Partirion !! \n"
			));
		continue;
	   }
	dprintf("[USBMS] Partition %d Start Sector = %08XH \n",slot,start);
	dprintf("[USBMS]                     Size = %d (%08XH) Sectors \n",
			size,size);

	#ifdef _ETCPIP
		state->next ++ ;
	#endif // _ETCPIP

	if (is_extended_partition(p)) {
		USBMS_DBG_PARTITION(iprintf(
			"[USBMS PARTITION] msdos_partition(6) : Extended Partition \n"
			));
		// dprintf("[USBMS]             Extended Partition \n") ;
		/* prevent someone doing mkfs or mkswap on an
		   extended partition, but leave room for LILO */
		/*
		#ifndef _ETCPIP
			//#ifdef _ETCPIP
			//if (!size) {
			//	USBMS_DBG_PARTITION(iprintf(
			//		"[USBMS PARTITION] msdos_partition(7) : Zero Partirion !! \n"
			//		));
			//   }
			//  else
			//#endif // _ETCPIP
			   {
				USBMS_DBG_FLOW(iprintf("[USBMS FLOW] msdos_partition(9) : put_partition() .. \n"));
				put_partition(state, slot, start, size == 1 ? 1 : 2);
			   }
		#endif // ndef _ETCPIP
		*/
		/*
		printk(" <");
		*/
		USBMS_DBG_PARTITION(iprintf(
			"[USBMS PARTITION] msdos_partition(8) : Parse Extended Partition .. \n"
			));
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] msdos_partition(10) : parse_extended() .. \n"));
		if (parse_extended(inode,state, bdev, start, size) <= 0 ) {
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] msdos_partition(11) End with Failed !! \n"));
			return -EIO ;
		   }
		/*
		printk(" >");
		*/
		continue;
	   } // if is_extended_partition

	USBMS_DBG_PARTITION(iprintf(
		"[USBMS PARTITION] msdos_partition(9) : put_partition() .. \n"
		));
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] msdos_partition(12) : put_partition() .. \n"));
	/*
	#ifndef _ETCPIP
		put_partition(state, slot, start, size);
	 #else
	*/
	PartitionType[state->next - 1] = (unsigned char) SYS_IND(p);
	put_partition(state, state->next, start, size);
	if (SYS_IND(p) == LINUX_RAID_PARTITION)
		state->parts[slot].flags = 1;
   } // for slot
 /*
 printk("\n");
 */
 /* second pass - output for each on a separate line */
 /*
 p = (struct partition *) (0x1be + data);
 */
 p = (struct hd_partition *) (0x1be + data);
 for (slot = 1 ; slot <= 4 ; slot++, p++) {
	unsigned char id = SYS_IND(p);
	int n;

	if (!NR_SECTS(p))
			continue;

	for (n = 0; subtypes[n].parse && id != subtypes[n].id; n++)
			;

	if (!subtypes[n].parse)
			continue;
	USBMS_DBG_PARTITION(iprintf(
		"[USBMS PARTITION] msdos_partition(10) : Parse Sub-Type .. \n"
		));
	subtypes[n].parse(state, bdev, START_SECT(p)*sector_size,
						NR_SECTS(p)*sector_size, slot);
   }

 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	put_dev_sector(sect);
  #else
 */
 /*
 #ifdef USBMS_ONLY_ONE_DISK
	put_dev_sector(disk_global_Sector_buf);
  #else
 */
	put_dev_sector(Sector_buf);
 /*
 #endif // nedf USBMS_ONLY_ONE_DISK
 */
 /*
 #endif
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] msdos_partition() End. \n"));
 return 1;
}
