
// drivers/usb/storage/initializers.c

/* Special Initializers for certain USB Mass Storage devices
 *
 * $Id: initializers.c,v 1.2 2000/09/06 22:35:57 mdharm Exp $
 *
 * Current development and maintenance by:
 *   (c) 1999, 2000 Matthew Dharm (mdharm-usb@one-eyed-alien.net)
 *
 * This driver is based on the 'USB Mass Storage Class' document. This
 * describes in detail the protocol used to communicate with such
 * devices.  Clearly, the designers had SCSI and ATAPI commands in
 * mind when they created this document.  The commands are all very
 * similar to commands in the SCSI-II and ATAPI specifications.
 *
 * It is important to note that in a number of cases this class
 * exhibits class-specific exemptions from the USB specification.
 * Notably the usage of NAK, STALL and ACK differs from the norm, in
 * that they are used to communicate wait, failed and OK on commands.
 *
 * Also, for certain devices, the interrupt endpoint is used to convey
 * status of a command.
 *
 * Please see http://www.one-eyed-alien.net/~mdharm/linux-usb for more
 * information about this driver.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

// #ifdef _ETCPIP
#include "feature.h"
#define __KERNEL__
        #ifdef _NEWLIB
          #include "string.h"
        #endif
   	#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
      // for deciding whether to compile
      //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"

#include "PoList.h" // osport_list.h

#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_TIMER
	#include "PoTimer.h" // osport_timer.h
#endif // ndef NO_TIMER
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
// #endif // _ETCPIP

#ifndef _ETCPIP
	#include <linux/sched.h>
#endif // ndef _ETCPIP
#ifndef _ETCPIP
   #include <linux/errno.h>
 #else
	#include "PoErrno.h" // osport_errno.h
#endif
/*
#include "initializers.h"
*/
#include "ms_init.h"
/*
#include "debug.h"
*/
#include "ms_debug.h"
/*
#include "transport.h"
*/
#include "ms_transport.h"

/* This places the Shuttle/SCM USB<->SCSI bridge devices in multi-target
 * mode */
int usb_stor_euscsi_init(struct us_data *us)
{
 int result;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_euscsi_init() Begin ..\n"));
 US_DEBUGP("Attempting to init eUSCSI bridge...\n");
	us->iobuf[0] = 0x1;

 // #ifdef USB_NO_VMM // no vitrtul memory management
 #ifdef USBMS_BS_ASSERT
	/*
	#if defined(_S3C2510)
		if (!(  (unsigned long) us
              > ((unsigned long)0x00800000L)
	         )
    	   )
	#elif defined(_ADM5120)
		if (!(     ((unsigned long) us
				 & (unsigned long)0x20000000L)
   	          ==((unsigned long)0x20000000L)
   		     )
	       )
	#endif
		   {
	*/
		if ( !(USB_IS_UN_CACHED(us)) ) {
			iprintf("[ASSERT][USBMS] usb_stor_euscsi_init() : " ) ;
   		    iprintf("Error : Not Un-Cached !!\n" ) ;
		    iprintf("[ASSERT][USBMS] us addr = %08XH \n", us ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
   		   }
 #endif // USBMS_BS_ASSERT
 // #endif // USB_NO_VMM

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_euscsi_init() : usb_stor_control_msg() ..\n"));
 result = usb_stor_control_msg(us, us->send_ctrl_pipe,
			0x0C, USB_RECIP_INTERFACE | USB_TYPE_VENDOR,
			/*
			0x01, 0x0, us->iobuf, 0x1, 5*HZ);
			*/
			0x01, 0x0, us->iobuf, 0x1, USBMS_EUSCSI_TIMEOUT);

 US_DEBUGP("-- result is %d\n", result);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_euscsi_init() End.\n"));
 return 0;
}

/* This function is required to activate all four slots on the UCR-61S2B
 * flash reader */
int usb_stor_ucr61s2b_init(struct us_data *us)
{
 struct bulk_cb_wrap *bcb = (struct bulk_cb_wrap*) us->iobuf;
 struct bulk_cs_wrap *bcs = (struct bulk_cs_wrap*) us->iobuf;
 int res, partial;
 static char init_string[] = "\xec\x0a\x06\x00$PCCHIPS";


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_ucr61s2b_init() Begin .. \n")) ;
 US_DEBUGP("Sending UCR-61S2B initialization packet...\n");

 bcb->Signature = cpu_to_le32(US_BULK_CB_SIGN);
 bcb->Tag = 0;
 bcb->DataTransferLength = cpu_to_le32(0);
 bcb->Flags = bcb->Lun = 0;
 bcb->Length = sizeof(init_string) - 1;
 memset(bcb->CDB, 0, sizeof(bcb->CDB));
 memcpy(bcb->CDB, init_string, sizeof(init_string) - 1);

 USBMS_DBG_FLOW(iprintf(
	"[USBMS FLOW] usb_stor_ucr61s2b_init() : usb_stor_bulk_transfer_buf() .. \n"
	)) ;

 res = usb_stor_bulk_transfer_buf(us, us->send_bulk_pipe, bcb,
			US_BULK_CB_WRAP_LEN, &partial);

 if(res==-ENODEV) { 
	USBMS_DBG_PLUG(iprintf(
		"[USBMS PLUG] usb_stor_ucr61s2b_init() : during disconnect !! \n"
		)) ;
	// return -ENODEV ;
   }
 if(res == -ENOMEM) { 
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] usb_stor_ucr61s2b_init() : Error : Out of Memory !! \n"
		)) ;
	// return -ENOMEM ;
   }
 if((res==-ENODEV) || (res==-ENOMEM)) { 
	return res ;
   }


 if(res) {
	return res;
   }

 US_DEBUGP("Getting status packet...\n");
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_ucr61s2b_init() : usb_stor_bulk_transfer_buf() .. \n")) ;
 res = usb_stor_bulk_transfer_buf(us, us->recv_bulk_pipe, bcs,
			US_BULK_CS_WRAP_LEN, &partial);


 if(res==-ENODEV) { 
	USBMS_DBG_PLUG(iprintf(
		"[USBMS PLUG] usb_stor_ucr61s2b_init() : during disconnect !! \n"
		)) ;
	// return -ENODEV ;
   }
 if (res == -ENOMEM) { 
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] usb_stor_ucr61s2b_init() : Error : Out of Memory !! \n"
		)) ;
	// return -ENOMEM ;
   }
 if ((res==-ENODEV) || (res==-ENOMEM)) { 
	return res ;
   }

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_ucr61s2b_init() End. \n")) ;
 return (res ? -1 : 0);
}
