
// Name of File : ms_usb.c
// ====================================================================

/* Driver for USB Mass Storage compliant devices
 *
 * $Id: usb.c,v 1.75 2002/04/22 03:39:43 mdharm Exp $
 *
 * Current development and maintenance by:
 *   (c) 1999-2003 Matthew Dharm (mdharm-usb@one-eyed-alien.net)
 *
 * Developed with the assistance of:
 *   (c) 2000 David L. Brown, Jr. (usb-storage@davidb.org)
 *   (c) 2003 Alan Stern (stern@rowland.harvard.edu)
 *
 * Initial work by:
 *   (c) 1999 Michael Gee (michael@linuxspecific.com)
 *
 * usb_device_id support by Adam J. Richter (adam@yggdrasil.com):
 *   (c) 2000 Yggdrasil Computing, Inc.
 *
 * This driver is based on the 'USB Mass Storage Class' document. This
 * describes in detail the protocol used to communicate with such
 * devices.  Clearly, the designers had SCSI and ATAPI commands in
 * mind when they created this document.  The commands are all very
 * similar to commands in the SCSI-II and ATAPI specifications.
 *
 * It is important to note that in a number of cases this class
 * exhibits class-specific exemptions from the USB specification.
 * Notably the usage of NAK, STALL and ACK differs from the norm, in
 * that they are used to communicate wait, failed and OK on commands.
 *
 * Also, for certain devices, the interrupt endpoint is used to convey
 * status of a command.
 *
 * Please see http://www.one-eyed-alien.net/~mdharm/linux-usb for more
 * information about this driver.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
          // for deciding whether to compile
          //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
   
#include "PoList.h" // osport_list.h
   
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsbMsInit.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
		#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
// #endif // _ETCPIP

/*
#include <linux/config.h>
#include <linux/sched.h>
#include <linux/errno.h>
*/
#include "PoErrno.h" // osport_errno.h

// #ifdef USB_MEM_LIB
#include "UsMemLib.h" // UsbMemLib.h
// #endif // USB_MEM_LIB
/*    
#include "usb.h"
*/
#include "ms_usb.h"
/*    
#include "scsiglue.h"
*/
#include "ms_scsiglue.h"
/*    
#include "transport.h"
*/
#include "ms_transport.h"
/*    
#include "protocol.h"
*/
#include "ms_protocol.h"
/*    
#include "debug.h"
*/
#include "ms_debug.h"
/*    
#include "initializers.h"
*/
#include "ms_init.h"

#include "hub.h"


#ifdef CONFIG_USB_STORAGE_HP8200e
#include "shuttle_usbat.h"
#endif
#ifdef CONFIG_USB_STORAGE_SDDR09
#include "sddr09.h"
#endif
#ifdef CONFIG_USB_STORAGE_SDDR55
#include "sddr55.h"
#endif
#ifdef CONFIG_USB_STORAGE_DPCM
#include "dpcm.h"
#endif
#ifdef CONFIG_USB_STORAGE_FREECOM
#include "freecom.h"
#endif
#ifdef CONFIG_USB_STORAGE_ISD200
#include "isd200.h"
#endif
#ifdef CONFIG_USB_STORAGE_DATAFAB
#include "datafab.h"
#endif
#ifdef CONFIG_USB_STORAGE_JUMPSHOT
#include "jumpshot.h"
#endif


/*
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/errno.h>
*/
#include "PoErrno.h" // osport_errno.h
/*
#include <linux/init.h>
#include <linux/slab.h>
*/

#include "config.h"
#include "part.h"
#include "nbios.h"

extern int validDiskCnt;
extern int validPartitionCnt;
extern unsigned char PartitionType[20];
extern unsigned long validPartitionID[20];
extern unsigned long partitionMask[32];

// #ifdef _ETCPIP
// B. S. Liaw Add these ..
static LIST_HEAD(msusb_shost_my_devices);
struct usb_device *global_usbms_usb_device = (struct usb_device *) NULL ;
#ifdef USBMS_SCSI_SHORTCUT  /* Charles 20040526 */
struct us_data *MSUsb_us;
#endif
// #endif // _ETCPIP


/* Some informational data */
/*
MODULE_AUTHOR("Matthew Dharm <mdharm-usb@one-eyed-alien.net>");
MODULE_DESCRIPTION("USB Mass Storage driver for Linux");
MODULE_LICENSE("GPL");
*/
static int storage_probe(struct usb_interface *iface,
			 const struct usb_device_id *id);

static void storage_disconnect(struct usb_interface *iface);

/* The entries in this table, except for final ones here
 * (USB_MASS_STORAGE_CLASS and the empty entry), correspond,
 * line for line with the entries of us_unsuaul_dev_list[].
 */

#define UNUSUAL_DEV(id_vendor, id_product, bcdDeviceMin, bcdDeviceMax, \
		    vendorName, productName,useProtocol, useTransport, \
		    initFunction, flags) \
{ USB_DEVICE_VER(id_vendor, id_product, bcdDeviceMin,bcdDeviceMax) }

/*
static struct usb_device_id storage_usb_ids [] = {
*/
static const struct usb_device_id storage_usb_ids [] = {

#	include "unusual_devs.h"

#undef UNUSUAL_DEV

	/* Control/Bulk transport for all SubClass values */
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_RBC, US_PR_CB) },
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_8020, US_PR_CB) },
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_QIC, US_PR_CB) },
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_UFI, US_PR_CB) },
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_8070, US_PR_CB) },
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_SCSI, US_PR_CB) },

	/* Control/Bulk/Interrupt transport for all SubClass values */
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_RBC, US_PR_CBI) },
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_8020, US_PR_CBI) },
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_QIC, US_PR_CBI) },
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_UFI, US_PR_CBI) },
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_8070, US_PR_CBI) },
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_SCSI, US_PR_CBI) },

	/* Bulk-only transport for all SubClass values */
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_RBC, US_PR_BULK) },
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_8020, US_PR_BULK) },
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_QIC, US_PR_BULK) },
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_UFI, US_PR_BULK) },
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_8070, US_PR_BULK) },
	{ USB_INTERFACE_INFO(USB_CLASS_MASS_STORAGE, US_SC_SCSI, US_PR_BULK) },

	/* Terminating entry */
	{ }
};

/*
MODULE_DEVICE_TABLE (usb, storage_usb_ids);
*/

/* This is the list of devices we recognize, along with their flag data */

/* The vendor name should be kept at eight characters or less, and
 * the product name should be kept at 16 characters or less. If a device
 * has the US_FL_FIX_INQUIRY flag, then the vendor and product names
 * normally generated by a device thorugh the INQUIRY response will be
 * taken from this list, and this is the reason for the above size
 * restriction. However, if the flag is not present, then you
 * are free to use as many characters as you like.
 */

#undef UNUSUAL_DEV
#define UNUSUAL_DEV(idVendor, idProduct, bcdDeviceMin, bcdDeviceMax, \
		    vendor_name, product_name, use_protocol, use_transport, \
		    init_function, Flags) \
{ \
	.vendorName = vendor_name,	\
	.productName = product_name,	\
	.useProtocol = use_protocol,	\
	.useTransport = use_transport,	\
	.initFunction = init_function,	\
	.flags = Flags, \
}

static struct us_unusual_dev us_unusual_dev_list[] = {
#	include "unusual_devs.h" 
#	undef UNUSUAL_DEV
	/* Control/Bulk transport for all SubClass values */
	{ .useProtocol = US_SC_RBC,
	  .useTransport = US_PR_CB},
	{ .useProtocol = US_SC_8020,
	  .useTransport = US_PR_CB},
	{ .useProtocol = US_SC_QIC,
	  .useTransport = US_PR_CB},
	{ .useProtocol = US_SC_UFI,
	  .useTransport = US_PR_CB},
	{ .useProtocol = US_SC_8070,
	  .useTransport = US_PR_CB},
	{ .useProtocol = US_SC_SCSI,
	  .useTransport = US_PR_CB},

	/* Control/Bulk/Interrupt transport for all SubClass values */
	{ .useProtocol = US_SC_RBC,
	  .useTransport = US_PR_CBI},
	{ .useProtocol = US_SC_8020,
	  .useTransport = US_PR_CBI},
	{ .useProtocol = US_SC_QIC,
	  .useTransport = US_PR_CBI},
	{ .useProtocol = US_SC_UFI,
	  .useTransport = US_PR_CBI},
	{ .useProtocol = US_SC_8070,
	  .useTransport = US_PR_CBI},
	{ .useProtocol = US_SC_SCSI,
	  .useTransport = US_PR_CBI},

	/* Bulk-only transport for all SubClass values */
	{ .useProtocol = US_SC_RBC,
	  .useTransport = US_PR_BULK},
	{ .useProtocol = US_SC_8020,
	  .useTransport = US_PR_BULK},
	{ .useProtocol = US_SC_QIC,
	  .useTransport = US_PR_BULK},
	{ .useProtocol = US_SC_UFI,
	  .useTransport = US_PR_BULK},
	{ .useProtocol = US_SC_8070,
	  .useTransport = US_PR_BULK},
	{ .useProtocol = US_SC_SCSI,
	  .useTransport = US_PR_BULK},

	/* Terminating entry */
	{ 0 }
};

struct usb_driver usb_storage_driver = {
 /*
 #ifndef _ETCPIP
	.owner =	THIS_MODULE,
 #endif // ndef _ETCPIP
 */
 
 #ifdef USB_DBGSTRING_ON
	.name =		"usb-storage",
 #endif // USB_DBGSTRING_ON
 
 .usbdev_probe =	storage_probe,
 .disconnect =	storage_disconnect,
 .id_table =	storage_usb_ids,
};

/*
 * fill_inquiry_response takes an unsigned char array (which must
 * be at least 36 characters) and populates the vendor name,
 * product name, and revision fields. Then the array is copied
 * into the SCSI command's response buffer (oddly enough
 * called request_buffer). data_len contains the length of the
 * data array, which again must be at least 36.
 */

void fill_inquiry_response(struct us_data *us, unsigned char *data,
		unsigned int data_len) {

 int i;
 #ifndef NO_SCATTER_LIST
	struct scatterlist *sg;
 #endif // ndef NO_SCATTER_LIST
	int len =
		us->srb->request_bufflen > data_len ? data_len :
		us->srb->request_bufflen;
	int transferred;
	int amt;

	if (data_len<36) // You lose.
		return;

	if(data[0]&0x20) { /* USB device currently not connected. Return
			      peripheral qualifier 001b ("...however, the
			      physical device is not currently connected
			      to this logical unit") and leave vendor and
			      product identification empty. ("If the target
			      does store some of the INQUIRY data on the
			      device, it may return zeros or ASCII spaces 
			      (20h) in those fields until the data is
			      available from the device."). */
		memset(data+8,0,28);
	} else {
		memcpy(data+8, us->unusual_dev->vendorName, 
			strlen(us->unusual_dev->vendorName) > 8 ? 8 :
			strlen(us->unusual_dev->vendorName));
		memcpy(data+16, us->unusual_dev->productName, 
			strlen(us->unusual_dev->productName) > 16 ? 16 :
			strlen(us->unusual_dev->productName));
		data[32] = 0x30 + ((us->pusb_dev->descriptor.bcdDevice>>12) & 0x0F);
		data[33] = 0x30 + ((us->pusb_dev->descriptor.bcdDevice>>8) & 0x0F);
		data[34] = 0x30 + ((us->pusb_dev->descriptor.bcdDevice>>4) & 0x0F);
		data[35] = 0x30 + ((us->pusb_dev->descriptor.bcdDevice) & 0x0F);
	}
 /*
 #ifndef NO_SCATTER_LIST
	if (us->srb->use_sg) {
			sg = (struct scatterlist *)us->srb->request_buffer;
			for (i=0; i<us->srb->use_sg; i++)
				memset(sg_address(sg[i]), 0, sg[i].length);
			for (i=0, transferred=0; 
					i<us->srb->use_sg && transferred < len;
					i++) {
				amt = sg[i].length > len-transferred ? 
					len-transferred : sg[i].length;
				memcpy(sg_address(sg[i]), data+transferred, amt);
				transferred -= amt;
				}
	} else
 #endif // ndef NO_SCATTER_LIST
 */
	  {
		memset(us->srb->request_buffer, 0, us->srb->request_bufflen);
		memcpy(us->srb->request_buffer, data, len);
	}
}

#if 1
// temporary solution - bonian chen 09.05.2005
void UsbMsStorLock(void)
{
	while (getres(MT_RES_USBMS_LOWER_IO, 0xffffffff) != SUCCESS) { scdtsk(); }
}

void UsbMsStorUnlock(void)
{
	relres(MT_RES_USBMS_LOWER_IO);
}
#endif
/*
static int usb_stor_control_thread(void * __us)
*/
#ifndef USBMS_NO_USB_TASK  /* Charles 20040520 */
static int usb_stor_task(void * __us)
#else
int usb_stor_task(void * __us)
#endif
{
 struct us_data *us = (struct us_data *)__us;
#ifdef USBMS_NO_USB_TASK  /* Charles 20040520 */
 int ret = 1;
#endif 
extern struct block_device *disk_global_bdev[USBMS_MAX_MINORS+1];

#if 1
	// temporary solution - bonian chen 09.05.2005
	UsbMsStorLock();
#endif
 USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_stor_task() Begin .. \n")) ;

 #ifdef USBMS_BS_ASSERT
	if ( ! (IS_VALID_ADDR(us)) ) {
		iprintf("[ASSERT][USBMS] usb_stor_task() : " ) ;
	    iprintf("Error : Invalid Address !!\n" ) ;
	    iprintf("[ASSERT][USBMS] us=%08XH \n", us ) ;
	    iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_ASSERT

 /*
 #ifndef _ETCPIP
	lock_kernel();
 #endif // ndef _ETCPIP
 */

 /*
  * This thread doesn't need any user-level access,
  * so get rid of all our resources.
  */
 /*
 #ifndef _ETCPIP
	daemonize("usb-storage");

	current->flags |= PF_IOTHREAD;

	unlock_kernel();
 #endif // ndef _ETCPIP
 */

 /* signal that we've started the thread */
 /*
 #ifndef _ETCPIP
    // struct completion in <linux/completion.h>
	complete(&(us->notify));
 #endif // ndef _ETCPIP
 */
 /*
 #ifdef NO_SEMPHORE
	// 2003/11/29 B. S. Liaw : it seems to use event rather than resource
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] usb_stor_task(1) : clrevt MT_EVT_USBMS_DOWN=%d ..\n",
		MT_EVT_USBMS_DOWN
		)) ;
	// MT_EVT_USBMS_DOWN is on kernel/local.h
    if (clrevt(MT_EVT_USBMS_DOWN) != SUCCESS) {
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] usb_stor_task(2) : Error : clrevt fail: Invalid MT_EVT_USBMS_DOWN=%d\n",
			MT_EVT_USBMS_DOWN
			)) ;
        dprintf("[USBMS] Error : clrevt fail: Invalid MT_EVT_USBMS_DOWN %d\n",
				MT_EVT_USBMS_DOWN);
    }
 #endif // NO_SEMPHORE
 */
#ifndef USBMS_NO_USB_TASK  /* Charles 20040520 */
 for(;;) {
#else
 {
#endif
	struct Scsi_Host *host;

	/*
	US_DEBUGP("*** thread sleeping.\n");
	*/
	USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_stor_task(1) *** task sleeping \n")) ;
	USBMS_DBG_URB(iprintf("[USBMS URB] usb_stor_task() *** task sleeping \n")) ;
	/*
	#ifndef NO_SEMPHORE
		if(down_interruptible(&us->sema))
			break;
	 #else
	*/
#ifndef USBMS_NO_USB_TASK  /* Charles 20040520 */
	#ifdef USBMS_USB_WAIT_BY_EVENT     // 2004/1/2 B. S. Liaw
		/*
    	getres(USBMS_RESOURCE,0); // add by B. S. Liaw 2003/11/29
								  // USBMS_RESOURCE is on kernel/local.h
								  // 0 : infinite
		*/
		// 2003/11/29 B. S. Liaw : it seems to use event rather than resource
		// MT_EVT_USBMS_DOWN is on kernel/local.h
        // wteset_dec(MT_EVT_USBMS_DOWN, MSECTOTICKS(500));
		/*
		i=wteset_dec(DHCP_ACT_EVT, timeout*CLOCKHZ);
		clrevt(DHCP_ACT_EVT);
		if (i==SUCCESS) return 0;
		*/

        (void)wteset_dec(MT_EVT_USBMS_DOWN, 0); // 0 : no timeout

		// MT_EVT_USBMS_DOWN is on kernel/local.h
	    if (clrevt(MT_EVT_USBMS_DOWN) != SUCCESS) {
			USBMS_DBG_INFO(iprintf(
				"[USBMS INFO] usb_stor_task(3) : Error : clrevt fail: Invalid MT_EVT_USBMS_DOWN=%d\n",
				MT_EVT_USBMS_DOWN
				)) ;
	        dprintf("[USBMS] Error : clrevt fail: Invalid MT_EVT_USBMS_DOWN %d\n",
					MT_EVT_USBMS_DOWN);
		   }

	#else
		USBMS_DBG_MEM(iprintf("[USBMS MEM] usb_stor_task() : us=%08XH .. \n",us)) ;
		// USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_stor_task(2) : us=%08XH .. \n",us)) ;
		for(;;) {
			// NULL ;
			if (us->evt_ms) {
				// NULL ;
				// break ;
				goto awakened ;
			   }
			// NULL ;
		   }

		// -----------
		awakened:
		// -----------

		USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_stor_task(3) : Set Value : us->evt_ms=0 .. \n")) ;
		// BLOCK_PREEMPTION;
		us->evt_ms = 0 ;
		// UNBLOCK_PREEMPTION;
	#endif // ndef USBMS_USB_WAIT_BY_EVENT
#endif
	/*
	#endif
	*/
	/*			
	US_DEBUGP("*** thread awakened.\n");
	*/
	USBMS_DBG_URB(iprintf("[USBMS URB] usb_stor_task() *** task awakened \n")) ;
	USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_stor_task(4) *** task awakened \n")) ;

	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] usb_stor_task() : us=%08XH \n",
		us
		)) ;
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] usb_stor_task() : us->srb=%08XH \n",
		us->srb
		)) ;

	/* if us->srb is NULL, we are being asked to exit */
	if (us->srb == NULL) {
		/*
		US_DEBUGP("-- exit command received\n");
		*/
		USBMS_DBG_PLUG(iprintf(
			"[USBMS PLUG] usb_stor_task() : exit command received \n"
			)) ;
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] usb_stor_task(4) : exit command received \n"
			)) ;
#ifdef USBMS_NO_USB_TASK  /* Charles 20040520 */
		goto end_loop ;
#else
		break;
#endif
	   }

	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] usb_stor_task() : us=%08XH \n",
		us
		)) ;
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] usb_stor_task() : us->srb=%08XH \n",
		us->srb
		)) ;
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] usb_stor_task() : us->srb->device=%08XH \n",
		us->srb->device
		)) ;
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] usb_stor_task() : us->srb->device->host=%08XH \n",
		us->srb->device->host
		)) ;

	host = us->srb->device->host;

	/* lock access to the state */
	/*
	#ifndef NO_SPINLOCK
		scsi_lock(host);
	#endif // ndef NO_SPINLOCK
	*/

	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] usb_stor_task() : &(us->sm_state)=%08XH us->sm_state=%d \n",
		&(us->sm_state),us->sm_state
		)) ;

	/* has the command been aborted *already* ? */
	if (us->sm_state == US_STATE_ABORTING) {
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] usb_stor_task(5) : Error : US_STATE_ABORTING \n"
			)) ;
		us->srb->result = DID_ABORT << 16;
		goto SkipForAbort;
	   }

	/* set the state and release the lock */
	us->sm_state = US_STATE_RUNNING;
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] usb_stor_task(5) : US_STATE_RUNNING .. \n"
		)) ;
	/*
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] usb_stor_task(6) : US_STATE_RUNNING .. \n"
		)) ;
	*/
	/*
	#ifndef NO_SPINLOCK
		scsi_unlock(host);
	#endif // ndef NO_SPINLOCK
	*/
	/* lock the device pointers */
	/*
	#ifndef NO_SEMPHORE
		down(&(us->dev_semaphore));
	#endif // ndef NO_SEMPHORE
	*/
	/* don't do anything if we are disconnecting */
	if (test_bit(US_FLIDX_DISCONNECTING, &us->flags)) {
		/*
		US_DEBUGP("No command during disconnect\n");
		*/
		USBMS_DBG_PLUG(iprintf(
			"[USBMS PLUG] usb_stor_task(7) : No command during disconnect \n"
			)) ;
		us->srb->result = DID_BAD_TARGET << 16;
	   }

	/* reject the command if the direction indicator 
	 * is UNKNOWN
	 */
	else if (us->srb->sc_data_direction == SCSI_DATA_UNKNOWN) {
		/*
		US_DEBUGP("UNKNOWN data direction\n");
		*/
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] usb_stor_task(8) : UNKNOWN data direction \n"
			)) ;
		us->srb->result = DID_ERROR << 16;
		}

	/* reject if target != 0 or if LUN is higher than
	 * the maximum known LUN
	 */
	else if (us->srb->device->id && 
				!(us->flags & US_FL_SCM_MULT_TARG)) {
		/*
			US_DEBUGP("Bad target number (%d:%d)\n",
				  us->srb->device->id, us->srb->device->lun);
		*/
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] usb_stor_task(9) : Bad target number (%d:%d)\n",
			us->srb->device->id, us->srb->device->lun
			)) ;
		us->srb->result = DID_BAD_TARGET << 16;
		}

	else if (us->srb->device->lun > us->max_lun) {
		/*
		US_DEBUGP("Bad LUN (%d:%d)\n",
				  us->srb->device->id, us->srb->device->lun);
		*/
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] usb_stor_task(10) : Bad LUN (%d:%d)\n",
				  us->srb->device->id, us->srb->device->lun
			)) ;
		us->srb->result = DID_BAD_TARGET << 16;
		}

	/* Handle those devices which need us to fake 
	 * their inquiry data */
	else if ((us->srb->cmnd[0] == INQUIRY) &&
			    (us->flags & US_FL_FIX_INQUIRY)) {
			unsigned char data_ptr[36] = {
			    0x00, 0x80, 0x02, 0x02,
			    0x1F, 0x00, 0x00, 0x00};
		/*
		US_DEBUGP("Faking INQUIRY command\n");
		*/
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] usb_stor_task(11) : Faking INQUIRY command\n"
			)) ;
		fill_inquiry_response(us, data_ptr, 36);
		us->srb->result = SAM_STAT_GOOD;
		}

	/* we've got a command, let's do it! */
	else {
		USBMS_DBG_URB(iprintf(
			"[USBMS URB] usb_stor_task() : got a command .. \n"
			)) ;
		US_DEBUG(usb_stor_show_command(us->srb));

		// 2004/1/12 B. S. Liaw
		if ( !us || !us->proto_handler || (us->flags & DONT_SUBMIT)) {
			USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_stor_task(12) : disconnect !! \n" )) ;
			goto end_loop ;
		   }

		if ( us && us->proto_handler) {
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_task() : us->proto_handler() .. \n")) ;
			if ( us->proto_handler(us->srb, us) == -ENODEV ) {
				USBMS_DBG_INFO(iprintf(
					"[USBMS INFO] usb_stor_task(13) : disconnect !! \n"
					)) ;
				goto end_loop ;
			   }
		   }
		  else {
			USBMS_DBG_INFO(iprintf(
				"[USBMS INFO] usb_stor_task(14) : Null Protocol !! \n"
				)) ;
			goto end_loop ;
		   }
		}

	/* unlock the device pointers */
	/*
	#ifndef NO_SEMPHORE
		up(&(us->dev_semaphore));
	#endif // ndef NO_SEMPHORE
	*/
	/* lock access to the state */
	/*
	#ifndef NO_SPINLOCK
		scsi_lock(host);
	#endif // ndef NO_SPINLOCK
	*/
	/* indicate that the command is done */
	if (us->srb->result != DID_ABORT << 16) {
		/*
		US_DEBUGP("scsi cmd done, result=0x%x\n", 
				   us->srb->result);
		*/
		USBMS_DBG_URB(iprintf(
			"[USBMS URB] usb_stor_task() : scsi cmd done, result=0x%x\n", 
				   us->srb->result
			)) ;
		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] usb_stor_task() : us->srb->scsi_done=%08XH \n",
			us->srb->scsi_done )) ;
		#ifdef USBMS_BS_ASSERT
			if ( ! (IS_VALID_ADDR(us->srb->scsi_done)) ) {
				iprintf("[ASSERT][USBMS] usb_stor_task() : " ) ;
			    iprintf("Error : Invalid Address !!\n" ) ;
		        iprintf("[ASSERT][USBMS] us->srb->scsi_done=%08XH \n",
						us->srb->scsi_done ) ; ;
		        iprintf("[ASSERT][USBMS] Continue ..\n") ;
			   }
		#endif // USBMS_BS_ASSERT
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_task() : us->srb->scsi_done() .. \n")) ;
		us->srb->scsi_done(us->srb);
#ifdef USBMS_NO_USB_TASK  /* Charles 20040520 */
		ret = 0;
#endif
//iprintf("usb_stor_task: scsi_done %u.%u\n", get_sys_time(), USB_CURRENT_TICKS);

	   } else {

		// ------------------------
		SkipForAbort:
		// ------------------------

		/*
		US_DEBUGP("scsi command aborted\n");
		*/
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] usb_stor_task(14) : scsi command aborted\n"
			)) ;
	   }

	/* If an abort request was received we need to signal that
	 * the abort has finished.  The proper test for this is
	 * sm_state == US_STATE_ABORTING, not srb->result == DID_ABORT,
	 * because an abort request might be received after all the
	 * USB processing was complete. */

	/*
	#ifndef _ETCPIP
		if (us->sm_state == US_STATE_ABORTING)
		    // struct completion in <linux/completion.h>
			complete(&(us->notify));
	#endif // ndef _ETCPIP
	*/
	/* empty the queue, reset the state, and release the lock */
	/*
	USBMS_DBG_URB(iprintf(
		"[USBMS URB] usb_stor_task() : Set Value : us->srb=NULL .. \n"
		)) ;
	us->srb = NULL;
	*/
	us->sm_state = US_STATE_IDLE;
#ifndef USBMS_NO_USB_TASK  /* Charles 20040520 */
	#ifdef USBMS_USB_WAIT_BY_EVENT     // 2004/1/2 B. S. Liaw
		USBMS_DBG_BLOCK(iprintf(
			"[USBMS BLOCK] usb_stor_task() : SIGNAL(MT_EVT_USBMS_RQ,%d) .. \n",
			MT_EVT_USBMS_RQ)) ;
		USBMS_DBG_INIT(iprintf(
			"[USBMS INIT] usb_stor_task(6) : SIGNAL(MT_EVT_USBMS_RQ,%d) .. \n",
			MT_EVT_USBMS_RQ )) ;
		// kernel/mtmacro.h : #define SIGNAL(signo) setevt(signo)
		if (SIGNAL(MT_EVT_USBMS_RQ) != SUCCESS) {
			USBMS_DBG_INFO(iprintf(
				"[USBMS INFO] usb_stor_task(15) : Error : SIGNAL MT_EVT_USBMS_RQ=%d failed !! \n",
				MT_EVT_USBMS_RQ
				)) ;
			dprintf("[USBMS] Error : SIGNAL %d failed !! \n",MT_EVT_USBMS_RQ);
		   }
	#else
		switch(waiting_evt) {
			case MT_EVT_USBMS_RQ :
				USBMS_DBG_INIT(iprintf(
					"[USBMS INIT] usb_stor_task(7) : Set Value : req->evt_blk_rq=1 .. \n"
					)) ;
				// BLOCK_PREEMPTION;
				req->evt_blk_rq = 1 ;
				// UNBLOCK_PREEMPTION;
				// iprintf("[USBMS] end_that_request_last() : FixMe : MT_EVT_USBMS_RQ !! \n") ;
				break ;
				/*
				#ifndef USBMS_NO_IOCTL
					case MT_EVT_USBMS_RQBI :
						USBMS_DBG_INIT(iprintf(
						"[USBMS INIT] usb_stor_task(8) : Set Value : req->evt_blk_rq_bi=1 .. \n"
						)) ;
					// BLOCK_PREEMPTION;
					req->evt_blk_rq_bi = 1 ;
					// UNBLOCK_PREEMPTION;
					break ;
				#endif // ndef USBMS_NO_IOCTL
				*/
			   }
	#endif // ndef USBMS_USB_WAIT_BY_EVENT
#endif
	/*
	#ifndef NO_SPINLOCK
		scsi_unlock(host);
	#endif // ndef NO_SPINLOCK
	*/
	USBMS_DBG_URB(iprintf("[USBMS URB] usb_stor_task() : Next Iteration .. \n")) ;
   } /* for (;;) */

 // -----------------
 end_loop: 
 // -----------------

 /* notify the exit routine that we're actually exiting now */
 /*
 #ifndef _ETCPIP
    // struct completion in <linux/completion.h>
	complete(&(us->notify));
 #endif // ndef _ETCPIP
 */
 
 // 2004/1/8 B. S. Liaw : Let the task keep going even Error Occur !!
#ifndef USBMS_NO_USB_TASK  /* Charles 20040520 */
 #ifdef USBMS_USB_WAIT_BY_EVENT     // 2004/1/2 B. S. Liaw
	// 2003/12/01 B. S. Liaw :
	USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_stor_task(9) : SIGNAL(MT_EVT_USBMS_RQ) .. \n")) ;
	// kernel/mtmacro.h : #define SIGNAL(signo) setevt(signo)
	if (SIGNAL(MT_EVT_USBMS_RQ) != SUCCESS) {
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] usb_stor_task(16) : Error : SIGNAL MT_EVT_USBMS_RQ=%d fail !! \n",
			MT_EVT_USBMS_RQ
			)) ;
		dprintf("[USBMS] Error : SIGNAL MT_EVT_USBMS_RQ %d fail !! \n",MT_EVT_USBMS_RQ);
	   }
  #else
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] usb_stor_task(10) : Set Value : req->evt_blk_rq=1 .. \n"
		)) ;
	req->evt_blk_rq = 1 ;
 #endif // ndef USBMS_USB_WAIT_BY_EVENT
#endif
 
 USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_stor_task() *** task end !! \n")) ;
// dprintf("[USBMS] usb_stor_task stop !! \n") ;
 USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_stor_task() End. \n")) ;
#if 1
	// temporary solution - bonian chen 09.05.2005
	UsbMsStorUnlock();
#endif
#ifndef USBMS_NO_USB_TASK  /* Charles 20040520 */
 return 0;
#else
 return ret;
#endif
}	

/***********************************************************************
 * Device probing and disconnecting
 ***********************************************************************/

/* Associate our private data with the USB device */
static int associate_dev(struct us_data *us, struct usb_interface *intf)
{
 US_DEBUGP("-- %s\n", __FUNCTION__);

	/* Fill in the device-related fields */
	us->pusb_dev = interface_to_usbdev(intf);
	us->pusb_intf = intf;
	us->ifnum = intf->altsetting->desc.bInterfaceNumber;

	/* Store our private data in the interface and increment the
	 * device's reference count */
	usb_set_intfdata(intf, us);
	usb_get_dev(us->pusb_dev);

	/* Allocate the device-related DMA-mapped buffers */
	us->cr = usb_buffer_alloc(us->pusb_dev, sizeof(*us->cr),
			GFP_KERNEL, &us->cr_dma);
	if (!us->cr) {
		US_DEBUGP("usb_ctrlrequest allocation failed\n");
		return -ENOMEM;
	}

 us->iobuf = usb_buffer_alloc(us->pusb_dev, US_IOBUF_SIZE,
			GFP_KERNEL, &us->iobuf_dma);
 if (!us->iobuf) {
	US_DEBUGP("I/O buffer allocation failed\n");
	return -ENOMEM;
   }
 return 0;
}

/* Get the unusual_devs entries and the string descriptors */
static void get_device_info(struct us_data *us, int id_index)
{
 struct usb_device *dev = us->pusb_dev;
 struct usb_interface_descriptor *idesc =
		&us->pusb_intf->altsetting[us->pusb_intf->act_altsetting].desc;
 struct us_unusual_dev *unusual_dev = &us_unusual_dev_list[id_index];
 /*
 struct usb_device_id *id = &storage_usb_ids[id_index];
 */
 const struct usb_device_id *id = &storage_usb_ids[id_index];


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] get_device_info() Begin .. \n"));

 if (unusual_dev->vendorName)
			US_DEBUGP("Vendor: %s\n", unusual_dev->vendorName);
 if (unusual_dev->productName)
			US_DEBUGP("Product: %s\n", unusual_dev->productName);

	/* Store the entries */
	us->unusual_dev = unusual_dev;
	us->subclass = (unusual_dev->useProtocol == US_SC_DEVICE) ?
			idesc->bInterfaceSubClass :
			unusual_dev->useProtocol;
	us->protocol = (unusual_dev->useTransport == US_PR_DEVICE) ?
			idesc->bInterfaceProtocol :
			unusual_dev->useTransport;
	us->flags = unusual_dev->flags;

	/* Log a message if a non-generic unusual_dev entry contains an
	 * unnecessary subclass or protocol override.  This may stimulate
	 * reports from users that will help us remove unneeded entries
	 * from the unusual_devs.h table.
	 */
	if (id->idVendor || id->idProduct) {
		static char *msgs[3] = {
			"an unneeded SubClass entry",
			"an unneeded Protocol entry",
			"unneeded SubClass and Protocol entries"};
		struct usb_device_descriptor *ddesc = &dev->descriptor;
		int msg = -1;

		if (unusual_dev->useProtocol != US_SC_DEVICE &&
			us->subclass == idesc->bInterfaceSubClass)
			msg += 1;
		if (unusual_dev->useTransport != US_PR_DEVICE &&
			us->protocol == idesc->bInterfaceProtocol)
			msg += 2;
		if (msg >= 0)
		   {
			/*
			printk(KERN_NOTICE USB_STORAGE "This device "
				"(%04x,%04x,%04x S %02x P %02x)"
				" has %s in unusual_devs.h\n"
				"   Please send a copy of this message to "
				"<linux-usb-devel@lists.sourceforge.net>\n",
				ddesc->idVendor, ddesc->idProduct,
				ddesc->bcdDevice,
				idesc->bInterfaceSubClass,
				idesc->bInterfaceProtocol,
				msgs[msg]);
			*/
			dprintf("[USBMS] NOTICE USB_STORAGE This device "
				"(%04x,%04x,%04x S %02x P %02x)"
				" has %s in unusual_devs.h\n"
				"   Please send a copy of this message to "
				"<linux-usb-devel@lists.sourceforge.net>\n",
				ddesc->idVendor, ddesc->idProduct,
				ddesc->bcdDevice,
				idesc->bInterfaceSubClass,
				idesc->bInterfaceProtocol,
				msgs[msg]);
		   }
	}

	/* Read the device's string descriptors */
	if (dev->descriptor.iManufacturer)
       {
		#ifdef USB_DBGSTRING_ON
			usb_string(dev, dev->descriptor.iManufacturer, 
			   us->vendor, sizeof(us->vendor));
         #else
			usb_string(dev, dev->descriptor.iManufacturer) ;
        #endif
       }
	if (dev->descriptor.iProduct)
       {
		#ifdef USB_DBGSTRING_ON
			usb_string(dev, dev->descriptor.iProduct, 
			   us->product, sizeof(us->product));
         #else
			usb_string(dev, dev->descriptor.iProduct) ;
        #endif
       }
	if (dev->descriptor.iSerialNumber)
       {
		#ifdef USB_DBGSTRING_ON
			usb_string(dev, dev->descriptor.iSerialNumber, 
			   us->serial, sizeof(us->serial));
         #else
			usb_string(dev, dev->descriptor.iSerialNumber);
        #endif
       }

	/* Use the unusual_dev strings if the device didn't provide them */
	if (strlen(us->vendor) == 0) {
		if (unusual_dev->vendorName)
			strlcpy(us->vendor, unusual_dev->vendorName,
				sizeof(us->vendor));
		else
			strcpy(us->vendor, "Unknown");
	}
	if (strlen(us->product) == 0) {
		if (unusual_dev->productName)
			strlcpy(us->product, unusual_dev->productName,
				sizeof(us->product));
		else
			strcpy(us->product, "Unknown");

	}
	if (strlen(us->serial) == 0)
		strcpy(us->serial, "None");
}

/* Get the transport settings */
static int get_transport(struct us_data *us)
{
	switch (us->protocol) {
	case US_PR_CB:
		us->transport_name = "Control/Bulk";
		us->transport = usb_stor_CB_transport;
		us->transport_reset = usb_stor_CB_reset;
		us->max_lun = 7;
		break;

	case US_PR_CBI:
		us->transport_name = "Control/Bulk/Interrupt";
		us->transport = usb_stor_CBI_transport;
		us->transport_reset = usb_stor_CB_reset;
		us->max_lun = 7;
		break;

	case US_PR_BULK:
		us->transport_name = "Bulk";
		us->transport = usb_stor_Bulk_transport;
		us->transport_reset = usb_stor_Bulk_reset;
		break;

#ifdef CONFIG_USB_STORAGE_HP8200e
	case US_PR_SCM_ATAPI:
		us->transport_name = "SCM/ATAPI";
		us->transport = hp8200e_transport;
		us->transport_reset = usb_stor_CB_reset;
		us->max_lun = 1;
		break;
#endif

#ifdef CONFIG_USB_STORAGE_SDDR09
	case US_PR_EUSB_SDDR09:
		us->transport_name = "EUSB/SDDR09";
		us->transport = sddr09_transport;
		us->transport_reset = usb_stor_CB_reset;
		us->max_lun = 0;
		break;
#endif

#ifdef CONFIG_USB_STORAGE_SDDR55
	case US_PR_SDDR55:
		us->transport_name = "SDDR55";
		us->transport = sddr55_transport;
		us->transport_reset = sddr55_reset;
		us->max_lun = 0;
		break;
#endif

#ifdef CONFIG_USB_STORAGE_DPCM
	case US_PR_DPCM_USB:
		us->transport_name = "Control/Bulk-EUSB/SDDR09";
		us->transport = dpcm_transport;
		us->transport_reset = usb_stor_CB_reset;
		us->max_lun = 1;
		break;
#endif

#ifdef CONFIG_USB_STORAGE_FREECOM
	case US_PR_FREECOM:
		us->transport_name = "Freecom";
		us->transport = freecom_transport;
		us->transport_reset = usb_stor_freecom_reset;
		us->max_lun = 0;
		break;
#endif

#ifdef CONFIG_USB_STORAGE_DATAFAB
	case US_PR_DATAFAB:
		us->transport_name  = "Datafab Bulk-Only";
		us->transport = datafab_transport;
		us->transport_reset = usb_stor_Bulk_reset;
		us->max_lun = 1;
		break;
#endif

#ifdef CONFIG_USB_STORAGE_JUMPSHOT
	case US_PR_JUMPSHOT:
		us->transport_name  = "Lexar Jumpshot Control/Bulk";
		us->transport = jumpshot_transport;
		us->transport_reset = usb_stor_Bulk_reset;
		us->max_lun = 1;
		break;
#endif

	default:
		return -EIO;
	}
 /*
 US_DEBUGP("Transport: %s\n", us->transport_name);
 */
 USBMS_DBG_INFO(iprintf("[USBMS INFO] get_transport() : Transport: %s \n",
	us->transport_name));
 dprintf("[USBMS] Transport: %s \n",us->transport_name) ;

	/* fix for single-lun devices */
	if (us->flags & US_FL_SINGLE_LUN)
		us->max_lun = 0;
	return 0;
}

/* Get the protocol settings */
static int get_protocol(struct us_data *us)
{
	switch (us->subclass) {
	case US_SC_RBC:
		us->protocol_name = "Reduced Block Commands (RBC)";
		us->proto_handler = usb_stor_transparent_scsi_command;
		break;

	case US_SC_8020:
		us->protocol_name = "8020i";
		us->proto_handler = usb_stor_ATAPI_command;
		us->max_lun = 0;
		break;

	case US_SC_QIC:
		us->protocol_name = "QIC-157";
		us->proto_handler = usb_stor_qic157_command;
		us->max_lun = 0;
		break;

	case US_SC_8070:
		us->protocol_name = "8070i";
		us->proto_handler = usb_stor_ATAPI_command;
		us->max_lun = 0;
		break;

	case US_SC_SCSI:
		us->protocol_name = "Transparent SCSI";
		us->proto_handler = usb_stor_transparent_scsi_command;
		break;

	case US_SC_UFI:
		us->protocol_name = "Uniform Floppy Interface (UFI)";
		us->proto_handler = usb_stor_ufi_command;
		break;

	#ifdef CONFIG_USB_STORAGE_ISD200
		case US_SC_ISD200:
			us->protocol_name = "ISD200 ATA/ATAPI";
			us->proto_handler = isd200_ata_command;
			break;
	#endif

	default:
		return -EIO;

	}
 /*
 US_DEBUGP("Protocol: %s\n", us->protocol_name);
 */
 USBMS_DBG_INFO(iprintf("[USBMS INFO] get_protocol() : Protocol: %s\n",
	us->protocol_name)) ;
 dprintf("[USBMS] Protocol: %s\n",
	us->protocol_name) ;
 return 0;
}

/* Get the pipe settings */
static int get_pipes(struct us_data *us)
{
	struct usb_host_interface *altsetting =
		&us->pusb_intf->altsetting[us->pusb_intf->act_altsetting];
	int i;
	struct usb_endpoint_descriptor *ep;
	struct usb_endpoint_descriptor *ep_in = NULL;
	struct usb_endpoint_descriptor *ep_out = NULL;
	struct usb_endpoint_descriptor *ep_int = NULL;

	/*
	 * Find the endpoints we need.
	 * We are expecting a minimum of 2 endpoints - in and out (bulk).
	 * An optional interrupt is OK (necessary for CBI protocol).
	 * We will ignore any others.
	 */
	for (i = 0; i < altsetting->desc.bNumEndpoints; i++) {
		ep = &altsetting->endpoint[i].desc;

		/* Is it a BULK endpoint? */
		if ((ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
				== USB_ENDPOINT_XFER_BULK) {
			/* BULK in or out? */
			if (ep->bEndpointAddress & USB_DIR_IN)
				ep_in = ep;
			else
				ep_out = ep;
		}

		/* Is it an interrupt endpoint? */
		else if ((ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
				== USB_ENDPOINT_XFER_INT) {
			ep_int = ep;
		}
	}
 /*
 US_DEBUGP("Endpoints: In: 0x%p Out: 0x%p Int: 0x%p (Period %d)\n",
		  ep_in, ep_out, ep_int, ep_int ? ep_int->bInterval : 0);
 */
 USBMS_DBG_INFO(iprintf(
	"[USBMS INFO] Endpoints: In:%08XH Out:%08XH Int:%08XH (Period %d)\n",
	ep_in, ep_out, ep_int, ep_int ? ep_int->bInterval : 0
	)) ;

 if (!ep_in || !ep_out || (us->protocol == US_PR_CBI && !ep_int)) {
	/*
	US_DEBUGP("Endpoint sanity check failed! Rejecting dev.\n");
	*/
	dprintf("[USBMS] Error : Endpoint sanity check failed! Rejecting dev.\n");
	return -EIO;
   }

	/* Calculate and store the pipe values */
	us->send_ctrl_pipe = usb_sndctrlpipe(us->pusb_dev, 0);
	us->recv_ctrl_pipe = usb_rcvctrlpipe(us->pusb_dev, 0);
	us->send_bulk_pipe = usb_sndbulkpipe(us->pusb_dev,
		ep_out->bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
	us->recv_bulk_pipe = usb_rcvbulkpipe(us->pusb_dev, 
		ep_in->bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
	if (ep_int) {
		us->recv_intr_pipe = usb_rcvintpipe(us->pusb_dev,
			ep_int->bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
		us->ep_bInterval = ep_int->bInterval;
	}
	return 0;
}

/* Initialize all the dynamic resources we need */
static int usb_stor_acquire_resources(struct us_data *us)
{
 /*
 #ifdef _ETCPIP
 */
	int status1 ;
 /*
  #else
	int p;
 #endif
 */


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_acquire_resources() Begin .. \n"));
 /*
 us->current_urb = usb_alloc_urb(0, GFP_KERNEL);
 */
 us->current_urb = usb_alloc_urb(0);
 if (!us->current_urb) {
	US_DEBUGP("URB allocation failed\n");
	dprintf("[USBMS] Error : out of memory !! \n") ;
	return -ENOMEM;
   }

 /* Lock the device while we carry out the next two operations */
 /*
 #ifndef NO_SEMPHORE
	down(&us->dev_semaphore);
 #endif // ndef NO_SEMPHORE
 */
 /* For bulk-only devices, determine the max LUN value */
 if(us->protocol == US_PR_BULK) {
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] usb_stor_acquire_resources(1) : determine the max LUN value .. \n"
		));
	
	#ifndef USBMS_NO_PROBE_MAXLUN
		us->max_lun = usb_stor_Bulk_max_lun(us);
	 #else
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] usb_stor_acquire_resources() : USBMS_NO_PROBE_MAXLUN !! \n"
			));
		dprintf("[USBMS] assuming device 1 LUN \n");
		us->max_lun = 1 ;
	#endif
   }

 /* Just before we start our control thread, initialize
  * the device if it needs initialization */
 if(us->unusual_dev->initFunction) {
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] usb_stor_acquire_resources(2) : initFunction() .. \n"
		));
	us->unusual_dev->initFunction(us);
   }
 /*
 #ifndef NO_SEMPHORE
	up(&us->dev_semaphore);
 #endif // ndef NO_SEMPHORE
 */
 /* Start up our control thread */
 us->sm_state = US_STATE_IDLE;

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] usb_stor_acquire_resources() : us=%08XH \n",
	us
	)) ;
 /*
 #ifndef _ETCPIP
	
	p = kernel_thread(usb_stor_control_thread, us, CLONE_VM);
	if (p < 0) {
		
		//printk(KERN_WARNING USB_STORAGE 
		//       "Unable to start control thread\n");
		
		dprintf("[USBMS] USB_STORAGE Unable to start control thread\n");
		return p;
		}
	
	us->pid = p;
	
  #else
 */
#ifndef USBMS_NO_USB_TASK  /* Charles 20040520 */
	USBMS_DBG_FLOW(iprintf(
		"[USBMS FLOW] usb_stor_acquire_resources() : runtsk usb_stor_task() .. \n"));

	#ifdef INT_SCHEDULING
		status1 = runtsk(100, (void (fAR *)(void))usb_stor_task, 6000,(unsigned long) us);
	 #else
		status1 = runtsk(80, (void (fAR *)(void))usb_stor_task, 6000,(unsigned long) us);
	#endif // INT_SCHEDULING

	if (status1 < SUCCESS ) {
        dprintf("RUNTASK usb_stor_task fails\n");
		return status1 ;
       }
      else
       {
        dprintf("RUNTASK id=%d usb_stor_task ...\n", status1);
		us->pid = status1 ;
       }
#endif
 /*
 #endif
 */

 /* Wait for the thread to start */
 /*
 #ifndef _ETCPIP
    // struct completion in <linux/completion.h>
	wait_for_completion(&(us->notify));
 #endif // ndef _ETCPIP
 */
 /*
  * Since this is a new device, we need to register a SCSI
  * host definition with the higher SCSI layers.
  */
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] usb_stor_acquire_resources(3) : register the scsi host .. \n"
	));
 /*
 #ifndef _ETCPIP
	us->host = scsi_host_alloc(&usb_stor_host_template, sizeof(us));
 #else
 */
	us->host = scsi_host_alloc(&usb_stor_host_template, sizeof(us) ,
							   &msusb_shost_my_devices ) ;
 /*
 #endif
 */
 if (!us->host) {
	/*
	printk(KERN_WARNING USB_STORAGE
			"Unable to register the scsi host\n");
	*/
	dprintf("[USBMS] Error : USB_STORAGE Unable to register the scsi host !! \n");
	return -EBUSY;
   }
		
 /* Set the hostdata to prepare for scanning */
 us->host->hostdata[0] = (unsigned long) us;
		
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_acquire_resources() End. \n"));
 return 0;
}

/* Dissociate from the USB device */
static void dissociate_dev(struct us_data *us)
{
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] dissociate_dev() Begin .. \n")) ;
 // US_DEBUGP("-- %s\n", __FUNCTION__);
 /*
 #ifndef NO_SEMPHORE
	down(&us->dev_semaphore);
 #endif // ndef NO_SEMPHORE
 */
 /* Free the device-related DMA-mapped buffers */
 if (us->cr) {
	USBMS_DBG_PLUG(iprintf("[USBMS PLUG] dissociate_dev(1) : usb_buffer_free(us->cr) .. \n")) ;
		usb_buffer_free(us->pusb_dev, sizeof(*us->cr), us->cr,
				us->cr_dma);
		us->cr = NULL;
	}

 if(us->iobuf) {
	USBMS_DBG_PLUG(iprintf("[USBMS PLUG] dissociate_dev(2) : usb_buffer_free(us->iobuf) .. \n")) ;
	usb_buffer_free(us->pusb_dev, US_IOBUF_SIZE, us->iobuf,
				us->iobuf_dma);
	us->iobuf = NULL;
   }

 /* Remove our private data from the interface and decrement the
  * device's reference count */
 usb_set_intfdata(us->pusb_intf, NULL);
 /*
 #ifndef _ETCPIP
	USB_DBG_FLOW(iprintf(
    	"[USB FLOW] dissociate_dev() Call usb_free_dev() ...\n"
        ));
	usb_put_dev(us->pusb_dev);
 #endif// ndef _ETCPIP
 */

 us->pusb_dev = NULL;
 us->pusb_intf = NULL;
 /*
 #ifndef NO_SEMPHORE
	up(&us->dev_semaphore);
 #endif // ndef NO_SEMPHORE
 */
}

/* Release all our static and dynamic resources */
void usb_stor_release_resources(struct us_data *us)
{
 /*
  * The host must already have been removed
  * and dissociate_dev() must have been called.
  */

 // USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_release_resources() Begin .. \n"));
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] usb_stor_release_resources() Begin .. \n"));

 /* Finish the SCSI host removal sequence */
 if (us->host) {
	(struct us_data *) us->host->hostdata[0] = NULL;
	scsi_host_put(us->host);
   }

 /* Kill the control thread
  *
  * Enqueue the command, wake up the thread, and wait for 
  * notification that it has exited.
  */
 if(us->pid) {
	US_DEBUGP("-- sending exit command to thread\n");
	/*
	BUG_ON(us->sm_state != US_STATE_IDLE);
	if (!( (us->sm_state != US_STATE_IDLE
		   )!=0
	     )
	   ) {
	// B. S. Liaw : it seems that BUB_ON(a) = a
	*/
	if (us->sm_state != US_STATE_IDLE) {
		// dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] usb_stor_release_resources(1) : Error : Not US_STATE_IDLE : %d !! \n" ,
			us->sm_state
			)) ;
	   }

	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] usb_stor_release_resources(2) : Set Value : us->srb=NULL .. \n"
		)) ;
	us->srb = NULL;
	/*
	#ifndef NO_SEMPHORE
		up(&(us->sema));
	 #else
	*/
#ifndef USBMS_NO_USB_TASK  /* Charles 20040520 */
	#ifdef USBMS_USB_WAIT_BY_EVENT     // 2004/1/2 B. S. Liaw
		/*
		relres(USBMS_RESOURCE); // add by B. S. Liaw 2003/11/29
						    // USBMS_RESOURCE is on kernel/local.h
		*/
		// 2003/11/29 B. S. Liaw : it seems to use event rather than resource
		// MT_EVT_USBMS_DOWN is on kernel/local.h
		USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_stor_release_resources() : SIGNAL(MT_EVT_USBMS_DOWN) .. \n")) ;
		// kernel/mtmacro.h : #define SIGNAL(signo) setevt(signo)
		if (SIGNAL(MT_EVT_USBMS_DOWN) != SUCCESS) {
			USBMS_DBG_INFO(iprintf(
				"[USBMS INFO] usb_stor_release_resources(3) : Error : SIGNAL MT_EVT_USBMS_DOWN=%d fail !! \n",
				MT_EVT_USBMS_DOWN
				)) ;
   			dprintf("[USBMS] Error : SIGNAL MT_EVT_USBMS_DOWN %d fail !! \n",MT_EVT_USBMS_DOWN);
		   }
	#else
		USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_stor_release_resources() : Set Value : us->evt_ms=1 .. \n")) ;
		// BLOCK_PREEMPTION;
		us->evt_ms = 1 ;
		// UNBLOCK_PREEMPTION;
	#endif // ndef USBMS_USB_WAIT_BY_EVENT
#endif
	/*
	#endif
	*/
	/*
	#ifndef _ETCPIP
	    // struct completion in <linux/completion.h>
		wait_for_completion(&(us->notify));
	#endif // ndef _ETCPIP
	*/
   }

 /* Call the destructor routine, if it exists */
 if(us->extra_destructor) {
	US_DEBUGP("-- calling extra_destructor()\n");
	us->extra_destructor(us->extra);
   }

 /* Free the extra data and the URB */
 if (us->extra) {
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] usb_stor_release_resources() : UsbFreeMemC(us_list->extra) , size=%d .. \n",
		sizeof(us->extra)
		));
	/*
	kfree(us->extra);
	*/
	UsbFreeMemC(USB_MEM_C_S,(unsigned char *) us->extra
			#if defined(USB_DBG_MEM_C_ON)
				,sizeof(us->extra)
			#endif
			);
   }

 if (us->current_urb) {
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] usb_stor_release_resources() : usb_free_urb() .. \n"
		));
	usb_free_urb(us->current_urb);
   }
 /* Free the structure itself */
 /*
	kfree(us);
 */
 /*
 UsbFreeMemC(USB_MEM_C_M,(unsigned char *) us
		#if defined(USB_DBG_MEM_C_ON)
			,sizeof(struct us_data) // Size : 236
		#endif
		);
 */
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] usb_stor_release_resources() : UsbFreeMemU(us) , size=%d .. \n",
	sizeof(struct us_data)
	));

 UsbFreeMemU(USB_MEM_U_M, (unsigned char *) us
				#if defined(USB_DBG_MEM_U_ON)
					,sizeof(struct us_data) // Size : 236
				#endif
				) ;

 US_DEBUGP("-- %s finished\n", __FUNCTION__);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_release_resources() End. \n"));
}



// 2004/2/10 adai s

static int storage_check_class(struct usb_host_interface *desc)
{
 /*
 struct usb_interface *if_alt;
 struct usb_host_interface *ifd;
 */
 /// struct usb_endpoint_descriptor *epd, *epwrite, *epread;
 // int p, i, e;
 // int i ;
 
 // struct usb_interface_descriptor *desc ;

 // desc = &intf->altsetting[intf->act_altsetting].desc;
 
 // USB_DBG_HUB(iprintf("[USB HUB] hub_check_class() : usb_dev->actconfig->desc.bNumInterfaces=%d \n", usb_dev->actconfig->desc.bNumInterfaces )) ;
 // desc->bInterfaceClass

 USB_DBG_INIT(iprintf("[USB INIT] storage_check_class() : desc->desc.bInterfaceClass=%d \n", desc->desc.bInterfaceClass)) ;

 if(desc->desc.bInterfaceClass == USB_CLASS_MASS_STORAGE ) { // 8
	// USB_DBG_INIT(iprintf("[USB INIT] hub_check_class() : Found HUB Class , usb_dev->actconfig->desc.bNumInterfaces=%d \n", usb_dev->actconfig->desc.bNumInterfaces)) ;
	USB_DBG_INIT(iprintf("[USB INIT] storage_check_class() : Found Storage Class \n")) ;
    return 1 ;
   }
 
 /*
 if_alt = &usb_dev->actconfig->interface[usb_dev->actconfig->desc.bNumInterfaces];
 
 // Find out what we have. 

 for(i = 0; i < if_alt->num_altsetting; i++) {
	ifd = &if_alt->altsetting[i];

	if (ifd->desc.bInterfaceClass == USB_CLASS_HUB // 9
    	// || ifd->desc.bInterfaceSubClass != 1
        )
       {
		USB_DBG_INIT(iprintf("[USB INIT] hub_check_class() : Found HUB Class , usb_dev->actconfig->desc.bNumInterfaces=%d , i=%d\n", usb_dev->actconfig->desc.bNumInterfaces, i)) ;
        return 1 ;
		// continue;
       }
	
	//if (ifd->desc.bInterfaceProtocol < USBLP_FIRST_PROTOCOL ||
	//	    ifd->desc.bInterfaceProtocol > USBLP_LAST_PROTOCOL)
    //   {
	//	continue;
    //   }
	
	// Look for bulk OUT and IN endpoints. 
    // return 1 ;
   } // for i 
 */
 
 return 0 ;
}
//adai e




/* Probe to see if we can drive a newly-connected USB device */
static int storage_probe(struct usb_interface *intf,
			 const struct usb_device_id *id)
{
 struct us_data *us;
 /*
 #ifdef _ETCPIP
 */
	int id_index = id - storage_usb_ids ;
 /*
  #else
	const int id_index = id - storage_usb_ids; 
 #endif
 */
 int result;
 //adai s
 struct usb_host_interface *desc;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] storage_probe() Begin .. \n"));
 desc = intf->altsetting + intf->act_altsetting;
//adai e

// 2004/2/10 adai s
//adai s

if( ! storage_check_class(desc) ) {
	USB_DBG_INIT(iprintf("[USB INIT] Not Storage Class !! \n")) ;
	//usb_hub_init_end = 1 ;
	return -EIO;
   }
//adai e




 // #ifdef _ETCPIP
 id = usb_match_id(intf, &(storage_usb_ids[0]));
 id_index = id - storage_usb_ids; 
 // #endif // _ETCPIP

 /*
 US_DEBUGP("USB Mass Storage device detected\n");
 US_DEBUGP("act_altsetting is %d, id_index is %d\n",
			intf->act_altsetting, id_index);
 */

// !!!!!
// malone, 12/Apr/2005
// Note: Should not remove this part because it is for MP test program check if USB is functional.
{
	char old_debug_act, old_debug_disp;

	// temporarily enable debugging console
	old_debug_act = set_debug_act(1);
	old_debug_disp = set_debug_rt_disp(1);

	dprintf("[USBMS] USB Mass Storage device detected\n");


	// restore the debugging console settings
	set_debug_act(old_debug_act);
	set_debug_rt_disp(old_debug_disp);
}

 dprintf("[USBMS] act_altsetting is %d, id_index is %d\n",
			intf->act_altsetting, id_index);

 /* Allocate the us_data structure and initialize the mutexes */

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] storage_probe() : us = UsbAllocMemC() .. \n"
		));

 /*
 us = (struct us_data *) kmalloc(sizeof(*us), GFP_KERNEL);
 */
 /*
 us = (struct us_data *) UsbAllocMemC(USB_MEM_C_M
					#ifdef USB_DBG_MEM_C_ON
						, sizeof(struct us_data) // Size : 236
		    	    #endif
	           		) ;
 */

 us = (struct us_data *) UsbAllocMemU(USB_MEM_U_M
					#ifdef USB_DBG_MEM_U_ON
						, sizeof(struct us_data) // Size : 236
			        #endif   
					);


 if (!us) {
		/*
		printk(KERN_WARNING USB_STORAGE "Out of memory\n");
		*/
		dprintf("[USBMS] USB_STORAGE Out of memory\n");
		return -ENOMEM;
	}

#ifdef USBMS_SCSI_SHORTCUT  /* Charles 20040526 */
 MSUsb_us = us;
#endif

 /*
 #ifndef _ETCPIP
	memset(us, 0, sizeof(struct us_data));
 #endif // ndef _ETCPIP
 */

 /*
 #ifndef NO_SEMPHORE
	init_MUTEX(&(us->dev_semaphore));
 #endif // ndef NO_SEMPHORE
 */
 /*
 #ifndef NO_SEMPHORE
	init_MUTEX_LOCKED(&(us->sema));
  #else
 */
#ifndef USBMS_NO_USB_TASK  /* Charles 20040520 */
 #ifdef USBMS_USB_WAIT_BY_EVENT     // 2004/1/2 B. S. Liaw
	// 2003/11/29 B. S. Liaw :
	// us->sema replace by USBMS_RESOURCE on kernel/local.h
	// 2003/11/29 B. S. Liaw : it seems to use event rather than resource
	// MT_EVT_USBMS_DOWN on kernel/local.h
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] storage_probe() : clrevt MT_EVT_USBMS_DOWN=%d ..\n",
		MT_EVT_USBMS_DOWN
		)) ;
	// MT_EVT_USBMS_DOWN is on kernel/local.h
    if (clrevt(MT_EVT_USBMS_DOWN) != SUCCESS) {
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] storage_probe() : Error : clrevt fail: Invalid MT_EVT_USBMS_DOWN=%d\n",
			MT_EVT_USBMS_DOWN
			)) ;
        dprintf("[USBMS] Error : clrevt fail: Invalid MT_EVT_USBMS_DOWN %d\n",
				MT_EVT_USBMS_DOWN);
	   }
  #else
	USBMS_DBG_INIT(iprintf("[USBMS INIT] storage_probe() : us=%08XH .. \n",us)) ;
	USBMS_DBG_INIT(iprintf("[USBMS INIT] storage_probe() : Set Value : us->evt_ms=0 .. \n")) ;
	// BLOCK_PREEMPTION;
	us->evt_ms = 0 ;
	// UNBLOCK_PREEMPTION;
 #endif // ndef USBMS_USB_WAIT_BY_EVENT
#endif
 /*
 #endif
 */
 /*
 #ifndef _ETCPIP
    // struct completion in <linux/completion.h>
	init_completion(&(us->notify));
 #endif // ndef _ETCPIP
 */
 /* Associate the us_data structure with the USB device */
 
 result = associate_dev(us, intf);

 if (result)
		goto BadDevice;

 /*
  * Get the unusual_devs entries and the descriptors
  *
  * id_index is calculated in the declaration to be the index number
  * of the match from the usb_device_id table, so we can find the
  * corresponding entry in the private table.
 */
 USBMS_DBG_INIT(iprintf("[USBMS INIT] storage_probe(): us = %08XH\n", us));

 get_device_info(us, id_index);

 #ifdef CONFIG_USB_STORAGE_SDDR09
	if (us->protocol == US_PR_EUSB_SDDR09 ||
			us->protocol == US_PR_DPCM_USB) {
		/* set the configuration -- STALL is an acceptable response here */
		if (us->pusb_dev->actconfig->desc.bConfigurationValue != 1) {
			US_DEBUGP("active config #%d != 1 ??\n", us->pusb_dev
				->actconfig->desc.bConfigurationValue);
		goto BadDevice;
	}
		result = usb_reset_configuration(us->pusb_dev);

		US_DEBUGP("Result of usb_reset_configuration is %d\n", result);
		if (result == -EPIPE) {
			US_DEBUGP("-- stall on control interface\n");
		} else if (result != 0) {
			/* it's not a stall, but another error -- time to bail */
			US_DEBUGP("-- Unknown error.  Rejecting device\n");
			goto BadDevice;
		}
	}
 #endif

 /* Get the transport, protocol, and pipe settings */
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] storage_probe() : Get the transport, protocol, and pipe settings .. \n"));
 result = get_transport(us);
 if (result)
		goto BadDevice;
 result = get_protocol(us);
 if (result)
		goto BadDevice;
 result = get_pipes(us);
 if (result)
		goto BadDevice;

 /* Acquire all the other resources */
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] storage_probe() : Acquire all the other resources .. \n"));
 result = usb_stor_acquire_resources(us);
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] storage_probe() : usb_stor_acquire_resources() result=%d \n",
	result
	));

 if(result)
   {
	goto BadDevice;
   }

 /* Finally, add the host (this does SCSI device scanning) */

 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] storage_probe() : add the host .. \n"));

 result = scsi_add_host(us->host, &intf->dev);
 if (result) {
	/*
	printk(KERN_WARNING USB_STORAGE
			"Unable to add the scsi host\n");
	*/
	dprintf("[USBMS] Error : USB_STORAGE Unable to add the scsi host !! \n");
	goto BadDevice;
   }

 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] storage_probe() : SCSI device scanning .. \n"));

 scsi_scan_host(us->host);

 /*
 printk(KERN_DEBUG 
	       "WARNING: USB Mass Storage data integrity not assured\n");
 */
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] storage_probe() : WARNING: USB Mass Storage data integrity not assured \n"));
 /*
 printk(KERN_DEBUG 
	       "USB Mass Storage device found at %d\n", us->pusb_dev->devnum);
 */
 global_usbms_usb_device = us->pusb_dev ;
 dprintf("[USBMS] USB Mass Storage device found at USB Device %d (port %d)\n",
	us->pusb_dev->devnum ,
	us->pusb_dev->hub_port+1  // Attached on root hub port
   );

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] storage_probe() End. \n"));
 return 0;

 /* We come here if there are any problems */

 // ---------------
 BadDevice:
 // ---------------
 /*
 US_DEBUGP("storage_probe() failed\n");
 */
 USBMS_DBG_INIT(iprintf("[USBMS INIT] storage_probe() : Bad Device !! \n"));
 dissociate_dev(us);
 usb_stor_release_resources(us);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] storage_probe() End with Fail !! \n"));
 return result;
}

/* Handle a disconnect event from the USB core */
static void storage_disconnect(struct usb_interface *intf)
{
	int i, j;
 struct us_data *us = usb_get_intfdata(intf);


 USB_DBG_PLUG(iprintf("[USB PLUG] storage_disconnect() Begin ..\n" )) ;
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] storage_disconnect() Begin ..\n" )) ;

 USB_DBG_INIT(iprintf("[USB INIT] storage_disconnect() : intf=%08XH , Set NULL ..\n",intf )) ;
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] storage_disconnect() : intf=%08XH , Set NULL ..\n",intf )) ;
 /*
 US_DEBUGP("storage_disconnect() called\n");
 */
	
	dprintf("[USBMS PLUG] total Logical Disk : %d \n", validPartitionCnt);

	// un-registe every partition.  fat_unmount() will judge which partition is valid.
	for (j = 0 ; j < validDiskCnt ; j++) {
		for (i = 0 ; i < validPartitionCnt ; i++) {
			USBMS_DBG_INFO(iprintf("[USBMS PLUG] mount(%d) %d .. \n", i, validPartitionID[i]));
			if (diskMount_status(j) & partitionMask[i]) {
				dprintf("[USBMS] un-mount partition %d\n", i);
				partition_umount(validPartitionID[i], PartitionType[i]);
				dprintf("[USBMS] partition %d unmounted\n", i);
			}
		}
	}
	validPartitionCnt = 0;
	memset(validPartitionID, 0, (sizeof(unsigned long) * 20));
	memset(PartitionType, 0, (sizeof(unsigned char) * 20));   // PTR-TI-0134 2004.11.24

	// update netbios share resource validation
	#if defined(_NB_SHARE) && defined(_NB_FILE)
	updateShareConfPlug(STYPE_DISKTREE, 1);
	#endif // defined(_NB_SHARE) && defined(_NB_FILE)

 /* Prevent new USB transfers and stop the current command */
 set_bit(US_FLIDX_DISCONNECTING, &us->flags);
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] storage_disconnect() : usb_stor_stop_transport() ..\n" )) ;
 usb_stor_stop_transport(us);
 
 /* Dissociate from the USB device */
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] storage_disconnect() : dissociate_dev() .. \n")) ;
 dissociate_dev(us);
 
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] storage_disconnect() : scsi_remove_host() .. \n")) ;
 scsi_remove_host(us->host);

 /* TODO: somehow, wait for the device to
  * be 'idle' (tasklet completion) */

 /* Release all our other resources */
 usb_stor_release_resources(us);
 global_usbms_usb_device = (struct usb_device *) NULL ;
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] storage_disconnect() End.\n" )) ;
}

int usb_ms_unplug(void)
{
 int rc;
 
 USB_DBG_PLUG(iprintf("[USB PLUG] usb_ms_unplug() Begin ..\n" )) ;
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] usb_ms_unplug() Begin ..\n" )) ;
 if( !global_usbms_usb_device) {
	return 0;
   }
 /*
 dprintf("[USB MS] UNPLUG , port %d disabled\n",
			global_usbms_usb_device->hub_port+1);
 usb_hub_port_disable(global_usbms_usb_device->hub_usb_device ,
					  global_usbms_usb_device->hub_port) ;
 */
 dprintf("[USBMS] Software Un-Plug , HUB port %d .. \n",
	global_usbms_usb_device->hub_port
	) ;

 USBMS_DBG_PLUG(iprintf(
	"[USBMS PLUG] usb_ms_unplug() : UNPLUG Reset the hub port .. \n"
	)) ;
    
 // Prevent the HUB to Re-Enable the Port

 global_usbms_usb_device->hub->port_enable
				[global_usbms_usb_device->hub_port] = 0 ;

 /*
 if(usb_hub_port_reset(global_usbms_usb_device->hub_usb_device,
		global_usbms_usb_device->hub_port, global_usbms_usb_device,HUB_SHORT_RESET_TIME)) {
	dprintf("[USBMS] Error : Reset the hub port Fail !! \n" ) ;
	// break;
   }
 */

#if defined(_AR7300) || defined(_ATHEROS)
 rc = usb_set_port_feature(global_usbms_usb_device->hub_usb_device,
		global_usbms_usb_device->hub_port + 1,
		USB_PORT_FEAT_SW_UNPLUG);
#else
 rc = usb_set_port_feature(global_usbms_usb_device->hub_usb_device,
		global_usbms_usb_device->hub_port + 1,
		USB_PORT_FEAT_RESET);
#endif

 USBMS_DBG_PLUG(iprintf(
	"[USBMS PLUG] usb_ms_unplug() : after Reset the hub port \n" )) ;

 /*
 dprintf("[USB MS] UNPLUG , port %d disabled\n",
			global_usbms_usb_device->hub_port+1);
 usb_hub_port_disable(global_usbms_usb_device->hub_usb_device ,
					  global_usbms_usb_device->hub_port) ;
 */
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] usb_ms_unplug() End.\n" )) ;
 
 return rc;
}


/***********************************************************************
 * Initialization and registration
 ***********************************************************************/
/*
static int __init usb_stor_init(void)
*/
int usb_stor_init(void)
{
 int retval;


 USBMS_DBG_FLOW(iprintf(
		"[USBMS FLOW] usb_stor_init() Begin .. \n"
		));
 /*
 printk(KERN_INFO "Initializing USB Mass Storage driver...\n");
 */
 dprintf("[USBMS] Initializing USB Mass Storage driver...\n");
 global_usbms_usb_device = (struct usb_device *) NULL ;
 /* register the driver, return usb_register return code if error */
 retval = usb_register(&usb_storage_driver);
 if (retval)
		goto out;

 /* we're all set */
 /*
 printk(KERN_INFO "USB Mass Storage support registered.\n");
 */
 dprintf("[USBMS] USB Mass Storage support registered.\n");
 out:
 return retval;
}

static void __exit usb_stor_exit(void)
{
	US_DEBUGP("usb_stor_exit() called\n");

	/* Deregister the driver
	 * This will cause disconnect() to be called for each
	 * attached unit
	 */
	US_DEBUGP("-- calling usb_deregister()\n");
	usb_deregister(&usb_storage_driver) ;
 /*
 #if 0
	// While there are still virtual hosts, unregister them
	// * Note that it's important to do this completely before removing
	// * the structures because of possible races with the /proc
	// * interface
	 
	for (next = us_list; next; next = next->next) {
		US_DEBUGP("-- calling scsi_unregister_host()\n");
		scsi_unregister_host(&usb_stor_host_template);
	}

	// While there are still structures, free them.  Note that we are
	// * now race-free, since these structures can no longer be accessed
	// * from either the SCSI command layer or the /proc interface
	 
	while (us_list) {
		// keep track of where the next one is 
		next = us_list->next;

		// If there's extra data in the us_data structure then
		// free that first 
		if (us_list->extra) {
			// call the destructor routine, if it exists 
			if (us_list->extra_destructor) {
				US_DEBUGP("-- calling extra_destructor()\n");
				us_list->extra_destructor(us_list->extra);
			}

			// destroy the extra data 
			US_DEBUGP("-- freeing the data structure\n");
		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] usb_stor_exit() : UsbFreeMemC(us_list->extra) , size=%d .. \n",
			sizeof(us_list->extra)
			));
		#ifndef _ETCPIP
			kfree(us_list->extra);
		#elif defined(USB_DBG_MEM_C_ON)
			UsbFreeMemC(USB_MEM_C_S,(unsigned char *) us_list->extra,sizeof(us_list->extra));
		#else
			UsbFreeMemC(USB_MEM_C_S,(unsigned char *) us_list->extra);
		#endif
		}

		// free the structure itself 
		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] usb_stor_exit() : UsbFreeMemC(us_list) , size=%d .. \n",
			sizeof(*us_list)
			));
		#ifndef _ETCPIP
			kfree(us_list);
		#elif defined(USB_DBG_MEM_C_ON)
			UsbFreeMemC(USB_MEM_C_S,(unsigned char *) us_list,sizeof(*us_list));
		#else
			UsbFreeMemC(USB_MEM_C_S,(unsigned char *) us_list);
		#endif

		// advance the list pointer 
		us_list = next;
	}
 #endif
 */
}

/*
module_init(usb_stor_init);
module_exit(usb_stor_exit);
*/

// B. S. Liaw TEST for compiler warning !!

// ====================================================================
// End of File : ms_usb.c

