
// drivers/scsi/scsicam.c

/*
 * scsicam.c - SCSI CAM support functions, use for HDIO_GETGEO, etc.
 *
 * Copyright 1993, 1994 Drew Eckhardt
 *      Visionary Computing 
 *      (Unix and Linux consulting and custom programming)
 *      drew@Colorado.EDU
 *      +1 (303) 786-7975
 *
 * For more information, please consult the SCSI-CAM draft.
 */

// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP


/*
#include <linux/module.h>
*/
#ifndef _ETCPIP
    #include <linux/fs.h>		/* for struct file_operations */
 #else
    #include "PoFile.h" // include/linux/file.h
    #include "radix-tree.h" // include/linux/radix-tree.h
    #include "PoFs.h" // osport_fs.h
#endif // ndef _ETCPIP

/*
#include <linux/genhd.h>
*/
#include "block/genhd.h"

/*
#include <linux/kernel.h>
*/
#include "PoKernel.h"
/*
#include <linux/blkdev.h>
*/
#include "bitops.h" // B. S. Liaw add for test_bit()
#include "blkdev.h"
/*
#include <linux/buffer_head.h>
*/
#include "buffer_head.h"

/*
#include <asm/unaligned.h>
*/
#include "unaligned-mips.h"

#include "scsi_ds.h"
#include "scsi_hosts.h"
/*
#include <scsi/scsicam.h>
*/
#include "scsi_cam.h"

static int setsize(unsigned long capacity, unsigned int *cyls, unsigned int *hds,
		   unsigned int *secs);

/*
unsigned char *scsi_bios_ptable(struct block_device *dev)
{
 
 // unsigned char *res = (unsigned char *) kmalloc(66, GFP_KERNEL);
 
 unsigned char *res ;


 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_bios_ptable() : res = UsbAllocMemC() .. \n"
		));

 res = (unsigned char *) UsbAllocMemC(USB_MEM_C_S
					#ifdef USB_DBG_MEM_C_ON
							, 66
		    	    #endif
	           		) ;

 if(res) {
	struct block_device *bdev = dev->bd_contains;

 
	// __bread is on linux/fs/buffer.c
	struct buffer_head *bh = __bread(bdev, 0, block_size(bdev));
	
	if (bh) {
		memcpy((void *) res, (void *)(bh->b_data + 0x1be), 66);
		brelse(bh);
	   }
      else
       {
		USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_bios_ptable() : UsbFreeMemC(res) .. \n"));
		
		// kfree(res);
		
		UsbFreeMemC(USB_MEM_C_S , (unsigned char *) res
						#if defined(USB_DBG_MEM_C_ON)
							,66
						#endif
					);

		res = NULL;
	   }
	
   }
 return res;
}
*/
/*
 * Function : int scsicam_bios_param (struct block_device *bdev, ector_t capacity, int *ip)
 *
 * Purpose : to determine the BIOS mapping used for a drive in a 
 *      SCSI-CAM system, storing the results in ip as required
 *      by the HDIO_GETGEO ioctl().
 *
 * Returns : -1 on failure, 0 on success.
 *
 */

/*
int scsicam_bios_param(struct block_device *bdev, sector_t capacity, int *ip)
{
	unsigned char *p;
	int ret;

	p = scsi_bios_ptable(bdev);
	if (!p)
		return -1;

	// try to infer mapping from partition table 
	ret = scsi_partsize(p, (unsigned long)capacity, (unsigned int *)ip + 2,
			       (unsigned int *)ip + 0, (unsigned int *)ip + 1);

 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsicam_bios_param() : UsbFreeMemC(p) .. \n"));
		#ifndef _ETCPIP
			kfree(p);
		#elif defined(USB_DBG_MEM_C_ON)
			UsbFreeMemC(USB_MEM_C_S,(unsigned char *) p, 66 ); // line # 94
		#else
			UsbFreeMemC(USB_MEM_C_S,(unsigned char *) p);
		#endif

	if (ret == -1) {
		// pick some standard mapping with at most 1024 cylinders,
		//   and at most 62 sectors per track - this works up to
		//   7905 MB 
		ret = setsize((unsigned long)capacity, (unsigned int *)ip + 2,
		       (unsigned int *)ip + 0, (unsigned int *)ip + 1);
	}

	// if something went wrong, then apparently we have to return
	//   a geometry with more than 1024 cylinders 
	if (ret || ip[0] > 255 || ip[1] > 63) {
		if ((capacity >> 11) > 65534) {
			ip[0] = 255;
			ip[1] = 63;
		} else {
			ip[0] = 64;
			ip[1] = 32;
		}

		if (capacity > 65535*63*255)
			ip[2] = 65535;
		else
			ip[2] = (unsigned long)capacity / (ip[0] * ip[1]);
	}

	return 0;
}
*/
/*
 * Function : static int scsi_partsize(unsigned char *buf, unsigned long 
 *     capacity,unsigned int *cyls, unsigned int *hds, unsigned int *secs);
 *
 * Purpose : to determine the BIOS mapping used to create the partition
 *      table, storing the results in *cyls, *hds, and *secs 
 *
 * Returns : -1 on failure, 0 on success.
 *
 */

#ifndef USBMS_NO_DBGHOST // SCSI Host for debug
int scsi_partsize(unsigned char *buf, unsigned long capacity,
	       unsigned int *cyls, unsigned int *hds, unsigned int *secs)
{
 
 // struct partition *p = (struct partition *)buf, *largest = NULL;
 
 struct hd_partition *p = (struct hd_partition *)buf, *largest = NULL;
 int i, largest_cyl;
 int cyl, ext_cyl, end_head, end_cyl, end_sector;
 unsigned int logical_end, physical_end, ext_physical_end;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_partsize() Begin .. \n"));
 if (*(unsigned short *) (buf + 64) == 0xAA55) {
	for (largest_cyl = -1, i = 0; i < 4; ++i, ++p) {
		if (!p->sys_ind)
				continue;
		#ifdef DEBUG
			printk("scsicam_bios_param : partition %d has system \n",
			       i);
		#endif
		cyl = p->cyl + ((p->sector & 0xc0) << 2);
		if (cyl > largest_cyl) {
			largest_cyl = cyl;
			largest = p;
		   }
	   }
   }
 if (largest) {
	end_cyl = largest->end_cyl + ((largest->end_sector & 0xc0) << 2);
	end_head = largest->end_head;
	end_sector = largest->end_sector & 0x3f;

	if (end_head + 1 == 0 || end_sector == 0)
			return -1;

	#ifdef DEBUG
		printk("scsicam_bios_param : end at h = %d, c = %d, s = %d\n",
		       end_head, end_cyl, end_sector);
	#endif

	physical_end = end_cyl * (end_head + 1) * end_sector +
		    end_head * end_sector + end_sector;

	//  This is the actual _sector_ number at the end 
	logical_end = get_unaligned(&largest->start_sect)
		    + get_unaligned(&largest->nr_sects);

	// This is for >1023 cylinders 
	ext_cyl = (logical_end - (end_head * end_sector + end_sector))
		    / (end_head + 1) / end_sector;
	ext_physical_end = ext_cyl * (end_head + 1) * end_sector +
		    end_head * end_sector + end_sector;

	#ifdef DEBUG
		printk("scsicam_bios_param : logical_end=%d physical_end=%d ext_physical_end=%d ext_cyl=%d\n"
		  ,logical_end, physical_end, ext_physical_end, ext_cyl);
	#endif

	if ((logical_end == physical_end) ||
		  (end_cyl == 1023 && ext_physical_end == logical_end)) {
		*secs = end_sector;
		*hds = end_head + 1;
		*cyls = capacity / ((end_head + 1) * end_sector);
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_partsize() End. \n"));
		return 0;
	   }
	#ifdef DEBUG
		printk("scsicam_bios_param : logical (%u) != physical (%u)\n",
		       logical_end, physical_end);
	#endif
   }
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_partsize() End with Failed !! \n"));
 return -1;
}
#endif // ndef USBMS_NO_DBGHOST


/*
 * Function : static int setsize(unsigned long capacity,unsigned int *cyls,
 *      unsigned int *hds, unsigned int *secs);
 *
 * Purpose : to determine a near-optimal int 0x13 mapping for a
 *      SCSI disk in terms of lost space of size capacity, storing
 *      the results in *cyls, *hds, and *secs.
 *
 * Returns : -1 on failure, 0 on success.
 *
 * Extracted from
 *
 * WORKING                                                    X3T9.2
 * DRAFT                                                        792D
 *
 *
 *                                                        Revision 6
 *                                                         10-MAR-94
 * Information technology -
 * SCSI-2 Common access method
 * transport and SCSI interface module
 * 
 * ANNEX A :
 *
 * setsize() converts a read capacity value to int 13h
 * head-cylinder-sector requirements. It minimizes the value for
 * number of heads and maximizes the number of cylinders. This
 * will support rather large disks before the number of heads
 * will not fit in 4 bits (or 6 bits). This algorithm also
 * minimizes the number of sectors that will be unused at the end
 * of the disk while allowing for very large disks to be
 * accommodated. This algorithm does not use physical geometry. 
 */
/*
static int setsize(unsigned long capacity, unsigned int *cyls, unsigned int *hds,
		   unsigned int *secs)
{
	unsigned int rv = 0;
	unsigned long heads, sectors, cylinders, temp;

	cylinders = 1024L;	// Set number of cylinders to max 
	sectors = 62L;		// Maximize sectors per track 

	temp = cylinders * sectors;	// Compute divisor for heads 
	heads = capacity / temp;	// Compute value for number of heads 
	if (capacity % temp) {	// If no remainder, done! 
		heads++;	// Else, increment number of heads 
		temp = cylinders * heads;	// Compute divisor for sectors 
		sectors = capacity / temp;	// Compute value for sectors per
						   // track 
		if (capacity % temp) {	// If no remainder, done! 
			sectors++;	// Else, increment number of sectors 
			temp = heads * sectors;		// Compute divisor for cylinders 
			cylinders = capacity / temp;	// Compute number of cylinders
		}
	}
	if (cylinders == 0)
		rv = (unsigned) -1;	// Give error if 0 cylinders 

	*cyls = (unsigned int) cylinders;	// Stuff return values
	*secs = (unsigned int) sectors;
	*hds = (unsigned int) heads;
	return (rv);
}
*/
