
// driver/scsi/hosts.c

/*
 *  hosts.c Copyright (C) 1992 Drew Eckhardt
 *          Copyright (C) 1993, 1994, 1995 Eric Youngdale
 *
 *  mid to lowlevel SCSI driver interface
 *      Initial versions: Drew Eckhardt
 *      Subsequent revisions: Eric Youngdale
 *
 *  <drew@colorado.edu>
 *
 *  Jiffies wrap fixes (host->resetting), 3 Dec 1998 Andrea Arcangeli
 *  Added QLOGIC QLA1280 SCSI controller kernel host support. 
 *     August 4, 1999 Fred Lewis, Intel DuPont
 *
 *  Updated to reflect the new initialization scheme for the higher 
 *  level of scsi drivers (sd/sr/st)
 *  September 17, 2000 Torben Mathiasen <tmm@image.dk>
 *
 *  Restructured scsi_host lists and associated functions.
 *  September 04, 2002 Mike Anderson (andmike@us.ibm.com)
 */


/*
 *  This file contains the medium level SCSI
 *  host interface initialization, as well as the scsi_hosts list of SCSI
 *  hosts currently present in the system.
 */

// #ifdef _ETCPIP
#define __KERNEL__
        #ifdef _NEWLIB
          #include "string.h"
        #endif
   	#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
          // for deciding whether to compile
          //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
		#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP


/*
#include <linux/module.h>
#include <linux/blkdev.h>
*/
#include "bitops.h" // B. S. Liaw add for test_bit()
#include "blkdev.h"
/*
#include <linux/kernel.h>
*/
#include "PoKernel.h"
/*
#include <linux/string.h>
*/
#include "PoString.h" // linux/string.h
#ifndef _ETCPIP
	#include <linux/mm.h>
	#include <linux/init.h>
#endif // ndef _ETCPIP
#ifndef _ETCPIP
   #include <linux/list.h>
 #else
	#include "PoList.h" // osport_list.h <linux/list.h>
#endif
#ifndef _ETCPIP
	#include <linux/completion.h>
#endif // ndef _ETCPIP

#define __KERNEL_SYSCALLS__

#ifndef _ETCPIP
	#include <linux/unistd.h>
#endif // ndef _ETCPIP

#include "scsi_ds.h"
#include "scsi_hosts.h"

#include "scsi_priv.h"
#include "scsi_logging.h"


static int scsi_host_next_hn = 0;		/* host_no for next new host */


static void scsi_host_cls_release(struct class_device *class_dev)
{
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_host_cls_release() Begin .. \n"));
 
 // #ifndef _ETCPIP
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_host_cls_release() : put_device() .. \n"));
 put_device(&class_to_shost(class_dev)->shost_gendev);
 // #endif // ndef _ETCPIP
 
}

static struct class shost_class = {
 
 #ifdef USB_DBGSTRING_ON
	.name		= "scsi_host",
 #endif // USB_DBGSTRING_ON
 
 .release	= scsi_host_cls_release,
};

/**
 * scsi_host_cancel - cancel outstanding IO to this host
 * @shost:	pointer to struct Scsi_Host
 * recovery:	recovery requested to run.
  **/
void scsi_host_cancel(struct Scsi_Host *shost, int recovery)
{
 unsigned long flags;


 // USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_host_cancel() Begin .. \n")) ;
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_host_cancel() Begin .. \n")) ;
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irqsave(shost->host_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 set_bit(SHOST_CANCEL, &shost->shost_state);
 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(shost->host_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 // USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_host_cancel() : scsi_device_cancel_cb() .. \n")) ;
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_host_cancel() : scsi_device_cancel_cb() .. \n")) ;
 device_for_each_child(&shost->shost_gendev, &recovery,
			      scsi_device_cancel_cb);
 /*
 #ifndef NO_WAITQUEUE
	wait_event(shost->host_wait, (!test_bit(SHOST_RECOVERY,
						&shost->shost_state)));
 #endif // ndef NO_WAITQUEUE
 */
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_host_cancel() End. \n")) ;
}

/**
 * scsi_remove_host - remove a scsi host
 * @shost:	a pointer to a scsi host to remove
 **/
void scsi_remove_host(struct Scsi_Host *shost)
{
 unsigned long flags;


 // USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_remove_host() Begin .. \n")) ;
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_remove_host() Begin .. \n")) ;

 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_remove_host() : scsi_host_cancel() .. \n")) ;
 scsi_host_cancel(shost, 0);
 scsi_proc_host_rm(shost);
 scsi_forget_host(shost);
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irqsave(shost->host_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 set_bit(SHOST_DEL, &shost->shost_state);
 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(shost->host_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 class_device_unregister(&shost->shost_classdev);
 USB_DBG_MEM(iprintf(
	"[USB MEM] scsi_remove_host() : &shost->shost_gendev=%08XH \n",
	&shost->shost_gendev));
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_remove_host() : device_del() .. \n"));
 device_del(&shost->shost_gendev);
 
 // #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	if(shost->shost_gendev.release) {
		USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_remove_host() : shost->shost_gendev.release() .. \n"));
		shost->shost_gendev.release(&shost->shost_gendev) ;
	   } ;
 // #endif // NO_ELEVATOR
 
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_remove_host() End. \n")) ;
}

/**
 * scsi_add_host - add a scsi host
 * @shost:	scsi host pointer to add
 * @dev:	a struct device of type scsi class
 *
 * Return value: 
 * 	0 on success / != 0 for error
 **/
int scsi_add_host(struct Scsi_Host *shost, struct device *dev)
{
 struct scsi_host_template *sht = shost->hostt;
 int error;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_add_host() Begin .. \n"));
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_add_host() Begin .. \n"));
 #ifdef USB_DBGSTRING_ON
	/*
	printk(KERN_INFO "scsi%d : %s\n", shost->host_no,
			sht->info ? sht->info(shost) : sht->name);
	*/
	dprintf("[USBMS] scsi%d : %s\n", shost->host_no,
			sht->info ? sht->info(shost) : sht->name);
  #else
	dprintf("[USBMS] scsi%d ..\n", shost->host_no );
 #endif // ndef USB_DBGSTRING_ON

 if (!shost->can_queue) {
	#ifdef USB_DBGSTRING_ON
		/*
		printk(KERN_ERR "%s: can_queue = 0 no longer supported\n",
				sht->name);
		*/
		dprintf("[USBMS] Error :%s: can_queue = 0 no longer supported\n",
				sht->name);
	 #else
		dprintf("[USBMS] Error : can_queue = 0 no longer supported\n");
	#endif // ndef USB_DBGSTRING_ON
	return -EINVAL;
   }

 if (!shost->shost_gendev.parent)
		shost->shost_gendev.parent = dev ? dev : &legacy_bus;
			
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_add_host() : device_add() .. \n"));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_add_host() call device_add() .. \n"));
 error = device_add(&shost->shost_gendev);
 if(error) {
	USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_add_host() : Error :  device_add() Fail !! \n"));
	goto out;
   }

 set_bit(SHOST_ADD, &shost->shost_state);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_add_host() call get_device() .. \n"));
 get_device(shost->shost_gendev.parent);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_add_host() call class_device_add() .. \n"));
 error = class_device_add(&shost->shost_classdev);
 if (error)
   {
	USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_add_host() : Error : class_device_add() Fail !! \n"));
	goto out_del_gendev;
   }

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_add_host() call get_device() .. \n"));
 get_device(&shost->shost_gendev);

 // #ifndef NO_SYS_FS
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_add_host() call scsi_sysfs_add_host() .. \n"));
 error = scsi_sysfs_add_host(shost);
 if (error)
   {
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] scsi_add_host() : Error : scsi_sysfs_add_host() Fail !! \n"
		));
	goto out_del_classdev;
   }
 // #endif // ndef NO_SYS_FS

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_add_host() call scsi_proc_host_add() .. \n"));
 scsi_proc_host_add(shost);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_add_host() End. \n"));
 return error;

 out_del_classdev:
	class_device_del(&shost->shost_classdev);
 out_del_gendev:
	device_del(&shost->shost_gendev);
 out:

 USBMS_DBG_FLOW(iprintf(
	"[USBMS FLOW] scsi_add_host() End with Fail !! \n"
	));
 return error;
}

static void scsi_host_dev_release(struct device *dev)
{
 struct Scsi_Host *shost = dev_to_shost(dev);
 struct device *parent = dev->parent;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_host_dev_release() Begin .. \n")) ;
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_host_dev_release() Begin .. \n"));
 /*
 #ifndef _ETCPIP
 if (shost->ehandler)
 #endif // ndef _ETCPIP
 */
  {
 /*
 #ifndef _ETCPIP
    // struct completion in <linux/completion.h>
		DECLARE_COMPLETION(sem);
		shost->eh_notify = &sem;
 #endif // ndef _ETCPIP
 */

 shost->eh_kill = 1;
 /*
 #ifndef NO_SEMPHORE
	up(shost->eh_wait);
  #else
 */
 #ifdef USBMS_EH_WAIT_BY_EVENT     // 2004/1/2 B. S. Liaw
	// 2003/12/01 B. S. Liaw :
	// shost->eh_wait replace by MT_EVT_USBMS_EH on kernel/local.h
	USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_host_dev_release() : SIGNAL(MT_EVT_USBMS_EH) .. \n")) ;
	// kernel/mtmacro.h : #define SIGNAL(signo) setevt(signo)
	if (SIGNAL(MT_EVT_USBMS_EH) != SUCCESS) {
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] scsi_host_dev_release() : Error : SIGNAL MT_EVT_USBMS_EH=%d fail !! \n",
			MT_EVT_USBMS_EH
			)) ;
		dprintf("[USBMS] Error : SIGNAL MT_EVT_USBMS_EH %d fail !! \n",MT_EVT_USBMS_EH);
	   }
  #else
	// BLOCK_PREEMPTION;
	req->evt_eh = 1 ;
	// UNBLOCK_PREEMPTION;
 #endif // ndef USBMS_EH_WAIT_BY_EVENT
 /*
 #endif
 */
 /*
 #ifndef _ETCPIP
    // wait_for_completion() is in <linux/completion.h>
		wait_for_completion(&sem);
 #endif // ndef _ETCPIP
 */
 /*
 #ifndef _ETCPIP
    // struct completion in <linux/completion.h>
		shost->eh_notify = NULL;
 #endif // ndef _ETCPIP
 */
	}

 scsi_proc_hostdir_rm(shost->hostt);

 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_host_dev_release() : scsi_destroy_command_freelist() .. \n"));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_host_dev_release() : scsi_destroy_command_freelist() .. \n"));
 scsi_destroy_command_freelist(shost);

 /*
  * Some drivers (eg aha1542) do scsi_register()/scsi_unregister()
  * during probing without performing a scsi_set_device() in between.
  * In this case dev->parent is NULL.
  */
 
 if (parent) {
	USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_host_dev_release() : put_device() .. \n"));
	put_device(parent);
   }
 
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_host_dev_release() : UsbFreeMemC(shost) , size=%d .. \n",
	sizeof(*shost)
	));
 /*
 kfree (shost);
 */
 UsbFreeMemC(USB_MEM_C_M,(unsigned char *) shost
			#if defined(USB_DBG_MEM_C_ON)
				,sizeof(*shost)  // size : 444
			#endif
			);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_host_dev_release() End. \n")) ;
}

/**
 * scsi_host_alloc - register a scsi host adapter instance.
 * @sht:	pointer to scsi host template
 * @privsize:	extra bytes to allocate for driver
 *
 * Note:
 * 	Allocate a new Scsi_Host and perform basic initialization.
 * 	The host is not published to the scsi midlayer until scsi_add_host
 * 	is called.
 *
 * Return value:
 * 	Pointer to a new Scsi_Host
 **/
/*
#ifndef _ETCPIP
	struct Scsi_Host *scsi_host_alloc(struct scsi_host_template *sht, int privsize)
 #else
*/
	struct Scsi_Host *scsi_host_alloc(struct scsi_host_template *sht, int privsize ,
		struct list_head *shost_my_devices )
/*
#endif
*/
{
 struct Scsi_Host *shost;
 int gfp_mask = GFP_KERNEL, rval;
 // #ifdef _ETCPIP
 int status1 ;
 // #endif
 /*
 #ifndef _ETCPIP
    // struct completion in <linux/completion.h>
	DECLARE_COMPLETION(complete);
 #endif // ndef _ETCPIP
 */

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_host_alloc() Begin .. \n"));
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_host_alloc() Begin .. \n")) ;

 if (sht->unchecked_isa_dma && privsize)
		gfp_mask |= __GFP_DMA;

 /* Check to see if this host has any error handling facilities */
 if (!sht->eh_strategy_handler && !sht->eh_abort_handler &&
	    !sht->eh_device_reset_handler && !sht->eh_bus_reset_handler &&
            !sht->eh_host_reset_handler) {
	/*
	printk(KERN_ERR "ERROR: SCSI host `%s' has no error handling\n"
			"ERROR: This is not a safe way to run your "
			        "SCSI host\n"
			"ERROR: The error handling must be added to "
			"this driver\n", sht->proc_name);
	*/
 #ifdef USB_DBGSTRING_ON
	dprintf("[USBMS] Error : SCSI host `%s' has no error handling\n"
			"ERROR: This is not a safe way to run your "
			        "SCSI host\n"
			"ERROR: The error handling must be added to "
			"this driver\n", sht->proc_name);
  #else
	dprintf("[USBMS] Error : SCSI host has no error handling\n"
			"ERROR: This is not a safe way to run your "
			        "SCSI host\n"
			"ERROR: The error handling must be added to "
			"this driver\n");
 #endif // ndef USB_DBGSTRING_ON
	/*
	dump_stack();
	*/
    } 

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_host_alloc() : shost = UsbAllocMemC() .. \n"
		));

 /*
 shost = kmalloc(sizeof(struct Scsi_Host) + privsize, gfp_mask);
 */
 shost = (struct Scsi_Host *) UsbAllocMemC(USB_MEM_C_M
				#ifdef USB_DBG_MEM_C_ON
						, sizeof(struct Scsi_Host) + privsize // size : 444
	    	    #endif
           		) ;

 if (!shost) {
	/*
	printk(KERN_ERR "%s: out of memory.\n", __FUNCTION__);
	*/
	USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_host_alloc() : Error : Out of memory !! \n"));
	dprintf("[USBMS] Error : out of memory.\n");
	return NULL;
   }

 /*
 memset(shost, 0, sizeof(struct Scsi_Host) + privsize);
 */

 /*
 #ifndef NO_SPINLOCK
	spin_lock_init(&shost->default_lock);
	scsi_assign_lock(shost, &shost->default_lock);
 #endif // ndef NO_SPINLOCK
 */

 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_host_alloc() : INIT_LIST_HEAD() .. \n"));
 /*
 #ifndef _ETCPIP
	INIT_LIST_HEAD(&shost->my_devices);
  #else
 */
	shost->my_devices = shost_my_devices ;
 /*
 #endif
 */
 INIT_LIST_HEAD(&shost->eh_cmd_q);
 INIT_LIST_HEAD(&shost->starved_list);
 /*
 #ifndef NO_WAITQUEUE
	init_waitqueue_head(&shost->host_wait);
 #endif // ndef NO_WAITQUEUE
 */
 /*
 #ifndef NO_SEMPHORE
	init_MUTEX(&shost->scan_mutex);
 #endif // ndef NO_SEMPHORE
 */
 shost->host_no = scsi_host_next_hn++; /* XXX(hch): still racy */
 USBMS_DBG_PLUG(iprintf(
	"[USBMS PLUG] scsi_host_alloc() : shost->host_no=%d \n",
	shost->host_no
	));
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] scsi_host_alloc() : shost->host_no=%d \n",
	shost->host_no
	));

 shost->dma_channel = 0xff;

 /* These three are default values which can be overridden */
 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_host_alloc() : Set default values .. \n"));
 shost->max_channel = 0;
 /*
 shost->max_id = 8;
 shost->max_lun = 8;
 */
 shost->max_id = USBMS_SCSI_MAX_ID ;
 shost->max_lun = USBMS_SCSI_MAX_LUM  ;

 /*
  * All drivers right now should be able to handle 12 byte
  * commands.  Every so often there are requests for 16 byte
  * commands, but individual low-level drivers need to certify that
  * they actually do something sensible with such commands.
  */

 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_host_alloc() : set 12 byte commands .. \n"));

 shost->max_cmd_len = 12;
 shost->hostt = sht;
	shost->this_id = sht->this_id;
	shost->can_queue = sht->can_queue;
	shost->sg_tablesize = sht->sg_tablesize;
	shost->cmd_per_lun = sht->cmd_per_lun;
	shost->unchecked_isa_dma = sht->unchecked_isa_dma;
	shost->use_clustering = sht->use_clustering;

	if (sht->max_host_blocked)
		shost->max_host_blocked = sht->max_host_blocked;
	else
		shost->max_host_blocked = SCSI_DEFAULT_HOST_BLOCKED;

 /*
  * If the driver imposes no hard sector transfer limit, start at
  * machine infinity initially.
  */

 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_host_alloc() : set hard sector transfer limit .. \n"));

 if (sht->max_sectors)
		shost->max_sectors = sht->max_sectors;
	else
		shost->max_sectors = SCSI_DEFAULT_MAX_SECTORS;

	/*
	 * assume a 4GB boundary, if not set
	 */
	if (sht->dma_boundary)
		shost->dma_boundary = sht->dma_boundary;
	else
		shost->dma_boundary = 0xffffffff;

 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_host_alloc() : scsi_setup_command_freelist() .. \n"));

 rval = scsi_setup_command_freelist(shost);
 if (rval)
		goto fail_kfree;

 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_host_alloc() : device_initialize() .. \n"));
 device_initialize(&shost->shost_gendev);
 snprintf(shost->shost_gendev.bus_id, BUS_ID_SIZE, "host%d",
		shost->host_no);
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_host_alloc() : &shost->shost_gendev=%08XH \n",
	&shost->shost_gendev));
 shost->shost_gendev.release = scsi_host_dev_release;

 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_host_alloc() : class_device_initialize() .. \n"));
 class_device_initialize(&shost->shost_classdev);
	shost->shost_classdev.dev = &shost->shost_gendev;
	shost->shost_classdev.class = &shost_class;
	snprintf(shost->shost_classdev.class_id, BUS_ID_SIZE, "host%d",
		  shost->host_no);

 /*
 #ifndef _ETCPIP
    // struct completion in <linux/completion.h>
	shost->eh_notify = &complete;
 #endif // ndef _ETCPIP
 */
 /*
 rval = kernel_thread(scsi_error_handler, shost, 0);
 if (rval < 0)
		goto fail_destroy_freelist;
 */

 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_host_alloc() : runtsk scsi_ErHandler() .. \n"));
 #ifdef INT_SCHEDULING
	status1 = runtsk_parm(100, (void (fAR *)(void))scsi_ErHandler, 6000,(unsigned long) shost );
 #else
	status1 = runtsk_parm(80, (void (fAR *)(void))scsi_ErHandler, 6000,(unsigned long) shost );
 #endif // INT_SCHEDULING

 if(status1 < SUCCESS )
   {
    dprintf("RUNTASK scsi_ErHandler fails\n");
	goto fail_destroy_freelist;
   }
  else
   {
    dprintf("RUNTASK id=%d scsi_ErHandler ...\n", status1);
   }

 /*
  * Now wait for the kernel error thread to initialize itself
  * as it might be needed when we scan the bus.
  */
 /*
 #ifndef _ETCPIP
    // wait_for_completion() is in <linux/completion.h>
	wait_for_completion(&complete);
 #endif // ndef _ETCPIP
 */
 /*
 #ifndef _ETCPIP
    // struct completion in <linux/completion.h>
	shost->eh_notify = NULL;
 #endif // ndef _ETCPIP
 */
 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_host_alloc() : scsi_proc_hostdir_add() .. \n"));
 scsi_proc_hostdir_add(shost->hostt);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_host_alloc() End. \n"));

 return shost;

 // ---------------------
 fail_destroy_freelist:
 // ---------------------

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_host_alloc() : scsi_destroy_command_freelist() .. \n"));
 scsi_destroy_command_freelist(shost);

 // ---------------------
 fail_kfree:
 // ---------------------

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_host_alloc() : UsbFreeMemC(shost) , size=%d .. \n",
	sizeof(*shost)
	));
 /*   
 kfree (shost);
 */
 UsbFreeMemC(USB_MEM_C_M,(unsigned char *) shost
			#if defined(USB_DBG_MEM_C_ON)
				,sizeof(*shost) // size : 444
			#endif
			);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_host_alloc() End with Fail !! \n"));
 return NULL;
}
/*

//#ifndef _ETCPIP
//	struct Scsi_Host *scsi_register(struct scsi_host_template *sht, int privsize)
// #else

	struct Scsi_Host *scsi_register(struct scsi_host_template *sht, int privsize ,
		struct list_head *shost_my_devices )

//#endif

{
 
 //#ifndef _ETCPIP
 //	struct Scsi_Host *shost = scsi_host_alloc(sht, privsize);
 // #else
 
	struct Scsi_Host *shost = scsi_host_alloc(sht, privsize , shost_my_devices ) ;
 
 // #endif
 

 if(!sht->detect) {
	#ifdef USB_DBGSTRING_ON
		
		// printk(KERN_WARNING "scsi_register() called on new-style "
		//		    "template for driver %s\n", sht->name);
		
		dprintf("[USBMS] scsi_register() called on new-style "
				    "template for driver %s\n", sht->name);
	 #else
		dprintf("[USBMS] scsi_register() called on new-style "
				    "template for driver .. \n");
	#endif // ndef USB_DBGSTRING_ON
		
		// dump_stack();
		
	}

 if(shost) {
	USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_register() : sht->legacy_hosts=%08XH \n",
			sht->legacy_hosts ));
	#ifdef USBMS_BS_ASSERT
		if ( ! (IS_VALID_ADDR(&sht->legacy_hosts)) ) {
			iprintf("[ASSERT][USBMS] scsi_register() : " ) ;
   		    iprintf("Error : Invlid address &sht->legacy_hosts=%08XH \n",
					&sht->legacy_hosts) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
   		   }
	#endif // USBMS_BS_ASSERT

	list_add_tail(&shost->sht_legacy_list, &sht->legacy_hosts);
   }
 return shost;
}

void scsi_unregister(struct Scsi_Host *shost)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_unregister() Begin .. \n"));
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_unregister() : list_del() .. \n"));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_unregister() : list_del() .. \n"));
 list_del(&shost->sht_legacy_list);
 scsi_host_put(shost);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_unregister() End. \n"));
}
*/

/**
 * scsi_host_lookup - get a reference to a Scsi_Host by host no
 *
 * @hostnum:	host number to locate
 *
 * Return value:
 *	A pointer to located Scsi_Host or NULL.
 **/
/* 2004/3/5 B. S. Liaw Comment out ..
struct Scsi_Host *scsi_host_lookup(unsigned short hostnum)
{
 struct class *class = class_get(&shost_class);
 struct class_device *cdev;
 #ifndef _ETCPIP
	struct Scsi_Host *shost = ERR_PTR(-ENXIO), *p;
  #else
	struct Scsi_Host *shost = (struct Scsi_Host *) 0x00L , *p;
 #endif


 if (class) {
	#ifndef _ETCPIP
		// down_read is on include/linux/rwsem.h
		down_read(&class->subsys.rwsem);
	#endif // ndef _ETCPIP
		list_for_each_entry(cdev, &class->children, node) {
			p = class_to_shost(cdev);
			if (p->host_no == hostnum) {
				shost = scsi_host_get(p);
				break;
}
	}
 #ifndef _ETCPIP
	// down_read is on include/linux/rwsem.h
		up_read(&class->subsys.rwsem);
 #endif // ndef _ETCPIP
	}

	class_put(&shost_class);
	return shost;
}
*/
/**
 * *scsi_host_get - inc a Scsi_Host ref count
 * @shost:	Pointer to Scsi_Host to inc.
 **/
struct Scsi_Host *scsi_host_get(struct Scsi_Host *shost)
{
	if (test_bit(SHOST_DEL, &shost->shost_state) ||
		!get_device(&shost->shost_gendev))
		return NULL;
	return shost;
}

/**
 * *scsi_host_put - dec a Scsi_Host ref count
 * @shost:	Pointer to Scsi_Host to dec.
 **/
void scsi_host_put(struct Scsi_Host *shost)
{
 
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_host_put() Begin .. \n"));
 // #ifndef _ETCPIP
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_host_put() : put_device() .. \n"));
 put_device(&shost->shost_gendev);
 // #endif // ndef _ETCPIP
}

int scsi_init_hosts(void)
{
	return class_register(&shost_class);
}

void scsi_exit_hosts(void)
{
	class_unregister(&shost_class);
}
