/*
 *  scsi_lib.c Copyright (C) 1999 Eric Youngdale
 *
 *  SCSI queueing library.
 *      Initial versions: Eric Youngdale (eric@andante.org).
 *                        Based upon conversations with large numbers
 *                        of people at Linux Expo.
 */

// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP

/*
#include <linux/bio.h>
*/
#include "bio.h"
/*
#include <linux/blkdev.h>
*/
#include "bitops.h" // B. S. Liaw add for test_bit()
#include "blkdev.h"
/*
	#include <linux/completion.h>
*/
/*
#include <linux/kernel.h>
*/
#include "PoKernel.h"
/*
#include <linux/mempool.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/pci.h>
*/
#include "PoDevice.h" // include/linux/device.h
#include "kdev_t.h"   // also need
#ifdef CONFIG_PCI
	#include "UsbPoPci.h" // usb_osport_pci.h
#endif	// CONFIG_PCI

#include "scsi_ds.h"
#include "scsi_hosts.h"

/*
#include <scsi/scsi_driver.h>
*/
#include "scsi_driver.h"
#include "scsi_priv.h"
#include "scsi_logging.h"

/*
#define SG_MEMPOOL_NR		5
#define SG_MEMPOOL_SIZE		32

struct scsi_host_sg_pool {
 size_t		size;
 #ifdef USB_DBGSTRING_ON
	char		*name; 
 #endif // USB_DBGSTRING_ON
 // kmem_cache_t is on include/linux/slab.h
 // kmem_cache_t	*slab;
 
 // mempool_t is on include/linux/mempool.h
 // mempool_t	*pool;
 
 void *pool_data;
};


#define SP(x) { x, "sgpool-" #x } 
struct scsi_host_sg_pool scsi_sg_pools[SG_MEMPOOL_NR] = { 
	SP(8), SP(16), SP(32), SP(64), SP(MAX_PHYS_SEGMENTS)
}; 	
#undef SP
*/

/*
#ifdef BLK_NO_QUEUE 
	// Transfer a buffer directly, without going through the request queue.
	static int sbull_make_request(request_queue_t *q, struct bio *bio) ;
#endif // def BLK_NO_QUEUE 
*/

/*
 * Function:    scsi_insert_special_req()
 *
 * Purpose:     Insert pre-formed request into request queue.
 *
 * Arguments:   sreq	- request that is ready to be queued.
 *              at_head	- boolean.  True if we should insert at head
 *                        of queue, false if we should insert at tail.
 *
 * Lock status: Assumed that lock is not held upon entry.
 *
 * Returns:     Nothing
 *
 * Notes:       This function is called from character device and from
 *              ioctl types of functions where the caller knows exactly
 *              what SCSI command needs to be issued.   The idea is that
 *              we merely inject the command into the queue (at the head
 *              for now), and then call the queue request function to actually
 *              process it.
 */
int scsi_insert_special_req(struct scsi_request *sreq, int at_head)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_insert_special_req() Begin .. \n"));
	/*
	 * Because users of this function are apt to reuse requests with no
	 * modification, we have to sanitise the request flags here
	 */
 sreq->sr_request->flags &= ~REQ_DONTPREP;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_insert_special_req() : blk_insert_request() .. \n"));
 if ( ! blk_insert_request(sreq->sr_device->request_queue, sreq->sr_request,
		       	   at_head, sreq, 0) ) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_insert_special_req() : End with Failed !! \n"));
	return 0 ;
   }
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_insert_special_req() End. \n"));
 return 1 ;
}

/*
 * Function:    scsi_queue_insert()
 *
 * Purpose:     Insert a command in the midlevel queue.
 *
 * Arguments:   cmd    - command that we are adding to queue.
 *              reason - why we are inserting command to queue.
 *
 * Lock status: Assumed that lock is not held upon entry.
 *
 * Returns:     Nothing.
 *
 * Notes:       We do this for one of two cases.  Either the host is busy
 *              and it cannot accept any more commands for the time being,
 *              or the device returned QUEUE_FULL and can accept no more
 *              commands.
 * Notes:       This could be called either from an interrupt context or a
 *              normal process context.
 */
int scsi_queue_insert(struct scsi_cmnd *cmd, int reason)
{
 struct Scsi_Host *host = cmd->device->host;
 struct scsi_device *device = cmd->device;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_queue_insert() Begin .. \n"));
 SCSI_LOG_MLQUEUE(1,
		 printk("Inserting command %p into mlqueue\n", cmd));

	/*
	 * We are inserting the command into the ml queue.  First, we
	 * cancel the timer, so it doesn't time out.
	 */
 /*
 #ifndef NO_TIMER
	scsi_delete_timer(cmd);
 #endif // ndef NO_TIMER
 */
	/*
	 * Next, set the appropriate busy bit for the device/host.
	 *
	 * If the host/device isn't busy, assume that something actually
	 * completed, and that we should be able to queue a command now.
	 *
	 * Note that the prior mid-layer assumption that any host could
	 * always queue at least one command is now broken.  The mid-layer
	 * will implement a user specifiable stall (see
	 * scsi_host.max_host_blocked and scsi_device.max_device_blocked)
	 * if a command is requeued with no other commands outstanding
	 * either for the device or for the host.
	 */
	if (reason == SCSI_MLQUEUE_HOST_BUSY)
		host->host_blocked = host->max_host_blocked;
	else if (reason == SCSI_MLQUEUE_DEVICE_BUSY)
		device->device_blocked = device->max_device_blocked;

	/*
	 * Register the fact that we own the thing for now.
	 */
	cmd->state = SCSI_STATE_MLQUEUE;
	cmd->owner = SCSI_OWNER_MIDLEVEL;

	/*
	 * Decrement the counters, since these commands are no longer
	 * active on the host/device.
	 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_queue_insert() : scsi_device_unbusy() .. \n")) ;
 scsi_device_unbusy(device);

	/*
	 * Insert this command at the head of the queue for it's device.
	 * It will go before all other commands that are already in the queue.
	 *
	 * NOTE: there is magic here about the way the queue is plugged if
	 * we have no outstanding commands.
	 * 
	 * Although this *doesn't* plug the queue, it does call the request
	 * function.  The SCSI request function detects the blocked condition
	 * and plugs the queue appropriately.
	 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_queue_insert() : blk_insert_request() .. \n"));
 if(! blk_insert_request(device->request_queue, cmd->request, 1, cmd, 1) ) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_queue_insert() : End with Failed !! \n"));
	return 0 ;
   }
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_queue_insert() End. \n"));
 return 1 ;
}

/*
 * Function:    scsi_do_req
 *
 * Purpose:     Queue a SCSI request
 *
 * Arguments:   sreq	  - command descriptor.
 *              cmnd      - actual SCSI command to be performed.
 *              buffer    - data buffer.
 *              bufflen   - size of data buffer.
 *              done      - completion function to be run.
 *              timeout   - how long to let it run before timeout.
 *              retries   - number of retries we allow.
 *
 * Lock status: No locks held upon entry.
 *
 * Returns:     Nothing.
 *
 * Notes:	This function is only used for queueing requests for things
 *		like ioctls and character device requests - this is because
 *		we essentially just inject a request into the queue for the
 *		device.
 */
void scsi_do_req(struct scsi_request *sreq, const void *cmnd,
		 void *buffer, unsigned bufflen,
		 void (*done)(struct scsi_cmnd *),
		 int timeout, int retries)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_do_req() Begin .. \n"));
 /*
  * If the upper level driver is reusing these things, then
  * we should release the low-level block now.  Another one will
  * be allocated later when this request is getting queued.
  */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_do_req() : __scsi_release_request() .. \n"));
 __scsi_release_request(sreq);

	/*
	 * Our own function scsi_done (which marks the host as not busy,
	 * disables the timeout counter, etc) will be called by us or by the
	 * scsi_hosts[host].queuecommand() function needs to also call
	 * the completion function for the high level driver.
	 */
 memcpy(sreq->sr_cmnd, cmnd, sizeof(sreq->sr_cmnd));
 sreq->sr_bufflen = bufflen;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_do_req() : Set Value : sreq->sr_bufflen=%d \n",
	sreq->sr_bufflen
	));

 // B. S. Liaw : request_buffer need uncached
 sreq->sr_buffer = buffer;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_do_req() : Set Value : sreq->sr_buffer=%08XH .. \n",
	sreq->sr_buffer
	));

 #ifdef USBMS_BS_ASSERT
	if (sreq->sr_bufflen) {
		if ( ! (IS_VALID_ADDR(sreq->sr_buffer)) ) {
			iprintf("[ASSERT][USBMS] scsi_do_req(1) : " ) ;
   		    iprintf("Error : Invalid Address !!\n" ) ;
	        iprintf("[ASSERT][USBMS] sreq->sr_buffer=%08XH \n",
					sreq->sr_buffer ) ; ;
	        iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
		if ( !(USB_IS_UN_CACHED(sreq->sr_buffer)) ) {
			iprintf("[ASSERT][USBMS] scsi_do_req(2) : " ) ;
		    iprintf("Error : Not Un-Cached !!\n" ) ;
		    iprintf("[ASSERT][USBMS] sreq->sr_buffer=%08XH \n", sreq->sr_buffer ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
	   } // if has sreq->sr_bufflen
 #endif // USBMS_BS_ASSERT

 sreq->sr_allowed = retries;
 sreq->sr_done = done;
 sreq->sr_timeout_per_command = timeout;

 if (sreq->sr_cmd_len == 0)
		sreq->sr_cmd_len = COMMAND_SIZE(sreq->sr_cmnd[0]);

 /*
  * At this point, we merely set up the command, stick it in the normal
  * request queue, and return.  Eventually that request will come to the
  * top of the list, and will be dispatched.
  */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_do_req() : scsi_insert_special_req() .. \n"));
 scsi_insert_special_req(sreq, 0);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_do_req() End. \n"));
}
 
static void scsi_wait_done(struct scsi_cmnd *cmd)
{
 struct request *req = cmd->request;
 struct request_queue *q = cmd->device->request_queue;
 unsigned long flags;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_wait_done() Begin .. \n"));

 // #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_wait_done() : q->req = cmd->request .. \n"));
		q->req = cmd->request ;
 // #endif // NO_ELEVATOR

 req->rq_status = RQ_SCSI_DONE;	/* Busy, but indicate request done */
 /*
 #ifndef NO_SPINLOCK
			spin_lock_irqsave(q->queue_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 if (blk_rq_tagged(req))
		blk_queue_end_tag(q, req);
 /*
 #ifndef NO_SPINLOCK
			spin_unlock_irqrestore(q->queue_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 /*
 #ifndef _ETCPIP
	if (req->waiting)
		complete(req->waiting);
  #else
 */

 /*
 // 2004/1/16  B. S. Liaw Comment Out !!
 // 2003/12/02 B. S. Liaw :
 // req->waiting replaced by MT_EVT_USBMS_RQ/MT_EVT_USBMS_RQBI on kernel/local.h
 if(req->waiting_evt) {
	
	#ifdef USBMS_USB_WAIT_BY_EVENT     // 2004/1/2 B. S. Liaw
		USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_wait_done() : SIGNAL(%d) .. \n",
			req->waiting_evt)) ;
		// kernel/mtmacro.h : #define SIGNAL(signo) setevt(signo)
		if (SIGNAL(req->waiting_evt) != SUCCESS) {
			USBMS_DBG_INFO(iprintf(
				"[USBMS INFO] scsi_wait_done() : Error : SIGNAL req->waiting_evt=%d fail !! \n",
				req->waiting_evt
				)) ;
			dprintf("[USBMS] Error : SIGNAL %d fail !! \n",req->waiting_evt);
		   }
	#else
		switch(req->waiting_evt) {
			case MT_EVT_USBMS_RQ :
				USBMS_DBG_INIT(iprintf(
					"[USBMS INIT] scsi_wait_done() : Set Value : req->evt_blk_rq=1 .. \n"
					)) ;
				// BLOCK_PREEMPTION;
				req->evt_blk_rq = 1 ;
				// UNBLOCK_PREEMPTION;
				break ;
				#ifndef USBMS_NO_IOCTL
					case MT_EVT_USBMS_RQBI :
						USBMS_DBG_INIT(iprintf(
							"[USBMS INIT] scsi_wait_done() : Set Value : req->evt_blk_rq_bi=1 .. \n"
							)) ;
						// BLOCK_PREEMPTION;
						req->evt_blk_rq_bi = 1 ;
						// UNBLOCK_PREEMPTION;
						break ;
				#endif // ndef USBMS_NO_IOCTL
			   }
	#endif // ndef USBMS_USB_WAIT_BY_EVENT
	
   }
 */

 /*
 #endif
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_wait_done() End. \n"));
}

void scsi_wait_req(struct scsi_request *sreq, const void *cmnd, void *buffer,
		   unsigned bufflen, int timeout, int retries)
{
 /*
 #ifndef _ETCPIP
    // struct completion in <linux/completion.h>
	DECLARE_COMPLETION(wait);
 #endif // ndef _ETCPIP
 */

	
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_wait_req() Begin .. \n"));

 // #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_wait_req() : q->req = sreq->sr_request .. \n"));
	sreq->sr_device->request_queue->req = sreq->sr_request ;
 // #endif // NO_ELEVATOR
 /*
 #ifndef _ETCPIP
	sreq->sr_request->waiting = &wait;
  #else
 */
#ifndef USBMS_NO_USB_TASK  /* Charles 20040520 */
 #ifdef USBMS_USB_WAIT_BY_EVENT     // 2004/1/2 B. S. Liaw
	// 2003/12/02 B. S. Liaw :
	// req->waiting replaced by MT_EVT_USBMS_RQ/MT_EVT_USBMS_RQBI on kernel/local.h
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] scsi_wait_req() : clrevt MT_EVT_USBMS_RQ=%d ..\n",
		MT_EVT_USBMS_RQ
		)) ;
    if (clrevt(MT_EVT_USBMS_RQ) != SUCCESS) {
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] scsi_wait_req() : Error : clrevt fail: Invalid MT_EVT_USBMS_RQ=%d\n",
			MT_EVT_USBMS_RQ
			)) ;
        dprintf("[USBMS] Error : clrevt fail: Invalid MT_EVT_USBMS_RQ %d\n",
				MT_EVT_USBMS_RQ);
	   }
  #else
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] scsi_wait_req() : Set Value : sreq->sr_request->evt_blk_rq=0 .. \n"
		)) ;
	// BLOCK_PREEMPTION;
	sreq->sr_request->evt_blk_rq = 0 ;
	// UNBLOCK_PREEMPTION;
 #endif // ndef USBMS_USB_WAIT_BY_EVENT
#endif
 sreq->sr_request->waiting_evt = MT_EVT_USBMS_RQ ;
 /*
 #endif
 */

 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_wait_req() : sreq=%08XH \n",
	sreq));
 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_wait_req() : sreq->sr_request=%08XH \n",
	sreq->sr_request));

 sreq->sr_request->rq_status = RQ_SCSI_BUSY;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_wait_req() : scsi_do_req() .. \n"));
 scsi_do_req(sreq, cmnd, buffer, bufflen, scsi_wait_done,
			timeout, retries);
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_wait_req() : generic_unplug_device() .. \n"));
	generic_unplug_device(sreq->sr_device->request_queue);
 #endif // ndef NO_ELEVATOR
 */
 /*
 #ifndef _ETCPIP
    // wait_for_completion() is in <linux/completion.h>
	wait_for_completion(&wait);
  #else
 */
 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_wait_req() *** task sleeping \n")) ;
 USBMS_DBG_SCSI(iprintf("[USBMS SCSI] scsi_wait_req() *** task sleeping \n")) ;
#ifndef USBMS_NO_USB_TASK  /* Charles 20040520 */
 #ifdef USBMS_USB_WAIT_BY_EVENT     // 2004/1/2 B. S. Liaw
	// 2003/12/02 B. S. Liaw :
	// req->waiting replaced by MT_EVT_USBMS_RQ/MT_EVT_USBMS_RQBI on kernel/local.h
    (void)wteset_dec(MT_EVT_USBMS_RQ, 0); // 0 : no timeout

    if (clrevt(MT_EVT_USBMS_RQ) != SUCCESS) {
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] scsi_wait_req() : Error : clrevt fail: Invalid MT_EVT_USBMS_RQ=%d\n",
			MT_EVT_USBMS_RQ
			)) ;
        dprintf("[USBMS] Error : clrevt fail: Invalid MT_EVT_USBMS_RQ %d\n",
				MT_EVT_USBMS_RQ);
	   }
  #else
	for(;;) {
		if (sreq->sr_request->evt_blk_rq) {
			break ;
		   }
	   }
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] scsi_wait_req() : Set Value : sreq->sr_request->evt_blk_rq=0 .. \n"
		)) ;
	// BLOCK_PREEMPTION;
	sreq->sr_request->evt_blk_rq = 0 ;
	// UNBLOCK_PREEMPTION;
 #endif // ndef USBMS_USB_WAIT_BY_EVENT
#endif
 USBMS_DBG_SCSI(iprintf("[USBMS SCSI] scsi_wait_req() *** task awakened \n")) ;
 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_wait_req() *** task awakened \n")) ;
 /*
 #endif
 */

 /*
 #ifndef _ETCPIP
    // wait_for_completion() is in <linux/completion.h>
	sreq->sr_request->waiting = NULL;
  #else
 */
	// 2003/12/02 B. S. Liaw :
	// req->waiting replaced by MT_EVT_USBMS_RQ/MT_EVT_USBMS_RQBI on kernel/local.h
	sreq->sr_request->waiting_evt = (uint)NULL_VALUE ;
 /*
 #endif
 */

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_wait_req() : __scsi_release_request() .. \n"));
 __scsi_release_request(sreq);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_wait_req() End. \n"));
}

/*
 * Function:    scsi_init_cmd_errh()
 *
 * Purpose:     Initialize cmd fields related to error handling.
 *
 * Arguments:   cmd	- command that is ready to be queued.
 *
 * Returns:     Nothing
 *
 * Notes:       This function has the job of initializing a number of
 *              fields related to error handling.   Typically this will
 *              be called once for each command, as required.
 */
static int scsi_init_cmd_errh(struct scsi_cmnd *cmd)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_init_cmd_errh() Begin .. \n"));

 cmd->owner = SCSI_OWNER_MIDLEVEL;
 cmd->serial_number = 0;
 cmd->serial_number_at_timeout = 0;
 cmd->abort_reason = 0;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_init_cmd_errh() : memset() .. \n"));
 /*
 memset(cmd->sense_buffer, 0, sizeof cmd->sense_buffer);
 */
 memset(cmd->sense_buffer, 0, SCSI_SENSE_BUFFERSIZE);

 if(cmd->cmd_len == 0)
	cmd->cmd_len = COMMAND_SIZE(cmd->cmnd[0]);

 /*
  * We need saved copies of a number of fields - this is because
  * error handling may need to overwrite these with different values
  * to run different commands, and once error handling is complete,
  * we will need to restore these values prior to running the actual
  * command.
  */
 #ifndef NO_SCATTER_LIST
	cmd->old_use_sg = cmd->use_sg;
 #endif // ndef NO_SCATTER_LIST
 cmd->old_cmd_len = cmd->cmd_len;
 cmd->sc_old_data_direction = cmd->sc_data_direction;
 cmd->old_underflow = cmd->underflow;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_init_cmd_errh() : memcpy() .. \n"));
 memcpy(cmd->data_cmnd, cmd->cmnd, sizeof(cmd->cmnd));
 cmd->buffer = cmd->request_buffer;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_init_cmd_errh() : Set Value : cmd->buffer=%08XH \n",
	cmd->buffer
	));
 cmd->bufflen = cmd->request_bufflen;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_init_cmd_errh() : Set Value : cmd->bufflen=%d \n",
	cmd->bufflen
	));
 cmd->internal_timeout = NORMAL_TIMEOUT;
 cmd->abort_reason = 0;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_init_cmd_errh() End. \n"));
 return 1;
}

/*
 * Function:   scsi_setup_cmd_retry()
 *
 * Purpose:    Restore the command state for a retry
 *
 * Arguments:  cmd	- command to be restored
 *
 * Returns:    Nothing
 *
 * Notes:      Immediately prior to retrying a command, we need
 *             to restore certain fields that we saved above.
 */
void scsi_setup_cmd_retry(struct scsi_cmnd *cmd)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_setup_cmd_retry() Begin .. \n"));
 memcpy(cmd->cmnd, cmd->data_cmnd, sizeof(cmd->data_cmnd));
 cmd->request_buffer = cmd->buffer;
 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_setup_cmd_retry() : Set Value : cmd->request_buffer=%08XH .. \n",
		cmd->request_buffer
		));
 cmd->request_bufflen = cmd->bufflen;
 #ifndef NO_SCATTER_LIST
	cmd->use_sg = cmd->old_use_sg;
 #endif // ndef NO_SCATTER_LIST
 cmd->cmd_len = cmd->old_cmd_len;
 cmd->sc_data_direction = cmd->sc_old_data_direction;
 cmd->underflow = cmd->old_underflow;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_setup_cmd_retry() End. \n"));
}

void scsi_device_unbusy(struct scsi_device *sdev)
{
 struct Scsi_Host *shost = sdev->host;
 unsigned long flags;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_device_unbusy() Begin .. \n")) ;
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irqsave(shost->host_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 shost->host_busy--;
 /*
 if (unlikely(test_bit(SHOST_RECOVERY, &shost->shost_state) &&
		     shost->host_failed))
 */
 if(test_bit(SHOST_RECOVERY, &shost->shost_state) &&
		     shost->host_failed)
   {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_device_unbusy() : scsi_eh_wakeup() .. \n")) ;
	scsi_eh_wakeup(shost);
   }
 /*
 #ifndef NO_SPINLOCK
	spin_unlock(shost->host_lock);
 #endif // ndef NO_SPINLOCK
 */
 /*
 #ifndef NO_SPINLOCK
	spin_lock(&sdev->sdev_lock);
 #endif // ndef NO_SPINLOCK
 */
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	sdev->device_busy--;
 #endif // ndef NO_ELEVATOR
 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(&sdev->sdev_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_device_unbusy() End. \n")) ;
}

/*
 * Called for single_lun devices on IO completion. Clear starget_sdev_user,
 * and call blk_run_queue for all the scsi_devices on the target -
 * including current_sdev first.
 *
 * Called with *no* scsi locks held.
 */
static void scsi_single_lun_run(struct scsi_device *current_sdev)
{
 struct scsi_device *sdev;
 unsigned long flags;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_single_lun_run() Begin .. \n"));
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irqsave(current_sdev->host->host_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 /*
 WARN_ON(!current_sdev->sdev_target->starget_sdev_user);
 */
 if (!((!current_sdev->sdev_target->starget_sdev_user)!=0)) {
	dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
   }

 current_sdev->sdev_target->starget_sdev_user = NULL;
 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(current_sdev->host->host_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 /*
	 * Call blk_run_queue for all LUNs on the target, starting with
	 * current_sdev. We race with others (to set starget_sdev_user),
	 * but in most cases, we will be first. Ideally, each LU on the
	 * target would get some limited time or requests on the target.
		 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_single_lun_run() : blk_run_queue() .. \n"));
 blk_run_queue(current_sdev->request_queue);
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irqsave(current_sdev->host->host_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 if (current_sdev->sdev_target->starget_sdev_user) {
		/*
		 * After unlock, this races with anyone clearing
		 * starget_sdev_user, but we (should) always enter this
		 * function again, avoiding any problems.
		 */
	/*
	#ifndef NO_SPINLOCK
		spin_unlock_irqrestore(current_sdev->host->host_lock, flags);
	#endif // ndef NO_SPINLOCK
	*/
	return;
   }
 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(current_sdev->host->host_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 list_for_each_entry(sdev, &current_sdev->same_target_siblings,
   			    same_target_siblings)
   {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_single_lun_run() : blk_run_queue() .. \n"));
	blk_run_queue(sdev->request_queue);
   }
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_single_lun_run() End. \n"));
}

/*
 * Function:	scsi_run_queue()
 *
 * Purpose:	Select a proper request queue to serve next
 *
 * Arguments:	q	- last request's queue
 *
 * Returns:     Nothing
 *
 * Notes:	The previous command was completely finished, start
 *		a new one if possible.
 */
static void scsi_run_queue(struct request_queue *q)
{
 struct scsi_device *sdev = q->queuedata;
 struct Scsi_Host *shost = sdev->host;
 unsigned long flags;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_run_queue() Begin .. \n"));
 if (sdev->single_lun)
   {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_run_queue() : scsi_single_lun_run() .. \n"));
	scsi_single_lun_run(sdev);
   }
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irqsave(shost->host_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 while (!list_empty(&shost->starved_list) &&
	       !shost->host_blocked && !shost->host_self_blocked &&
		!((shost->can_queue > 0) &&
		  (shost->host_busy >= shost->can_queue))) {
		/*
		 * As long as shost is accepting commands and we have
		 * starved queues, call blk_run_queue. scsi_request_fn
		 * drops the queue_lock and can add us back to the
		 * starved_list.
		 *
		 * host_lock protects the starved_list and starved_entry.
		 * scsi_request_fn must get the host_lock before checking
		 * or modifying starved_list or starved_entry.
		 */
	sdev = list_entry(shost->starved_list.next,
					  struct scsi_device, starved_entry);
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_run_queue() : list_del_init() .. \n"));
	list_del_init(&sdev->starved_entry);
	/*
	#ifndef NO_SPINLOCK
		spin_unlock_irqrestore(shost->host_lock, flags);
	#endif // ndef NO_SPINLOCK
	*/

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_run_queue() : blk_run_queue() .. \n"));
	blk_run_queue(sdev->request_queue);
	/*
	#ifndef NO_SPINLOCK
		spin_lock_irqsave(shost->host_lock, flags);
	#endif // ndef NO_SPINLOCK
	*/
	/*
	if (unlikely(!list_empty(&sdev->starved_entry)))
	*/
	if (!list_empty(&sdev->starved_entry))
	   {
		/*
		 * sdev lost a race, and was put back on the
		 * starved list. This is unlikely but without this
		 * in theory we could loop forever.
		 */
		break;
	   }
   }
 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(shost->host_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_run_queue() : blk_run_queue() .. \n"));
 blk_run_queue(q);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_run_queue() End. \n"));
}

/*
 * Function:	scsi_requeue_command()
 *
 * Purpose:	Handle post-processing of completed commands.
 *
 * Arguments:	q	- queue to operate on
 *		cmd	- command that may need to be requeued.
 *
 * Returns:	Nothing
 *
 * Notes:	After command completion, there may be blocks left
 *		over which weren't finished by the previous command
 *		this can be for a number of reasons - the main one is
 *		I/O errors in the middle of the request, in which case
 *		we need to request the blocks that come after the bad
 *		sector.
 */
/*
static void scsi_requeue_command(struct request_queue *q, struct scsi_cmnd *cmd)
*/

/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
static int scsi_requeue_command(struct request_queue *q, struct scsi_cmnd *cmd)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_requeue_command() : Begin .. \n"));

 // #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_requeue_command() : q->req = cmd->request .. \n"));
	q->req = cmd->request ;
 // #endif // NO_ELEVATOR

 cmd->request->flags &= ~REQ_DONTPREP;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_requeue_command() : blk_insert_request() .. \n"));
 if(! blk_insert_request(q, cmd->request, 1, cmd, 1) ) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_requeue_command() : End with Failed !! \n"));
	return 0 ;
   }

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_requeue_command() : scsi_run_queue() .. \n"));
 scsi_run_queue(q);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_requeue_command() : End. \n"));
 return 1 ;
}
#endif // ndef NO_ELEVATOR
*/
void scsi_next_command(struct scsi_cmnd *cmd)
{
 struct request_queue *q = cmd->device->request_queue;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_next_command() Begin .. \n"));
 // #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_next_command() : q->req = cmd->request .. \n"));
		q->req = cmd->request ;
 // #endif // NO_ELEVATOR
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_next_command() : scsi_put_command() .. \n"));
 scsi_put_command(cmd);
 /*
 #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
 */
	USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_next_command() : Not Call scsi_run_queue() !! \n"));
 /*
  #else
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_next_command() : scsi_run_queue() .. \n"));
	scsi_run_queue(q);
 #endif
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_next_command() End. \n"));
}

void scsi_run_host_queues(struct Scsi_Host *shost)
{
 struct scsi_device *sdev;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_run_host_queues() Begin .. \n"));
 shost_for_each_device(sdev, shost)
   {
	// #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	if (!sdev->request_queue->req)
	   {
		USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_run_host_queues() : Null sdev->request_queue->req !! \n"));
	   }
	  else
	// #endif // NO_ELEVATOR
	   {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_run_host_queues() : scsi_run_queue() .. \n"));
		scsi_run_queue(sdev->request_queue);
	   }
   }
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_run_host_queues() End. \n"));
}

/*
 * Function:    scsi_end_request()
 *
 * Purpose:     Post-processing of completed commands called from interrupt
 *              handler or a bottom-half handler.
 *
 * Arguments:   cmd	 - command that is complete.
 *              uptodate - 1 if I/O indicates success, 0 for I/O error.
 *              sectors  - number of sectors we want to mark.
 *		requeue  - indicates whether we should requeue leftovers.
 *		frequeue - indicates that if we release the command block
 *			   that the queue request function should be called.
 *
 * Lock status: Assumed that lock is not held upon entry.
 *
 * Returns:     Nothing
 *
 * Notes:       This is called for block device requests in order to
 *              mark some number of sectors as complete.
 * 
 *		We are guaranteeing that the request queue will be goosed
 *		at some point during this call.
 */
static struct scsi_cmnd *scsi_end_request(struct scsi_cmnd *cmd, int uptodate,
					  int sectors, int requeue)
{
 request_queue_t *q = cmd->device->request_queue;
 struct request *req = cmd->request;
 unsigned long flags;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_end_request() Begin .. \n"));

 // #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_end_request() : q->req = cmd->request .. \n"));
		q->req = cmd->request ;
 // #endif // NO_ELEVATOR
 /*
  * If there are blocks left over at the end, set up the command
  * to queue the remainder of them.
  */
 if (end_that_request_first(req, uptodate, sectors)) {
	int leftover = req->hard_nr_sectors - sectors;

	/* kill remainder if no retrys */
	if (!uptodate && blk_noretry_request(req)) {
			end_that_request_first(req, 0, leftover);
	   }
	  else {
		/*
		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			if (requeue) {
				
				// * Bleah.  Leftovers again.  Stick the
				// * leftovers in the front of the
				// * queue, and goose the queue again.
				
				USBMS_DBG_FLOW(iprintf(
					"[USBMS FLOW] scsi_end_request() : scsi_requeue_command() .. \n"
					));
				scsi_requeue_command(q, cmd);
			   }

			return cmd;
		#endif // ndef NO_ELEVATOR
		*/
	   }
   }

 /*
 // drivers/char/random.c 
 add_disk_randomness(req->rq_disk);
 */
 /*
 #ifndef NO_SPINLOCK
			spin_lock_irqsave(q->queue_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 if (blk_rq_tagged(req))
		blk_queue_end_tag(q, req);
 end_that_request_last(req);
 /*
 #ifndef NO_SPINLOCK
			spin_unlock_irqrestore(q->queue_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 /*
  * This will goose the queue request function at the end, so we don't
  * need to worry about launching another command.
  */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_end_request() : scsi_next_command() .. \n"));
 scsi_next_command(cmd);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_end_request() End. \n"));
 return NULL;
}

/*
#ifndef NO_SCATTER_LIST
   static struct scatterlist *scsi_alloc_sgtable(struct scsi_cmnd *cmd, int gfp_mask)
   {
	struct scsi_host_sg_pool *sgp;
	struct scatterlist *sgl;
	
	BUG_ON(!cmd->use_sg);
	
	if (!cmd->use_sg) {
		dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
	   }

	switch (cmd->use_sg) {
	case 1 ... 8:
		cmd->sglist_len = 0;
		break;
	case 9 ... 16:
		cmd->sglist_len = 1;
		break;
	case 17 ... 32:
		cmd->sglist_len = 2;
		break;
	case 33 ... 64:
		cmd->sglist_len = 3;
		break;
	case 65 ... MAX_PHYS_SEGMENTS:
		cmd->sglist_len = 4;
		break;
	default:
		return NULL;
	}

	sgp = scsi_sg_pools + cmd->sglist_len;

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_alloc_sgtable() : sgl = UsbAllocMemC() , size=%d.. \n",
		sizeof(struct scatterlist)
		));

 #ifndef USB_NO_VMM // no vitrtul memory management
	// mempool_t is on include/linux/mempool.h
	sgl = mempool_alloc(sgp->pool, gfp_mask);
 #endif // ndef USB_NO_VMM
	
	sgl = (struct scatterlist *) UsbAllocMemC(USB_MEM_C_S
			#ifdef USB_DBG_MEM_C_ON
 		  		,sizeof(struct scatterlist)
			#endif
			) ;

	if (sgl)
		memset(sgl, 0, sgp->size);
	return sgl;
}
#endif // ndef NO_SCATTER_LIST
*/
/*
#ifndef NO_SCATTER_LIST
static void scsi_free_sgtable(
	#ifndef NO_SCATTER_LIST
		struct scatterlist *sgl,
	#endif // ndef NO_SCATTER_LIST
	int index)
{
	struct scsi_host_sg_pool *sgp;
 
	BUG_ON(index > SG_MEMPOOL_NR);
 
 if (index > SG_MEMPOOL_NR) {
		dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
	   }

	sgp = scsi_sg_pools + index;

 #ifndef NO_SCATTER_LIST
	#ifndef USB_NO_VMM // no vitrtul memory management
		// mempool_t is on include/linux/mempool.h
		mempool_free(sgl, sgp->pool);
	#endif // ndef USB_NO_VMM
	
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_free_sgtable() : UsbFreeMemC(sgl) , size=%d .. \n",
		sizeof(struct scatterlist)
		));
	UsbFreeMemC(USB_MEM_C_S,(unsigned char *) sgl
			#if defined(USB_DBG_MEM_C_ON)
				, sizeof(struct scatterlist)
			#endif
			) ;
 #endif // ndef NO_SCATTER_LIST
}
#endif // ndef NO_SCATTER_LIST
*/

/*
 * Function:    scsi_release_buffers()
 *
 * Purpose:     Completion processing for block device I/O requests.
 *
 * Arguments:   cmd	- command that we are bailing.
 *
 * Lock status: Assumed that no lock is held upon entry.
 *
 * Returns:     Nothing
 *
 * Notes:       In the event that an upper level driver rejects a
 *		command, we must release resources allocated during
 *		the __init_io() function.  Primarily this would involve
 *		the scatter-gather table, and potentially any bounce
 *		buffers.
 */
static void scsi_release_buffers(struct scsi_cmnd *cmd)
{
 struct request *req = cmd->request;

 /*
  * Free up any indirection buffers we allocated for DMA purposes. 
  */
 /*
 #ifndef NO_SCATTER_LIST
	if (cmd->use_sg)
		scsi_free_sgtable(
				cmd->request_buffer,
			cmd->sglist_len);

 #endif // ndef NO_SCATTER_LIST
 */
 // B. S. Liaw : request_buffer need uncached
 if(cmd->request_buffer != req->buffer)
   {
	#ifndef NO_SCATTER_LIST
		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] scsi_release_buffers() : UsbFreeMemU(cmd->request_buffer) .. \n"
			));
		/*
		kfree(cmd->request_buffer);
		*/
		// B. S. Liaw : request_buffer need uncached
		UsbFreeMemU(USB_MEM_U_S, (unsigned char *) cmd->request_buffer
				#if defined(USB_DBG_MEM_C_ON)
					,sizeof(*cmd->request_buffer)
				#endif
				);
	#endif // ndef NO_SCATTER_LIST
   }

 /*
  * Zero these out.  They now point to freed memory, and it is
  * dangerous to hang onto the pointers.
  */
 cmd->buffer  = NULL;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_release_buffers() : Set Value : cmd->buffer=%08XH \n",
	cmd->buffer
	));
	cmd->bufflen = 0;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_release_buffers() : Set Value : cmd->bufflen=%d \n",
	cmd->bufflen
	));
	cmd->request_buffer = NULL;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_release_buffers() : Set Value : cmd->request_buffer=%08XH .. \n",
	cmd->request_buffer
	));
	cmd->request_bufflen = 0;
}

/*
 * Function:    scsi_io_completion()
 *
 * Purpose:     Completion processing for block device I/O requests.
 *
 * Arguments:   cmd   - command that is finished.
 *
 * Lock status: Assumed that no lock is held upon entry.
 *
 * Returns:     Nothing
 *
 * Notes:       This function is matched in terms of capabilities to
 *              the function that created the scatter-gather list.
 *              In other words, if there are no bounce buffers
 *              (the normal case for most drivers), we don't need
 *              the logic to deal with cleaning up afterwards.
 *
 *		We must do one of several things here:
 *
 *		a) Call scsi_end_request.  This will finish off the
 *		   specified number of sectors.  If we are done, the
 *		   command block will be released, and the queue
 *		   function will be goosed.  If we are not done, then
 *		   scsi_end_request will directly goose the queue.
 *
 *		b) We can just use scsi_requeue_command() here.  This would
 *		   be used if we just wanted to retry, for example.
 */
void scsi_io_completion(struct scsi_cmnd *cmd, int good_sectors,
			int block_sectors)
{
 int result = cmd->result;
 int this_count;
 request_queue_t *q = cmd->device->request_queue;
 struct request *req = cmd->request;
 int clear_errors = 1;

 this_count = cmd->bufflen / q->hardsect_size;
 
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_io_completion() Begin .. \n"));

 // #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_io_completion() : q->req = cmd->request .. \n"));
	q->req = cmd->request ;
 // #endif // NO_ELEVATOR
 /*
  * Free up any indirection buffers we allocated for DMA purposes. 
  * For the case of a READ, we need to copy the data out of the
  * bounce buffer and into the real buffer.
  */
 /*
 #ifndef NO_SCATTER_LIST
	if (cmd->use_sg)
		scsi_free_sgtable(
				cmd->buffer,
			cmd->sglist_len);
	else 
 #endif // ndef NO_SCATTER_LIST
 */
 if (cmd->buffer != req->buffer) {
	if (rq_data_dir(req) == READ) {
			unsigned long flags;
		/*
		#ifndef USB_NO_VMM // vitrtul memory management
			char *to = bio_kmap_irq(req->bio, &flags);
			memcpy(to, cmd->buffer, cmd->bufflen);
			bio_kunmap_irq(to, &flags);
		#endif // ndef USB_NO_VMM
		*/
	   }

	/*        
	USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] scsi_io_completion() : UsbFreeMemC(cmd->buffer) .. \n"
			));
	 #ifndef _ETCPIP
			kfree(cmd->buffer);
		#elif defined(USB_DBG_MEM_C_ON)
			UsbFreeMemC(USB_MEM_C_S,(unsigned char *) cmd->buffer,sizeof(*cmd->buffer));
		#else
			UsbFreeMemC(USB_MEM_C_S,(unsigned char *) cmd->buffer);
	#endif
	cmd->buffer = NULL ; // B. S. Liaw add this
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_io_completion() : Set Value : cmd->buffer=%08XH \n",
		cmd->buffer
		));
	*/
   }

 if(blk_pc_request(req)) { /* SG_IO ioctl from block level */
	req->errors = (driver_byte(result) & DRIVER_SENSE) ?
			      (CHECK_CONDITION << 1) : (result & 0xff);
	if (result) {
		clear_errors = 0;
		if (cmd->sense_buffer[0] & 0x70) {
			int len = 8 + cmd->sense_buffer[7];

			if (len > SCSI_SENSE_BUFFERSIZE)
					len = SCSI_SENSE_BUFFERSIZE;
			memcpy(req->sense, cmd->sense_buffer,  len);
			req->sense_len = len;
		   }
	   } else
		req->data_len -= cmd->bufflen;
   }

 /*
  * Zero these out.  They now point to freed memory, and it is
  * dangerous to hang onto the pointers.
  */
 cmd->buffer  = NULL;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_io_completion() : Set Value : cmd->buffer=%08XH \n",
	cmd->buffer
	));
	cmd->bufflen = 0;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_io_completion() : Set Value : cmd->bufflen=%d \n",
	cmd->bufflen
	));
 cmd->request_buffer = NULL;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_io_completion() : Set Value : cmd->request_buffer=%08XH .. \n",
	cmd->request_buffer
	));
 cmd->request_bufflen = 0;

 /*
  * Next deal with any sectors which we were able to correctly
  * handle.
  */
 if(good_sectors >= 0) {
	/*
	#ifndef NO_SCATTER_LIST
			// SCSI_LOG_HLCOMPLETE(1, printk("%ld sectors total, %d sectors done.\n",
			USBMS_DBG_SCSI(iprintf("[USBMS SCSI] %ld sectors total, %d sectors done.\n",
					      req->nr_sectors, good_sectors));
	 #else
	*/
	USBMS_DBG_SCSI(iprintf(
		"[USBMS SCSI] scsi_io_completion() : %d sector(s) done.\n",
		good_sectors));
	USBMS_DBG_BLOCK_INFO(iprintf(
		"[USBMS BLOCK_INFO] scsi_io_completion() : %d sector(s) done.\n",
		good_sectors));
	USBMS_DBG_RD_CACHE(iprintf(
		"[USBMS RD_CACHE] scsi_io_completion() : %d sector(s) done.\n",
		good_sectors));
	USBMS_DBG_WR_CACHE(iprintf(
		"[USBMS WR_CACHE] scsi_io_completion() : %d sector(s) done.\n",
		good_sectors));
	/*
	#endif
	*/

	/*
	#ifndef NO_SCATTER_LIST
		SCSI_LOG_HLCOMPLETE(1, printk("use_sg is %d\n ", cmd->use_sg));
	#endif // ndef NO_SCATTER_LIST
	*/
	if (clear_errors)
			req->errors = 0;
	/*
	 * If multiple sectors are requested in one buffer, then
	 * they will have been finished off by the first command.
	 * If not, then we have a multi-buffer command.
	 *
	 * If block_sectors != 0, it means we had a medium error
	 * of some sort, and that we want to mark some number of
	 * sectors as not uptodate.  Thus we want to inhibit
	 * requeueing right here - we will requeue down below
	 * when we handle the bad sectors.
	 */
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_io_completion() : scsi_end_request() .. \n"));
	cmd = scsi_end_request(cmd, 1, good_sectors, result == 0);

	/*
	 * If the command completed without error, then either finish off the
	 * rest of the command, or start a new one.
	 */
	if (result == 0 || cmd == NULL ) {
		return;
	   }
   }
  else {
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] scsi_io_completion() : Error : %d sector(s) done. !! \n",
	    good_sectors));
   }
 /*
  * Now, if we were good little boys and girls, Santa left us a request
  * sense buffer.  We can extract information from this, so we
  * can choose a block to remap, etc.
  */
 if (driver_byte(result) != 0) {
		if ((cmd->sense_buffer[0] & 0x7f) == 0x70) {
			/*
			 * If the device is in the process of becoming ready,
			 * retry.
			 */
			if (cmd->sense_buffer[12] == 0x04 &&
			    cmd->sense_buffer[13] == 0x01) {
		dprintf("[USBMS] SCSI : Error : device is in the process !! \n");
		/*
		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_io_completion() : scsi_requeue_command() .. \n"));
				scsi_requeue_command(q, cmd);
		#endif // ndef NO_ELEVATOR
		*/
				return;
			}
			if ((cmd->sense_buffer[2] & 0xf) == UNIT_ATTENTION) {
				if (cmd->device->removable) {
					/* detected disc change.  set a bit 
					 * and quietly refuse further access.
		 			 */
					cmd->device->changed = 1;
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_io_completion() : scsi_end_request() .. \n"));
					cmd = scsi_end_request(cmd, 0,
							this_count, 1);
					return;
				} else {
					/*
				 	* Must have been a power glitch, or a
				 	* bus reset.  Could not have been a
				 	* media change, so we just retry the
				 	* request and see what happens.  
				 	*/
		dprintf("[USBMS] SCSI : Error : power glitch ?? !! \n");
		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_io_completion() : scsi_requeue_command() .. \n"));
					scsi_requeue_command(q, cmd);
		#endif // ndef NO_ELEVATOR
					return;
				}
			}
		}
		/*
		 * If we had an ILLEGAL REQUEST returned, then we may have
		 * performed an unsupported command.  The only thing this
		 * should be would be a ten byte read where only a six byte
		 * read was supported.  Also, on a system where READ CAPACITY
		 * failed, we may have read past the end of the disk.
		 */

		switch (cmd->sense_buffer[2]) {
		case ILLEGAL_REQUEST:
			if (cmd->device->use_10_for_rw &&
			    (cmd->cmnd[0] == READ_10 ||
			     cmd->cmnd[0] == WRITE_10)) {
				USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_io_completion() : Set ->use_10_for_rw = 0 .. \n")) ;
				cmd->device->use_10_for_rw = 0;
				/*
				 * This will cause a retry with a 6-byte
				 * command.
				 */
				dprintf("[USBMS] SCSI : Error : ILLEGAL_REQUEST !! \n");
				#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
					USBMS_DBG_FLOW(iprintf(
						"[USBMS FLOW] scsi_io_completion() : scsi_requeue_command() .. \n"));
					scsi_requeue_command(q, cmd);
				#endif // ndef NO_ELEVATOR
				result = 0;
			} else {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_io_completion() : scsi_end_request() .. \n"));
				cmd = scsi_end_request(cmd, 0, this_count, 1);
				return;
			}
			break;
		case NOT_READY:
			#ifdef USB_DBGSTRING_ON
			/*
			printk(KERN_INFO "Device %s not ready.\n",
			       req->rq_disk ? req->rq_disk->disk_name : "");
			*/
			dprintf("[USBMS] Error : Device %s not ready.\n",
			       req->rq_disk ? req->rq_disk->disk_name : "");
			 #else
				dprintf("[USBMS] Error : Device not ready !! \n");
			#endif
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_io_completion() : scsi_end_request() .. \n"));
			cmd = scsi_end_request(cmd, 0, this_count, 1);
			return;
			break;
		case MEDIUM_ERROR:
		case VOLUME_OVERFLOW:
			printk("scsi%d: ERROR on channel %d, id %d, lun %d, CDB: ",
			       cmd->device->host->host_no, (int) cmd->device->channel,
			       (int) cmd->device->id, (int) cmd->device->lun);
			print_command(cmd->data_cmnd);
			print_sense("", cmd);
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_io_completion() : scsi_end_request() .. \n"));
			cmd = scsi_end_request(cmd, 0, block_sectors, 1);
			return;
		default:
			break;
		}
	}			/* driver byte != 0 */
	if (host_byte(result) == DID_RESET) {
		dprintf("[USBMS] SCSI : Error : DID_RESET !! \n");
		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		/*
		 * Third party bus reset or reset for error
		 * recovery reasons.  Just retry the request
		 * and see what happens.  
		 */
		USBMS_DBG_FLOW(iprintf(
			"[USBMS FLOW] scsi_io_completion() : scsi_requeue_command() .. \n"
			));
		scsi_requeue_command(q, cmd);
		#endif // ndef NO_ELEVATOR
		return;
	}
	if (result) {
		/*
		printk("SCSI error : <%d %d %d %d> return code = 0x%x\n",
		*/
		dprintf("[USBMS] SCSI error : <%d %d %d %d> return code = 0x%x\n",
		       cmd->device->host->host_no,
		       cmd->device->channel,
		       cmd->device->id,
		       cmd->device->lun, result);

		if (driver_byte(result) & DRIVER_SENSE)
			print_sense("", cmd);
		/*
		 * Mark a single buffer as not uptodate.  Queue the remainder.
		 * We sometimes get this cruft in the event that a medium error
		 * isn't properly reported.
		 */
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_io_completion() : scsi_end_request() .. \n"));
		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			cmd = scsi_end_request(cmd, 0, req->current_nr_sectors, 1);
		 #else
			cmd = scsi_end_request(cmd, 0, 1, 1);
		#endif
		return;
	}
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_io_completion() End. \n"));
}

/*
 * Function:    scsi_init_io()
 *
 * Purpose:     SCSI I/O initialize function.
 *
 * Arguments:   cmd   - Command descriptor we wish to initialize
 *
 * Returns:     0 on success
 *		BLKPREP_DEFER if the failure is retryable
 *		BLKPREP_KILL if the failure is fatal
 */
static int scsi_init_io(struct scsi_cmnd *cmd)
{
 struct request     *req = cmd->request;
 /*
 #ifndef NO_SCATTER_LIST
	struct scatterlist *sgpnt;
 #endif // ndef NO_SCATTER_LIST
 */
 int		   count;
 /*
 struct scsi_request *sreq ;
 */


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_init_io() Begin .. \n")) ;

 /*
  * if this is a rq->data based REQ_BLOCK_PC, setup for a non-sg xfer
  */
 if((req->flags & REQ_BLOCK_PC)
		// #ifndef USB_NO_VMM // vitrtul memory management
			 && !req->bio
		// #endif // ndef USB_NO_VMM
   ) {
	cmd->request_bufflen = req->data_len;
	// B. S. Liaw : request_buffer need uncached
	cmd->request_buffer = req->data;
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_init_io(1) : Set Value : cmd->request_buffer=%08XH .. \n",
		cmd->request_buffer
		));
	req->buffer = req->data;
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_init_io(2) : Set Value : req->buffer=%08XH .. \n",
		req->buffer
		));
	/*
	#ifndef NO_SCATTER_LIST
		cmd->use_sg = 0;
	#endif // ndef NO_SCATTER_LIST
	*/
	return 0;
   }

 /*
  * we used to not use scatter-gather for single segment request,
  * but now we do (it makes highmem I/O easier to support without
  * kmapping pages)
  */
 /*
 #ifndef NO_SCATTER_LIST
	cmd->use_sg = req->nr_phys_segments;

	// if sg table allocation fails, requeue request later.
	sgpnt = scsi_alloc_sgtable(cmd, GFP_ATOMIC);

	// if (unlikely(!sgpnt)) {
	if (!sgpnt) {
		req->flags |= REQ_SPECIAL;
		return BLKPREP_DEFER;
	}
 #endif // ndef NO_SCATTER_LIST
 */
 // USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_init_io() : Not Set cmd->request_buffer !! \n")) ;
 /*
 #ifndef NO_SCATTER_LIST
	cmd->request_buffer = (char *) sgpnt;
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_init_io(3) : Set Value : cmd->request_buffer=%08XH .. \n",
		cmd->request_buffer
		));
 #else
 */
	// B. S. Liaw : request_buffer need uncached
	cmd->request_buffer = req->data;
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_init_io(4) : Set Value : cmd->request_buffer=%08XH .. \n",
		cmd->request_buffer
		));
 /*
 #endif // NO_SCATTER_LIST
 */
 /*
 #ifndef NO_SCATTER_LIST
	cmd->request_bufflen = req->nr_sectors << 9;
 #else
 */
 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_init_io(): req->q->hardsect_size=%d\n",req->q->hardsect_size)) ;
 cmd->request_bufflen = req->hard_nr_sectors * req->q->hardsect_size;
	// cmd->request_bufflen = 1 << 9;
 /*
 #endif // NO_SCATTER_LIST
 */

 #ifdef USBMS_BS_ASSERT
	if (cmd->request_bufflen) {
		if ( ! (IS_VALID_ADDR(cmd->request_buffer)) ) {
			iprintf("[ASSERT][USBMS] scsi_init_io(1) : " ) ;
   		    iprintf("Error : Invalid Address !!\n" ) ;
	        iprintf("[ASSERT][USBMS] cmd->request_buffer=%08XH \n",
					cmd->request_buffer ) ; ;
	        iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
		if ( !(USB_IS_UN_CACHED(cmd->request_buffer)) ) {
			iprintf("[ASSERT][USBMS] scsi_init_io(2) : " ) ;
		    iprintf("Error : Not Un-Cached !!\n" ) ;
		    iprintf("[ASSERT][USBMS] cmd->request_buffer=%08XH \n",
					cmd->request_buffer ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
	   } // if has cmd->request_bufflen
 #endif // USBMS_BS_ASSERT

 if (blk_pc_request(req))
		cmd->request_bufflen = req->data_len;
 req->buffer = NULL;

	/* 
	 * Next, walk the list, and fill in the addresses and sizes of
	 * each segment.
	 */
 /*
 #ifndef NO_SCATTER_LIST
	count = blk_rq_map_sg(req->q, req, cmd->request_buffer);

	// mapped well, send it off
	//if (likely(count <= cmd->use_sg)) {
	if ((count <= cmd->use_sg)) {
		cmd->use_sg = count;
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_init_io(1) End. \n")) ;
		return 0;
	}
  #else
 */
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_init_io(2) End. \n")) ;
	return 0 ;
 /*
 #endif // NO_SCATTER_LIST
 */

	/*
	printk(KERN_ERR "Incorrect number of segments after building list\n");
	printk(KERN_ERR "counted %d, received %d\n", count, cmd->use_sg);
	printk(KERN_ERR "req nr_sec %lu, cur_nr_sec %u\n", req->nr_sectors,
			req->current_nr_sectors);
	*/

	dprintf("[USBMS] Error :Incorrect number of segments after building list\n");
	dprintf("[USBMS] Error :counted %d \n", count);
 /*
 #ifndef NO_SCATTER_LIST
	dprintf("[USBMS] Error :counted %d, received %d\n", count, cmd->use_sg);
	dprintf("[USBMS] Error :req nr_sec %lu, cur_nr_sec %u\n", req->nr_sectors,
			req->current_nr_sectors);
 #endif // ndef NO_SCATTER_LIST
 */

 /* release the command and kill it */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_init_io(3) : scsi_put_command() .. \n"));
 scsi_put_command(cmd);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_init_io(4) End. \n")) ;
 return BLKPREP_KILL;
}


static int scsi_prep_fn(struct request_queue *q, struct request *req)
{
 /*
 struct scsi_device *sdev = q->queuedata;
 */
 struct scsi_device *sdev ;
 struct scsi_cmnd *cmd;
 struct scsi_request *sreq ;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_prep_fn() Begin .. \n"));

 USBMS_DBG_BLOCK(iprintf("[USBMS BLOCK] scsi_prep_fn(1) : q=%08XH \n",q)) ;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_prep_fn() : q->queuedata=%08XH \n",q->queuedata
	)) ;
 #ifdef USBMS_BS_ASSERT
	/*
	if (!(   (unsigned long) q->queuedata
		   > (unsigned long) 0x200000L
         )
	   )
	   {
	*/
	if ( ! (IS_VALID_ADDR(q->queuedata)) ) {
		iprintf("[ASSERT][USBMS] scsi_prep_fn(1) : " ) ;
	    iprintf("Error : Invlid address q->queuedata=%08XH \n",
			q->queuedata ) ;
	    iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_ASSERT

 sdev = q->queuedata;

 /*
  * Find the actual device driver associated with this command.
  * The SPECIAL requests are things like character device or
  * ioctls, which did not originate from ll_rw_blk.  Note that
  * the special field is also used to indicate the cmd for
  * the remainder of a partially fulfilled request that can 
  * come up when there is a medium error.  We have to treat
  * these two cases differently.  We differentiate by looking
  * at request->cmd, as this tells us the real story.
  */

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_prep_fn() : req=%08XH \n",req
	)) ;

 #ifdef USBMS_BS_ASSERT
	/*
	if (!(   (unsigned long) req
		   > (unsigned long) 0x200000L
         )
	   )
	   {
	*/
	if ( ! (IS_VALID_ADDR(req)) ) {
		iprintf("[ASSERT][USBMS] scsi_prep_fn(2) : " ) ;
	    iprintf("Error : Invlid address req=%08XH \n",
			req ) ;
	    iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_ASSERT

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_prep_fn() : req->flags=%08XH \n",req->flags
	)) ;

 #ifdef USBMS_BS_ASSERT
	/*
	if (!(   (unsigned long) &(req->flags)
		   > (unsigned long) 0x200000L
         )
	   )
	   {
	*/
	if ( ! (IS_VALID_ADDR(&(req->flags))) ) {
		iprintf("[ASSERT][USBMS] scsi_prep_fn(3) : " ) ;
	    iprintf("Error : Invlid address &(req->flags)=%08XH \n",
			&(req->flags) ) ;
	    iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_ASSERT

 #ifdef USBMS_DBG_BLOCK_ON
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_prep_fn(1) : blk_dump_rq_flags() \n"));
	blk_dump_rq_flags(req, "SCSI req");
 #endif // USBMS_DBG_BLOCK_ON

 if(req->flags & REQ_SPECIAL) {
	USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_prep_fn(2) : REQ_SPECIAL \n"));
    /*
	struct scsi_request *sreq = req->special;
    */
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_prep_fn() : req->special=%08XH \n",req->special
		)) ;
	sreq = req->special;
	if (sreq->sr_magic == SCSI_REQ_MAGIC) {
		USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_prep_fn(3) : SCSI_REQ_MAGIC \n"));
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_prep_fn(2) : scsi_get_command() .. \n" )) ;
		cmd = scsi_get_command(sreq->sr_device, GFP_ATOMIC);
		/*
		if (unlikely(!cmd))
		*/
		if (!cmd)
		   {
			goto defer;
		   }
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_prep_fn(3) : scsi_init_cmd_from_req() .. \n"));
		scsi_init_cmd_from_req(cmd, sreq);
	   }
	  else
		cmd = req->special;
   }
  else if (req->flags & (REQ_CMD | REQ_BLOCK_PC)) {
	USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_prep_fn(4) : REQ_CMD | REQ_BLOCK_PC \n"));
	/*
	 * Just check to see if the device is online.  If
	 * it isn't, we refuse to process ordinary commands
	 * (we will allow specials just in case someone needs
	 * to send a command to an offline device without bringing
	 * it back online)
	 */
	if(!sdev->online) {
		/*
		printk(KERN_ERR "scsi%d (%d:%d): rejecting I/O to offline device\n",
		       sdev->host->host_no, sdev->id, sdev->lun);
		*/
		dprintf("[USBMS] Error :scsi%d (%d:%d): rejecting I/O to offline device\n",
			       sdev->host->host_no, sdev->id, sdev->lun);
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_prep_fn(3) End with BLKPREP_KILL. \n"));
		return BLKPREP_KILL;
	   }
	/*
	 * Now try and find a command block that we can use.
	 */
	if (!req->special) {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_prep_fn(4) : scsi_get_command() .. \n" )) ;
		cmd = scsi_get_command(sdev, GFP_ATOMIC);
		/*
		if (unlikely(!cmd))
		*/
		if (!cmd) {
			USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_prep_fn() : scsi_get_command() Failed !! \n" )) ;
			goto defer;
		   }
	   }
	  else
		cmd = req->special;
		
	/* pull a tag out of the request if we have one */
	cmd->tag = req->tag;
   }
  else {
	dprintf("[USBMS] Error : SCSI bad req !! \n");
	#ifdef USBMS_DBG_BLOCK_ON
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_prep_fn(5) : blk_dump_rq_flags() \n"));
		blk_dump_rq_flags(req, "Error : SCSI bad req");
	#endif // USBMS_DBG_BLOCK_ON
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_prep_fn(6) End with BLKPREP_KILL. \n"));
	return BLKPREP_KILL;
   }
	
 /* note the overloading of req->special.  When the tag
  * is active it always means cmd.  If the tag goes
  * back for re-queueing, it may be reset */
 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_prep_fn() : cmd=%08XH \n",cmd)) ;
 req->special = cmd;
 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_prep_fn() : Set Value : req->special=%08XH \n",
		req->special
		)) ;
 USBMS_DBG_BLOCK(iprintf("[USBMS BLOCK] scsi_prep_fn() : cmd->request = req .. \n"));
 cmd->request = req;
	
 /*
  * FIXME: drop the lock here because the functions below
  * expect to be called without the queue lock held.  Also,
  * previously, we dequeued the request before dropping the
  * lock.  We hope REQ_STARTED prevents anything untoward from
  * happening now.
  */
 if (req->flags & (REQ_CMD | REQ_BLOCK_PC)) {
	struct scsi_driver *drv;
	int ret;

	USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_prep_fn(5) : REQ_CMD | REQ_BLOCK_PC \n"));
	/*
	 * This will do a couple of things:
	 *  1) Fill in the actual SCSI command.
	 *  2) Fill in any other upper-level specific fields
	 * (timeout).
	 *
	 * If this returns 0, it means that the request failed
	 * (reading past end of disk, reading offline device,
	 * etc).   This won't actually talk to the device, but
	 * some kinds of consistency checking may cause the	
	 * request to be rejected immediately.
	 */

	/* 
	 * This sets up the scatter-gather table (allocating if
	 * required).
	 */
	ret = scsi_init_io(cmd);
	if (ret)	/* BLKPREP_KILL return also releases the command */
			return ret;
		
	/*
	 * Initialize the actual SCSI command for this request.
	 */
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_prep_fn() : req=%08XH \n",
		req
		));
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_prep_fn() : req->rq_disk=%08XH \n",
		req->rq_disk
		));

	#ifdef USBMS_BS_ASSERT
		if ( ! (IS_VALID_ADDR(req->rq_disk)) ) {
			iprintf("[ASSERT][USBMS] scsi_prep_fn(4) : " ) ;
   		    iprintf("Error : Invalid Address !!\n" ) ;
	        iprintf("[ASSERT][USBMS] req->rq_disk=%08XH \n",
					req->rq_disk ) ; ;
	        iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
	#endif // USBMS_BS_ASSERT

	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_prep_fn() : req->rq_disk->private_data=%08XH \n",
		req->rq_disk->private_data
		));

	drv = *(struct scsi_driver **)req->rq_disk->private_data;

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_prep_fn(7) : drv->init_command() .. \n"));
	/*
	if (unlikely(!drv->init_command(cmd))) {
	*/
	if (!drv->init_command(cmd)) {
		scsi_release_buffers(cmd);
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_prep_fn(8) : scsi_put_command() .. \n"));
		scsi_put_command(cmd);
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_prep_fn(9) End with BLKPREP_KILL. \n"));
		return BLKPREP_KILL;
	   }
   }

 /*
  * The request is now prepped, no need to come back here
  */
 req->flags |= REQ_DONTPREP;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_prep_fn() End. \n"));
 return BLKPREP_OK;

 // ---------
 defer:
 // ---------

 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_prep_fn(6) : defer \n"));
 /* If we defer, the elv_next_request() returns NULL, but the
  * queue must be restarted, so we plug here if no returning
  * command will automatically do that. */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	if (sdev->device_busy == 0) {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_prep_fn(10) : blk_plug_device() .. \n"));
		blk_plug_device(q);
	   }
 #endif // ndef NO_ELEVATOR
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_prep_fn(11) End with BLKPREP_DEFER. \n"));
 return BLKPREP_DEFER;
}



/*
 * scsi_dev_queue_ready: if we can send requests to sdev, return 1 else
 * return 0.
 *
 * Called with the queue_lock held.
 */
/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
static inline int scsi_dev_queue_ready(struct request_queue *q,
				  struct scsi_device *sdev)
{
 if (sdev->device_busy >= sdev->queue_depth)
		return 0;
 if (sdev->device_busy == 0 && sdev->device_blocked) {
	
	// unblock after device_blocked iterates to zero

	if (--sdev->device_blocked == 0) {
			SCSI_LOG_MLQUEUE(3,
				printk("scsi%d (%d:%d) unblocking device at"
				       " zero depth\n", sdev->host->host_no,
				       sdev->id, sdev->lun));
		} else {
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_dev_queue_ready() : blk_plug_device() .. \n"));
			blk_plug_device(q);
			return 0;
		}
	}
	if (sdev->device_blocked)
		return 0;

	return 1;
}
#endif // ndef NO_ELEVATOR
*/
/*
 * scsi_host_queue_ready: if we can send requests to shost, return 1 else
 * return 0. We must end up running the queue again whenever 0 is
 * returned, else IO can hang.
 *
 * Called with host_lock held.
 */
/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
static inline int scsi_host_queue_ready(struct request_queue *q,
				   struct Scsi_Host *shost,
				   struct scsi_device *sdev)
{
	if (test_bit(SHOST_RECOVERY, &shost->shost_state))
		return 0;
	if (shost->host_busy == 0 && shost->host_blocked) {
		
		// * unblock after host_blocked iterates to zero
		 
		if (--shost->host_blocked == 0) {
			SCSI_LOG_MLQUEUE(3,
				printk("scsi%d unblocking host at zero depth\n",
					shost->host_no));
		} else {
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_host_queue_ready() : blk_plug_device() .. \n"));
			blk_plug_device(q);
			return 0;
		}
	}
	if ((shost->can_queue > 0 && shost->host_busy >= shost->can_queue) ||
	    shost->host_blocked || shost->host_self_blocked) {
		if (list_empty(&sdev->starved_entry))
		   {
 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_host_queue_ready() : shost->starved_list=%08XH \n",
			shost->starved_list ));
		#ifdef USBMS_BS_ASSERT
			if (!(   (unsigned long) &shost->starved_list
				   > (unsigned long) 0x200000L
		         )
			   )
			   {
				iprintf("[ASSERT][USBMS] scsi_host_queue_ready() : " ) ;
    		    iprintf("Error : Invlid address &shost->starved_list=%08XH \n",
						&shost->starved_list) ;
			    iprintf("[ASSERT][USBMS] Continue ..\n") ;
    		   }
		#endif // USBMS_BS_ASSERT
			list_add_tail(&sdev->starved_entry, &shost->starved_list);
		   }
		return 0;
	}

 // We're OK to process the command, so we can't be starved 
 if(!list_empty(&sdev->starved_entry)) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_host_queue_ready() : list_del_init() .. \n"));
	list_del_init(&sdev->starved_entry);
   }
 return 1;
}
#endif // ndef NO_ELEVATOR
*/

/*
 * Function:    scsi_request_fn()
 *
 * Purpose:     Main strategy routine for SCSI.
 *
 * Arguments:   q       - Pointer to actual queue.
 *
 * Returns:     Nothing
 * Returns:     1 : ok , 0 : error
 *
 * Lock status: IO request lock assumed to be held when called.
 */
// request_queue_t *blk_init_queue(request_fn_proc *rfn)
// so , rfn = scsi_request_fn
// q->request_fn		= rfn;

// static void scsi_request_fn(struct request_queue *q)
static int scsi_request_fn(struct request_queue *q)
{
 struct scsi_device *sdev = q->queuedata;
 struct Scsi_Host *shost = sdev->host;
 struct scsi_cmnd *cmd;
 struct request *req;
 // #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	int ret;
 // #endif // NO_ELEVATOR


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_request_fn() Begin .. \n")) ;
 /*
  * To start with, we keep looping until the queue is empty, or until
  * the host is no longer able to accept any more requests.
  */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
 while (!blk_queue_plugged(q))
 #endif // ndef NO_ELEVATOR
 */
   {
	int rtn;

	/*
	 * get next queueable request.  We do this early to make sure
	 * that the request is fully prepared even if we cannot 
	 * accept it.
	 */
	/*
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		// elv_next_request() is on include/linux/elevator.h
		req = elv_next_request(q);
	 #else
	*/
		USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_request_fn() : Bug on req !! \n"));

		USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_request_fn() : q->req=%08XH \n",
			q->req)) ;

		#ifdef USBMS_BS_ASSERT
			if (!(   (unsigned long) q->req
			   > (unsigned long) 0x200000L
    	     )
		   )
		   {
			iprintf("[ASSERT][USBMS]  scsi_request_fn() : " ) ;
		    iprintf("Error : Invlid address q->req=%08XH \n",
				q->req ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
		#endif // USBMS_BS_ASSERT

		req = q->req;
		/*
		if ((rq->flags & REQ_DONTPREP) || !q->prep_rq_fn)
			break;
		*/
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_request_fn(1) : q->prep_rq_fn() .. \n"));
		/*
		ret = q->prep_rq_fn(q, req);
		if (ret == BLKPREP_OK) {
		*/
		if (q->prep_rq_fn(q, req) == BLKPREP_OK) {
			// break;
		   } 
		/*
		  else if (ret == BLKPREP_DEFER) {
			req = NULL;
			break;
		   }
		
		  else if (ret == BLKPREP_KILL) {
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_request_fn(2) : blkdev_dequeue_request() .. \n"));
			blkdev_dequeue_request(req);
			rq->flags |= REQ_QUIET;
			while (end_that_request_first(req, 0, rq->nr_sectors))
				;
			end_that_request_last(req);
		   }
		*/
		 else {
			/*
			printk("%s: q->prep_rq_fn() Failed !! \n", __FUNCTION__);
			*/
			USBMS_DBG_INFO(iprintf(
				"[USBMS INFO] scsi_request_fn() : Error : q->prep_rq_fn() Failed !! \n"
				));
			#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
				break;
			 #else
				goto break_while ;
			#endif
		   }
	/*	
	#endif // NO_ELEVATOR 
	*/

	#ifdef USBMS_DBG_INFO_ON
		if (!req) {
			USBMS_DBG_INFO(iprintf(
				"[USBMS INFO] scsi_request_fn() : Error : !req !! \n"
				));
		   }
		/*
		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			if ( !scsi_dev_queue_ready(q, sdev)) {
				USBMS_DBG_INFO(iprintf(
					"[USBMS INFO] scsi_request_fn() : Error : !scsi_dev_queue_ready(q, sdev) !! \n"
					));
			   }
		#endif // ndef NO_ELEVATOR
		*/
	#endif //  USBMS_DBG_INFO_ON

	if (    !req
		/*
		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		 || !scsi_dev_queue_ready(q, sdev)
		#endif // ndef NO_ELEVATOR
		*/
	   ) {
		/*
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] scsi_request_fn() : Error : !req || !scsi_dev_queue_ready(q, sdev) !! \n"
			));
		*/
		dprintf(
			"[USBMS] Error : SCSI queue not ready !! \n"
			);
		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			break;
		 #else
			// goto break_while ;
			goto not_ok ;
		#endif
	   }

	/*
	 * Remove the request from the request list.
	 */
	if (!(blk_queue_tagged(q) && !blk_queue_start_tag(q, req)))
	   {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_request_fn(3) : blkdev_dequeue_request() .. \n"));
		blkdev_dequeue_request(req);
	   }
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		sdev->device_busy++;
	#endif // ndef NO_ELEVATOR
	/*
	#ifndef NO_SPINLOCK
		spin_unlock(q->queue_lock);
		spin_lock(shost->host_lock);
	#endif // ndef NO_SPINLOCK
	*/
	/*
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_request_fn(4) : scsi_host_queue_ready() .. \n"));
	if (!scsi_host_queue_ready(q, shost, sdev)) {
		USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_request_fn() : Error : Host Not Ready !! \n")) ;
		dprintf("[USBMS] SCSI : Error : Host Not Ready !! \n") ;
		goto not_ready;
	   }
	#endif // ndef NO_ELEVATOR
	*/
	if (sdev->single_lun) {
		if (sdev->sdev_target->starget_sdev_user &&
		    sdev->sdev_target->starget_sdev_user != sdev) {
			USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_request_fn() : Error : targe Not Ready !! \n")) ;
			dprintf("[USBMS INFO] SCSI : Error : targe Not Ready !! \n") ;
			goto not_ready;
		   }
		sdev->sdev_target->starget_sdev_user = sdev;
	   }
	shost->host_busy++;

	/*
	 * XXX(hch): This is rather suboptimal, scsi_dispatch_cmd will
	 *		take the lock again.
	 */
	/*
	#ifndef NO_SPINLOCK
		spin_unlock_irq(shost->host_lock);
	#endif // ndef NO_SPINLOCK
	*/

	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_request_fn() : req->special=%08XH \n",req->special
		)) ;
	cmd = req->special;
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_request_fn() : cmd=%08XH \n",cmd
		)) ;
	/*
	if (unlikely(cmd == NULL)) {
	*/
	if (cmd == NULL) {
		/*
		printk(KERN_CRIT "impossible request in %s.\n"
					 "please mail a stack trace to "
					 "linux-scsi@vger.kernel.org",
					 __FUNCTION__);
		*/
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] scsi_request_fn() : Error : impossible request !! \n"
			)) ;
		dprintf("[USBMS] Error : impossible request !! \n") ;
		/*
		BUG();
		*/
	   }

	/*
	 * Finally, initialize any error handling parameters, and set up
	 * the timers for timeouts.
	 */
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_request_fn(5) : scsi_init_cmd_errh() .. \n"));
	scsi_init_cmd_errh(cmd);

	/*
	 * Dispatch the command to the low-level driver.
	 */
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_request_fn(6) : scsi_dispatch_cmd() .. \n"));
	rtn = scsi_dispatch_cmd(cmd);
	/*
	#ifndef NO_SPINLOCK
		spin_lock_irq(q->queue_lock);
	#endif // ndef NO_SPINLOCK
	*/
	if (rtn) {
		/* we're refusing the command; because of
		 * the way locks get dropped, we need to 
		 * check here if plugging is required */

		USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_request_fn() : scsi_dispatch_cmd() Failed !! \n"));

		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			if(sdev->device_busy == 0) {
				USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_request_fn(7) : blk_plug_device() .. \n"));
				blk_plug_device(q);
			   }
		#endif // ndef NO_ELEVATOR

		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			break;
		 #else
			// goto break_while ;
			goto not_ok ;
		#endif
	   }
	/*
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_request_fn() : Next Iteration .. \n"));
	#endif // ndef NO_ELEVATOR
	*/
   } // while loop

 // -----------------
 break_while:
 // -----------------

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_request_fn() End. \n")) ;
 return 1 ;

 // -----------------
 not_ready:
 // -----------------
 USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_request_fn() : Error : not_ready !! \n")) ;
 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irq(shost->host_lock);
 #endif // ndef NO_SPINLOCK
 */
 /*
  * lock q, handle tag, requeue req, and decrement device_busy. We
  * must return with queue_lock held.
  *
  * Decrementing device_busy without checking it is OK, as all such
  * cases (host limits or settings) should run the queue at some
  * later time.
 */
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irq(q->queue_lock);
 #endif // ndef NO_SPINLOCK
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_request_fn(8) : blk_requeue_request() .. \n"));
 blk_requeue_request(q, req);
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	sdev->device_busy--;
	if(sdev->device_busy == 0) {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_request_fn(9) : blk_plug_device() .. \n"));
		blk_plug_device(q);
	   }
 #endif // ndef NO_ELEVATOR

 // -----------------
 not_ok:
 // -----------------
#ifndef USBMS_NO_USB_TASK  /* Charles 20040520 */
 // 2004/1/8 B. S. Liaw : Let the task keep going even Error Occur !!
 #ifdef USBMS_USB_WAIT_BY_EVENT     // 2004/1/2 B. S. Liaw
	// 2003/12/01 B. S. Liaw :
	USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_request_fn() : SIGNAL(MT_EVT_USBMS_RQ) .. \n")) ;
	// kernel/mtmacro.h : #define SIGNAL(signo) setevt(signo)
	if (SIGNAL(MT_EVT_USBMS_RQ) != SUCCESS) {
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] scsi_request_fn() : Error : SIGNAL MT_EVT_USBMS_RQ=%d fail !! \n",
			MT_EVT_USBMS_RQ
			)) ;
		dprintf("[USBMS] Error : SIGNAL MT_EVT_USBMS_RQ %d fail !! \n",MT_EVT_USBMS_RQ);
	   }
  #else
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] scsi_wait_req() : Set Value : req->evt_blk_rq=1 .. \n"
		)) ;
	req->evt_blk_rq = 1 ;
 #endif // ndef USBMS_USB_WAIT_BY_EVENT
#endif
 USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_request_fn() : Error : Failed !! \n")) ;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_request_fn(10) End with Failed !! \n")) ;
 return 0 ;
}

/*
#ifdef CONFIG_HIGHMEM
u64 scsi_calculate_bounce_limit(struct Scsi_Host *shost)
{
	struct device *host_dev;


	if (shost->unchecked_isa_dma)
		return BLK_BOUNCE_ISA;

 host_dev = scsi_get_device(shost);

 #ifdef CONFIG_PCI
	if (PCI_DMA_BUS_IS_PHYS && host_dev && host_dev->dma_mask)
			return *host_dev->dma_mask;
 #endif	// CONFIG_PCI

 
 // * Platforms with virtual-DMA translation
 // * hardware have no practical limit.
  
 return BLK_BOUNCE_ANY;
}
#endif // CONFIG_HIGHMEM
*/
struct request_queue *scsi_alloc_queue(struct scsi_device *sdev)
{
 struct Scsi_Host *shost = sdev->host;
 struct request_queue *q;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_alloc_queue() Begin .. \n"));
 /*
 #ifndef BLK_NO_QUEUE 
 */
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_alloc_queue() : blk_init_queue() .. \n"));
	q = blk_init_queue(scsi_request_fn
						/*
						#ifndef NO_SPINLOCK
							, &sdev->sdev_lock
						#endif // ndef NO_SPINLOCK
						*/
						);
	USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_alloc_queue() : q=%08XH \n",q)) ;
	if (!q)
		return NULL;
 /*
  #else
 
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_alloc_queue() : blk_alloc_queue() .. \n"));
	#ifndef _ETCPIP
		q = blk_alloc_queue(GFP_KERNEL);
	#else
		q = blk_alloc_queue();
	#endif
	if (!q)
		return NULL;

	if (blk_init_free_list(q))
		goto out_init;

	USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_alloc_queue() : Not Using request queue \n"));

    
	// q->request_fn		= rfn;
 	// q->back_merge_fn       	= ll_back_merge_fn;
	// q->front_merge_fn      	= ll_front_merge_fn;
	// q->merge_requests_fn	= ll_merge_requests_fn;
	// q->prep_rq_fn		= NULL;
	// q->unplug_fn		= generic_unplug_device;
	// q->queue_flags		= (1 << QUEUE_FLAG_CLUSTER);
	// blk_queue_segment_boundary(q, 0xffffffff);
	
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_alloc_queue() : blk_queue_make_request() .. \n")) ;
	scsi_request_fn(q) ;
 
	
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_alloc_queue() : blk_queue_make_request() .. \n")) ;
	blk_queue_make_request(q, sbull_make_request);
	
 #endif
 */

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_alloc_queue() : blk_queue_prep_rq() .. \n"));
 blk_queue_prep_rq(q, scsi_prep_fn);
 blk_queue_max_hw_segments(q, shost->sg_tablesize);
 blk_queue_max_phys_segments(q, MAX_PHYS_SEGMENTS);
 blk_queue_max_sectors(q, shost->max_sectors);
 #ifdef CONFIG_HIGHMEM
	blk_queue_bounce_limit(q, scsi_calculate_bounce_limit(shost));
 #endif // CONFIG_HIGHMEM
 blk_queue_segment_boundary(q, shost->dma_boundary);

 if (!shost->use_clustering)
		clear_bit(QUEUE_FLAG_CLUSTER, &q->queue_flags);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_alloc_queue() End. \n"));
 return q;
 /*
 #ifdef BLK_NO_QUEUE 
 out_init:

	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_alloc_queue() : UsbFreeMemC(q) , size=%d .. \n",
		sizeof(request_queue_t)
		));
 
	 // kfree(q);
 
	 UsbFreeMemC(USB_MEM_C_M,(unsigned char *) q
					#if defined(USB_DBG_MEM_C_ON)
						,sizeof(request_queue_t) // size : 148
					#endif
					);
	#endif // ndef NO_ELEVATOR

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_alloc_queue() End with Fail !! \n"));
 return NULL;
 #endif // BLK_NO_QUEUE 
 */
}

void scsi_free_queue(struct request_queue *q)
{
 blk_cleanup_queue(q);
}

/*
 * Function:    scsi_block_requests()
 *
 * Purpose:     Utility function used by low-level drivers to prevent further
 *		commands from being queued to the device.
 *
 * Arguments:   shost       - Host in question
 *
 * Returns:     Nothing
 *
 * Lock status: No locks are assumed held.
 *
 * Notes:       There is no timer nor any other means by which the requests
 *		get unblocked other than the low-level driver calling
 *		scsi_unblock_requests().
 */
void scsi_block_requests(struct Scsi_Host *shost)
{
	shost->host_self_blocked = 1;
}

/*
 * Function:    scsi_unblock_requests()
 *
 * Purpose:     Utility function used by low-level drivers to allow further
 *		commands from being queued to the device.
 *
 * Arguments:   shost       - Host in question
 *
 * Returns:     Nothing
 *
 * Lock status: No locks are assumed held.
 *
 * Notes:       There is no timer nor any other means by which the requests
 *		get unblocked other than the low-level driver calling
 *		scsi_unblock_requests().
 *
 *		This is done as an API function so that changes to the
 *		internals of the scsi mid-layer won't require wholesale
 *		changes to drivers that use this feature.
 */
void scsi_unblock_requests(struct Scsi_Host *shost)
{
 shost->host_self_blocked = 0;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_unblock_requests() : scsi_run_host_queues() .. \n"));
 scsi_run_host_queues(shost);
}

int __init scsi_init_queue(void)
{
 /*
	int i;
 
	for (i = 0; i < SG_MEMPOOL_NR; i++) {
		struct scsi_host_sg_pool *sgp = scsi_sg_pools + i;
	#ifndef NO_SCATTER_LIST
		int size = sgp->size * sizeof(struct scatterlist);
	#endif // ndef NO_SCATTER_LIST

	// kmem_cache_t is on include/linux/slab.h
	  sgp->slab = kmem_cache_create(sgp->name, size, 0,
				SLAB_HWCACHE_ALIGN, NULL, NULL);
	if (!sgp->slab) {
			
			//printk(KERN_ERR "SCSI: can't init sg slab %s\n",
			//		sgp->name);
			
			dprintf("[USBMS] Error :SCSI: can't init sg slab %s\n",
					sgp->name);
		}
	
	// kmem_cache_t is on include/linux/slab.h
		sgp->pool = mempool_create(SG_MEMPOOL_SIZE,
				mempool_alloc_slab, mempool_free_slab,
				sgp->slab);
		
		// mempool_t is on include/linux/mempool.h
		if (!sgp->pool) {
		
		
		if (!sgp->pool_data) {
			
			//printk(KERN_ERR "SCSI: can't init sg mempool %s\n",
			//		sgp->name);
			
			dprintf("[USBMS] Error :SCSI: can't init sg mempool %s\n",
					sgp->name);
			}
		
	}
 */
 return 0;
}

void scsi_exit_queue(void)
{
 /*
	int i;

	for (i = 0; i < SG_MEMPOOL_NR; i++) {
		struct scsi_host_sg_pool *sgp = scsi_sg_pools + i;

		// mempool_t is on include/linux/mempool.h
		mempool_destroy(sgp->pool);
		
		
		UsbFreeMemC(USB_MEM_C_S,(unsigned char *) sgp->pool_data
			#if defined(USB_DBG_MEM_C_ON)
				sizeof(sgp->pool_data)
			#endif
			) ;
		sgp->pool_data = NULL ; // B. S. Liaw add this
		
	// kmem_cache_t is on include/linux/slab.h
		kmem_cache_destroy(sgp->slab);
	
	}
 */
}
/**
 *	__scsi_mode_sense - issue a mode sense, falling back from 10 to 
 *		six bytes if necessary.
 *	@sreq:	SCSI request to fill in with the MODE_SENSE
 *	@dbd:	set if mode sense will allow block descriptors to be returned
 *	@modepage: mode page being requested
 *	@buffer: request buffer (may not be smaller than eight bytes)
 *	@len:	length of request buffer.
 *	@timeout: command timeout
 *	@retries: number of retries before failing
 *	@data: returns a structure abstracting the mode header data
 *
 *	Returns zero if unsuccessful, or the header offset (either 4
 *	or 8 depending on whether a six or ten byte command was
 *	issued) if successful.
 **/
int
__scsi_mode_sense(struct scsi_request *sreq, int dbd, int modepage,
		  unsigned char *buffer, int len, int timeout, int retries,
		  struct scsi_mode_data *data) {
 unsigned char cmd[12];
 int use_10_for_ms;
 int header_length;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __scsi_mode_sense() Begin .. \n"));
 memset(data, 0, sizeof(*data));
 memset(&cmd[0], 0, 12);
 cmd[1] = dbd & 0x18;	/* allows DBD and LLBA bits */
 cmd[2] = modepage;

 retry:
	use_10_for_ms = sreq->sr_device->use_10_for_ms;

	if (use_10_for_ms) {
		if (len < 8)
			len = 8;

		cmd[0] = MODE_SENSE_10;
		cmd[8] = len;
		header_length = 8;
	} else {
		if (len < 4)
			len = 4;

		cmd[0] = MODE_SENSE;
		cmd[4] = len;
		header_length = 4;
	}

	sreq->sr_cmd_len = 0;
	sreq->sr_sense_buffer[0] = 0;
	sreq->sr_sense_buffer[2] = 0;
	sreq->sr_data_direction = DMA_FROM_DEVICE;

 memset(buffer, 0, len);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __scsi_mode_sense() : scsi_wait_req() .. \n"));
 scsi_wait_req(sreq, cmd, buffer, len, timeout, retries);

	/* This code looks awful: what it's doing is making sure an
	 * ILLEGAL REQUEST sense return identifies the actual command
	 * byte as the problem.  MODE_SENSE commands can return
	 * ILLEGAL REQUEST if the code page isn't supported */
	if (use_10_for_ms && ! scsi_status_is_good(sreq->sr_result) &&
	    (driver_byte(sreq->sr_result) & DRIVER_SENSE) &&
	    sreq->sr_sense_buffer[2] == ILLEGAL_REQUEST &&
	    (sreq->sr_sense_buffer[4] & 0x40) == 0x40 &&
	    sreq->sr_sense_buffer[5] == 0 &&
	    sreq->sr_sense_buffer[6] == 0 ) {
 USBMS_DBG_INIT(iprintf("[USBMS INIT] __scsi_mode_sense() : Set ->use_10_for_ms=0 .. \n")) ;
		sreq->sr_device->use_10_for_ms = 0;
		goto retry;
	}

	if(scsi_status_is_good(sreq->sr_result)) {
		data->header_length = header_length;
		if(use_10_for_ms) {
			data->length = buffer[0]*256 + buffer[1] + 2;
			data->medium_type = buffer[2];
			data->device_specific = buffer[3];
			data->longlba = buffer[4] & 0x01;
			data->block_descriptor_length = buffer[6]*256
				+ buffer[7];
		} else {
			data->length = buffer[0] + 1;
			data->medium_type = buffer[1];
			data->device_specific = buffer[2];
			data->block_descriptor_length = buffer[3];
		}
	}

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __scsi_mode_sense() End. \n"));
 return sreq->sr_result;
}

/**
 *	scsi_mode_sense - issue a mode sense, falling back from 10 to 
 *		six bytes if necessary.
 *	@sdev:	scsi device to send command to.
 *	@dbd:	set if mode sense will disable block descriptors in the return
 *	@modepage: mode page being requested
 *	@buffer: request buffer (may not be smaller than eight bytes)
 *	@len:	length of request buffer.
 *	@timeout: command timeout
 *	@retries: number of retries before failing
 *
 *	Returns zero if unsuccessful, or the header offset (either 4
 *	or 8 depending on whether a six or ten byte command was
 *	issued) if successful.
 **/

/*
B. S. Liw : None Call This ??

int
scsi_mode_sense(struct scsi_device *sdev, int dbd, int modepage,
		unsigned char *buffer, int len, int timeout, int retries,
		struct scsi_mode_data *data)
{
 
 // struct scsi_request *sreq = scsi_allocate_request(sdev, GFP_KERNEL);
 
 struct scsi_request *sreq ;
 int ret;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_mode_sense() Begin .. \n"));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_mode_sense() : scsi_allocate_request() .. \n"));
 
 // sreq = scsi_allocate_request(sdev, GFP_KERNEL);
 
 sreq = scsi_allocate_request(sdev);
 if (!sreq)
		return -1;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_mode_sense() : __scsi_mode_sense() .. \n"));
 ret = __scsi_mode_sense(sreq, dbd, modepage, buffer, len,
				timeout, retries, data);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_mode_sense() : scsi_release_request() .. \n"));
 scsi_release_request(sreq);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_mode_sense() End. \n"));
 return ret;
}
*/
/*
#ifdef BLK_NO_QUEUE 

// Transfer a buffer directly, without going through the request queue.

// static int sbull_make_request(request_queue_t *queue, int rw, struct buffer_head *bh)
static int sbull_make_request(request_queue_t *q, struct bio *bio)
{
 u8 *ptr;
 int rw ;
 struct buffer_head *bh ;
 
 // Sbull_Dev *device = sbull_devices + MINOR(bh->b_rdev);
 
 struct block_device *device ;
 int nr_sectors, cur_nr_sectors ;
 sector_t sector;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sbull_make_request() Begin .. \n"));

 rw = bio_data_dir(bio);

 // Figure out what we are doing 
 
 // device = sbull_devices + MINOR(bh->b_rdev);
 
 // device = sbull_devices + MINOR(bh->b_bdev->bd_dev) ;
 device = bio->bi_bdev ;
 
 // ptr = device->data + bh->b_rsector * sbull_hardsect;
 
 sector = bio->bi_sector;
 nr_sectors = bio_sectors(bio);
 cur_nr_sectors = bio_cur_sectors(bio);
 ptr = device->data + bh->b_rsector * queue_hardsect_size(q) ;

 // Paranoid check, this apparently can really happen 
 
 //if (ptr + bh->b_size > device->data + sbull_blksize*sbull_size) {
 //       static int count = 0;
 //       if (count++ < 5)
 //           printk(KERN_WARNING "sbull: request past end of device\n");
 //       bh->b_end_io(bh, 0);
 //       return 0;
 //   }
 
 // This could be a high memory buffer, shift it down 
 #ifdef CONFIG_HIGHMEM
    bh = create_bounce(rw, bh);
 #endif // CONFIG_HIGHMEM

 // Do the transfer 
 switch(rw) {
        case READ:
        case READA:  // Readahead 
            memcpy(bh->b_data, ptr, bh->b_size); // from sbull to buffer 
			
            // bh->b_end_io(bh, 1);
			
 // #ifndef USB_NO_VMM
			bio_endio(bio, bio->bi_size, 0); // 0 on success
 // #endif // ndef USB_NO_VMM
            break;
        case WRITE:
            refile_buffer(bh);
            memcpy(ptr, bh->b_data, bh->b_size); // from buffer to sbull
            mark_buffer_uptodate(bh, 1);
			
            // bh->b_end_io(bh, 1);
			
 // #ifndef USB_NO_VMM
			bio_endio(bio, bio->bi_size, 0); // 0 on success
 // #endif // ndef USB_NO_VMM
            break;
        default:
            // can't happen 
			
            // bh->b_end_io(bh, 0);
			
 // #ifndef USB_NO_VMM
			bio_endio(bio, bio->bi_size, -EIO);
 // #endif // ndef USB_NO_VMM
            break;
    }

 // Nonzero return means we're done 
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sbull_make_request() End. \n"));
 return 0;
}


void sbull_unused_request(request_queue_t *q)
{
    static int count = 0;
    if (count++ < 5)
        printk(KERN_WARNING "sbull: unused_request called\n");
}
#endif // BLK_NO_QUEUE 
*/

// B. S. Liaw TEST for compiler warning !!
