/*
 * scsi_sysfs.c
 *
 * SCSI sysfs interface routines.
 *
 * Created to pull SCSI mid layer sysfs routines into one file.
 */

// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
          // for deciding whether to compile
          //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
// #endif // _ETCPIP

/*
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/blkdev.h>
*/
#include "bitops.h" // B. S. Liaw add for test_bit()
#include "blkdev.h"
/*
#include <linux/device.h>
*/
#include "PoDevice.h" // include/linux/device.h
#include "kdev_t.h"   // also need
// B. S. Liaw Add this
#include "base/PoDeviceSc.h" // #include/linux/device.h , 2.6.0-test7
/*
#include "scsi.h"
*/
#include "scsi_ds.h"
/*
#include "hosts.h"
*/
#include "scsi_hosts.h"

#include "scsi_priv.h"
#include "scsi_logging.h"

// #ifdef _ETCPIP
// B. S. Liaw add this
LIST_HEAD(scsi_bus_drvlist);
LIST_HEAD(scsi_bus_devlist);
// #endif // _ETCPIP



static int check_set(unsigned int *val, char *src)
{
	char *last;

	if (strncmp(src, "-", 20) == 0) {
		*val = SCAN_WILD_CARD;
	} else {
		/*
		 * Doesn't check for int overflow
		 */
		*val = simple_strtoul(src, &last, 0);
		if (*last != '\0')
			return 1;
	}
	return 0;
}

static int scsi_scan(struct Scsi_Host *shost, const char *str)
{
 char s1[15], s2[15], s3[15], junk;
 unsigned int channel, id, lun;
 int res;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan() Begin .. \n")) ;

 USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_scan() : sscanf() .. \n")) ;
 res = sscanf((char *)str, "%10s %10s %10s %c", s1, s2, s3, &junk);
	if (res != 3)
		return -EINVAL;
	if (check_set(&channel, s1))
		return -EINVAL;
	if (check_set(&id, s2))
		return -EINVAL;
	if (check_set(&lun, s3))
		return -EINVAL;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan() : scsi_scan_host_selected() .. \n")) ;
 res = scsi_scan_host_selected(shost, channel, id, lun, 1);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan() End. \n")) ;
 return res;
}

/*
 * shost_show_function: macro to create an attr function that can be used to
 * show a non-bit field.
 */
#define shost_show_function(field, format_string)			\
static ssize_t								\
show_##field (struct class_device *class_dev, char *buf)		\
{									\
	struct Scsi_Host *shost = class_to_shost(class_dev);		\
	return snprintf (buf, 20, format_string, shost->field);	\
}

/*
 * shost_rd_attr: macro to create a function and attribute variable for a
 * read only field.
 */
#define shost_rd_attr(field, format_string)				\
	shost_show_function(field, format_string)				\
static CLASS_DEVICE_ATTR(field, S_IRUGO, show_##field, NULL)

/*
 * Create the actual show/store functions and data structures.
 */

static ssize_t store_scan(struct class_device *class_dev, const char *buf,
			  size_t count)
{
	struct Scsi_Host *shost = class_to_shost(class_dev);
	int res;

	res = scsi_scan(shost, buf);
	if (res == 0)
		res = count;
	return res;
};
static CLASS_DEVICE_ATTR(scan, S_IWUSR, NULL, store_scan);

shost_rd_attr(unique_id, "%u\n");
shost_rd_attr(host_busy, "%hu\n");
shost_rd_attr(cmd_per_lun, "%hd\n");
shost_rd_attr(sg_tablesize, "%hu\n");
shost_rd_attr(unchecked_isa_dma, "%d\n");

static struct class_device_attribute *scsi_sysfs_shost_attrs[] = {
	&class_device_attr_unique_id,
	&class_device_attr_host_busy,
	&class_device_attr_cmd_per_lun,
	&class_device_attr_sg_tablesize,
	&class_device_attr_unchecked_isa_dma,
	&class_device_attr_scan,
	NULL
};

static void scsi_device_cls_release(struct class_device *class_dev)
{
 struct scsi_device *sdev;

 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_device_cls_release() Begin .. \n"));
 sdev = class_to_sdev(class_dev);
 
 // #ifndef _ETCPIP
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_device_cls_release() : put_device() .. \n"));
 put_device(&sdev->sdev_gendev);
 // #endif // ndef _ETCPIP
 
}
/*
#ifndef USBMS_ONLY_ONE_DISK
static void scsi_device_dev_release(struct device *dev)
{
 struct scsi_device *sdev;
 struct device *parent;


 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_device_dev_release() Begin .. \n"));
 parent = dev->parent;
 sdev = to_scsi_device(dev);
 scsi_free_sdev(sdev);
 
 // #ifndef _ETCPIP
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_device_dev_release() : put_device() .. \n"));
 put_device(parent);
 // #endif // ndef _ETCPIP
}
#endif // ndef USBMS_ONLY_ONE_DISK
*/
struct class sdev_class = {
 
 #ifdef USB_DBGSTRING_ON
	.name		= "scsi_device",
 #endif // USB_DBGSTRING_ON
 
	.release	= scsi_device_cls_release,
};

/* all probing is done in the individual ->probe routines */
static int scsi_bus_match(struct device *dev, struct device_driver *gendrv)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_bus_match() Begin .. \n"));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_bus_match() End. \n"));
 return 1;
}

struct bus_type scsi_bus_type = {
 #ifdef USB_DBGSTRING_ON
	.name		= "scsi",
 #endif // USB_DBGSTRING_ON
 .match		= scsi_bus_match,
};

// #ifndef NO_SYS_FS
int scsi_sysfs_register(void)
{
 int error;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_sysfs_register() Begin .. \n"));
 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_sysfs_register() : bus_register() .. \n"));
 #ifndef NO_KSET
	error = bus_register(&scsi_bus_type);
  #else
	error = bus_register(&scsi_bus_type,&scsi_bus_drvlist,&scsi_bus_devlist) ;
 #endif
	if (!error) {
		error = class_register(&sdev_class);
		if (error)
			bus_unregister(&scsi_bus_type);
	}

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_sysfs_register() End. \n"));
 return error;
}
// #endif // ndef NO_SYS_FS

// #ifndef NO_SYS_FS
void scsi_sysfs_unregister(void)
{
	class_unregister(&sdev_class);
	bus_unregister(&scsi_bus_type);
}
// #endif // ndef NO_SYS_FS

/*
 * sdev_show_function: macro to create an attr function that can be used to
 * show a non-bit field.
 */

#define sdev_show_function(field, format_string)				\
static ssize_t								\
sdev_show_##field (struct device *dev, char *buf)				\
{									\
	struct scsi_device *sdev;					\
	sdev = to_scsi_device(dev);					\
	return snprintf (buf, 20, format_string, sdev->field);		\
}									\

/*
 * sdev_rd_attr: macro to create a function and attribute variable for a
 * read only field.
 */
#define sdev_rd_attr(field, format_string)				\
	sdev_show_function(field, format_string)				\
static DEVICE_ATTR(field, S_IRUGO, sdev_show_##field, NULL)


/*
 * sdev_rd_attr: create a function and attribute variable for a
 * read/write field.
 */
#define sdev_rw_attr(field, format_string)				\
	sdev_show_function(field, format_string)				\
									\
static ssize_t								\
sdev_store_##field (struct device *dev, const char *buf, size_t count)	\
{									\
	struct scsi_device *sdev;					\
	sdev = to_scsi_device(dev);					\
	snscanf (buf, 20, format_string, &sdev->field);			\
	return count;							\
}									\
static DEVICE_ATTR(field, S_IRUGO | S_IWUSR, sdev_show_##field, sdev_store_##field)

/*
 * sdev_rd_attr: create a function and attribute variable for a
 * read/write bit field.
 */
#define sdev_rw_attr_bit(field)						\
	sdev_show_function(field, "%d\n")					\
									\
static ssize_t								\
sdev_store_##field (struct device *dev, const char *buf, size_t count)	\
{									\
	int ret;							\
	struct scsi_device *sdev;					\
	ret = scsi_sdev_check_buf_bit(buf);				\
	if (ret >= 0)	{						\
		sdev = to_scsi_device(dev);				\
		sdev->field = ret;					\
		ret = count;						\
	}								\
	return ret;							\
}									\
static DEVICE_ATTR(field, S_IRUGO | S_IWUSR, sdev_show_##field, sdev_store_##field)

/*
 * scsi_sdev_check_buf_bit: return 0 if buf is "0", return 1 if buf is "1",
 * else return -EINVAL.
 */
static int scsi_sdev_check_buf_bit(const char *buf)
{
	if ((buf[1] == '\0') || ((buf[1] == '\n') && (buf[2] == '\0'))) {
		if (buf[0] == '1')
			return 1;
		else if (buf[0] == '0')
			return 0;
		else 
			return -EINVAL;
	} else
		return -EINVAL;
}

/*
 * Create the actual show/store functions and data structures.
 */
sdev_rd_attr (device_blocked, "%d\n");
sdev_rd_attr (queue_depth, "%d\n");
sdev_rd_attr (type, "%d\n");
sdev_rd_attr (scsi_level, "%d\n");
sdev_rd_attr (vendor, "%.8s\n");
sdev_rd_attr (model, "%.16s\n");
sdev_rd_attr (rev, "%.4s\n");
sdev_rw_attr_bit (online);

static ssize_t
store_rescan_field (struct device *dev, const char *buf, size_t count) 
{
	scsi_rescan_device(dev);
	return count;
}

static DEVICE_ATTR(rescan, S_IWUSR, NULL, store_rescan_field)

static ssize_t sdev_store_delete(struct device *dev, const char *buf,
				 size_t count)
{
 struct scsi_device *sdev = to_scsi_device(dev);

 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] sdev_store_delete() Begin .. \n"));
 /*
  * FIXME and scsi_proc.c: racey use of access_count,
  */
 if (atomic_read(&sdev->access_count))
		return -EBUSY;
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] sdev_store_delete() : scsi_remove_device() .. \n"));
 scsi_remove_device(sdev);
 return count;
};
static DEVICE_ATTR(delete, S_IWUSR, NULL, sdev_store_delete);

/* Default template for device attributes.  May NOT be modified */
static struct device_attribute *scsi_sysfs_sdev_attrs[] = {
	&dev_attr_device_blocked,
	&dev_attr_queue_depth,
	&dev_attr_type,
	&dev_attr_scsi_level,
	&dev_attr_vendor,
	&dev_attr_model,
	&dev_attr_rev,
	&dev_attr_online,
	&dev_attr_rescan,
	&dev_attr_delete,
	NULL
};


static struct device_attribute *attr_overridden(
		struct device_attribute **attrs,
		struct device_attribute *attr)
{
	int i;

	if (!attrs)
		return NULL;
 #ifndef NO_SYS_FS
	for (i = 0; attrs[i]; i++)
		if (!strcmp(attrs[i]->attr.name, attr->attr.name))
			return attrs[i];
 #endif // ndef NO_SYS_FS
	return NULL;
}

static int attr_add(struct device *dev, struct device_attribute *attr)
{
	struct device_attribute *base_attr;

	/*
	 * Spare the caller from having to copy things it's not interested in.
	 */
	base_attr = attr_overridden(scsi_sysfs_sdev_attrs, attr);
	if (base_attr) {
		/* extend permissions */
		#ifndef NO_SYS_FS
	    // struct attribute on sysfs.h
			attr->attr.mode |= base_attr->attr.mode;
		#endif // ndef NO_SYS_FS

		/* override null show/store with default */
		if (!attr->show)
			attr->show = base_attr->show;
		if (!attr->store)
			attr->store = base_attr->store;
	}

	return device_create_file(dev, attr);
}

/**
 * scsi_device_register - register a scsi device with the scsi bus
 * @sdev:	scsi_device to register
 *
 * Return value:
 * 	0 on Success / non-zero on Failure
 **/
int scsi_device_register(struct scsi_device *sdev)
{
 int error = 0, i;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_device_register() Begin .. \n"));
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_device_register() Begin .. \n"));
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_device_register() : sdev=%08XH \n",
	sdev));

 set_bit(SDEV_ADD, &sdev->sdev_state);
 device_initialize(&sdev->sdev_gendev);
 sprintf(sdev->sdev_gendev.bus_id,"%d:%d:%d:%d",
		sdev->host->host_no, sdev->channel, sdev->id, sdev->lun);

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_device_register() : &sdev->sdev_gendev=%08XH \n",
	&sdev->sdev_gendev));

 sdev->sdev_gendev.parent = &sdev->host->shost_gendev;
 sdev->sdev_gendev.bus = &scsi_bus_type;
 #ifndef USBMS_ONLY_ONE_DISK
	sdev->sdev_gendev.release = scsi_device_dev_release;
 #endif // ndef USBMS_ONLY_ONE_DISK

 class_device_initialize(&sdev->sdev_classdev);
 sdev->sdev_classdev.dev = &sdev->sdev_gendev;
 sdev->sdev_classdev.class = &sdev_class;
 snprintf(sdev->sdev_classdev.class_id, BUS_ID_SIZE, "%d:%d:%d:%d",
		sdev->host->host_no, sdev->channel, sdev->id, sdev->lun);

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_device_register() : &sdev->sdev_gendev=%08XH \n",
	&sdev->sdev_gendev));
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_device_register() : device_add().. \n"));
 error = device_add(&sdev->sdev_gendev);
 if (error) {
	/*
	printk(KERN_INFO "error 1\n");
	*/
	dprintf("[USBMS] error on add device !! \n");
	return error;
	}

 get_device(sdev->sdev_gendev.parent);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_device_register() call class_device_add() .. \n"));
 error = class_device_add(&sdev->sdev_classdev);
 if (error) {
	/*
	printk(KERN_INFO "error 2\n");
	*/
	dprintf("[USBMS] error on add class device !! \n");
	goto clean_device;
	return error;
   }

 get_device(&sdev->sdev_gendev);

 if(sdev->host->hostt->sdev_attrs) {
	for(i = 0; sdev->host->hostt->sdev_attrs[i]; i++) {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_device_register() : attr_add() .. \n"));
		error = attr_add(&sdev->sdev_gendev,
					sdev->host->hostt->sdev_attrs[i]);
		if (error)
		   {
			dprintf("[USBMS] error on add attr !! \n");
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_device_register() : scsi_remove_device() .. \n"));
			scsi_remove_device(sdev);
		   }
	   }
   }
	
 for(i = 0; scsi_sysfs_sdev_attrs[i]; i++) {
	if (!attr_overridden(sdev->host->hostt->sdev_attrs,
				scsi_sysfs_sdev_attrs[i])) {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_device_register() : device_create_file() .. \n"));
		error = device_create_file(&sdev->sdev_gendev,
					scsi_sysfs_sdev_attrs[i]);
		if (error)
		   {
			dprintf("[USBMS] error on add create file !! \n");
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_device_register() : scsi_remove_device() .. \n"));
			scsi_remove_device(sdev);
		   }
	   }
   }

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_device_register() End. \n"));
 return error;

 clean_device:
	device_del(&sdev->sdev_gendev);
 
 // #ifndef _ETCPIP
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_device_register() : put_device() .. \n"));
 put_device(&sdev->sdev_gendev);
 // #endif // ndef _ETCPIP
 
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_device_register() End with Fail !! \n"));
 return error;

}

/**
 * scsi_remove_device - unregister a device from the scsi bus
 * @sdev:	scsi_device to unregister
 **/
void scsi_remove_device(struct scsi_device *sdev)
{
 struct class *class = class_get(&sdev_class);


 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_remove_device() Begin .. \n"));
 class_device_unregister(&sdev->sdev_classdev);

 if (class) {
	/*
	#ifndef _ETCPIP
	// down_read is on include/linux/rwsem.h
		down_write(&class->subsys.rwsem);
	#endif // ndef _ETCPIP
	*/
	set_bit(SDEV_DEL, &sdev->sdev_state);
	if (sdev->host->hostt->slave_destroy)
			sdev->host->hostt->slave_destroy(sdev);
		if (atomic_read(&sdev->access_count)) {
			USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_remove_device() : &sdev->sdev_gendev=%08XH \n",
				&sdev->sdev_gendev));
			USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_remove_device() : device_del() .. \n"));
			device_del(&sdev->sdev_gendev);
		   }
	/*
	#ifndef _ETCPIP
		// down_read is on include/linux/rwsem.h
		up_write(&class->subsys.rwsem);
	#endif // ndef _ETCPIP
	*/
   }

 
 // #ifndef _ETCPIP
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_remove_device() : put_device() .. \n"));
 put_device(&sdev->sdev_gendev);
 // #endif // ndef _ETCPIP
 

 class_put(&sdev_class);
}

int scsi_register_driver(struct device_driver *drv)
{
 drv->bus = &scsi_bus_type;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_register_driver() Begin .. \n"));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_register_driver() End with call driver_register() .. \n")) ;
 return driver_register(drv);
}

int scsi_register_interface(struct class_interface *intf)
{
	intf->class = &sdev_class;

	return class_interface_register(intf);
}


static struct class_device_attribute *class_attr_overridden(
		struct class_device_attribute **attrs,
		struct class_device_attribute *attr)
{
	int i;

	if (!attrs)
		return NULL;
 #ifndef NO_SYS_FS
	for (i = 0; attrs[i]; i++)
	   {
		if (!strcmp(attrs[i]->attr.name, attr->attr.name))
			return attrs[i];
	   }
 #endif // ndef NO_SYS_FS
	return NULL;
}

static int class_attr_add(struct class_device *classdev,
		struct class_device_attribute *attr)
{
 struct class_device_attribute *base_attr;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] class_attr_add() Begin .. \n"));
	/*
	 * Spare the caller from having to copy things it's not interested in.
	 */
	base_attr = class_attr_overridden(scsi_sysfs_shost_attrs, attr);
	if (base_attr) {
		/* extend permissions */
		#ifndef NO_SYS_FS
		    // struct attribute on sysfs.h
			attr->attr.mode |= base_attr->attr.mode;
		#endif // ndef NO_SYS_FS

		/* override null show/store with default */
		if (!attr->show)
			attr->show = base_attr->show;
		if (!attr->store)
			attr->store = base_attr->store;
	}

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] class_attr_add() End with call class_device_create_file() .. \n"));
 return class_device_create_file(classdev, attr);
}

/**
 * scsi_sysfs_add_host - add scsi host to subsystem
 * @shost:     scsi host struct to add to subsystem
 * @dev:       parent struct device pointer
 **/
// #ifndef NO_SYS_FS
int scsi_sysfs_add_host(struct Scsi_Host *shost)
{
 int error, i;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_sysfs_add_host() Begin .. \n"));
 if(shost->hostt->shost_attrs) {
	for(i = 0; shost->hostt->shost_attrs[i]; i++) {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_sysfs_add_host() : class_attr_add() .. \n"));
		error = class_attr_add(&shost->shost_classdev,
					shost->hostt->shost_attrs[i]);
		if (error)
		   {
			USBMS_DBG_INFO(iprintf("[USBMS FLOW] scsi_sysfs_add_host() : class_attr_add() Fail !! \n"));
			return error;
		   }
	   }
   }

 for (i = 0; scsi_sysfs_shost_attrs[i]; i++) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_sysfs_add_host() : class_attr_overridden() .. \n"));
	if (!class_attr_overridden(shost->hostt->shost_attrs,
					scsi_sysfs_shost_attrs[i])) {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_sysfs_add_host() : class_device_create_file() .. \n"));
		error = class_device_create_file(&shost->shost_classdev,
					scsi_sysfs_shost_attrs[i]);
		if (error)
		   {
			USBMS_DBG_INFO(iprintf("[USBMS FLOW] scsi_sysfs_add_host() : class_device_create_file() Fail !! \n"));
			return error;
		   }
	   }
   }

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_sysfs_add_host() End. \n"));
 return 0;
}
// #endif // ndef NO_SYS_FS

// B. S. Liaw TEST for compiler warning !!

