

// Name of file : UsbLptIo.c
// ====================================================================

/*
 * usblp.c  Version 0.13
 *
 * Copyright (c) 1999 Michael Gee	<michael@linuxspecific.com>
 * Copyright (c) 1999 Pavel Machek	<pavel@suse.cz>
 * Copyright (c) 2000 Randy Dunlap	<rddunlap@osdl.org>
 * Copyright (c) 2000 Vojtech Pavlik	<vojtech@suse.cz>
 # Copyright (c) 2001 Pete Zaitcev	<zaitcev@redhat.com>
 # Copyright (c) 2001 David Paschal	<paschal@rcsis.com>
 *
 * USB Printer Device Class driver for USB printers and printer cables
 *
 * Sponsored by SuSE
 *
 * ChangeLog:
 *	v0.1 - thorough cleaning, URBification, almost a rewrite
 *	v0.2 - some more cleanups
 *	v0.3 - cleaner again, waitqueue fixes
 *	v0.4 - fixes in unidirectional mode
 *	v0.5 - add DEVICE_ID string support
 *	v0.6 - never time out
 *	v0.7 - fixed bulk-IN read and poll (David Paschal)
 *	v0.8 - add devfs support
 *	v0.9 - fix unplug-while-open paths
 *	v0.10- remove sleep_on, fix error on oom (oliver@neukum.org)
 *	v0.11 - add proto_bias option (Pete Zaitcev)
 *	v0.11 - add hpoj.sf.net ioctls, MODULE_*_USE_COUNT, 2.2 backport (David Paschal)
 *	v0.12 - add hpoj.sourceforge.net ioctls (David Paschal)
 *	v0.13 - alloc space for statusbuf (<status> not on stack);
 *		use usb_buffer_alloc() for read buf & write buf;
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifdef _ETCPIP
#ifdef _NEWLIB
  #include "sys_type.h"
  #include "time.h"
#endif
	#include "feature.h"
   #define __KERNEL__
           #ifdef _NEWLIB
          #include "string.h"
        #endif
   	#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
          // for deciding whether to compile
          //  Debug Version or Release Version
   #include "etcpip.h"
   #include "UsbDbg.h"
   #include "UsbLptDbg.h"
   
   #include "PoList.h" // osport_list.h
   
   #include "UsbCfg.h" // Config and Setting
   #include "UsOsport.h" // "usb_osport.h"
   #ifndef NO_WAITQUEUE
		#include  "PoWait.h" // osport_wait.h
   #endif // ndef NO_WAITQUEUE
   #include "UsMemLib.h"
#endif // _ETCPIP

#ifndef _ETCPIP
   #include <linux/module.h>
   #include <linux/kernel.h>
   #include <linux/sched.h>
   #include <linux/smp_lock.h>
   #include <linux/signal.h>
   #include <linux/poll.h>
   #include <linux/init.h>
   #include <linux/slab.h>
#endif // _ETCPIP

#ifndef _ETCPIP
   #include <linux/lp.h>
 #else
   #include "lp.h"
#endif
#ifndef _ETCPIP
   #include <linux/devfs_fs_kernel.h>
#endif // ndef _ETCPIP

/*
#ifndef _ETCPIP
   #undef DEBUG
 #else
   #ifdef CONFIG_USB_DEBUG
      #define DEBUG
    #else
	  #undef DEBUG
   #endif
#endif
*/
#ifndef _ETCPIP
   #include <linux/usb.h>
 #else
   #include "usb.h"
	#include "UsbSysInit.h"
#endif

#ifdef USB_NO_ISR_FREE
	#ifndef NO_TIMER
		#include "PoTimer.h" // osport_timer.h
	#endif // ndef NO_TIMER
	#include "PoList.h" // osport_list.h
	#include "UsbHcd.h"  // ../core/hcd.h
#endif // USB_NO_ISR_FREE

#ifdef RH_INT_BY_POLL
	#include "hub.h"
#endif

// #ifdef _ETCPIP
#include "prdrv.h"

#include "UsbLpt.h" // it is modified from w83877tf.h
                    // Printer Hardware Layer driver Header for USB
#include "UsbLptIo.h"
#include "UsbSysInit.h"
// #endif // _ETCPIP
/*
#if defined(USB_OHCI)
	#include "ohci.h" // not usb-ohci.h
#elif defined(USB_SOTA)   // Sota USB 1.1 Host
	// 2003/10/23 B. S. Liaw
	#include "sota.h"
#endif
*/
#include "hcdx.h" // ohci.h not usb-ohci.h , sota.h

#ifdef _USB_SYSLOG
	#include "syslog.h"
	#include "logStr.h"
	#include "config.h"
	extern struct III_Config_t gSetting;
 #else
	// #error: Please Make Sure : Not _USB_SYSLOG !!
	//#warning: Not _USB_SYSLOG !!
#endif // _SYSLOG

#ifdef _NB_SHARE
#ifndef _USB_SYSLOG
#include "config.h"
#endif // _USB_SYSLOG
#include "nbios.h"
#endif // _NB_SHARE

// #define USBLPT_READ_INFINITE

// ====================================================================

#ifndef _ETCPIP
	/*
	static int usblp_release(struct inode *inode, struct file *file) ;
    */
 #else
	static int usblp_release(struct usblp *usblp) ;
	// static int usblp_release(struct inode *inode, 
    // 							struct usblp *usblp)
#endif

// ====================================================================

/*
 * Version Information
 */
#ifndef _ETCPIP
    #define DRIVER_VERSION "v0.13"
	#define DRIVER_AUTHOR "Michael Gee, Pavel Machek, Vojtech Pavlik, Randy Dunlap, Pete Zaitcev, David Paschal"
	#define DRIVER_DESC "USB Printer Device Class driver"
#endif // ndef _ETCPIP

/*  Move to  UsbLptIo.h
#define USBLP_BUF_SIZE		8192
#define USBLP_DEVICE_ID_SIZE	1024
*/

/*
#ifdef _ETCPIP
	unsigned char device_id_array[DEVICE_ID_SIZE] ;
	unsigned char usblp_buf_array[sizeof(struct usblp)] ;
	unsigned char usblp_buf_array[USBLP_BUF_SIZE * 2] ;
#endif // _ETCPIP
*/

// #ifdef _ETCPIP
struct usblp *global_usblp = 0;
#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	static int volatile usblp_rd_done = 0 ;
#endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw
#ifndef USBLPT_PRINTERS  /* Charles 20040706 */
static int volatile usblp_wr_done = 0 ;
static int volatile usb_lpt_released = 0 ;
static int volatile usb_lpt_access = 0 ;
#endif
#ifdef USBLPT_WAIT_FOR_ADSL
	#error: Please Donot Enable USBLPT_WAIT_FOR_ADSL !!
	// int	volatile usblpt_want_write = 0 ;
	int	volatile usblpt_write = 0 ;
#endif // USBLPT_WAIT_FOR_ADSL
// #endif // _ETCPIP


// #ifdef DEBUG
#ifdef USBLPT_DBG_INIT_ON
   static void usblp_dump(struct usblp *usblp)
   {
	int p;

	iprintf("[USBLPT INIT] ====== LPT Dump ===============\n");
	iprintf("[USBLPT INIT] usblp=0x%p\n", usblp);
	iprintf("[USBLPT INIT] dev=0x%p\n", usblp->usb_dev);
	#ifdef CONFIG_USB_DEVICEFS
		dbg("[USBLPT INIT] devfs=0x%p\n", usblp->usb_devfs);
	#endif // CONFIG_USB_DEVICEFS
	iprintf("[USBLPT INIT] writebuf=0x%p\n", usblp->writebuf);
	#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
		iprintf("[USBLPT INIT] readbuf=0x%p\n", usblp->readbuf);
	#endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	iprintf("[USBLPT INIT] statusbuf=0x%p\n", usblp->statusbuf);
	#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
		iprintf("[USBLPT INIT] readcount=%d\n", usblp->readcount);
	#endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	iprintf("[USBLPT INIT] ifnum=%d\n", usblp->ifnum);
    for(p = USBLPT_FIRST_PROTOCOL; p <= USBLPT_LAST_PROTOCOL; p++) {
		iprintf("[USBLPT INIT] protocol[%d].alt_setting=%d ", p, usblp->protocol[p].alt_setting);
		iprintf("protocol[%d].epwrite=%p ", p, usblp->protocol[p].epwrite);
		#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
			iprintf("protocol[%d].epread=%p ", p, usblp->protocol[p].epread);
		#endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw
		iprintf("\n");
       }
	iprintf("[USBLPT INIT] current_protocol=%d\n", usblp->current_protocol);
	iprintf("[USBLPT INIT] minor=%d\n", usblp->minor);
	iprintf("[USBLPT INIT] wcomplete=%d ", usblp->wcomplete);
	#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
		iprintf(" rcomplete=%d ", usblp->rcomplete);
	#endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	iprintf(" quirks=%d ", usblp->quirks);
	iprintf(" opened=%d \n", usblp->opened);
	#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
		iprintf(" bidir=%d\n", usblp->bidir);
	#endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	iprintf("[USBLPT INIT] device_id_string=[%s] \n",
		usblp->device_id_string ?
			usblp->device_id_string + 2 :
			(unsigned char *)"(null)");
	iprintf("[USBLPT INIT] ===============================\n");
   }
#endif

/*
#ifdef CONFIG_USB_DEVICEFS
   extern devfs_handle_t usb_devfs_handle;		// /dev/usb dir. 
#endif // CONFIG_USB_DEVICEFS
*/
/*
#ifndef _ETCPIP
	struct usblp *usblp_table[USBLP_MINORS];
#endif // ndef _ETCPIP
*/


/* Quirks: various printer quirks are handled by this table & its flags. */
/*
struct quirk_printer_struct {
	__u16 vendorId;
	__u16 productId;
	unsigned int quirks;
};

#define USBLP_QUIRK_BIDIR		0x1	// reports bidir but requires unidirectional mode (no INs/reads) 
#define USBLP_QUIRK_USB_INIT	0x2	// needs vendor USB init string 
*/

static struct quirk_printer_struct quirk_printers[] = {
 { 0x03f0, 0x0004, USBLPT_QUIRK_BIDIR }, /* HP DeskJet 895C */
 { 0x03f0, 0x0104, USBLPT_QUIRK_BIDIR }, /* HP DeskJet 880C */
 { 0x03f0, 0x0204, USBLPT_QUIRK_BIDIR }, /* HP DeskJet 815C */
 { 0x03f0, 0x0304, USBLPT_QUIRK_BIDIR }, /* HP DeskJet 810C/812C */
 { 0x03f0, 0x0404, USBLPT_QUIRK_BIDIR }, /* HP DeskJet 830C */
 { 0x03f0, 0x0504, USBLPT_QUIRK_BIDIR }, /* HP DeskJet 885C */
 { 0x03f0, 0x0604, USBLPT_QUIRK_BIDIR }, /* HP DeskJet 840C */   
 { 0x03f0, 0x0804, USBLPT_QUIRK_BIDIR }, /* HP DeskJet 816C */   
 { 0x03f0, 0x1104, USBLPT_QUIRK_BIDIR }, /* HP Deskjet 959C */
 { 0x0409, 0xefbe, USBLPT_QUIRK_BIDIR }, /* NEC Picty900 (HP OEM) */
 { 0x0409, 0xbef4, USBLPT_QUIRK_BIDIR }, /* NEC Picty760 (HP OEM) */
 { 0x0409, 0xf0be, USBLPT_QUIRK_BIDIR }, /* NEC Picty920 (HP OEM) */
 { 0x0409, 0xf1be, USBLPT_QUIRK_BIDIR }, /* NEC Picty800 (HP OEM) */

 { 0x04b8, 0x0005, USBLPT_QUIRK_USB_INIT // EPSON STYLUS C41UX
                  | USBLPT_QUIRK_INIT_EPSON  }, // EPSON STYLUS Color 980

 { 0x04b8, 0x0001, USBLPT_QUIRK_USB_INIT // EPSON Stylus Scan 2500
                  | USBLPT_QUIRK_INIT_EPSON  },
 { 0x04b8, 0x0801,  USBLPT_QUIRK_INTF_1      // EPSON Stylus CX5100
                  | USBLPT_QUIRK_USB_INIT
                  | USBLPT_QUIRK_INIT_EPSON 
                  | USBLPT_QUIRK_NO_1284ID } ,

 /*
 // HP LaserJet 1000 need use LaserJet 4L/5L/6L Driver 
 //                  LJ1000 Driver Cannot Set to lan port !!
 { 0x03f0, 0x0517 , // USBLP_QUIRK_BIDIR  // HP LaserJet 1000
                  // | 
                  USBLP_QUIRK_USB_INIT
                  | USBLP_QUIRK_INIT_HP1000 },
 */
 { 0x03f0, 0x0517 , USBLPT_QUIRK_UNSUPPORT},  // HP LaserJet 1000
 
 { 0x043D, 0x0059, USBLPT_QUIRK_BIDIR }, // Lexmark E210 Laser Printer
 
 { 0x043D, 0x0096, USBLPT_QUIRK_INTF_1 | USBLPT_QUIRK_BIDIR }, // Lexmark X2250
 
 { 0x04F9, 0x0016, // USBLP_QUIRK_NORESET
				USBLPT_QUIRK_RESET11
 },					// Brother HL-5040
 
 { 0x04F9, 0x000D, USBLPT_QUIRK_NORESET },
 /* Brother HL-1440 can't be reset! Or the hub port will be disabled. Even after we
    re-enable the port, it still can't work! Charles 20050104 */

 /*
 { 0x03F0,0x1204, // USBLP_QUIRK_NORESET
				USBLP_QUIRK_RESET11
 }, 									// HP Deskjet 930C/935C
 */
 // { 0x03f0, 0x1017, USBLP_QUIRK_GET_ID }, // HP LaserJet 1300 
 /*
 { 0x04b8, 0x0005, USBLP_QUIRK_USB_INIT // EPSON STYLUS Color 980
                  | USBLP_QUIRK_INIT_EPSON  },
 */
 // ------------------------------------                
 { 0, 0 }
};

// ====================================================================

static int usblp_select_alts(struct usblp *usblp);
static int usblp_set_protocol(struct usblp *usblp, int protocol);
static int usblp_get_device_id_string(struct usblp *usblp);
static int usblp_cache_device_id_string(struct usblp *usblp);

// ====================================================================
/* forward reference to make our lives easier */
extern struct usb_driver usblp_driver;

// ====================================================================


// ====================================================================

// 2003/4/14 B. S. Liaw , This is important !!
// vendor USB init string 
// at least EPSON STYLUS C41UX Need this for exit packet mode
// for USBLP_QUIRK_INIT_EPSON
static char hdr_epson[] 
	= "\000\000\000\033\001@EJL 1284.4\n@EJL     \n\033@";
    
/*    // for USBLP_QUIRK_INIT_HP1000
#include "hp1000.c" // for hdr_hp1000
*/
#ifdef USBLPT_INIT_STR    // 2003/10/13 B. S. Liaw
   // static 
   unsigned char usblpt_init_string(struct usblp *usblp)
   // exit_packet_mode(void)
   {
	/*
	// 00 00 00 1b 01 40 45 4a 4c 20 31 32 38 34 2e 34 ..
	// static char hdr[] 
	static char hdr_epson[] 
		= "\000\000\000\033\001@EJL 1284.4\n@EJL     \n\033@";
	*/
	char *hdr ;
	// size_t hdr_size ;
	int hdr_size ;
	// unsigned long i;
	int i ;
	/*
	#ifdef _USB_SYSLOG
	    char usbsyslog_buf[100];
	#endif // _SYSLOG
	*/

 
	USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblpt_init_string() Begin ..\n" ));
 
	// memcpy(printer_cmd + bufpos, hdr, sizeof(hdr) - 1); 
	// DON'T include null! 
 
	// bufpos += sizeof(hdr) - 1;
 
	if (usblp->quirks & USBLPT_QUIRK_INIT_EPSON) {
		USBLPT_DBG_INFO(iprintf("[USBLPT INFO] QUIRK EPSON ..\n" )) ;
    
	    hdr = hdr_epson ;
	    hdr_size = sizeof(hdr_epson)-1 ; // string , need minus 1
    
		USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Epson Init String ..\n" )) ;

		iprintf("[USBLPT] Init Epson (or OEM) Printer ..\n") ;
		#ifdef _USB_SYSLOG
		    // sprintf(usbsyslog_buf,"[USBLPT] Init Epson Printer .. ");
						#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_PR_EPSON_INIT
				//sprintf_SetSystemLog(USB_SYS_LOG_PR_EPSON_INIT);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_PR_EPSON_INIT);
#else
				//sprintf_SetSystemLog("[USBLPT] Init Epson (or OEM) Printer .. ");
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] Init Epson (or OEM) Printer .. ");
#endif // USB_SYS_LOG_PR_EPSON_INIT
			}
		#endif // _SYSLOG
	   }
	/*
	else if (usblp->quirks & USBLP_QUIRK_INIT_HP1000)
	   {
		USBLPT_DBG_INFO(iprintf("[USBLPT INFO] QUIRK HP1000 ..\n" )) ;
    
	    hdr = hdr_hp1000 ;
	    hdr_size = HDR_HP1000_DATA_SZ ; // array , donnot need minus 1
    
	    USBLPT_DBG_INFO(iprintf("[USBLPT INFO] HP LJ 1000 Init String ..\n" )) ;
	   }
	*/
 
	USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblpt_init_string() call usblp_write() ..\n"));
	/*   
	i = usblp_write( usblp , hdr , sizeof(hdr)-1 ) ;
	*/   
	i = usblp_write( usblp , hdr , hdr_size ) ;
	// dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;

	USBLPT_DBG_INIT(iprintf(
		"[USBLPT INIT] usblpt_init_string() : hdr_size = %d , i = %d\n",
		hdr_size , i
	    ));
        
	// if (i < sizeof(hdr)-1)
	if (i < hdr_size) { 
		USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Epson Init String Fail !!\n" )) ;
		USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblpt_init_string() End with Fail !!\n"));
		return(D_False);
	   }
       
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Vendor Init String OK. \n")) ;
	 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblpt_init_string() End.\n")) ;
	return(D_OK);
   } // function usb_init_string()
#endif // USBLPT_INIT_STR


// ====================================================================

// ====================================================================

/*
 * Functions for usblp control messages.
 */

static int usblp_ctrl_msg(struct usblp *usblp, int request, int type,
						 int dir,
						 int recip, int value, void *buf, int len)
{
 int retval ;
 
 
 USBLPT_DBG_FLOW(iprintf(
    "[USBLPT FLOW] usblp_ctrl_msg() Begin ..\n"));
    
 if (dir)
   {
	USBLPT_DBG_PIPE(iprintf(
    	"[USB PIPE] usblp_ctrl_msg() : usb_rcvctrlpipe"
        "(dev,0) = %08XH \n",
    	usb_rcvctrlpipe(usblp->usb_dev, 0))) ;
        
 	USBLPT_DBG_PIPE(iprintf(
    	"[USB PIPE] usblp_ctrl_msg() : usb_pipedevice(usb_rcvctrlpipe"
        "(dev, 0)) = %d \n",
    	usb_pipedevice(usb_rcvctrlpipe(usblp->usb_dev, 0))
    	)) ;
   }
  else
   {
	USBLPT_DBG_PIPE(iprintf(
    	"[USB PIPE] usb_sndctrlpipe(dev, 0) = %08XH \n",
    	usb_sndctrlpipe(usblp->usb_dev, 0))) ;
        
 	USBLPT_DBG_PIPE(iprintf(
    	"[USB PIPE] usb_pipedevice(usb_sndctrlpipe(dev, 0)) = %d \n",
    	usb_pipedevice(usb_sndctrlpipe(usblp->usb_dev, 0))
    	)) ;
    }
 /*   
 USBLPT_DBG_INIT(iprintf(
    "[USBLPT INIT] usblp_ctrl_msg() : HZ=%d , HZ*5=%d \n",HZ,HZ*5));
 */   
 USBLPT_DBG_URB(iprintf("[USBLPT URB] usblp_ctrl_msg() : USBLPT_CTLMSG_TIMEOUT=%d \n",
    USBLPT_CTLMSG_TIMEOUT
    ));
    
 USBLPT_DBG_URB(iprintf("[USB URB] usblp_ctrl_msg() : Control Message .. \n"
    )) ;

 #ifdef USB_NO_VMM // no vitrtul memory management
	#ifdef USB_BS_ASSERT
		if (   (unsigned long)buf 
    	     > (unsigned long)0x00L
		   ) {
			/*
			#if defined(_S3C2510)
				if (!(  (unsigned long)buf
        	          > ((unsigned long)0x00800000L)
    			     )
		    	   )
			#elif defined(_ADM5120)
				if (!(  (  (unsigned long)buf
						 & (unsigned long)0x20000000L)
    		          ==((unsigned long)0x20000000L)
    			     )
		    	   )
			#endif
				   {
			*/
		if ( !(USB_IS_UN_CACHED(buf)) ) {
					iprintf("[ASSERT][USB] usblp_ctrl_msg() : " ) ;
    			    iprintf("Error : Not Un-Cached !!\n" ) ;
				    iprintf("[ASSERT][USB] buf addr = %08XH \n", buf ) ;
				    iprintf("[ASSERT][USB] Continue ..\n") ;
    			   }
		   }
	#endif // USB_BS_ASSERT
 #endif // USB_NO_VMM

 USBLPT_DBG_FLOW(iprintf(
    "[USBLPT FLOW] usblp_ctrl_msg() Call usb_control_msg() ..\n"));
 
 // urb->complete = usb_api_blocking_completion ;
    
 retval = usb_control_msg(usblp->usb_dev,
		dir ? usb_rcvctrlpipe(usblp->usb_dev, 0)
            : usb_sndctrlpipe(usblp->usb_dev, 0),
		// request, USB_TYPE_CLASS | dir | recip, value, usblp->ifnum,
        // buf, len, HZ * 5);
		request, type | dir | recip, value, usblp->ifnum,
        buf, len, USBLPT_CTLMSG_TIMEOUT);
        
 USBLPT_DBG_INIT(iprintf(
	"[USBLPT INIT] usblp_control_msg() : rq:%02XH dir:%d recip:%d "
	"value:%d len:%d result:%d \n",
	request, !!dir, recip, value, len, retval
    ));
    
 USBLPT_DBG_FLOW(iprintf(
    "[USBLPT FLOW] usblp_ctrl_msg() End.\n"));
 return retval < 0 ? retval : 0;
} // function usblp_ctrl_msg()

// ====================================================================

#define usblp_read_status(usblp, status) \
	usblp_ctrl_msg(usblp, USBLPT_REQ_GET_STATUS, USB_TYPE_CLASS, \
    USB_DIR_IN, \
    USB_RECIP_INTERFACE, 0, status, 1)
    
// ====================================================================

#define usblp_get_id(usblp, config, id, maxlen) \
	usblp_ctrl_msg(usblp, USBLPT_REQ_GET_ID, USB_TYPE_CLASS, \
    USB_DIR_IN, \
    USB_RECIP_INTERFACE, config, id, maxlen)

// ====================================================================

/*    
#define usblp_reset(usblp)\
	usblp_ctrl_msg(usblp, USBLP_REQ_RESET, USB_TYPE_CLASS, \
    USB_DIR_OUT, \
    USB_RECIP_OTHER, 0, NULL, 0)
*/

#ifdef USBLPT_RESET_CMD // 2003/10/14 B. S. Liaw
   // return : < 0 : error

   int usblp_reset(struct usblp *the_usblp) {
	/*
	#ifdef _USB_SYSLOG
	    char usbsyslog_buf[100];
	#endif // _SYSLOG
	*/


    if (!the_usblp)
       {
		USBLPT_DBG_INFO(iprintf(
			"[USBLPT INFO] usblp_reset() : Null the_usblp !! \n"));
		return -1 ;
	   }

	// if (!the_usblp->support_reset) {
    if (the_usblp->quirks & USBLPT_QUIRK_NORESET) {
		USBLPT_DBG_INFO(iprintf(
			"[USBLPT INFO] usblp_reset() : Not Do Anything !! \n"));
		iprintf("[USBLPT] Dummy Reset .. \n");

		#ifdef _USB_SYSLOG
		    //sprintf(usbsyslog_buf,"[USBLPT] Dummy Reset .. ");
						#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_PR_RESER_DUMMY
				//sprintf_SetSystemLog(USB_SYS_LOG_PR_RESER_DUMMY);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_PR_RESER_DUMMY);
#else
				//sprintf_SetSystemLog("[USBLPT] Dummy Reset .. ");
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] Dummy Reset .. ");
#endif // USB_SYS_LOG_PR_RESER_DUMMY
			}
		#endif // _SYSLOG

		return 1 ;
	   }
	else if (the_usblp->quirks & USBLPT_QUIRK_RESET11) { // 2003/10/17 B. S. Liaw
	   
		// USB Device Class Definition for Printing Devices
		// 4.2.3 SOFT_RESET command
		// Version 1.0 :  0x23 , USB_RECIP_OTHER     0x03
		// Version 1.1 :  0x21 , USB_RECIP_INTERFACE 0x01
		// 1.1 , [USB URB] setup(8): 21 02 00 00 00 00 00 00
		return usblp_ctrl_msg(the_usblp, USBLPT_REQ_RESET, USB_TYPE_CLASS,
			 USB_DIR_OUT,   USB_RECIP_INTERFACE, 0, NULL, 0) ;
       }
	else {
		// Now 1.0 , [USB URB] setup(8): 23 02 00 00 00 00 00 00
		return usblp_ctrl_msg(the_usblp, USBLPT_REQ_RESET, USB_TYPE_CLASS,
			 USB_DIR_OUT,   USB_RECIP_OTHER, 0, NULL, 0) ;
       }
   } // function usblp_reset()
#endif // USBLPT_RESET_CMD

// ====================================================================

#define usblp_hp_channel_change_request(usblp, channel, buffer) \
	usblp_ctrl_msg(usblp, USBLPT_REQ_HP_CHANNEL_CHANGE_REQUEST, \
    USB_TYPE_VENDOR, USB_DIR_IN, USB_RECIP_INTERFACE, channel, \
    buffer, 1)

// ====================================================================

/*
 * See the description for usblp_select_alts() below for the usage
 * explanation.  Look into your /proc/bus/usb/devices and dmesg in
 * case of any trouble.
 */
static int proto_bias = -1;


// ====================================================================

/*
 * URB callback.
 */

#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
#ifndef _ETCPIP
	/*
	static void usblp_bulk_read(struct urb *urb, struct pt_regs *regs)
    */
#else
	static void usblp_bulk_read(struct urb *urb)
#endif
{
 struct usblp *usblp = urb->context;

 // USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_bulk_read() Begin ..\n"));
 USB_DBG_CALLBACK(iprintf( "[USB CALLBACK] usblp_bulk_read() Begin ..\n"));
 USBLPT_DBG_CALLBACK(iprintf("[USBLPT CALLBACK] usblp_bulk_read() Begin ..\n"));
 /*
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_bulk_read() : urb = %08XH *urb = %08XH\n",
	urb , *urb
    )) ;
 */
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_bulk_read() : urb= %08XH , "
    "&(urb->count)= %08XH urb->count= %08XH\n",
    urb , &(urb->count) , urb->count
    ));
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_bulk_read() : &urb->context = %08XH urb->context = "
    "%08XH \n",
	&urb->context , urb->context
    )) ;
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_bulk_read() : usblp = %08XH \n",
	usblp )) ;
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_bulk_read() : &usblp->usb_dev  = %08XH , usblp->usb_dev ="
    " %08XH \n",
	&usblp->usb_dev,usblp->usb_dev
    )) ;
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_bulk_read() : &usblp->opened = %08XH , usblp->opened ="
    " %d \n",
	&usblp->opened,usblp->opened
    )) ;
    
 if (!usblp || !usblp->usb_dev || (usblp->opened==0) )
   {
	dprintf("[USBLPT] Error : usblp_bulk_read() : !usblp !usblp->usb_dev "
    	"(usblp->opened==0) \n"
        );
        
	USBLPT_DBG_FLOW(iprintf(
    	"[USBLPT FLOW] usblp_bulk_read() End With Fail !!\n"));
	return;
   }

 #ifndef _ETCPIP
	if (unlikely(urb->status))
  #else
	if (urb->status)
 #endif
   {
	dprintf("[USBLPT] Warning : UsbLpt-%d: nonzero read bulk "
    	"status received: %d",
		usblp->minor, urb->status);
   }

 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	usblp->rcomplete = 1;
 #endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw
 
 // #ifdef _ETCPIP
 #ifdef USB_BS_ASSERT
		// 2003/7/14 B. S. Liaw
		if (!( usblp_rd_done == 0
    	     )
	       )
		   {
			iprintf("[ASSERT][USB] usblp_bulk_read() : " ) ;
    	    iprintf("Error : Called More than once !!\n" ) ;
	        iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
 #endif
 USBLPT_DBG_CTLMSG(iprintf("[USBLPT CTLMSG] usblp_bulk_read() : usblp_rd_done=%d\n"
   	,usblp_rd_done
    ));

 usblp_rd_done = 1 ;

 // #endif // _ETCPIP
 
 #ifndef NO_WAITQUEUE
	wake_up_interruptible(&usblp->wait);
 #endif // ndef NO_WAITQUEUE
 
 // USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_bulk_read() End.\n"));
 USBLPT_DBG_CALLBACK(iprintf("[USBLPT CALLBACK] usblp_bulk_read() End.\n"));
}
#endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw


#ifndef _ETCPIP
	static void usblp_bulk_write(struct urb *urb, struct pt_regs *regs)
#else
	static void usblp_bulk_write(struct urb *urb)
#endif
{
 struct usblp *usblp = urb->context;

 // USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_bulk_write() Begin ..\n"));
 USB_DBG_CALLBACK(iprintf("[USB CALLBACK] usblp_bulk_write() Begin ..\n"));
 USBLPT_DBG_CALLBACK(iprintf("[USBLPT CALLBACK] usblp_bulk_write() Begin ..\n"));
 /* 
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_bulk_write() : urb = %08XH *urb = %08XH\n",
	urb , *urb
    )) ;
 */
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_bulk_write() : urb= %08XH , "
    "&(urb->count)= %08XH urb->count= %08XH\n",
    urb , &(urb->count) , urb->count
    ));
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_bulk_write() : &urb->context = %08XH urb->context = "
    "%08XH \n",
	&urb->context , urb->context
    )) ;
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_bulk_write() : usblp = %08XH \n",
	usblp )) ;
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_bulk_write() : &usblp->usb_dev  = %08XH , usblp->usb_dev ="
    " %08XH \n",
	&usblp->usb_dev,usblp->usb_dev
    )) ;
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_bulk_write() : &usblp->opened = %08XH , usblp->opened ="
    " %d \n",
	&usblp->opened,usblp->opened
    )) ;
    
 if (!usblp || !usblp->usb_dev || (usblp->opened==0) )
   {
	dprintf("[USBLPT] Error : usblp_bulk_write() : !usblp !usblp->usb_dev (usblp->opened==0) \n" ) ;
        
	USBLPT_DBG_FLOW(iprintf(
    	"[USBLPT FLOW] usblp_bulk_write() End With Fail !!\n"));
	return;
   }

 #ifndef _ETCPIP
	if (unlikely(urb->status))
  #else
	if (urb->status)
 #endif
      {
		warn("UsbLpt-%d: nonzero write bulk status received: %d",
			usblp->minor, urb->status);
      }
      
 usblp->wcomplete = 1;
 
 #ifdef _ETCPIP
	#ifdef USB_BS_ASSERT
		// 2003/7/14 B. S. Liaw
		if (!( usblp_wr_done == 0
    	     )
	       )
		   {
			iprintf("[ASSERT][USB] usblp_bulk_write() : " ) ;
    	    iprintf("Error : Called More than once !!\n" ) ;
	        iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
	#endif
	usblp_wr_done = 1 ;
	USBLPT_DBG_CTLMSG(iprintf("[USBLPT CTLMSG] usblp_bulk_write() : usblp_wr_done=%d\n"
	    	,usblp_wr_done
    	    ));
 #endif // _ETCPIP
 
 #ifndef NO_WAITQUEUE
	wake_up_interruptible(&usblp->wait);
 #endif // ndef NO_WAITQUEUE
 // USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_bulk_write() End.\n"));
 USBLPT_DBG_CALLBACK(iprintf("[USBLPT CALLBACK] usblp_bulk_write() End.\n"));
}

// ====================================================================
// ====================================================================

/*
 * Get and print printer errors.
 */

static char *usblp_messages[] =
{
 "ok", "out of paper", "off-line", "on fire"
};

// 1 : 	LP_POUTPA
// 2 :  LP_PSELECD
// 3 :  LP_PERRORP

#ifndef _ETCPIP
	static 
#endif // ndef _ETCPIP
int usblp_check_status(struct usblp *usblp, int err)
{
 unsigned char status, newerr = 0;
 int error;
 /*
 #ifdef _USB_SYSLOG
    char usbsyslog_buf[100];
 #endif // _SYSLOG
 */


 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_check_status() Begin ..\n"));
    
 error = usblp_read_status (usblp, usblp->statusbuf);
 
 if (error < 0)
   {
	dprintf("[USBLPT] Error : UsbLpt-%d: error %d reading printer "
     	"status\n",
		usblp->minor, error);
	#ifndef _ETCPIP
        /*
		return 0 ;
        */
     #else
		return -1 ;
    #endif
   }
   
 status = *usblp->statusbuf;
 
 /* the followings are on lp.h
 #define LP_POUTPA	0x20  // unchanged input, active high 
						// USB : PAPER_EMPTY 1 : paper empty
                        
 #define LP_PSELECD	0x10  // unchanged input, active high 
						// USB : SELECT , 1 : selected
                        
 #define LP_PERRORP	0x08  // unchanged input, active low 
						// USB : NOT_ERROR , 1 : no error
 */
 
 // Normal Value : 18H = 0001 10000 , paper not empty ,
 //                                 selected , no error
 
 USBLPT_DBG_INIT(iprintf(
	"[USBLPT INIT] UsbLpt-%d status=%02XH \n", usblp->minor,
    status
    ));
    
 #ifdef USBLPT_DBG_INIT_ON
	if (status & LP_POUTPA)
	   {
		iprintf("[USBLPT INIT] UsbLpt-%d Paper Empty !! \n",usblp->minor);
	   }
	if (!(status & LP_PSELECD))
	   {
		iprintf("[USBLPT INIT] UsbLpt-%d Status : Not Selected !! \n",usblp->minor);
	   }
 #endif // USBLPT_DBG_INIT_ON
       
	if (!(status & LP_PERRORP))
	   {
		USBLPT_DBG_INFO(iprintf(
			"[USBLPT INFO] UsbLpt-%d Status : Not LP_PERRORP , Error !! \n",	usblp->minor));
		iprintf("[USBLPT] Printer Status Error , Out of Ink / Paper End / Paper Jam ?? \n");
		#ifdef _USB_SYSLOG
		    //sprintf(usbsyslog_buf,"[USBLPT] Printer Status Error , Out of Ink / Paper End / Paper Jam ?? ");
						#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_PR_STATUS_ERR
				//sprintf_SetSystemLog(USB_SYS_LOG_PR_STATUS_ERR);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_PR_STATUS_ERR);
#else
				//sprintf_SetSystemLog("[USBLPT] Printer Status Error , Out of Ink / Paper End / Paper Jam ?? ");
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] Printer Status Error , Out of Ink / Paper End / Paper Jam ?? ");
#endif // USB_SYS_LOG_PR_STATUS_ERR
			}
		#endif // _SYSLOG
	   }
 
 if (~status & LP_PERRORP)
   {
	newerr = 3;
	if (status & LP_POUTPA) newerr = 1;
	if (~status & LP_PSELECD) newerr = 2;
   }
   
 /*
 USBLPT_DBG_INFO(iprintf(
	"[USBLPT INFO] usblp%d Status : %s \n", usblp->minor,
    usblp_messages[newerr]
    ));
 */
 
 if (newerr != err)
    {
	 USBLPT_DBG_INIT(iprintf("[USBLPT INIT] UsbLpt-%d Status : %s \n", usblp->minor,	usblp_messages[newerr] ));
    }

 USBLPT_DBG_FLOWLOOP(iprintf("[USBLPT FLOWLOOP] usblp_check_status() End.\n"));
 return newerr;
}

// ====================================================================
// ====================================================================

/*
 * File op functions.
 */

#ifndef _ETCPIP
    /*
	static int usblp_open(struct inode *inode, struct file *file)
    */
 #else
	int usblp_open(struct usblp *usblp)
	// static int usblp_open(struct inode *inode, struct usblp *usblp)
#endif
{
 #ifndef _ETCPIP
	// int minor = MINOR(inode->i_rdev) - USBLP_MINOR_BASE;
	int minor = minor(inode->i_rdev);
  #else
	int minor = 0 ;
 #endif // ndef _ETCPIP
 #ifndef _ETCPIP
    struct usblp *usblp;
 #endif // ndef _ETCPIP
 struct usb_interface *intf;
 int retval;
 
 #ifdef _ETCPIP
	// 2003/4/8 B. S. Liaw
	int  retv ;
	int  actual_length ;
 #endif // _ETCPIP
 #ifndef USBLPT_READ_BY_URBTASK
	#ifdef USBLPT_READ_ON_PROBE  // 2003/10/02 B. S. Liaw
		unsigned long read_timeout ;
	#endif // USBLPT_READ_ON_PROBE
	struct usb_hcd		*hcd ;
 #endif // ndef USBLPT_READ_BY_URBTASK
 /*
 #ifdef _USB_SYSLOG
    char usbsyslog_buf[100];
 #endif // _SYSLOG
 */


 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_open() Begin ..\n"));
 /*
 #ifdef _ETCPIP
	usb_lpt_ok = 0 ;
 #endif // _ETCPIP
 */
 #ifdef USBLPT_PRINTERS
 if (usblp->usb_lpt_released)
 #else
 if (usb_lpt_released)
 #endif
   {
    USBLPT_DBG_INIT(iprintf(
			"[USBLPT INIT] Error : usblp_open() : Printer Driver "
            "Released !! \n"
        	)) ;
	return (-ENODEV) ;
   }

 #ifndef _ETCPIP
	if (minor < 0 || minor >= USBLPT_MINORS)
		return -ENODEV;
 #endif // ndef _ETCPIP

 #ifndef _ETCPIP
	lock_kernel();
 #endif // ndef _ETCPIP
 
 /* 
 #ifndef _ETCPIP
	usblp  = usblp_table[minor];
  #else
	// usblp  = usblp_table[USBLP_MINOR_BASE];
 #endif
 */
 USBLPT_DBG_MEM(iprintf("[USBLPT MEM] usblp_open() : usblp = %08XH \n",
	usblp
    )) ;
 #ifdef USBLPT_BS_ASSERT
	if ( !( usblp != 0
          )
       )
       {
		iprintf("[ASSERT][USBLPT] usblp_open() : " ) ;
        iprintf("Error : pointer usblp NULL !! \n") ;
		iprintf("[ASSERT][USBLPT] Continue .. \n" ) ;
	   }   
 #endif
 
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_open() : &usblp->usb_dev  = %08XH , usblp->usb_dev ="
    " %08XH \n",
	&usblp->usb_dev,usblp->usb_dev )) ;
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_open() : &usblp->opened = %08XH , usblp->opened ="
    " %d \n",
	&usblp->opened,usblp->opened )) ;
    
    
 retval = -ENODEV;
 
 #ifndef NO_SYS_FS
	intf = usb_find_interface(&usblp_driver, mk_kdev(USB_MAJOR,minor));
	if (!intf) {
		goto out;
	   }
	usblp = usb_get_intfdata (intf);
 #endif // ndef NO_SYS_FS
 
 if (!usblp || !usblp->usb_dev)
   {
    dprintf("[USBLPT] Error : No usblp or usblp->usb_dev !!\n") ;
	goto out;
   }

 retval = -EBUSY;
 if (usblp->opened)
   {
    USBLPT_DBG_INFO(iprintf("[USBLPT INFO] usblp_open() : Usb LPT already Opened !!\n")) ;
	retval = 0;
	goto out;
   }

 /*
  * TODO: need to implement LP_ABORTOPEN + O_NONBLOCK as 
          in drivers/char/lp.c ???
  * This is #if 0-ed because we *don't* want to fail an open
  * just because the printer is off-line.
  */
  
 #if 0
 	/*
	if ((retval = usblp_check_status(usblp, 0))) {
		retval = retval > 1 ? -EIO : -ENOSPC;
		goto out;
	}
    */
 #else
	retval = 0;
 #endif

 
 usblp->opened = -1 ;
 
 USBLPT_DBG_INIT(iprintf(
	"[USBLPT INIT] usblp_open() : Set Value : usblp->opened = %d \n",
	usblp->opened )) ;


 #ifdef _ETCPIP
 	#ifdef USBLPT_PRINTERS
 	usblp->usb_lpt_access = 1 ;
 	#else
	usb_lpt_access = 1 ;
	#endif
 #endif
     
 #ifndef _ETCPIP
	/*
	file->private_data = usblp;
    */
 #endif // ndef _ETCPIP

 usblp->writeurb->transfer_buffer_length = 0;
 usblp->writeurb->status = 0;
 
 usblp->wcomplete = 1; /* we begin writeable */
 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	usblp->rcomplete = 0;
 #endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw

 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	usblp_rd_done = 0 ;

	USBLPT_DBG_CTLMSG(iprintf("[USBLPT CTLMSG] usblp_open() : usblp_rd_done=%d\n"
	    	,usblp_rd_done
    	    ));
 #endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw

 if(usblp->quirks & USBLPT_QUIRK_UNSUPPORT) { // 2003/10/14 B. S. Liaw
	USBLPT_DBG_INFO(iprintf(
		"[USBLPT INFO] usblp_open() : QUIRK UNSUPPORT !! \n"));
	iprintf("[USBLPT] Error : This Printer is Not Supported by Firmware !! \n");
	#ifdef _USB_SYSLOG
	    //sprintf(usbsyslog_buf,"[USBLPT] Error : This Printer is Not Supported by Firmware !! ");
					#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_PR_UNSUPP
			//sprintf_SetSystemLog(USB_SYS_LOG_PR_UNSUPP);
			sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_PR_UNSUPP);
#else
			//sprintf_SetSystemLog("[USBLPT] Error : This Printer is Not Supported by Firmware !! ");
			sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] Error : This Printer is Not Supported by Firmware !! ");
#endif // USB_SYS_LOG_PR_UNSUPP
		}
	#endif // _SYSLOG
   }

 #ifdef USBLPT_RESET_CMD // 2003/10/14 B. S. Liaw
    if (usblp->quirks & USBLPT_QUIRK_NORESET) // 2003/10/14 B. S. Liaw
       {
		USBLPT_DBG_INFO(iprintf(
			"[USBLPT INFO] usblp_open() : QUIRK NO RESET !! \n"));
		iprintf("[USBLPT] This Printer Has not Reset Command \n");
		#ifdef _USB_SYSLOG
		    //sprintf(usbsyslog_buf,"[USBLPT] This Printer Has not Reset Command ");
						#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_PR_NO_RESET
				//sprintf_SetSystemLog(USB_SYS_LOG_PR_NO_RESET);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_PR_NO_RESET);
#else
				//sprintf_SetSystemLog("[USBLPT] This Printer Has not Reset Command ");
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] This Printer Has not Reset Command ");
#endif // USB_SYS_LOG_PR_NO_RESET
			}
		#endif // _SYSLOG
	   }
    if (usblp->quirks & USBLPT_QUIRK_RESET11) { // 2003/10/17 B. S. Liaw
		USBLPT_DBG_INFO(iprintf(
			"[USBLPT INFO] usblp_open() : QUIRK Version 1.1 RESET !! \n"));

		iprintf("[USBLPT] This Printer Only Accept Version 1.1 Reset Command \n");

		#ifdef _USB_SYSLOG
		    //sprintf(usbsyslog_buf,"[USBLPT] This Printer Only Accept Version 1.1 Reset Command ");
						#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_PR_RESET_11
				//sprintf_SetSystemLog(USB_SYS_LOG_PR_RESET_11);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_PR_RESET_11);
#else
				//sprintf_SetSystemLog("[USBLPT] This Printer Only Accept Version 1.1 Reset Command ");
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] This Printer Only Accept Version 1.1 Reset Command ");
#endif // USB_SYS_LOG_PR_RESET_11
			}
		#endif // _SYSLOG
	   }
 #endif // USBLPT_RESET_CMD

 #ifdef USBLPT_INIT_STR    // 2003/10/13 B. S. Liaw
	if (usblp->quirks & USBLPT_QUIRK_USB_INIT) {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] usblp_open() : QUIRK USB INIT ..\n" )) ;
		USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_open() call usblpt_init_string() ..\n" )) ;
		/*
		dlytsk(cur_task, DLY_TICKS, USBLPT_INIT_DLY_TSK_TICK) ;
		*/
		wait_ms(USBLPT_INIT_WAIT_MS);

		if (usblpt_init_string(usblp) != D_OK) {
		    dprintf("[USBLPT] Error : init string Fail !!\n");
			usblp->opened = 0 ;
			USBLPT_DBG_INIT(iprintf(
				"[USBLPT INIT] usblp_open() : Set Value : usblp->opened = %d \n",
				usblp->opened )) ;
			// retval = -EIO;
	        goto open_end ;
	       }
	   }
	  else
 #endif // USBLPT_INIT_STR
	   {
		USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Not USBLPT_QUIRK_USB_INIT !!\n")) ;
	   }

 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	if (usblp->bidir) {
		usblp->readcount = 0;
		usblp->readurb->dev = usblp->usb_dev;
	    /*
		usblp->readurb->hcpriv = 0 ;
	    */
    
		 // #ifndef _ETCPIP
 
		USBLPT_DBG_CTLMSG(iprintf(
		    "[USBLPT CTLMSG] usblp_open() : &(usblp->writeurb->hcpriv) = %08XH usblp->writeurb->hcpriv=%08XH\n",
		    &(usblp->writeurb->hcpriv) , usblp->writeurb->hcpriv
		    ));
		USBLPT_DBG_CTLMSG(iprintf(
		    "[USBLPT CTLMSG] usblp_open() : &(usblp->writeurb->callback) = %08XH usblp->writeurb->callback=%08XH\n",
		    &(usblp->writeurb->callback) , usblp->writeurb->callback
		    ));

		USBLPT_DBG_CTLMSG(iprintf(
		    "[USBLPT CTLMSG] usblp_open() : &(usblp->readurb->hcpriv) = %08XH usblp->readurb->hcpriv=%08XH\n",
		    &(usblp->readurb->hcpriv) , usblp->readurb->hcpriv
		    ));
		USBLPT_DBG_CTLMSG(iprintf(
		    "[USBLPT CTLMSG] usblp_open() : &(usblp->readurb->callback) = %08XH usblp->readurb->callback=%08XH\n",
		    &(usblp->readurb->callback) , usblp->readurb->callback
		    ));
		/*
		#ifdef _ETCPIP
		*/
		usblp_rd_done = 0 ;
		USBLPT_DBG_CTLMSG(iprintf("[USBLPT CTLMSG] usblp_open() : usblp_rd_done=%d\n"
	    	,usblp_rd_done
    	    ));
        usblp_wr_done = 0 ;
		USBLPT_DBG_CTLMSG(iprintf("[USBLPT CTLMSG] usblp_open() : usblp_wr_done=%d\n"
	    	,usblp_wr_done
    	    ));
	/*
	#endif // _ETCPIP
	*/

	#ifdef USBLPT_READ_ON_PROBE  // 2003/10/02 B. S. Liaw
		#error : Please Donot Define USBLPT_READ_ON_PROBE
		/*
		dlytsk(cur_task, DLY_TICKS, USBLPT_INIT_DLY_TSK_TICK) ;
		*/
		wait_ms(USBLPT_INIT_WAIT_MS);

		/*
		#ifndef USBLPT_READ_NO_DELAY
			dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
			dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
			dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
		#endif // ndef USBLPT_READ_NO_DELAY
	    */

		USBLPT_DBG_INFO(iprintf("[USBLPT INFO] usblp_open() : Read ..\n" ));
		USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_open() : Read ..\n" ));
       
		// 2003/4/8 B. S. Liaw
		read_timeout = USBLPT_READ_TIMEOUT;
    
		USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_open() : read_timeout = %d \n" , read_timeout ));
    
		//retv =  usb_start_wait_urb(&usblp->readurb,timeout,
	    //							&actual_length);
    
    
		// This Donot free urb
		//retv =  usblp_start_wait_urb(&usblp->readurb,timeout,
	    //							&actual_length);
		// retv =  usblp_start_wait_urb(usblp->readurb,read_timeout,
	    //							&actual_length);
    
		// This Donot free urb
		retv =  usb_start_wait_urb_nf(usblp->readurb,read_timeout, // 1 ,
    							&actual_length);
    
		USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_open() : retv = %d , actual_length = %d \n" ,
			retv , actual_length
	        ));
        
	    if (retv < 0 ) {
			dprintf("[USBLPT] Error : usblp I/O Fail !!\n") ;
			retval = -EIO;
        
			usblp->opened = 0;
			USBLPT_DBG_INIT(iprintf(
				"[USBLPT INIT] usblp_open() : Set Value : usblp->opened = %d \n" ,
				usblp->opened
				));
		   }
	      else
	       {
			USBLPT_DBG_INFO(iprintf("[USBLPT INFO] usblp_open() : Read OK.\n" ));
	       }
	#endif // USBLPT_READ_ON_PROBE
    
	#ifndef USBLPT_READ_BY_START_WAIT
	#ifdef USBLPT_READ_BY_URBTASK
	// Begin 2003/4/11 B. S. Liaw
    
		usblp->readcount = 0;
		usblp->readurb->dev = usblp->usb_dev;
		usblp->readurb->count = 0 ;
	    usblp->rcomplete = 0;
		usblp_rd_done = 0 ;
		USBLPT_DBG_CTLMSG(iprintf("[USBLPT CTLMSG] usblp_open() : usblp_rd_done=%d\n"
	    	,usblp_rd_done
    	    ));
        
		// usblp->readurb->kill = 0 ;
		// usblp->readurb->next_step = NEXT_STEP_SUBMIT 
		//					        | NEXT_STEP_FREE
		//					         ;
        
		BLOCK_PREEMPTION;
		usblp->readurb->next_step = 
								#ifndef USBLPT_READ_INFINITE
							          NEXT_STEP_PAUSE
									// | NEXT_STEP_CONTI
								#endif // ndef USBLPT_READ_INFINITE
                                	/*
                                    | NEXT_STEP_SUBMIT
							        | NEXT_STEP_FREE
                                    // B. S. Liaw : Need Free
                                    */
    							    ;
                                    
		UNBLOCK_PREEMPTION;
        
        
		usb_fill_bulk_urb(usblp->readurb, usblp->usb_dev,
			usb_rcvbulkpipe(usblp->usb_dev,
			usblp->protocol[protocol].epread->bEndpointAddress),
			usblp->readbuf, USBLPT_BUF_SIZE,
			usblp_bulk_read, usblp);

    	/*
		if (usb_submit_urb(usblp->readurb, GFP_KERNEL) < 0)
     	*/
        // if (usb_submit_urb(usblp->readurb) < 0)
		retv = usb_submit_urb(usblp->readurb);

	#ifdef USBLPT_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
	   ) ) {
			iprintf("[ASSERT][USB] usblp_open() : " ) ;
    	    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
		    iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
	#endif // USB_BS_ASSERT
		#ifndef USBLPT_READ_NO_DELAY
		dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
		#endif // ndef USBLPT_READ_NO_DELAY
            
		if (retv<0)
        /*
     #else
		
		// 2003/4/8 B. S. Liaw
		read_timeout = USBLP_READ_TIMEOUT;
        
		USBLPT_DBG_INIT(iprintf(
		   	"[USBLPT INIT] usblp_open() : read_timeout = %d \n" ,
			read_timeout
            ));
        
		//retv =  usb_start_wait_urb(&usblp->readurb,timeout,
        //							&actual_length);
        
        
		// This Donot free urb
		//retv =  usblp_start_wait_urb(&usblp->readurb,timeout,
        //							&actual_length);
		// retv =  usblp_start_wait_urb(usblp->readurb,read_timeout,
        //							&actual_length);
        
		// This Donot free urb
	    retv =  usb_start_wait_urb_nf(usblp->readurb,read_timeout, 1 ,
        							&actual_length);
        
        USBLPT_DBG_INIT(iprintf(
		   	"[USBLPT INIT] usblp_open() : retv = %d , actual_length ="
            " %d \n" ,
			retv , actual_length
            ));
            
        if (retv < 0 )
        */
    #endif
    
	#ifdef USBLPT_READ_BY_CHECK
 
		hcd = usblp->readurb->dev->bus->hcpriv_hcd ;
        
		// if (urb->dev != hcd->self.root_hub)
        if (usblp->readurb->hcpriv)
		   {
			USBLPT_DBG_URB(iprintf("[USBLPT URB] usblp_open(2) : free priv .. \n" )) ;
			USBLPT_DBG_FLOW(iprintf( "[USBLPT FLOW] usblp_open() : Call hcd->driver->urb_free_priv() ...\n" )) ;
	
			hcd->driver->urb_free_priv(hcd, usblp->readurb->hcpriv);
            
			USBLPT_DBG_URB(iprintf("[USBLPT URB] usblp_open() : Set Vaue : usblp->readurb->hcpriv = 0 ..\n" )) ;
            
			usblp->readurb->hcpriv = NULL;
           }

		usblp->readurb->count = 0 ;
	    usblp->rcomplete = 0;
		usblp->readurb->dev = usblp->usb_dev;
	 	usblp->readcount = 0;
		usblp_rd_done = 0 ;
		USBLPT_DBG_CTLMSG(iprintf("[USBLPT CTLMSG] usblp_open() : usblp_rd_done=%d\n"
	    	,usblp_rd_done
    	    ));

		/*
		usb_fill_bulk_urb(usblp->readurb, usblp->usb_dev,
			usb_rcvbulkpipe(usblp->usb_dev,
			usblp->protocol[protocol].epread->bEndpointAddress),
			usblp->readbuf, USBLP_BUF_SIZE,
			usblp_bulk_read, usblp);
		*/

		USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_open(1) : usb_submit_urb() .. \n" )) ;
        
		retv = usb_submit_urb(usblp->readurb);

	#ifdef USBLPT_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
	   ) ) {
			iprintf("[ASSERT][USB] usblp_open() : " ) ;
    	    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
		    iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
	#endif // USB_BS_ASSERT
		#ifndef USBLPT_READ_NO_DELAY
		dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
		#endif // ndef USBLPT_READ_NO_DELAY
            
		if (retv<0)

	#endif // USBLPT_READ_BY_CHECK
    
           {
			dprintf("[USBLPT] Error : usblp I/O Fail !!\n") ;
			retval = -EIO;
            
			usblp->opened = 0;
			USBLPT_DBG_INIT(iprintf(
				"[USBLPT INIT] usblp_open() : Set Value : usblp->opened "
                "= %d \n" ,
				usblp->opened
				));
			#ifndef _ETCPIP
            	/*
				file->private_data = NULL;
                */
			#endif // ndef _ETCPIP
	       }
           
	#ifdef USBLPT_READ_BY_URBTASK
          else
		   {
			int status1 ;
        
			#ifdef INT_SCHEDULING
			    status1 = runtsk(100,(void (fAR *)(void))urb_task, 
                				 6000,(unsigned long)usblp->readurb);
			#else
			    status1 = runtsk(80, (void (fAR *)(void))urb_task,
                				 6000,(unsigned long)usblp->readurb);
                                 // 100 too high !!
			#endif // INT_SCHEDULING

		    if (status1 < SUCCESS )
    		   {
		        dprintf("[USB] Error : RUNTASK urb_task fails !!\n");
    		   }
		      else
    		   {
		        USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_open() : RUNTASK id=%d urb_task ...\n", status1));
		       }
               
	#ifdef USBLPT_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
	   ) ) {
			iprintf("[ASSERT][USB] usblp_write(2) : " ) ;
    	    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
		    iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
	#endif // USB_BS_ASSERT
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
	    	NULL ;
            
		   }
	#endif // def USBLPT_READ_BY_URBTASK
	#endif // ndef USBLPT_READ_BY_START_WAIT
   } //  if usblp->bidir
 #endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw
   
 // dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
 
 // -------------------    
 open_end:
 // -------------------    
 #ifdef USBLPT_DBG_INIT_ON
	USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_open() call usblp_dump() ..\n"));
    usblp_dump(usblp);
     /*
		USBLPT_DBG_INIT(iprintf(
		   "[USBLPT INIT] usblp_open() : Check Status ..\n"
		   ));
		usblp_check_status(usblp, 0);
     */
  #else
     // #error : Please Make Sure : Not DEBUG !!
 #endif
    
 /*    
	#ifdef _ETCPIP
		usb_lpt_ok = 1 ;
	#endif // _ETCPIP
 */
 if(usblp->opened == -1) {
	/* 2003/90/27 B. S. Liaw , comment out
	usblp->opened = 1 ;
	USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_open() : Set Value : usblp->opened = %d \n" , usblp->opened )) ;
	*/
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Printer Open OK. \n"));
   }
  else {
	usblp->opened = 0 ;
	USBLPT_DBG_INIT(iprintf(
		"[USBLPT INIT] usblp_open() : Set Value : usblp->opened = %d \n",
		usblp->opened )) ;
	USBLPT_DBG_INFO(iprintf(
		"[USBLPT INFO] Printer Open Fail , need Re-Open Later. \n"));
   }
   
 /*  
 iprintf("[USB] Found Printer , V/P = %04XH/%04XH \n" , usblp->usb_dev->descriptor.idVendor,
					usblp->usb_dev->descriptor.idProduct);
                    
 sprintf(usb_info_str, "Printer , V/P = %04XH/%04XH ", usblp->usb_dev->descriptor.idVendor,
					usblp->usb_dev->descriptor.idProduct);

 // usb_device_count ++ ; // >= 1 : enable ISR

 dprintf("[USB] Total Supported External Device = %d \n",usb_device_count) ;
 */
 /*
 ParallelTest() ;
 */
 // -------------------    
 out:
 // -------------------    
 
 #ifndef _ETCPIP
	/*
	#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
		if (!retval) MOD_INC_USE_COUNT;
	#endif
		unlock_kernel();
    */
 #endif // ndef _ETCPIP
    
 // #ifdef _ETCPIP
 #ifdef USBLPT_PRINTERS
 if(usblp->usblp_plug_out) {
 #else
 if(usblp_plug_out >= 1) {
 #endif 	
	usblp->opened = 0 ;
	USBLPT_DBG_INIT(iprintf(
		"[USBLPT INIT] usblp_open() : Set Value : usblp->opened = %d \n",
		usblp->opened )) ;
	usblp_release(usblp) ;
   }
 // #endif // _ETCPIP
   
 #ifdef _ETCPIP
 #ifdef USBLPT_PRINTERS
 	usblp->usb_lpt_access = 0 ;
 #else
	usb_lpt_access = 0 ;
 #endif
 #endif

 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_open() End.\n"));
 return retval;
} // function usblp_open()

// ====================================================================

static void usblp_cleanup(struct usblp *usblp)
{
 struct usb_hcd		*hcd ;
 unsigned int next_step ;
 /*
 #ifdef _USB_SYSLOG
    char usbsyslog_buf[100];
 #endif // _SYSLOG
 */


 USBLPT_DBG_FLOW(iprintf(
	"[USBLPT FLOW] usblp_cleanup() Begin ..\n"));
 // dprintf("[USBLPT] usblp_cleanup() Begin ..\n") ;
 
 // #ifdef USBLPT_INIT_ON
 // #if defined(USB_DBG_MEM_C_ON) || defined(USB_DBG_MEM_U_ON)
 #ifdef USB_DBG_MEM_COUNT_ON
	iprintf("[USB MEM_COUNT] usblp_cleanup(1) ====== show USB Mem Count .. \n"
    	) ;
	showUsbMemC() ;
	showUsbMemU() ;
 #endif
 /*
 BLOCK_PREEMPTION;
 usblp->readurb->next_step = NEXT_STEP_KILL ;
 next_step = NEXT_STEP_KILL ;
 UNBLOCK_PREEMPTION;
 
 USB_DBG_INIT(iprintf("[USB INIT] usblp_release() : wait urb_task stop .. \n" )) ;
 while( next_step != NEXT_STEP_NONE) 
   {
	dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
    
	BLOCK_PREEMPTION;
	next_step = usblp->readurb->next_step ;
	UNBLOCK_PREEMPTION;
   }
   
 USB_DBG_INIT(iprintf("[USB INIT] usblp_release() : wait urb_task stop ok. \n" )) ;
 */
 #ifdef CONFIG_USB_DEVICEFS
	devfs_unregister (usblp->usb_devfs);
 #endif // CONFIG_USB_DEVICEFS
 
 #ifndef _ETCPIP
	usb_deregister_dev (1, usblp->minor);
 #endif // ndef _ETCPIP
 /*
 #ifndef _ETCPIP
 	
	info ("UsbLpt-%d: removed", usblp->minor);
    
  #else
 */
 USBLPT_DBG_INIT(iprintf(
		"[USBLPT INIT] UsbLpt-%d removed \n", usblp->minor
        ));

 dprintf("[USBLPT] USB Printer Removed \n");

 #ifdef _USB_SYSLOG
		    //sprintf(usbsyslog_buf,"[USBLPT] USB Printer Removed ");
				#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_PR_REMOVE
		//sprintf_SetSystemLog(USB_SYS_LOG_PR_REMOVE);
		sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_PR_REMOVE);
#else
		//sprintf_SetSystemLog("[USBLPT] USB Printer Removed ");
		sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] USB Printer Removed ");
#endif // USB_SYS_LOG_PR_REMOVE
	}
 #endif // _USB_SYSLOG

 /*
 #endif
 */
 // usb_device_count -- ; // >= 1 : enable ISR

 usb_buffer_free(usblp->usb_dev, USBLPT_BUF_SIZE,
		usblp->writebuf,
		#ifndef USB_NO_VMM // no vitrtul memory management
	       (dma_addr_t) usblp->writeurb->transfer_dma
	     #else
           (dma_addr_t) usblp->writeurb->transfer_buffer
	    #endif
       );

 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
 usb_buffer_free(usblp->usb_dev, USBLPT_BUF_SIZE,
		usblp->readbuf,
		#ifndef USB_NO_VMM // no vitrtul memory management
	       (dma_addr_t) usblp->readurb->transfer_dma
	     #else
           (dma_addr_t) usblp->readurb->transfer_buffer
	    #endif
       );
 #endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw

 /*
 kfree (usblp->usb_device_id_string);
 */
 UsbFreeMemU(USB_MEM_U_L, (unsigned char *) // 1024
 		usblp->device_id_string
			 #if defined(USB_DBG_MEM_U_ON)
    	    	,  USBLPT_DEVICE_ID_SIZE
             #endif
        ) ;
 
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_cleanup() : free usblp->statusbuf .. \n"
    ));
    
 #ifndef USB_MEM_LIB
    /*
 	kfree (usblp->statusbuf);
    */
 #else
 	/*
    UsbFreeMemC(USB_MEM_C_S,(unsigned char *)
                 usblp->statusbuf
 		#if defined USB_DBG_MEM_C_ON
                 ,STATUS_BUF_SIZE
		#endif
		) ;
    */
	UsbFreeMemU(USB_MEM_U_S, (unsigned char *) usblp->statusbuf
					 #if defined(USB_DBG_MEM_U_ON) // size : 8
	    		    	,  STATUS_BUF_SIZE
    	    	     #endif
                     ) ;
 #endif
 
 if(usblp->writeurb->hcpriv)
   {
	hcd = usblp->writeurb->dev->bus->hcpriv_hcd;
        
	// if (urb->dev != hcd->self.root_hub)
	   {
		USB_DBG_INIT(iprintf( "[USB INIT] usblp_cleanup() urb_free_priv() ...\n" )) ;
	
		hcd->driver->urb_free_priv(hcd, usblp->writeurb->hcpriv);
            
		USB_DBG_URB(iprintf("[USB URB] usblp_cleanup() : Set Vaue : urb->hcpriv = 0 ..\n" )) ;
            
		usblp->writeurb->hcpriv = NULL;
	   }
   }
   
 USBLPT_DBG_INIT(iprintf(
 	"[USBLPT INIT] usblp_cleanup() : free usblp->writeurb=%08XH\n",
	usblp->writeurb
	)) ;
    
 usb_free_urb(usblp->writeurb);
 
 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
 if(usblp->readurb->hcpriv)
   {
	hcd = usblp->readurb->dev->bus->hcpriv_hcd;
        
	// if (urb->dev != hcd->self.root_hub)
	   {
		USB_DBG_INIT(iprintf( "[USB INIT] usblp_cleanup() urb_free_priv() ...\n" )) ;
	
		hcd->driver->urb_free_priv(hcd, usblp->readurb->hcpriv);
            
		USB_DBG_URB(iprintf("[USB URB] usblp_cleanup() : Set Vaue : urb->hcpriv = 0 ..\n" )) ;
            
		usblp->readurb->hcpriv = NULL;
	   }
   }
   
 USBLPT_DBG_INIT(iprintf(
 	"[USBLPT INIT] usblp_cleanup() : free usblp->readurb=%08XH\n", usblp->readurb
	)) ;
    
 usb_free_urb(usblp->readurb);

 #endif // USBLPT_BIDIR

 USBLPT_DBG_MEM(iprintf(
 	"[USBLPT MEM] usblp_cleanup() : free usblp->writebuf=%08XH\n",
	usblp->writebuf
	)) ;
 /*    
 UsbFreeMemC(USB_MEM_C_L,       // 8192
 	(unsigned char *) usblp->writebuf
 	#if defined USB_DBG_MEM_C_ON
	    	, USBLP_BUF_SIZE
	 	#endif
     ) ;
 */
 UsbFreeMemU(USB_MEM_U_L, (unsigned char *) // 8192
			usblp->writebuf
			#if defined(USB_DBG_MEM_U_ON)
	    	   	,  USBLPT_BUF_SIZE
    	    #endif
		   ) ;

 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
 USBLPT_DBG_MEM(iprintf(
 	"[USBLPT MEM] usblp_cleanup() : free usblp->readbuf=%08XH\n",
	usblp->readbuf
	)) ;
    
 /*    
 UsbFreeMemC(USB_MEM_C_L,       // 8192
	 		(unsigned char *) usblp->readbuf
		 	#if defined USB_DBG_MEM_C_ON
		    	, USBLP_BUF_SIZE
		 	#endif
		    ) ;
 */
 UsbFreeMemU(USB_MEM_U_L, (unsigned char *) // 8192
			usblp->readbuf
			#if defined(USB_DBG_MEM_U_ON)
	    	   	,  USBLPT_BUF_SIZE
    	    #endif
		   ) ;

 #endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw

 USB_DBG_MEM(iprintf(
	"[USB MEM] usblp_cleanup(): free usblp .. \n"
    ));
 #ifndef USB_MEM_LIB
 	kfree (usblp);
 #else
    UsbFreeMemC(USB_MEM_C_S,(unsigned char *) usblp
					#if defined(USB_DBG_MEM_C_ON)
                    	,sizeof(struct usblp)
					#endif
                    ) ;
 #endif
 
 global_usblp = 0 ;
 USBLPT_DBG_INFO(iprintf("[USBLPT INFO] usblp_cleanup() : global_usblp=%08XH \n",global_usblp));

 // #ifdef USB_DBG_INIT_ON
 // #if defined(USB_DBG_MEM_C_ON) || defined(USB_DBG_MEM_U_ON)
 // #ifdef USBLPT_INIT_ON
 #ifdef USB_DBG_MEM_COUNT_ON
	iprintf("[USB MEM_COUNT] usblp_cleanup(2) ====== show USB Mem Count .. \n");
	showUsbMemC() ;
	showUsbMemU() ;
 #endif
 
 // dprintf("[USBLPT] usblp_cleanup() End.\n") ;
 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_cleanup() End.\n"));
}

// ====================================================================

static void usblp_unlink_urbs(struct usblp *usblp)
{
 usb_unlink_urb(usblp->writeurb);
 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
 if (usblp->bidir)
   {
    /*
    if (usblp->readurb->hcpriv)
	   {
		USB_DBG_INIT(iprintf("[USB INIT] usblp_unlink_urbs() : free priv .. \n" )) ;
		USB_DBG_FLOW(iprintf( "[USB FLOW] usblp_unlink_urbs() Call hcd->driver->urb_free_priv() ...\n" )) ;
	
		hcd->driver->urb_free_priv(hcd, usblp->readurb->hcpriv);
    
		USB_DBG_URB(iprintf("[USB URB] usblp_unlink_urbs() : Set Vaue : urb->hcpriv = 0 ..\n" )) ;
    
		usblp->readurb->hcpriv = NULL;
	   }
    */
		usb_unlink_urb(usblp->readurb);
   }
 #endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw
}

// ====================================================================

#ifndef _ETCPIP
	/*
	static int usblp_release(struct inode *inode, struct file *file)
    */
 #else
	static int usblp_release(struct usblp *usblp)
	// static int usblp_release(struct inode *inode, struct usblp *usblp)
#endif
{
 #ifndef _ETCPIP
    /*
	struct usblp *usblp = file->private_data;
    */
 #endif // ndef _ETCPIP
 unsigned int next_step ;
 

 #ifndef NO_SEMPHORE
 	/*
	down (&usblp->sem);
    */
 #endif // ndef NO_SEMPHORE
 
 #ifndef _ETCPIP
 	/*
	lock_kernel();
    */
 #endif // ndef _ETCPIP
 
 usblp->opened = 0;
 usb_lpt_ok = 0 ;
 /*
 BLOCK_PREEMPTION;
 usblp->readurb->next_step = NEXT_STEP_KILL ;
 next_step = NEXT_STEP_KILL ;
 UNBLOCK_PREEMPTION;
 
 USB_DBG_INIT(iprintf("[USB INIT] usblp_release() : wait urb_task stop .. \n" )) ;
 while( next_step != NEXT_STEP_NONE) 
   {
	dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
    
	BLOCK_PREEMPTION;
	next_step = usblp->readurb->next_step ;
	UNBLOCK_PREEMPTION;
   }
 
 USB_DBG_INIT(iprintf("[USB INIT] usblp_release() : wait urb_task stop ok. \n" )) ;
 */
 if(usblp->usb_dev)
   {
	//if (usblp->bidir)
	//	usb_unlink_urb(&usblp->readurb);
        
	// usb_unlink_urb(&usblp->writeurb);
    
	usblp_unlink_urbs(usblp);
    
    #ifndef NO_SEMPHORE
		up(&usblp->sem);
    #endif // ndef NO_SEMPHORE
    
   }
  else 		/* finish cleanup from disconnect */
   {
	if (usb_lpt_released == 0) // prevent double clean up
       {
        if (usblp)
           {
			USB_DBG_INIT(iprintf("[USB INIT] usblp_release() : usblp_cleanup() ..\n")) ;
			usblp_cleanup(usblp);
	       }
   }
   }
 /*
 #ifndef _ETCPIP
	#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
		MOD_DEC_USE_COUNT;
	#endif
	unlock_kernel();
 #endif // ndef _ETCPIP
 */
 usb_lpt_released = 1 ;

 return 0;
} // function usblp_release()

/* No kernel lock - fine */
/*

#ifndef _ETCPIP
	static unsigned int usblp_poll(struct file *file ,
    						 struct poll_table_struct *wait)
#elif !defined(NO_WAITQUEUE)
	static unsigned int usblp_poll( struct usblp *usblp ,
    						 struct poll_table_struct *wait)
#else
	static unsigned int usblp_poll( struct usblp *usblp)
#endif   
{
 #ifndef _ETCPIP
	struct usblp *usblp = file->private_data;
 #endif // ndef _ETCPIP
 
 #ifndef NO_WAITQUEUE
	poll_wait(file, &usblp->wait, wait);
 #endif // ndef NO_WAITQUEUE
 
 return    ((   !usblp->bidir 
		     || usblp->readurb->status  == -EINPROGRESS
            ) ? 0 : POLLIN  | POLLRDNORM
           )
 		 | ( usblp->writeurb->status == -EINPROGRESS 
             ? 0 : POLLOUT
            | POLLWRNORM
           ) ;
}
*/

/*
#ifndef _ETCPIP
  
   static int usblp_ioctl( struct inode *inode, struct file *file,
    					    unsigned int cmd, unsigned long arg)
   {
	struct usblp *usblp = file->private_data;
	int length, err, i;
	unsigned char lpstatus, newChannel;
	int status;
	int twoints[2];
	int retval = 0;

	#ifndef NO_SEMPHORE
		down (&usblp->sem);
	#endif // ndef NO_SEMPHORE
	if (!usblp->usb_dev) {
		retval = -ENODEV;
		goto done;
	}

	if (_IOC_TYPE(cmd) == 'P')	// new-style ioctl number 

	switch (_IOC_NR(cmd))
       {

		case IOCNR_GET_DEVICE_ID: //  get the DEVICE_ID string
				if (_IOC_DIR(cmd) != _IOC_READ) {
					retval = -EINVAL;
					goto done;
				}

				//length = (usblp->usb_device_id_string[0] << 8) + usblp->usb_device_id_string[1];
                // big-endian 
                
				//if (length < DEVICE_ID_SIZE)
				//	usblp->usb_device_id_string[length] = '\0';
				//else
				//	usblp->usb_device_id_string[DEVICE_ID_SIZE - 1] = '\0';
                
				// length = usblp_cache_device_id_string(usblp);
				length = usblp_get_device_id_string(usblp);
				if (length < 0) {
					retval = length;
					goto done;
				}
				dprintf("[USBLPT] UsbLpt-%d Device ID string [%d/max %d]='%s' \n",
					usblp->minor, length, _IOC_SIZE(cmd), &usblp->usb_device_id_string[2]);

				if (length > _IOC_SIZE(cmd))
					length = _IOC_SIZE(cmd); // truncate 

				if (copy_to_user((unsigned char *) arg,
						usblp->usb_device_id_string,
						(unsigned long) length)) {
					retval = -EFAULT;
					goto done;
				}

				break;

		case IOCNR_GET_PROTOCOLS:
				if (_IOC_DIR(cmd) != _IOC_READ ||
				    _IOC_SIZE(cmd) < sizeof(twoints)) {
					retval = -EINVAL;
					goto done;
				}

				twoints[0] = usblp->current_protocol;
				twoints[1] = 0;
				for (i = USBLP_FIRST_PROTOCOL;
				     i <= USBLP_LAST_PROTOCOL; i++) {
					if (usblp->protocol[i].alt_setting >= 0)
						twoints[1] |= (1<<i);
				}

				if (copy_to_user((unsigned char *)arg,
						(unsigned char *)twoints,
						sizeof(twoints))) {
					retval = -EFAULT;
					goto done;
				}

				break;

		case IOCNR_SET_PROTOCOL:
			if (_IOC_DIR(cmd) != _IOC_WRITE)
               {
				retval = -EINVAL;
				goto done;
			   }

			#ifdef DEBUG
				if (arg == -10) {
					usblp_dump(usblp);
					break;
				}
			#endif

			usblp_unlink_urbs(usblp);
			retval = usblp_set_protocol(usblp, arg);
			if (retval < 0) {
					usblp_set_protocol(usblp,
						usblp->current_protocol);
				}
			// At this point getting the device ID is safe
			// even on buggy printers, because the bulk I/O
			//  is not in progress (URBs are unlinked). 
			if (usblp->quirks & USBLP_QUIRK_GET_ID)
					usblp_cache_device_id_string(usblp);
			break;

		case IOCNR_HP_SET_CHANNEL:
			if (_IOC_DIR(cmd) != _IOC_WRITE ||
				    usblp->usb_dev->descriptor.idVendor != 0x03F0 ||
				    usblp->quirks & USBLP_QUIRK_BIDIR) {
					retval = -EINVAL;
					goto done;
				}

				err = usblp_hp_channel_change_request(usblp,
					arg, &newChannel);
				if (err < 0) {
					//dprintf("[USBLPT] usblp%d: error = %d reading "
                    //	"IEEE-1284 Device ID string \n",
					//	usblp->minor, err);
					dprintf("[USBLPT] Error : UsbLpt-%d: error = %d setting "
						"HP channel",
						usblp->minor, err);
					// usblp->usb_device_id_string[0] = usblp->usb_device_id_string[1] = '\0';
					retval = -EIO;
					goto done;
				}

				dbg("UsbLpt-%d requested/got HP channel %ld/%d",
					usblp->minor, arg, newChannel);
				break;

		case IOCNR_GET_BUS_ADDRESS:
				if (_IOC_DIR(cmd) != _IOC_READ ||
				    _IOC_SIZE(cmd) < sizeof(twoints)) {
					retval = -EINVAL;
					goto done;
				}

				twoints[0] = usblp->usb_dev->bus->busnum;
				twoints[1] = usblp->usb_dev->devnum;
				if (copy_to_user((unsigned char *)arg,
						(unsigned char *)twoints,
						sizeof(twoints))) {
					retval = -EFAULT;
					goto done;
				}

				dbg("UsbLpt-%d is bus=%d, device=%d",
					usblp->minor, twoints[0], twoints[1]);
				break;

		case IOCNR_GET_VID_PID:
				if (_IOC_DIR(cmd) != _IOC_READ ||
				    _IOC_SIZE(cmd) < sizeof(twoints)) {
					retval = -EINVAL;
					goto done;
				}

				twoints[0] = usblp->usb_dev->descriptor.idVendor;
				twoints[1] = usblp->usb_dev->descriptor.idProduct;
				if (copy_to_user((unsigned char *) arg,
						(unsigned char *)twoints,
						sizeof(twoints))) {
					retval = -EFAULT;
					goto done;
				}

				dbg("UsbLpt-%d is VID=0x%4.4X, PID=0x%4.4X",
					usblp->minor, twoints[0], twoints[1]);
				break;

		default:
				retval = -EINVAL;
	   }
	else	// old-style ioctl value
	switch (cmd)
       {

		case LPGETSTATUS:
				if (usblp_read_status(usblp, &lpstatus)) {
					err("UsbLpt-%d: failed reading printer status", usblp->minor);
					retval = -EIO;
					goto done;
				}
				status = lpstatus;
				if (copy_to_user ((int *)arg, &status, sizeof(int)))
					retval = -EFAULT;
			break;

		default:
			retval = -EINVAL;
	   }

	done:
    	#ifndef NO_SEMPHORE
			up (&usblp->sem);
	    #endif // ndef NO_SEMPHORE
		return retval;
        
   } // function usblp_ioctl()
  
#endif // ndef _ETCPIP
*/

// ====================================================================


// ====================================================================
/*
#ifndef _ETCPIP

	static ssize_t usblp_write( struct file *file, const char *buffer,
    							size_t count, loff_t *ppos)
    
 #else
    */
// ssize_t usblp_write( struct usblp *usblp , 
int usblp_write( struct usblp *usblp , unsigned char *buffer,
    							size_t input_count)
/*
#endif
*/
{
 /*
 #ifndef NO_WAITQUEUE
	#ifndef _ETCPIP
		DECLARE_WAITQUEUE(wait, current);
     #else
		DECLARE_WAITQUEUE(wait, 0);
	#endif
 #endif // ndef NO_WAITQUEUE
 */
 #ifndef _ETCPIP
 	/*
	struct usblp *usblp = file->private_data;
    */
 #endif // ndef _ETCPIP
 
 unsigned long timeout, err = 0;
 size_t writecount = 0;
 
 // #ifdef _ETCPIP
 // 2003/4/11 B. S. Liaw
 int retv ;
 long wait_urb_timeout ;
 int actual_length ;
 int ret_val ;
 struct usb_hcd		*hcd ;
 int result;
 // #endif // _ETCPIP
 /*
 #ifdef _USB_SYSLOG
    char usbsyslog_buf[100];
 #endif // _SYSLOG
 */

 // USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_write() Begin ..\n"));
 USBLPT_DBG_WRITE(iprintf("[USBLPT WRITE] usblp_write() Begin ..\n"));

 ret_val = 0 ;
 #ifdef USBLPT_PRINTERS
 if (usblp->usblp_plug_out) {
 #else
 if (usblp_plug_out >= 1) {
 #endif
	ret_val = -EIO ;
 	 goto write_end ;
    }
 
 USBLPT_DBG_URB(iprintf("[USBLPT URB] usblp_write() : input_count = %d \n" , input_count	)) ;
    
 USBLPT_DBG_MEM(iprintf("[USBLPT MEM] usblp = %08XH \n" , usblp )) ;
 
 #ifdef _ETCPIP
	#ifdef USBLPT_PRINTERS
 	if (usblp->usb_lpt_released)
 	#else 
	if (usb_lpt_released)
	#endif
	   {
	    USBLPT_DBG_INIT(iprintf(
			"[USBLPT INIT] Error : usblp_write() : Printer Driver "
            "Released !! \n"
        	)) ;
		return 0 ;
	   }
	#ifdef USBLPT_PRINTERS
 	if (usblp->usblp_plug_out)
 	#else 
	if (usblp_plug_out >= 1)
	#endif
        {
        USBLPT_DBG_INIT(iprintf(
			"[USBLPT INIT] Error : usblp_write() : Printer Plug "
            "Out !! \n"
        	)) ;
		ret_val = -EIO ;
        goto write_end ;
        /*
		USBLPT_DBG_FLOW(iprintf(
    		"[USBLPT FLOW] usblp_write() End 1 With Fail !!\n"));
		return 0 ;
        */
       }
    
	// 2003/4/8 B. S. Liaw
	if (!usblp || !usblp->usb_dev || (usblp->opened==0) )
	   {
        USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Error : usblp_write() : !usblp !usblp->usb_dev (usblp->opened==0) \n" )) ;
        
	#ifdef USBLPT_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
	   ) ) {
			iprintf("[ASSERT][USB] usblp_write(2) : " ) ;
    	    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
		    iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
	#endif // USB_BS_ASSERT
		dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
		USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_write() End 2 With Fail !!\n"));
		return 0 ;
      }
 #endif // _ETCPIP

 #ifdef _ETCPIP
 #ifdef USBLPT_PRINTERS
 	usblp->usb_lpt_access = 1 ;
 #else
	usb_lpt_access = 1 ;
 #endif
 #endif

 #ifdef _ETCPIP
 	/*
	USBLPT_DBG_INIT(iprintf(
		"[USBLPT INIT] usblp_write() 1 : Check Status ..\n"
		));
	 
	// 1 : 	LP_POUTPA // USB : PAPER_EMPTY 1 : paper empty
	// 2 :  LP_PSELECD	// USB : SELECT , 1 : selected
	// 3 :  LP_PERRORP	// USB : NOT_ERROR , 1 : no error
	
	err = usblp_check_status(usblp, 0);
    
    if ( err == 1 ) // paper empty
       {
        // writecount = -EIO;
        writecount = 0 ;
		goto write_fail1 ;
       }
    else if ( err == 2 ) // not selected
       {
        writecount = -EIO;
		goto write_fail1 ;
       }
    else if ( err == 3 ) // error
       {
        writecount = -EIO;
		goto write_fail1 ;
       }
    */
 #endif // def _ETCPIP
 
 while (writecount < input_count)
   {
    /*
    // 2003/4/10 B. S. Liaw
    int i ;
 	for (i=0 ; i< 10 ; i++)   
       {
      	if ( usblp_check_status(usblp, 0)== 0)
           {
        	break ;
	       }
       }
    */   

	// FIXME:  only use urb->status inside completion
	// callbacks; this way is racey...

	USBLPT_DBG_WRITE(iprintf("[USBLPT WRITE] usblp_write() : input_count = %d , writecount = %d\n", input_count, writecount )) ;
	// USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_write() : count = %d , writecount = %d\n", count, writecount )) ;
    /*
    
    // -115 : EINPROGRESS
	USBLPT_DBG_WRITE(iprintf(
	   	"[USBLPT WRITE] usblp_write() : usblp->writeurb->status = %d \n",
		usblp->writeurb->status
        ));
        
	if (!usblp->wcomplete)
       {
		#ifndef _ETCPIP
			barrier();
		#endif // ndef _ETCPIP

		#ifndef _ETCPIP
			if (file->f_flags & O_NONBLOCK)
				return -EAGAIN;
		#endif // ndef _ETCPIP

		timeout = USBLP_WRITE_TIMEOUT;
		#ifndef NO_WAITQUEUE
			add_wait_queue(&usblp->wait, &wait);
		#endif // ndef NO_WAITQUEUE
        
		USBLPT_DBG_WRITE(iprintf(
		   	"[USBLPT WRITE] usblp_write() : timeout = %d \n" ,
			timeout
            ));
            
		USBLPT_DBG_WRITE(iprintf(
		   	"[USBLPT WRITE] usblp_write() : Wait .. \n"
            ));
            
		#ifndef _ETCPIP
			while ( 1==1 ) 
		  #else
		    while (timeout && usblp->writeurb->status == -EINPROGRESS)
		#endif
           {

			#ifndef _ETCPIP
				if (signal_pending(current)) {
				#ifndef NO_WAITQUEUE
					remove_wait_queue(&usblp->wait, &wait);
				#endif // ndef NO_WAITQUEUE
					return writecount ? writecount : -EINTR;
				}
			#endif // ndef _ETCPIP
            
			#ifndef _ETCPIP
				set_current_state(TASK_INTERRUPTIBLE);
			#endif // ndef _ETCPIP

			#ifndef NO_WAITQUEUE
				//  timeout = interruptible_sleep_on_timeout(
                //              &usblp->wait, timeout);
             #else
             	// 2003/4/8 B. S. Liaw
				// dlytsk(cur_task, DLY_TICKS, 1+TIME_PLUS) ;
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
			#endif // ndef NO_WAITQUEUE
			if (timeout && !usblp->wcomplete) {
					timeout = usb_schedule_timeout(timeout);
				} else {
					#ifndef _ETCPIP
						set_current_state(TASK_RUNNING);
					#endif // ndef _ETCPIP
					break;
				}
		   } // while
           
		#ifndef NO_WAITQUEUE
			remove_wait_queue(&usblp->wait, &wait);
		#endif // ndef NO_WAITQUEUE
		USBLPT_DBG_WRITE(iprintf(
		   	"[USBLPT WRITE] usblp_write() : Wait End. \n"
            ));
	   } // if
      else
       {
		USBLPT_DBG_WRITE(iprintf(
		   	"[USBLPT WRITE] usblp_write() : racey Donot Wait \n"
            ));
       }
    
	#ifndef NO_SEMPHORE
		down (&usblp->sem);
	#endif // ndef NO_SEMPHORE
    */
    
	if (!usblp->usb_dev)
	   {
		USBLPT_DBG_INFO(iprintf("[USBLPT INFO] usblp_write() : Error : No usblp->usb_dev !!\n")) ;
		#ifndef NO_SEMPHORE
			up (&usblp->sem);
		#endif // ndef NO_SEMPHORE
        
		#ifdef _ETCPIP
	        #ifdef USBLPT_PRINTERS
			usblp->usb_lpt_access = 0 ;
 		#else
			usb_lpt_access = 0 ;
		#endif
		#endif
        
		return -ENODEV;
	   }
       
	/*
	if (usblp->writeurb->status != 0)
       {
		if (usblp->quirks & USBLP_QUIRK_BIDIR)
           {
			if (!usblp->wcomplete)
               {
				dprintf("[USBLPT] UsbLpt-%d: error %d writing to printer \n",
						usblp->minor, usblp->writeurb->status);
               }
			err = usblp->writeurb->status;
		   }
          else
           {
			USBLPT_DBG_INIT(iprintf(
					"[USBLPT INIT] usblp_write() 1 : write fail !!\n"
					));
			goto write_fail ;
           }
            
		#ifndef NO_SEMPHORE
        	
			up (&usblp->sem);
            
		#endif // ndef NO_SEMPHORE

		// if the fault was due to disconnect, let khubd's
		// call to usblp_disconnect() grab usblp->sem ...
		// 
        
		#ifndef _ETCPIP
			schedule ();
         #else
			#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
				// dlytsk(cur_task, DLY_TICKS, 1+TIME_PLUS) ;
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
             #else
            	YIELD();
            #endif
        #endif
        
		continue;
	   } //	if (usblp->writeurb.status != 0)
	*/
    
	#ifndef _ETCPIP
    /*
		writecount += usblp->writeurb->transfer_buffer_length;
		USBLPT_DBG_INIT(iprintf(
	   		"[USBLPT INIT] usblp_write() : writecount = %d\n",
			writecount
        	));
        
		usblp->writeurb->transfer_buffer_length = 0;

	if (writecount == count)
       {
		#ifndef NO_SEMPHORE
			up (&usblp->sem);
		#endif // ndef NO_SEMPHORE
        
		USBLPT_DBG_WRITE(iprintf(
	   		"[USBLPT WRITE] usblp_write() : write ok \n"
        	));
            
		break;
	   }
    */
	#endif // ndef _ETCPIP
    
	usblp->writeurb->transfer_buffer_length
    			 = (input_count - writecount) < USBLPT_BUF_SIZE ?
					 (input_count - writecount) : USBLPT_BUF_SIZE;

	USBLPT_DBG_WRITE(iprintf(
	   	"[USBLPT WRITE] usblp_write() : usblp->writeurb->transfer_buffer_length = %d\n",
		usblp->writeurb->transfer_buffer_length
        ));
	USBLPT_DBG_WRITE(iprintf(
    	"[USBLPT WRITE] usblp_write() : Copy ..\n"));
        
	/* Prototype: unsigned long __arch_copy_from_user(void *to,
                             const void *from,unsigned long n);
	 * Purpose  : copy a block from user memory to kernel memory
	 * Params   : to   - kernel memory
	 *          : from - user memory
	 *          : n    - number of bytes to copy
	 * Returns  : Number of bytes NOT copied.
	 */

	if (copy_from_user(usblp->writeurb->transfer_buffer,
    			buffer + writecount,
				usblp->writeurb->transfer_buffer_length))
       {
		#ifndef NO_SEMPHORE
        	/*
			up(&usblp->sem);
            */
		#endif // ndef NO_SEMPHORE
		return writecount ? writecount : -EFAULT;
       }

	USBLPT_DBG_WRITE(iprintf(
    	"[USBLPT WRITE] usblp_write() : Write ..\n"));
        
	usblp->writeurb->dev = usblp->usb_dev;
	usblp->wcomplete = 0;
    /*
	usblp->writeurb->hcpriv = 0 ;
    */
	hcd = usblp->writeurb->dev->bus->hcpriv_hcd ;
        
	// if (urb->dev != hcd->self.root_hub)
    if (usblp->writeurb->hcpriv) {
			USBLPT_DBG_URB(iprintf("[USBLPT URB] usblp_write() : free priv .. \n" )) ;
			USBLPT_DBG_FLOW(iprintf( "[USBLPT FLOW] usblp_write() : Call hcd->driver->urb_free_priv() ...\n" )) ;
	
			hcd->driver->urb_free_priv(hcd, usblp->writeurb->hcpriv);
            
			USBLPT_DBG_URB(iprintf("[USBLPT URB] usblp_write() : Set Vaue : usblp->writeurb->hcpriv = 0 ..\n" )) ;
            
			usblp->writeurb->hcpriv = NULL;
           }


	// 2003/9/9 B. S. Liaw
	if(	usb_endpoint_halted(usblp->writeurb->dev,
				usb_pipeendpoint (usblp->writeurb->pipe),
				usb_pipeout (usblp->writeurb->pipe))
      ) {
		dprintf("[USBLPT] Clear Halted Write Pipe ..\n") ;
		result = usb_clear_halt(usblp->writeurb->dev,usblp->writeurb->pipe);
		if (result < 0) {
		    // goto Done;
			dprintf("[USBLPT] Clear Halted Write Pipe Fail !!\n") ;
           }
      }
	

	USBLPT_DBG_CTLMSG(iprintf(
    "[USBLPT CTLMSG] usblp_write() : &(usblp->writeurb->hcpriv) = %08XH usblp->writeurb->hcpriv=%08XH\n",
    &(usblp->writeurb->hcpriv) , usblp->writeurb->hcpriv
    ));
	USBLPT_DBG_CTLMSG(iprintf(
    "[USBLPT CTLMSG] usblp_write() : &(usblp->writeurb->callback) = %08XH usblp->writeurb->callback=%08XH\n",
    &(usblp->writeurb->callback) , usblp->writeurb->callback
    ));
    
	USBLPT_DBG_CTLMSG(iprintf(
    "[USBLPT CTLMSG] usblp_write() : &(usblp->readurb->hcpriv) = %08XH usblp->readurb->hcpriv=%08XH\n",
    &(usblp->readurb->hcpriv) , usblp->readurb->hcpriv
    ));
	USBLPT_DBG_CTLMSG(iprintf(
    "[USBLPT CTLMSG] usblp_write() : &(usblp->readurb->callback) = %08XH usblp->readurb->callback=%08XH\n",
    &(usblp->readurb->callback) , usblp->readurb->callback
    ));
    
	#ifdef _ETCPIP
	#ifdef USBLPT_PRINTERS
		usblp->usblp_wr_done = 0 ;
	#else
		usblp_wr_done = 0 ;
	#endif
		USBLPT_DBG_CTLMSG(iprintf("[USBLPT CTLMSG] usblp_write() : usblp_wr_done=%d\n"
	    	,usblp_wr_done
    	    ));
	#endif // _ETCPIP
    /*
	#ifndef _ETCPIP
    	
		if (usb_submit_urb(usblp->writeurb, GFP_KERNEL)) {
			count = -EIO;
			#ifndef NO_SEMPHORE
			 up (&usblp->sem);
            #endif // ndef NO_SEMPHORE
			break;
		}
        
	#else
    */   
	// 2003/4/8 B. S. Liaw
	wait_urb_timeout = USBLPT_WRITE_TIMEOUT;
    
	USBLPT_DBG_URB(iprintf("[USBLPT URB] usblp_write() : wait_urb_timeout = %d \n" ,
		wait_urb_timeout
        ));
	/*
	// dprintf("[USBLPT URB] usblp_write() : set usblpt_want_write .. \n") ;
	USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_write() : set usblpt_want_write .. \n")) ;

	// BLOCK_PREEMPTION ;
	usblpt_want_write = 1 ;
	// UNBLOCK_PREEMPTION ;
	*/

		// #ifdef USB_BS_ASSERT
		#ifdef USBLPT_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		   ) ) {
				iprintf("[ASSERT][USB] usblp_write(1) : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT

	#ifdef USBLPT_WAIT_FOR_ADSL
		#error: Please Donot Define USBLPT_WAIT_FOR_ADSL !!
		#if defined(_S3C2510)
		#ifdef _ADSL
		// dprintf("[USBLPT URB] usblp_write() : set usblpt_write .. \n") ;
		USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_write(1) : set usblpt_write .. \n")) ;
		usblpt_write = 1 ;
	   {
		unsigned long wait_adsl_timeout ;
		int x , y ;

		// wait for adsl free
        for(wait_adsl_timeout = 0 ;
			wait_adsl_timeout < USBLPT_WRITE_TIMEOUT ;
			wait_adsl_timeout++ ) {
			// for (x=0; x<100; x++) {
				// for (y=0; y<10; x++) {
					if (adsl_access <=0) {
						goto loop_out ;
					   }
    	    	//   } // for y
    	    //    } // for x
		   } // for wait_adsl_timeout

		// ----------
		loop_out:
		// ----------

        if ( wait_adsl_timeout >= USBLPT_WRITE_TIMEOUT) {
			USB_DBG_INFO(iprintf("[USB INFO] usblp_write() : wait for ADSL free Time Out !! \n")) ;
			// USBLPT_DBG_INFO(iprintf("[USBLPT INFO] usblp_write() : wait for ADSL free Time Out !! \n")) ;
			// dprintf("[USBLPT] wait for ADSL free Time Out !! \n") ;
	       }
		  else {
			// dprintf("[USBLPT] wait for ADSL free Ok. \n") ;
			USBLPT_DBG_INIT(iprintf("[USBLPT INIT] wait for ADSL free Ok. \n")) ;
		   }
	      }
		#endif // _ADSL
		#endif // _S3C2510
	#endif //USBLPT_WAIT_FOR_ADSL
	/*
	// dprintf("[USBLPT URB] usblp_write() : clear usblpt_want_write .. \n") ;
	USBLPT_DBG_INIT(iprintf("[USBLPT INIT]  usblp_write() : clear usblpt_want_write .. \n")) ;
	usblpt_want_write = 0 ;
	*/
	#ifdef USBLPT_WAIT_FOR_ADSL
		// dprintf("[USBLPT URB] usblp_write() : set usblpt_write .. \n") ;
		USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_write(2) : set usblpt_write .. \n")) ;
		// BLOCK_PREEMPTION ;
		usblpt_write = 1 ;
		// UNBLOCK_PREEMPTION ;
	#endif // USBLPT_WAIT_FOR_ADSL

	//retv =  usb_start_wait_urb(&usblp->readurb,timeout,
    //							&actual_length);
    
    
	// This Donot free urb
	//retv =  usblp_start_wait_urb(&usblp->writeurb,wait_urb_timeout,
    //							&actual_length);
	//retv =  usblp_start_wait_urb(usblp->writeurb,wait_urb_timeout,
    //							&actual_length);
	// This Donot free urb
	retv =  usb_start_wait_urb_nf(usblp->writeurb,wait_urb_timeout, // 1 ,
    							&actual_length);
	// BLOCK_PREEMPTION ;

	#ifdef USBLPT_WAIT_FOR_ADSL
	// dprintf("[USBLPT URB] usblp_write() : clear usblpt_write .. \n") ;
	USBLPT_DBG_INIT(iprintf("[USBLPT INIT]  usblp_write() : clear usblpt_write .. \n")) ;
	usblpt_write = 0 ;
	#endif // USBLPT_WAIT_FOR_ADSL
	/*
	// dprintf("[USBLPT URB] usblp_write() : clear usblpt_want_write .. \n") ;
	USBLPT_DBG_INIT(iprintf("[USBLPT INIT]  usblp_write() : clear usblpt_want_write .. \n")) ;
	usblpt_want_write = 0 ;
	*/
	// UNBLOCK_PREEMPTION ;

	// 2003/10/13 B. S. Liaw
	#ifdef USBLPT_PRINTERS
	if (usblp->usblp_plug_out) {
	#else
	if (usblp_plug_out >= 1) {
	#endif
		goto write_end ;
       }

	// dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;

    USBLPT_DBG_WRITE(iprintf("[USBLPT WRITE] usblp_write() : retv = %d , actual_length = %d \n" ,
		retv , actual_length
        ));
    USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_write(3) : retv = %d , actual_length = %d \n" ,
		retv , actual_length
        ));
        
    if (retv < 0 )
       {
        // ---------------
        write_fail:
        // ---------------
        
		USBLPT_DBG_INIT(iprintf(
			"[USBLPT INIT] usblp_write(4) : Error : usblp write Fail !!\n" )) ;

		iprintf("[USBLPT] Error : Print Data write Fail !! \n") ;
		#ifdef _USB_SYSLOG
		    // sprintf(usbsyslog_buf,"[USBLPT] Error : Print Data write Fail !! ");
						#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_PR_WR_DATA_FAIL
				//sprintf_SetSystemLog(USB_SYS_LOG_PR_WR_DATA_FAIL);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_PR_WR_DATA_FAIL);
#else
				//sprintf_SetSystemLog("[USBLPT] Error : Print Data write Fail !! ");
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] Error : Print Data write Fail !! ");
#endif // USB_SYS_LOG_PR_WR_DATA_FAIL
			}
		#endif // _SYSLOG
            
		if (!usblp || !usblp->usb_dev || (usblp->opened==0) )
		   {
			dprintf("[USBLPT] Error : usblp_bulk_write() : !usblp !usblp->usb_dev (usblp->opened==0) \n" ) ;
    
			USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_write() End With Fail !!\n"));
			goto write_end ;
			// return;
		   }
		#ifdef USBLPT_PRINTERS
		if (usblp->usblp_plug_out)
		#else
		if (usblp_plug_out >= 1)
		#endif
		   {
			goto write_end ;
           }
                   
		USBLPT_DBG_INIT(iprintf(
				"[USBLPT INIT] usblp_write(5) : Check Status ..\n"
				));
		err = usblp_check_status(usblp, err);
		// dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
            
        // ---------------
        write_fail1:
        // ---------------
		#ifdef _ETCPIP
			if ( err == 1 ) // paper empty
			   {
			    // writecount = -EIO;
			    // writecount = 0 ;
				USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_write(6) : Paper Out !!\n"));
			   }
			else if ( err == 2 ) // not selected
			   {
			    writecount = -EIO;
				ret_val = -EIO ;
			   }
			else if ( err == 3 ) // error
			   {
			    writecount = -EIO;
				ret_val = -EIO ;
			   }
		#endif // def _ETCPIP
        
	    // 2003/5/2 B. S. Liaw
		// usblp->writeurb->status = 0 ;
        
        #ifndef _ETCPIP
        	/*
        	count = -EIO;
            */
        #else
        	/*
			USBLPT_DBG_WRITE(iprintf("[USBLPT WRITE] usblp_write() : usblp->writeurb->"
    		    "actual_length = %d\n",
				usblp->writeurb->actual_length
	        	));
	    	writecount += usblp->writeurb->actual_length ;
            */
            // count = writecount ;
        	// count = -EIO;
            ret_val = -EIO ;
        #endif

		#ifdef _ETCPIP
        	/*
			usb_lpt_ok = 0 ;
            */
		#endif // _ETCPIP
		break;
	   }
       
    // Begin 2003/4/11 B. S. Liaw
    
	if (usblp->writeurb->status != 0)
       {
		if (usblp->quirks & USBLPT_QUIRK_BIDIR)
	       {
			#ifndef _ETCPIP
			    if (usblp->writeurb.status != -EINPROGRESS)
			  #else
			    if (usblp->writeurb->status != -EINPROGRESS)
			#endif
    	       {
				dprintf("[USBLPT] UsbLpt-%d: error %d writing to printer \n",
					usblp->minor, usblp->writeurb->status);
	           }
			err = usblp->writeurb->status;
		   }
	      else
           {
			USBLPT_DBG_INIT(iprintf(
				"[USBLPT INIT] usblp_write(7) : write fail !!\n"
				));
			/*
			#ifndef _ETCPIP
				goto write_fail ;
             #else
			*/
			#ifdef USBLPT_PRINTERS
			if (usblp->usblp_plug_out) {
			#else
			if (usblp_plug_out >= 1) {
			#endif
				goto write_end ;
               }
			/*
			#endif
			*/
    	   }
	   } //	if (usblp->writeurb.status != 0)
       
	USBLPT_DBG_WRITE(iprintf("[USBLPT WRITE] usblp_write() : usblp->writeurb->"
        "transfer_buffer_length = %d\n",
		usblp->writeurb->transfer_buffer_length
    	));
    /*
	writecount += usblp->writeurb->transfer_buffer_length;
    */
	USBLPT_DBG_WRITE(iprintf("[USBLPT WRITE] usblp_write() : usblp->writeurb->"
        "actual_length = %d\n",
		usblp->writeurb->actual_length
    	));
    writecount += usblp->writeurb->actual_length ;
    
	USBLPT_DBG_WRITE(iprintf(
	   	"[USBLPT WRITE] usblp_write() : writecount = %d\n",
		writecount
	    ));
	USBLPT_DBG_SPOOL(iprintf(
	   	"[USBLPT SPOOL] usblp_write() : writecount = %d\n",
		writecount
	    ));
    
	usblp->writeurb->transfer_buffer_length = 0;

	if (writecount == input_count)
	   {
		#ifndef NO_SEMPHORE
            /*
			up (&usblp->sem);
            */
		#endif // ndef NO_SEMPHORE
    
		// dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
		USBLPT_DBG_WRITE(iprintf(
			"[USBLPT WRITE] usblp_write() : write ok \n"
    		));
        
		break;
	   }
       
    // End 2003/4/11 B. S. Liaw
    /*    
    #endif
    */
    #ifndef NO_SEMPHORE
    	/*
		up (&usblp->sem);
        */
    #endif // ndef NO_SEMPHORE


	#ifdef USBLPT_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
	   ) ) {
			iprintf("[ASSERT][USB] usblp_write(2) : " ) ;
    	    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
		    iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
	#endif // USB_BS_ASSERT
	dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
   } // while

 // #ifdef DEBUG
 #ifdef USBLPT_DBG_INIT_ON
    /*
	USBLPT_DBG_WRITE(iprintf(
	   "[USBLPT WRITE] usblp_write() : Check Status ..\n"
	   ));
	usblp_check_status(usblp, 0);
    */
 #endif

 // ------------------
 write_end:
 // ------------------
 
 // #ifdef _ETCPIP
 #ifdef USBLPT_PRINTERS
 if (usblp->usblp_plug_out)
 #else
 if (usblp_plug_out >= 1)
 #endif
   {
	/*
		BLOCK_PREEMPTION;
		usblp->readurb->next_step = NEXT_STEP_KILL ;
		UNBLOCK_PREEMPTION;
 
		while( next_step != NEXT_STEP_NONE) 
		   {
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
    
			BLOCK_PREEMPTION;
			next_step = usblp->readurb->next_step ;
			UNBLOCK_PREEMPTION;
		   }
    */   
    // if (usblp)
	if (usblp->opened) // opened
       {
		usblp_release(usblp) ;
	   }
   }
 // #endif // _ETCPIP
   
 // #ifdef _ETCPIP
 #ifdef USBLPT_PRINTERS
 usblp->usb_lpt_access = 0 ;
 #else
 usb_lpt_access = 0 ;
 #endif
 // #endif

 // USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_write() End.\n"));
 USBLPT_DBG_WRITE(iprintf("[USBLPT WRITE] usblp_write() End.\n"));

 // return count;
 // if (ret_val == 0) { // 2003/10/13 B. S. Liaw comment out
 if (ret_val >= 0) {
	// ret_val = input_count ;
	ret_val = writecount ;
   }

 return ret_val ;
}

// ====================================================================

#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
#ifndef _ETCPIP
	/*
	static ssize_t usblp_read(struct file *file, char *buffer,
						  size_t count, loff_t *ppos)
    */
 #else
	ssize_t usblp_read(struct usblp *usblp , char *buffer,
						  size_t count)
#endif
{
 #ifndef _ETCPIP
	struct usblp *usblp = file->private_data;
 #endif // ndef _ETCPIP
 #ifndef NO_WAITQUEUE
	#ifndef _ETCPIP
		DECLARE_WAITQUEUE(wait, current);
     #else
		DECLARE_WAITQUEUE(wait, 0);
	#endif
 #endif // ndef NO_WAITQUEUE
 // #ifdef _ETCPIP
	// 2003/4/11 B. S. Liaw
	int retv ;
 #ifdef USBLPT_READ_BY_START_WAIT
	unsigned long read_timeout ;
 #endif // USBLPT_READ_BY_START_WAIT
	int actual_length ;
    
	unsigned int next_step ;
    
 // #endif // _ETCPIP

 #ifndef USBLPT_READ_BY_URBTASK
	struct usb_hcd		*hcd ;
 #endif // ndef USBLPT_READ_BY_URBTASK

 #ifdef USBLPT_READ_BY_CHECK
	int ret ;
 #endif
 
 // USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_read() Begin ..\n"));
 USBLPT_DBG_READ(iprintf("[USBLPT READ] usblp_read() Begin ..\n"));

 if( usblp_plug_out >= 1 ){
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] usblp_read() : Usb Lpt Plug Out !!\n" )) ;
	dprintf("[USBLPT] Printer Plug Out !!\n");
    return -EINVAL ; 
   }

 if (usb_lpt_released)
   {
    USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Error : usblp_read() : Printer Driver Released !! \n" )) ;
	return -EINVAL ;
   }
       
 USBLPT_DBG_MEM(iprintf("[USBLPT MEM] usblp = %08XH \n" , usblp )) ;

 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
 if (!usblp->bidir)
   {
	return -EINVAL;
   }
 #else
	return -EINVAL;
 #endif

 #ifndef NO_SEMPHORE
	down (&usblp->sem);
 #endif // ndef NO_SEMPHORE
 
 #ifdef _ETCPIP
	usb_lpt_access = 1 ;
 #endif
 
 #ifdef _ETCPIP
	// 2003/4/8 B. S. Liaw
    if (!usblp || !usblp->usb_dev || (usblp->opened==0) )
       {
    	dprintf("[USBLPT] Error : usblp_read() : !usblp !usblp->usb_dev (usblp->opened==0) \n") ;
        
	#ifdef USBLPT_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
	   ) ) {
			iprintf("[ASSERT][USB] usblp_write(2) : " ) ;
    	    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
		    iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
	#endif // USB_BS_ASSERT
		dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
        USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_read() End With Fail !!\n")) ;
		count = -ENODEV;
		goto done;
      }
 #endif // _ETCPIP
 
 /*
 if (!usblp->usb_dev)
   {
	count = -ENODEV;
	goto done;
   }
 */

 #ifdef USBLPT_READ_BY_CHECK
 
	if (! usblp->readurb->hcpriv)
	   {
		usblp->readurb->dev = usblp->usb_dev;
	 	usblp->readcount = 0;
		usblp->rcomplete = 0;
		USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_read(1) : usb_submit_urb() .. \n" )) ;
		usblp_rd_done = 0 ;
		USBLPT_DBG_CTLMSG(iprintf("[USBLPT CTLMSG] usblp_read(1) : usblp_rd_done=%d\n"
	    	,usblp_rd_done
    	    ));
        /*
		usb_fill_bulk_urb(usblp->readurb, usblp->usb_dev,
			usb_rcvbulkpipe(usblp->usb_dev,
			usblp->protocol[protocol].epread->bEndpointAddress),
			usblp->readbuf, USBLP_BUF_SIZE,
			usblp_bulk_read, usblp);
		*/
		ret = usb_submit_urb(usblp->readurb);
        
		#ifdef USBLPT_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
			   ) ) {
				iprintf("[ASSERT][USB] usblp_read(2) : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
	    	   }
		#endif // USB_BS_ASSERT

		#ifndef USBLPT_READ_NO_DELAY
			dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
		#endif // ndef USBLPT_READ_NO_DELAY
            
		if (ret<0)
		   {
			USB_DBG_INFO(iprintf("[USB INFO] usblp_read(1) : usb_submit_urb failed (%d) !!\n", ret )) ;
			count = -EIO;
		   }
		// dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
		goto done;
       }
       
	#ifdef USB_BS_ASSERT
		if (!(     (usblp->readurb->awd_done_old > 0)
				&& (   (    usblp->readurb->awd_done
                         == usblp->readurb->awd_done_old
                       )
                    || (    usblp->readurb->awd_done
                         == ~usblp->readurb->awd_done_old
                       )
                   )
    	     )
	       )
		   {
			iprintf("[ASSERT][USB] usblp_read(1) : " ) ;
    	    iprintf("Error : Invalid call seguence : %d %d !!\n", usblp->readurb->awd_done , usblp->readurb->awd_done_old) ;
	        iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
	#endif
       
	if (usblp->readurb->awd_done == ~usblp->readurb->awd_done_old)
	   {
        
		// urb->next_step = NEXT_STEP_NONE ; // 0
		if (usblp->readurb->callback)
		   {
			USB_DBG_CALLBACK(iprintf("[USB CALLBACK] usblp_read(1) : Call urb->callback() ...\n" )) ;
			USBLPT_DBG_CALLBACK(iprintf("[USBLPT CALLBACK] usblp_read(1) : Call urb->callback() ...\n" )) ;
			usblp->readurb->callback(usblp->readurb);
		   }
		  else
		   {
			USB_DBG_CALLBACK(iprintf("[USB CALLBACK] usblp_read(1) : Null urb->callback() !! \n" )) ;
			USBLPT_DBG_CALLBACK(iprintf("[USBLPT CALLBACK] usblp_read(1) : Null urb->callback() !! \n" )) ;
		   }
		usblp->readurb->awd_done = 0 ;
        
		#ifdef USB_NO_ISR_UNLINK
			// 2003/9/12 B. S. Liaw BugFix
			USBLPT_DBG_INFO(iprintf(
				"[USBLPT INFO] usblp_read() : usb_unlink_urb() .. \n"
				)) ;
		       {
				int status_unlink ;
				status_unlink = usb_unlink_urb(usblp->readurb);
				if (status_unlink && status_unlink != -EINPROGRESS) {
					USB_DBG_INFO(iprintf("[USB INFO] usblp_read() : usb_unlink_urb fail : %d !!\n", status_unlink));
				   }
			   }
			/*
			USBLPT_DBG_INFO(iprintf(
				"[USBLPT INFO] usblp_read() : wait_sof() .. \n"
				)) ;

			if (!wait_sof()) {
				USBLPT_DBG_INFO(iprintf(
				"[USBLPT INFO] usblp_read() :  wait_sof() time out !!\n" )) ;
		      }
			*/
		#endif // USB_NO_ISR_UNLINK

       } // if complete
      else
       {
		// count = 0 ;
		USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_read(1) : Not Read In Yet !! \n" )) ;
		USBLPT_DBG_INFO(iprintf("[USBLPT INFO] usblp_read(1) : Not Read In Yet !! \n" )) ;
		// dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
		goto done;
        // return 0 ;
       }
	/*
 if (!usblp->rcomplete) 
   {
	// FIXME:  only use urb->status inside completion
	// callbacks; this way is racey...
    
    USBLPT_DBG_READ(iprintf(
        "[USBLPT READ] usblp_read() : Wait .. \n"
        ));

	USBLPT_DBG_MEM(iprintf(
    	"[USBLPT MEM] &(usblp->readurb->status) = %08XH \n" ,
    	&(usblp->readurb->status)
        )) ;
        
	USBLPT_DBG_MEM(iprintf(
    	"[USBLPT MEM] usblp->readurb->status = %08XH \n" ,
    	usblp->readurb->status
        )) ;
        
	#ifdef _ETCPIP
    	usblp->readurb->status = 0 ;
	#endif
        
	#ifndef _ETCPIP
		while (1==1) 
	  #else
		while (usblp->readurb->status == -EINPROGRESS)
	#endif
       {
		#ifndef _ETCPIP
		  if (signal_pending(current))
           {
			count = -EINTR;
			#ifndef NO_WAITQUEUE
				remove_wait_queue(&usblp->wait, &wait);
			#endif // ndef NO_WAITQUEUE
			goto done;
		   }
		#endif // ndef _ETCPIP
        
		#ifndef NO_SEMPHORE
			up (&usblp->sem);
		#endif // ndef NO_SEMPHORE
		#ifndef _ETCPIP
			set_current_state(TASK_INTERRUPTIBLE);
		#endif // ndef _ETCPIP
        
		#ifndef NO_WAITQUEUE
         #else
            // 2003/4/8 B. S. Liaw
			// dlytsk(cur_task, DLY_TICKS, 1+TIME_PLUS) ;
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
		#endif 
		if (!usblp->rcomplete) {
		#ifndef _ETCPIP
			schedule ();
         #else
			#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
				// dlytsk(cur_task, DLY_TICKS, 1+TIME_PLUS) ;
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
             #else
            	YIELD();
            #endif
        #endif
			} else {
			#ifndef _ETCPIP
				set_current_state(TASK_RUNNING);
			#endif // ndef _ETCPIP
				break;
			}
        
		#ifndef NO_SEMPHORE
			down (&usblp->sem);
		#endif // ndef NO_SEMPHORE
	   } // while
	#ifndef NO_WAITQUEUE
		remove_wait_queue(&usblp->wait, &wait);
	#endif // ndef NO_WAITQUEUE
   } // if
  else
   {
     USBLPT_DBG_READ(iprintf(
          "[USBLPT READ] usblp_read() : racey Donot Wait \n"
          ));
   }
	 */

 if (!usblp->usb_dev)
   {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] usblp_read() : Error : No usblp->usb_dev !!\n")) ;

	count = -ENODEV;
	goto done;
   }
 
 //#ifdef _ETCPIP
 //   usblp->readurb->status = 0 ;
 //#endif
 
 if (usblp->readurb->status)
   {
	USB_DBG_INFO(iprintf("[USB INFO] usblp_read(2) : Error reading from printer !!\n" )) ;
	dprintf("[USBLPT] UsbLpt-%d: error %d reading from printer \n",
			usblp->minor, usblp->readurb->status);
		usblp->readurb->dev = usblp->usb_dev;
 	usblp->readcount = 0;
	usblp->rcomplete = 0;
	usblp_rd_done = 0 ;
	USBLPT_DBG_CTLMSG(iprintf("[USBLPT CTLMSG] usblp_read() : usblp_rd_done=%d\n"
		,usblp_rd_done
        ));
    
		// 2003/9/12 B. S. Liaw , BugFix    	
		hcd = usblp->readurb->dev->bus->hcpriv_hcd ;
    	if (usblp->readurb->hcpriv)
		   {
			USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_read(2) : free priv .. \n" )) ;
			USBLPT_DBG_FLOW(iprintf( "[USBLPT FLOW] usblp_read(2) : Call hcd->driver->urb_free_priv() ...\n" )) ;
	
			hcd->driver->urb_free_priv(hcd, usblp->readurb->hcpriv);
	            
			USBLPT_DBG_URB(iprintf("[USBLPT URB] usblp_read(2) : Set Vaue : usblp->readurb->hcpriv = 0 ..\n" )) ;
            
			usblp->readurb->hcpriv = NULL;
	       }
	/*
		usb_fill_bulk_urb(usblp->readurb, usblp->usb_dev,
			usb_rcvbulkpipe(usblp->usb_dev,
			usblp->protocol[protocol].epread->bEndpointAddress),
			usblp->readbuf, USBLP_BUF_SIZE,
			usblp_bulk_read, usblp);
	*/

	//#ifndef _ETCPIP
	//	if (usb_submit_urb(usblp->readurb, GFP_KERNEL) < 0)
	//		dbg("error submitting urb"); 
			USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_read(2) : usb_submit_urb() .. \n" )) ;
        
		ret = usb_submit_urb(usblp->readurb);

	#ifdef USBLPT_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
	   ) ) {
			iprintf("[ASSERT][USB] usblp_write(2) : " ) ;
    	    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
		    iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
	#endif // USB_BS_ASSERT
		#ifndef USBLPT_READ_NO_DELAY
		dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
		#endif // ndef USBLPT_READ_NO_DELAY
            
		if (ret<0)
		   {
			USB_DBG_INFO(iprintf("[USB INFO] usblp_read(3) : usb_submit_urb failed (%d) !!\n", ret )) ;
			count = -EIO;
			goto done;
		   }

		// dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
        
	//	// 2003/4/8 B. S. Liaw
	//	timeout = USBLP_READ_TIMEOUT;
    //    
	//	USBLPT_DBG_INIT(iprintf(
	//	   	"[USBLPT INIT] usblp_read() : timeout = %d \n" ,
	//		timeout
    //        ));
        
	//	//retv =  usb_start_wait_urb(&usblp->readurb,timeout,
    //    //							&actual_length);
        
        
	//	// This Donot free urb
	//	retv =  usblp_start_wait_urb(&usblp->readurb,timeout,
	    //    							&actual_length);
	//	USBLPT_DBG_INIT(iprintf(
	//	   	"[USBLPT INIT] usblp_read() : retv = %d , actual_length ="
    //        " %d \n" ,
	//		retv , actual_length
    //        ));
            
    //    if (retv < 0 )
    //       {
	//		dprintf("[USBLPT] Error : usblp read Fail !!\n") ;
	//       }
        
	count = -EIO;
	goto done;
   }
 #endif // USBLPT_READ_BY_CHECK

 #ifdef USBLPT_READ_BY_START_WAIT
    #error: Please Donot Define USBLPT_READ_BY_START_WAIT !!
	// Begin 2003/4/11 B. S. Liaw
    
	usblp->readcount = 0;
	usblp->readurb->dev = usblp->usb_dev;
	usblp->readurb->count = 0 ;
    usblp->rcomplete = 0;
    
	// usblp->readurb->hcpriv = 0 ;
    
    
	USBLPT_DBG_MEM(iprintf(
    "[USBLPT MEM] usblp_read() : &(usblp->writeurb->hcpriv) = %08XH usblp->writeurb->hcpriv=%08XH\n",
    &(usblp->writeurb->hcpriv) , usblp->writeurb->hcpriv
    ));
	USBLPT_DBG_MEM(iprintf(
    "[USBLPT MEM] usblp_read() : &(usblp->writeurb->callback) = %08XH usblp->writeurb->callback=%08XH\n",
    &(usblp->writeurb->callback) , usblp->writeurb->callback
    ));
    
	USBLPT_DBG_MEM(iprintf(
    "[USBLPT MEM] usblp_read() : &(usblp->readurb->hcpriv) = %08XH usblp->readurb->hcpriv=%08XH\n",
    &(usblp->readurb->hcpriv) , usblp->readurb->hcpriv
    ));
	USBLPT_DBG_MEM(iprintf(
    "[USBLPT MEM] usblp_read() : &(usblp->readurb->callback) = %08XH usblp->readurb->callback=%08XH\n",
    &(usblp->readurb->callback) , usblp->readurb->callback
    ));


	// #ifdef _ETCPIP
		usblp_rd_done = 0 ;
		USBLPT_DBG_CTLMSG(iprintf("[USBLPT CTLMSG] usblp_read() : usblp_rd_done=%d\n"
	    	,usblp_rd_done
    	    ));
	// #endif // _ETCPIP
    
	read_timeout = USBLPT_READ_TIMEOUT;
    
	USBLPT_DBG_READ(iprintf(
	   	"[USBLPT READ] usblp_read() : Before , read_timeout = %d \n" ,
		read_timeout
        ));
    
	hcd = usblp->readurb->dev->bus->hcpriv_hcd ;
        
	// if (urb->dev != hcd->self.root_hub)
    if (usblp->readurb->hcpriv)
	   {
		USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_read(1) : free priv .. \n" )) ;
		USBLPT_DBG_FLOW(iprintf( "[USBLPT FLOW] usblp_read() : Call hcd->driver->urb_free_priv() ...\n" )) ;
	
		hcd->driver->urb_free_priv(hcd, usblp->readurb->hcpriv);
            
		USBLPT_DBG_URB(iprintf("[USBLPT URB] usblp_read() : Set Vaue : usblp->readurb->hcpriv = 0 ..\n" )) ;
            
		usblp->readurb->hcpriv = NULL;
       }

	//retv =  usb_start_wait_urb(&usblp->readurb,timeout,
    //							&actual_length);
    
    
	// This Donot free urb
	//retv =  usblp_start_wait_urb(&usblp->readurb,timeout,
    //							&actual_length);
	//retv =  usblp_start_wait_urb(usblp->readurb,read_timeout,
    //							&actual_length);
	// This Donot free urb
    retv =  usb_start_wait_urb_nf(usblp->readurb,read_timeout, // 1 ,
    							&actual_length);                            
    // -19	: No such device 
	// -22	: Invalid argument
	USBLPT_DBG_READ(iprintf(
	   	"[USBLPT READ] usblp_read() : retv = %d , actual_length ="
        " %d \n" ,
		retv , actual_length
        ));
        
    if (retv < 0 )
       {
		dprintf("[USBLPT] Error : usblp read Fail !!\n") ;
        count = -EIO;
		goto done;
       }
       
	// End 2003/4/11 B. S. Liaw
 
 #endif // USBLPT_READ_BY_START_WAIT
 
 #ifdef USBLPT_READ_BY_URBTASK
	#error : Please Donot Define USBLPT_READ_BY_URBTASK
	USBLPT_DBG_READ(iprintf("[USBLPT READ] usblp_read() : Wait ..\n"));
	USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_read() : Wait ..\n"));
    
	next_step = NEXT_STEP_NONE ;

	read_timeout = USBLPT_READ_TIMEOUT;
    while(!(next_step & NEXT_STEP_WC))
       {
        read_timeout -- ;
        if (read_timeout <=0)
           {
			USBLPT_DBG_INIT(iprintf("[USBLPT INT] usblp_read() : Error : time out !!\n"));
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
			count = -EFAULT;
			goto done;
           }
           
	#ifdef USBLPT_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
	   ) ) {
			iprintf("[ASSERT][USB] usblp_write(2) : " ) ;
    	    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
		    iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
	#endif // USB_BS_ASSERT
		dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
        NULL ;
        
		BLOCK_PREEMPTION;
		next_step = usblp->readurb->next_step ;
		UNBLOCK_PREEMPTION;
    	NULL ;
        
       }
 #endif // USBLPT_READ_BY_URBTASK

 count = count < usblp->readurb->actual_length - usblp->readcount ?
		count :	usblp->readurb->actual_length - usblp->readcount;

 USBLPT_DBG_READ(iprintf("[USBLPT READ] usblp_read() : Copy ..\n"));
     
 /* Prototype: int __arch_copy_to_user(void *to, const char *from,
                                       size_t n)
  * Purpose  : copy a block to user memory from kernel memory
  * Params   : to   - user memory
  *          : from - kernel memory
  *          : n    - number of bytes to copy
  * Returns  : Number of bytes NOT copied.
  */

 if (copy_to_user(buffer, usblp->readurb->transfer_buffer + usblp->readcount, count))
   {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] usblp_read() : Copy Fail !!\n"));
	count = -EFAULT;
	goto done;
   }

 // if ((usblp->readcount += count) == usblp->readurb->actual_length) 
 usblp->readcount += count ;
 
 USBLPT_DBG_READ(iprintf("[USBLPT READ] usblp_read() : usblp->readcount = %d\n",usblp->readcount )) ;
 USBLPT_DBG_READ(iprintf("[USBLPT READ] usblp_read() : usblp->readurb->actual_length = %d\n",usblp->readurb->actual_length )) ;
 
 #ifdef USB_BS_ASSERT
	if (!(    usblp->readcount 
           <= usblp->readurb->actual_length
	     )
	   )
	   {
		iprintf("[ASSERT][USB] usblp_read(2) : " ) ;
        iprintf("Error : Read Buffer Overflow %d > %d !!\n",usblp->readcount, usblp->readurb->actual_length) ;
	    iprintf("[ASSERT][USB] Continue ..\n") ;
       }
 #endif

 #ifndef USBLPT_READ_BY_CHECK
 // if (usblp->readcount  == usblp->readurb->actual_length) 
 if (usblp->readcount  >= usblp->readurb->actual_length) 
 #endif // def USBLPT_READ_BY_CHECK
   {
	usblp->readcount = 0;
	usblp->readurb->dev = usblp->usb_dev;
	usblp->rcomplete = 0;
    
	#ifdef USBLPT_READ_BY_URBTASK
		usblp_rd_done = 0 ;
		USBLPT_DBG_CTLMSG(iprintf("[USBLPT CTLMSG] usblp_read() : usblp_rd_done=%d\n"
	    	,usblp_rd_done
    	    ));
                
		USBLPT_DBG_READ(iprintf("[USBLPT READ] usblp_read() : Continue ..\n"));
		BLOCK_PREEMPTION;
		usblp->readurb->next_step = 
						#ifndef USBLPT_READ_INFINITE
        							  NEXT_STEP_PAUSE
									| NEXT_STEP_CONTI
									| 
						#endif // ndef USBLPT_READ_INFINITE
                                      NEXT_STEP_SUBMIT
							        | NEXT_STEP_FREE 
                                    // B. S. Liaw : Need Free
    							    ;
                                    
		// usblp->readurb->kill = 0 ;
		UNBLOCK_PREEMPTION;
        
	#ifdef USBLPT_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
	   ) ) {
			iprintf("[ASSERT][USB] usblp_write(2) : " ) ;
    	    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
		    iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
	#endif // USB_BS_ASSERT
		dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
    	NULL ;
        
	#endif // def USBLPT_READ_BY_URBTASK

	#ifdef USBLPT_READ_BY_CHECK
    /*    
	#ifndef _ETCPIP
		if (usb_submit_urb(usblp->readurb, GFP_KERNEL)) {
			count = -EIO;
			goto done;
		}
    */
     	
		hcd = usblp->readurb->dev->bus->hcpriv_hcd ;
        
		// if (urb->dev != hcd->self.root_hub)
        if (usblp->readurb->hcpriv)
		   {
			USBLPT_DBG_URB(iprintf("[USBLPT URB] usblp_read(2) : free priv .. \n" )) ;
			USBLPT_DBG_FLOW(iprintf( "[USBLPT FLOW] usblp_read() : Call hcd->driver->urb_free_priv() ...\n" )) ;
	
			hcd->driver->urb_free_priv(hcd, usblp->readurb->hcpriv);
            
			USBLPT_DBG_URB(iprintf("[USBLPT URB] usblp_read() : Set Vaue : usblp->readurb->hcpriv = 0 ..\n" )) ;
            
			usblp->readurb->hcpriv = NULL;
           }
        
		usblp->readurb->dev = usblp->usb_dev;
	 	usblp->readcount = 0;
		usblp->rcomplete = 0;
		usblp_rd_done = 0 ;
		USBLPT_DBG_CTLMSG(iprintf("[USBLPT CTLMSG] usblp_open() : usblp_rd_done=%d\n"
	    	,usblp_rd_done
    	    ));
		/*
		usb_fill_bulk_urb(usblp->readurb, usblp->usb_dev,
			usb_rcvbulkpipe(usblp->usb_dev,
			usblp->protocol[protocol].epread->bEndpointAddress),
			usblp->readbuf, USBLP_BUF_SIZE,
			usblp_bulk_read, usblp);
		*/
		USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_read(3) : usb_submit_urb() .. \n" )) ;
        
		ret = usb_submit_urb(usblp->readurb);

	#ifdef USBLPT_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
	   ) ) {
			iprintf("[ASSERT][USB] usblp_write(2) : " ) ;
    	    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
		    iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
	#endif // USB_BS_ASSERT
		#ifndef USBLPT_READ_NO_DELAY
		dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
		#endif // ndef USBLPT_READ_NO_DELAY
            
		if (ret<0)
		   {
			USB_DBG_INFO(iprintf("[USB INFO] usblp_read(3) : usb_submit_urb failed (%d) !!\n", ret )) ;
			// count = -EIO;
			goto done;
		   }

		// dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
               
   		/*
		// 2003/4/8 B. S. Liaw
		timeout = USBLP_READ_TIMEOUT;
        
		USBLPT_DBG_INIT(iprintf(
		   	"[USBLPT INIT] usblp_read() : timeout = %d \n" ,
			timeout
            ));
        
		//retv =  usb_start_wait_urb(&usblp->readurb,timeout,
        //							&actual_length);
        
        
		// This Donot free urb
		retv =  usblp_start_wait_urb(&usblp->readurb,timeout,
        							&actual_length);
		USBLPT_DBG_INIT(iprintf(
		   	"[USBLPT INIT] usblp_read() : retv = %d , actual_length ="
            " %d \n" ,
			retv , actual_length
            ));
            
        if (retv < 0 )
           {
			dprintf("[USBLPT] Error : usblp read Fail !!\n") ;
	       }
        */
    #endif
   }
   
 // ---------
 done:
 // ---------

	#ifndef NO_SEMPHORE
		up (&usblp->sem);
    #endif // ndef NO_SEMPHORE
    
 #ifdef _ETCPIP
 	
	if (usblp_plug_out >= 1)
	   {
        /*
		BLOCK_PREEMPTION;
		usblp->readurb->next_step = NEXT_STEP_KILL ;
		UNBLOCK_PREEMPTION;
 
		while( next_step != NEXT_STEP_NONE) 
		   {
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
    
			BLOCK_PREEMPTION;
			next_step = usblp->readurb->next_step ;
			UNBLOCK_PREEMPTION;
		   }
		*/
        
		// if (usblp->usb_dev)
		   {
			usblp_release(usblp) ;
    	   }    
	   }
 #endif // _ETCPIP

 #ifdef _ETCPIP
	usb_lpt_access = 0 ;
 #endif
 USBLPT_DBG_READ(iprintf("[USBLPT READ] usblp_read() End.\n"));
 // USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_read() End.\n"));
	return count;
} // function usblp_read()
#endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw


/*
 * Checks for printers that have quirks, such as requiring unidirectional
 * communication but reporting bidirectional; currently some HP printers
 * have this flaw (HP 810, 880, 895, etc.), or needing an init string
 * sent at each open (like some Epsons).
 // at least EPSON STYLUS C41UX Need this
 * Returns 1 if found, 0 if not found.
 *
 * HP recommended that we use the bidirectional interface but
 * don't attempt any bulk IN transfers from the IN endpoint.
 * Here's some more detail on the problem:
 * The problem is not that it isn't bidirectional though. The problem
 * is that if you request a device ID, or status information, while
 * the buffers are full, the return data will end up in the print data
 * buffer. For example if you make sure you never request the device ID
 * while you are sending print data, and you don't try to query the
 * printer status every couple of milliseconds, you will probably be OK.
 */
static unsigned int usblp_quirks (__u16 vendor, __u16 product)
{
 int i;

 for (i = 0; quirk_printers[i].vendorId; i++)
   {
	if (vendor == quirk_printers[i].vendorId &&
	    product == quirk_printers[i].productId)
       {
        /*
		dprintf("[USBLPT] Disabling reads from problem bidirectional "
    		"printer on usblp%d \n",
			minor);
        */
		return quirk_printers[i].quirks;
       }
   }
 return 0;
}

#ifndef _ETCPIP
   /*
   static struct file_operations usblp_fops = 
   {
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
		.owner=		THIS_MODULE,
	#endif
	.read =		usblp_read,
	.write =	usblp_write,
	.poll =		usblp_poll,
	.ioctl =	usblp_ioctl,
	.open =		usblp_open,
	.release =	usblp_release,
   };

 static struct usb_class_driver usblp_class = {
	.name =		"usb/lp%d",
	.fops =		&usblp_fops,
	.mode =		S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP,
	.minor_base =	USBLP_MINOR_BASE,
};
   */
#endif // ndef _ETCPIP

// ====================================================================

// ====================================================================
// ======= usblp_probe() ==============================================
// ====================================================================

#ifndef _ETCPIP
	/*
   #if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	static void *usblp_probe(struct usb_device *dev, unsigned int ifnum)
   #else
    static int usblp_probe(struct usb_interface *intf,
		       const struct usb_device_id *id)
   #endif
   */
 #else
	static int usblp_probe(struct usb_interface *intf,
			const struct usb_device_id *id)
#endif
{
 // struct usb_device *dev = interface_to_usbdev (intf);
 struct usb_device *dev ;
 // struct usblp *usblp = 0;
 struct usblp *local_usblp ;
 int protocol;
 int retval;
 int error;
 
 #ifdef USB_DBGSTRING_ON
	char name[10];
 #endif // USB_DBGSTRING_ON
 /*
 #ifdef _USB_SYSLOG
    char usbsyslog_buf[100];
 #endif // _SYSLOG
 */


 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_probe() Begin ..\n"));
 
 // #if defined(USB_DBG_MEM_C_ON) || defined(USB_DBG_MEM_U_ON)
 // #ifdef USBLPT_INIT_ON
 #ifdef USB_DBG_MEM_COUNT_ON
	iprintf("[USB MEM_COUNT] usblp_probe(1) ====== show USB Mem Count .. \n"
    	) ;
	showUsbMemC() ;
	showUsbMemU() ;
 #endif

 #ifdef USBLPT_PRINTERS  /* Charles 20040705 */ 
 iprintf("sizeof(struct usblp)=%d\n", sizeof(struct usblp));
 usb_lpt_ok++;
 #else
 if (usb_lpt_ok)
   {
    USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Printer Aready Plug-In !! \n")) ;
	USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_probe() 1 End with Fail !!\n" )) ;
	return -EIO;
   }
 #endif
 
 #ifdef _ETCPIP
 	dev = interface_to_usbdev (intf);
 	local_usblp = 0;
    
	usb_lpt_ok = 0 ;
	#ifndef USBLPT_PRINTERS
		usb_lpt_released = 0 ;
	#endif
 #endif // _ETCPIP
 
 /* Malloc and start initializing usblp structure so we can use it
  * directly. */
 #ifndef _ETCPIP
	if (!(local_usblp = kmalloc(sizeof(struct usblp), GFP_KERNEL)))
       {
		err("out of memory for usblp");
		goto abort;
	   }
  #else
	// usblp = usblp_array ;
	// usblp = NonCache_contigmalloc(sizeof(struct usblp)) ;
    
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

    // BLOCK_PREEMPTION;

	local_usblp = (struct usblp *)
		#ifdef USB_DBG_MEM_C_ON
			UsbAllocMemC(USB_MEM_C_S, sizeof(struct usblp) ) ;
         #else
            UsbAllocMemC(USB_MEM_C_S) ;
        #endif
                 
    // UNBLOCK_PREEMPTION;

    USBLPT_DBG_MEM(iprintf(
    	"[USBLPT MEM] usblp_probe() : local_usblp = %08XH \n",local_usblp));

    if ( local_usblp == NULL )
       {
		dprintf("[USBLPT] Error : usblp_probe() : out of memory "
        	"for local_usblp\n");
		goto abort;
	   }
 #endif
    
 #ifndef _ETCPIP
    // alloc_mem is on mt_??/USFSup1.c , has built-in fill all 0s
	memset(local_usblp, 0, sizeof(struct usblp));
 #endif // ndef _ETCPIP

 local_usblp->usb_dev = dev;
    
 #ifndef NO_SEMPHORE
	init_MUTEX (&local_usblp->sem);
 #endif // ndef NO_SEMPHORE
 
 
 #ifndef NO_WAITQUEUE
 	/*
	init_waitqueue_head(&usblp->wait);
    */
 #endif // ndef NO_WAITQUEUE
 
 local_usblp->ifnum = intf->altsetting->desc.bInterfaceNumber;
 USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_probe() local_usblp->ifnum=%d \n", local_usblp->ifnum));

 local_usblp->quirks = usblp_quirks(
					dev->descriptor.idVendor,
					dev->descriptor.idProduct);

 if (local_usblp->quirks & USBLPT_QUIRK_INTF_1)
   {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] QUIRK INTF 1 ..\n" )) ;
	local_usblp->ifnum = 1 ;
	USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_probe() : Set Value : local_usblp->ifnum=%d \n", local_usblp->ifnum));
   }
  else
   {
	USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Not USBLPT_QUIRK_INTF_1 \n")) ;
   }
 
 if(local_usblp->ifnum >= local_usblp->usb_dev->actconfig->desc.bNumInterfaces)
   {
    dprintf("[USB LPT] Error : local_usblp->ifnum too large !!\n") ;
   }
 
 /* Look for a free usblp_table entry. */
 #ifdef CONFIG_USB_DEVICEFS
	retval = usb_register_dev(&usblp_fops, USBLPT_MINOR_BASE, 1, &local_usblp->minor);
	if (retval) {
		err("Not able to get a minor for this device.");
			goto abort;
		}
 #endif // CONFIG_USB_DEVICEFS

 #ifndef _ETCPIP
    /*
	usblp->writeurb = usb_alloc_urb(0, GFP_KERNEL);
    */
  #else
	local_usblp->writeurb = usb_alloc_urb(0);
 #endif
	if (!local_usblp->writeurb) {
		err("out of memory");
		goto abort_minor;
	}
    
 USBLPT_DBG_MEM(iprintf(
 	"[USBLPT MEM] usblp_probe() : &local_usblp->writeurb = %08XH\n",
	&(local_usblp->writeurb)
	)) ;
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_probe() : local_usblp->writeurb = %08XH\n",
	local_usblp->writeurb
	)) ;
    
 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	iprintf("[USBLPT] Set Input Connection .. \n");
	#ifdef _USB_SYSLOG
	    //sprintf(usbsyslog_buf,"[USBLPT] Set Input Connection .. ");
					#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_SET_INPUT_CONNECT
			//sprintf_SetSystemLog(USB_SYS_LOG_SET_INPUT_CONNECT);
			sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_SET_INPUT_CONNECT);
#else
			//sprintf_SetSystemLog("[USBLPT] Set Input Connection .. ");
			sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] Set Input Connection .. ");
#endif // USB_SYS_LOG_SET_INPUT_CONNECT
		}
	#endif // _SYSLOG
	/*
	#ifndef _ETCPIP
		local_usblp->readurb = usb_alloc_urb(0, GFP_KERNEL);
	 #else
	*/
	local_usblp->readurb = usb_alloc_urb(0);
	/*
	#endif
	*/
	 if (!local_usblp->readurb) {
		dprintf("[USBLPT] Error : out of memory \n");
		goto abort_minor;
	   }
  #else
	iprintf("[USBLPT] Set Output Only .. \n");
	#ifdef _USB_SYSLOG
	    //sprintf(usbsyslog_buf,"[USBLPT] Set Output Only .. ");
					#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_SET_OUTPUT_ONLY
			//sprintf_SetSystemLog(USB_SYS_LOG_SET_OUTPUT_ONLY);
			sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_SET_OUTPUT_ONLY);
#else
			//sprintf_SetSystemLog("[USBLPT] Set Output Only .. ");
			sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] Set Output Only .. ");
#endif // USB_SYS_LOG_SET_OUTPUT_ONLY
		}
	#endif // _SYSLOG
 #endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw

 USBLPT_DBG_MEM(iprintf(
 	"[USBLPT MEM] usblp_probe() : &local_usblp->readurb = %08XH\n",
	&(local_usblp->readurb)
	)) ;
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_probe() : local_usblp->readurb = %08XH\n",
	local_usblp->readurb
	)) ;
 USBLPT_DBG_MEM(iprintf(
    "[USBLPT MEM] usblp_probe() : &(local_usblp->readurb->pipe) = %08XH "
    "local_usblp->readurb->pipe=%08XH\n",
    &(local_usblp->readurb->pipe) , local_usblp->readurb->pipe
    ));
    
 
 /* Malloc device ID string buffer to the largest expected length,
	 * since we can re-query it on an ioctl and a dynamic string
	 * could change in length. */
     
 #ifndef _ETCPIP
    /*
	if (!(usblp->usb_device_id_string 
    			= kmalloc(USBLP_DEVICE_ID_SIZE, GFP_KERNEL))
       )
       {
		err("out of memory for device_id_string");
		goto abort_minor;
	   }
    */
  #else
    // usblp->usb_device_id_string = device_id_array ;
	// usblp->usb_device_id_string = NonCache_contigmalloc(DEVICE_ID_SIZE)) ;
    
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

    // BLOCK_PREEMPTION;

	local_usblp->device_id_string = (unsigned char *) // 1024
    	/*
		#ifdef USB_DBG_MEM_C_ON
			UsbAllocMemC(USB_MEM_C_L, USBLP_DEVICE_ID_SIZE) ;
         #else
    			 UsbAllocMemC(USB_MEM_C_L) ;
        #endif
        */
		UsbAllocMemU(USB_MEM_U_L
					#ifdef USB_DBG_MEM_U_ON
    	        		, USBLPT_DEVICE_ID_SIZE
	                #endif
                    );
                 
    // UNBLOCK_PREEMPTION;

    USBLPT_DBG_MEM(iprintf(
      "[USBLPT MEM] usblp_probe() : local_usblp->device_id_string = %08XH\n",
      local_usblp->device_id_string));

    if ( local_usblp->device_id_string == NULL )
       {
		dprintf("[USBLPT] Error : usblp_probe() : out of memory for "
        	"device_id_string\n");
		goto abort_minor;
	   }
 #endif
 
 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	local_usblp->readbuf = NULL;
 #endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw
 local_usblp->writebuf = NULL;
 local_usblp->writeurb->transfer_flags = URB_NO_DMA_MAP;
 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	local_usblp->readurb->transfer_flags = URB_NO_DMA_MAP;
 #endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw
 
 /* Malloc write & read buffers.  We somewhat wastefully
  * malloc both regardless of bidirectionality, because the
  * alternate setting can be changed later via an ioctl. */
 
 
 USBLPT_DBG_INIT(iprintf(
	"[USBLPT INIT] usblp_probe() : alloc local_usblp->writebuf .. \n"
    ));
    
 #ifndef _ETCPIP
	/*
	if (!(usblp->writebuf = usb_buffer_alloc(dev, USBLP_BUF_SIZE,
				GFP_KERNEL, &usblp->writeurb->transfer_dma))) {
		err("out of memory for write buf");
		goto abort_minor;
	}
	*/
  #else
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

	local_usblp->writebuf = (char *) // 8192
    	/*
			UsbAllocMemC(USB_MEM_C_L
				#ifdef USB_DBG_MEM_C_ON
					, USBLP_BUF_SIZE
		        #endif
	            ) ;
        */
		UsbAllocMemU(USB_MEM_U_L
					#ifdef USB_DBG_MEM_U_ON
    	        		, USBLPT_BUF_SIZE
	                #endif
                    );

    USBLPT_DBG_MEM(iprintf(
    	"[USBLPT MEM] usblp_probe() : local_usblp->writebuf = %08XH \n",
        local_usblp->writebuf
        ));

    if ( local_usblp->writebuf == NULL )
       {
		dprintf("[USBLPT] Error : usblp_probe() : out of memory "
        	"for writebuf\n");
		goto abort_minor;
	   }
 #endif
 
 
 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
 USBLPT_DBG_INIT(iprintf(
	"[USBLPT INIT] usblp_probe() : alloc local_usblp->readbuf .. \n"
    ));
    
 #ifndef _ETCPIP
	/*
	if (!(usblp->readbuf = usb_buffer_alloc(dev, USBLP_BUF_SIZE,
				GFP_KERNEL, &usblp->readurb->transfer_dma))) {
		err("out of memory for read buf");
		goto abort_minor;
	}
	*/
  #else
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

	local_usblp->readbuf = (char *)   // 8192
    	/*
			UsbAllocMemC(USB_MEM_C_L
				#ifdef USB_DBG_MEM_C_ON
					, USBLP_BUF_SIZE
		        #endif
                ) ;
        */
		UsbAllocMemU(USB_MEM_U_L
					#ifdef USB_DBG_MEM_U_ON
    	        		, USBLPT_BUF_SIZE
	                #endif
                    );

    USBLPT_DBG_MEM(iprintf(
    	"[USBLPT MEM] usblp_probe() : local_usblp->readbuf = %08XH \n",
        local_usblp->readbuf
        ));

    if ( local_usblp->readbuf == NULL )
       {
		dprintf("[USBLPT] Error : usblp_probe() : out of memory "
        	"for readbuf\n");
		goto abort_minor;
	   }
 #endif
 #endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw

 USBLPT_DBG_MEM(iprintf(
    "[USBLPT MEM] usblp_probe() : &(local_usblp->writeurb->hcpriv) = %08XH local_usblp->writeurb->hcpriv=%08XH\n",
    &(local_usblp->writeurb->hcpriv) , local_usblp->writeurb->hcpriv
    ));
 USBLPT_DBG_MEM(iprintf(
    "[USBLPT MEM] usblp_probe() : &(local_usblp->writeurb->callback) = %08XH local_usblp->writeurb->callback=%08XH\n",
    &(local_usblp->writeurb->callback) , local_usblp->writeurb->callback
    ));
    
 USBLPT_DBG_MEM(iprintf(
    	"[USBLPT MEM] usblp_probe() : &(local_usblp->readurb->status) = %08XH \n" ,
    	&(local_usblp->readurb->status)
        )) ;
 USBLPT_DBG_MEM(iprintf(
    	"[USBLPT MEM] usblp_probe() : local_usblp->readurb->status = %08XH \n" ,
    	local_usblp->readurb->status
        )) ;

 USBLPT_DBG_MEM(iprintf(
    "[USBLPT MEM] usblp_probe() : &(local_usblp->readurb->hcpriv) = %08XH local_usblp->readurb->hcpriv=%08XH\n",
    &(local_usblp->readurb->hcpriv) , local_usblp->readurb->hcpriv
    ));
 USBLPT_DBG_MEM(iprintf(
    "[USBLPT MEM] usblp_probe() : &(local_usblp->readurb->callback) = %08XH local_usblp->readurb->callback=%08XH\n",
    &(local_usblp->readurb->callback) , local_usblp->readurb->callback
    ));

 
 
 /* Allocate buffer for printer status */
 /*
 #ifndef _ETCPIP
	
	local_usblp->statusbuf = kmalloc(STATUS_BUF_SIZE, GFP_KERNEL);
	if (!usblp->statusbuf) {
		err("out of memory for statusbuf");
		goto abort_minor;
	}
    
  #else
    */
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

	local_usblp->statusbuf = (char *) // 8 
       /*
		#ifdef USB_DBG_MEM_C_ON
			UsbAllocMemC(USB_MEM_C_S, STATUS_BUF_SIZE) ;
         #else
    		UsbAllocMemC(USB_MEM_C_S) ;
        #endif
        */
		UsbAllocMemU(USB_MEM_U_S
					#ifdef USB_DBG_MEM_U_ON
    	        		, STATUS_BUF_SIZE
	                #endif
                    );

    USBLPT_DBG_MEM(iprintf(
    	"[USBLPT MEM] usblp_probe() : local_usblp->statusbuf = %08XH \n",
        local_usblp->statusbuf
        ));

    if ( local_usblp->statusbuf == NULL )
       {
		dprintf("[USBLPT] Error : usblp_probe() : out of memory "
        	"for statusbuf\n");
		goto abort_minor;
	   }
 /*
 #endif
 */
 
 // #ifdef _ETCPIP
 #ifdef USBLPT_PRINTERS
 	local_usblp->usblp_plug_out = 0 ;
	local_usblp->usb_lpt_access = 1 ;
 #else
	usblp_plug_out = 0 ;
	usb_lpt_access = 1 ;
 #endif
 // #endif
  
 /* Lookup quirks for this printer. */
 /*
 usblp->quirks = usblp_quirks(
					dev->descriptor.idVendor,
					dev->descriptor.idProduct);
 */
 /* Analyze and pick initial alternate settings and endpoints. */
 protocol = usblp_select_alts(local_usblp);
 if(protocol < 0) {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] usblp_probe() : incompatible printer-class device , V/P = %04XH/%04XH \n",
			dev->descriptor.idVendor, dev->descriptor.idProduct));
            
	goto abort_minor;
   }

 /* Setup the selected alternate setting and endpoints. */
    
 USBLPT_DBG_INIT(iprintf(
	"[USBLPT INIT] usblp_probe() : Setup alternate setting and "
       "endpoints ..\n"
	   ));
       
 if (usblp_set_protocol(local_usblp, protocol) < 0)
       {
		dprintf(
	    	"[USBLPT] Error : usblp_probe() : alternate setting "
            "Fail !!\n"
	        );
		goto abort_minor;
       }

 USBLPT_DBG_CTLMSG(iprintf(
    "[USBLPT CTLMSG] usblp_probe() : &(local_usblp->readurb->hcpriv) = %08XH local_usblp->readurb->hcpriv=%08XH\n",
    &(local_usblp->readurb->hcpriv) , local_usblp->readurb->hcpriv
    ));
 USBLPT_DBG_MEM(iprintf(
    "[USBLPT MEM] usblp_probe() : &(local_usblp->readurb->pipe) = %08XH "
    "local_usblp->readurb->pipe=%08XH\n",
    &(local_usblp->readurb->pipe) , local_usblp->readurb->pipe
    ));

 /* Retrieve and store the device ID string. */
    
 if (local_usblp->quirks & USBLPT_QUIRK_NO_1284ID)
   {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] QUIRK NO 1284ID ..\n" )) ;
   }
  else
   {
	usblp_cache_device_id_string(local_usblp);
	// usblp_get_device_id_string(local_usblp);
   }
    
 USBLPT_DBG_CTLMSG(iprintf(
    "[USBLPT CTLMSG] usblp_probe() : &(local_usblp->readurb->hcpriv) = %08XH local_usblp->readurb->hcpriv=%08XH\n",
    &(local_usblp->readurb->hcpriv) , local_usblp->readurb->hcpriv
    ));
 USBLPT_DBG_MEM(iprintf(
    "[USBLPT MEM] usblp_probe() : &(local_usblp->readurb->pipe) = %08XH "
    "local_usblp->readurb->pipe=%08XH\n",
    &(local_usblp->readurb->pipe) , local_usblp->readurb->pipe
    ));

 /*
 #ifdef DEBUG
	USBLPT_DBG_INIT(iprintf(
		"[USBLPT INIT] usblp_probe() : Check Status ..\n"
		));
	usblp_check_status(usblp, 0);
 #endif
 */

 /* If we have devfs, create with perms=660. */
 #ifdef CONFIG_USB_DEVICEFS
	sprintf(name, "usb/lp%d", local_usblp->minor);
 #endif // USB_DBGSTRING_ON

 #ifdef CONFIG_USB_DEVICEFS
	#ifdef USB_DBGSTRING_ON
		local_usblp->usb_devfs = devfs_register(NULL, name,
				      DEVFS_FL_DEFAULT, USB_MAJOR,
				      local_usblp->minor,
				      S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP |
				      S_IWGRP, &usblp_fops, NULL);
	 #else
		local_usblp->usb_devfs = devfs_register(NULL,
				      DEVFS_FL_DEFAULT, USB_MAJOR,
				      local_usblp->minor,
				      S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP |
				      S_IWGRP, &usblp_fops, NULL);
	#endif
 #endif // CONFIG_USB_DEVICEFS

 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	USBLPT_DBG_INFO(iprintf(
		"[USBLPT INFO] UsbLpt-%d: USB %sdirectional printer dev %d "
		"if %d alt %d proto %d vid 0x%4.4X pid 0x%4.4X \n",
		local_usblp->minor, local_usblp->bidir ? "Bi" : "Uni", dev->devnum,
		local_usblp->ifnum,
		local_usblp->protocol[local_usblp->current_protocol].alt_setting,
		local_usblp->current_protocol, local_usblp->usb_dev->descriptor.idVendor,
		local_usblp->usb_dev->descriptor.idProduct
        ));
 #else
	USBLPT_DBG_INFO(iprintf(
		"[USBLPT INFO] UsbLpt-%d: USB UNI-directional printer dev %d "
		"if %d alt %d proto %d \n",
		local_usblp->minor, dev->devnum,
		local_usblp->ifnum,
		local_usblp->protocol[local_usblp->current_protocol].alt_setting,
		local_usblp->current_protocol
        ));

	USBLPT_DBG_INFO(iprintf(
		"[USBLPT INFO] usblp_probe() : vid %04XH pid %04XH \n",
		local_usblp->usb_dev->descriptor.idVendor,
		local_usblp->usb_dev->descriptor.idProduct
        ));
 #endif        
 /*
 // FixMe : Why Be NULL ??
 USBLPT_DBG_INIT(iprintf(
	"[USBLPT INIT] usblp_probe() : intf=%08XH usblp=%08XH \n" ,
    intf, usblp
    )) ;
 dprintf(
	"[USBLPT] usblp_probe() : intf=%08XH usblp=%08XH \n" ,
    intf, usblp
    ) ;
 */   
 // FixMe : Why Be NULL ??
 usb_set_intfdata (intf, local_usblp);
    
 #ifdef USBLPT_BS_ASSERT
	if ( !( local_usblp != 0
          )
       )
       {
		iprintf("[ASSERT][USBLPT] usblp_probe() : " ) ;
        iprintf("Error : pointer local_usblp NULL !! \n") ;
		iprintf("[ASSERT][USBLPT] Continue .. \n" ) ;
	   }   
 #endif
    
 
 /* add device id so the device works when advertised */
 intf->kdev = mk_kdev(USB_MAJOR,local_usblp->minor);
 
 // #ifdef _ETCPIP
 BLOCK_PREEMPTION;
 #ifdef USBLPT_PRINTERS
 	if (!global_usblp)
 		global_usblp = local_usblp ;
 	local_usblp->usb_lpt_access = 0;
 #else
 	global_usblp = local_usblp ;
	usb_lpt_ok = 1 ;
	usb_lpt_access = 0 ;
 #endif
 UNBLOCK_PREEMPTION;
 // #endif // _ETCPIP
 
 USBLPT_DBG_MEM(iprintf(
    "[USBLPT MEM] usblp_probe() : &(local_usblp->readurb->hcpriv) = %08XH local_usblp->readurb->hcpriv=%08XH\n",
    &(local_usblp->readurb->hcpriv) , local_usblp->readurb->hcpriv
    ));
 USBLPT_DBG_MEM(iprintf(
    "[USBLPT MEM] usblp_probe() : &(local_usblp->readurb->callback) = %08XH local_usblp->readurb->callback=%08XH\n",
    &(local_usblp->readurb->callback) , local_usblp->readurb->callback
    ));
 USBLPT_DBG_MEM(iprintf(
    "[USBLPT MEM] usblp_probe() : &(local_usblp->readurb->pipe) = %08XH local_usblp->readurb->pipe=%08XH\n",
    &(local_usblp->readurb->pipe) , local_usblp->readurb->pipe
    ));
    
 if(usblp_open(local_usblp) != 0 )
   {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] usblp_probe() : Open Fail !!\n"));
    
	dprintf("[USBLPT] Printer Probe : Open Fail !!\n");
    
    // return(D_Fail); 
    goto abort ;
   }
    
 #ifdef USB_DBG_MEM_AREA_VERIFY_ON
   {
	extern int	enable_area6(int i, int j);
 
	dprintf("[USBLPT] usblp_probe() : 1234567890\n");
	dprintf("[USBLPT] usblp_probe() : enable area6 ...\n");
	MASK_INTS();
	enable_area6(1, 1);
	dprintf("[USBLPT] usblp_probe() : 1234567890\n");
	dprintf("[USBLPT] usblp_probe() : 1234567890\n");
	UNMASK_INTS();
	dprintf("[USBLPT] usblp_probe() : OK\n");
			
	//dprintf("[USBLPT] usblp_probe() : writing address 0x10114\n");
	//*((unsigned long*)0x10114) = 0;
	//dprintf("[USBLPT] usblp_probe() : OK\n");
   }		
 #endif // USB_DBG_MEM_AREA_VERIFY_ON
 
 // #ifdef USBLPT_DBG_INIT_ON
	USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_probe() : Check Status ..\n"));
	usblp_check_status(local_usblp, 0);
 // #endif

 #ifdef USBLPT_PROBE_RESET // 2003/10/13 B. S. Liaw
 #ifdef USBLPT_RESET_CMD // 2003/10/14 B. S. Liaw
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] usblp_probe() : usblp_reset() .. \n"));
	// dprintf("[USBLPT] Printer Reset ..\n");
    
	wait_ms(USBLPT_INIT_WAIT_MS);

	error = usblp_reset(local_usblp);
 
	if(error < 0) {
		USBLPT_DBG_INFO(iprintf(
			"[USBLPT INFO] Error : Initial Cannot Reset Printer : %d !! \n",
			error));

		iprintf("[USBLPT] Error : Initial Cannot Reset Printer : %d !! \n",
			error);

		#ifdef _USB_SYSLOG
		    //sprintf(usbsyslog_buf,"[USBLPT] Error : Initial Cannot Reset Printer : %d !! \n",
			//	error);
						#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_PR_INIT_RESET_FAIL
				//sprintf_SetSystemLog("%s : %d !! \n", USB_SYS_LOG_PR_INIT_RESET_FAIL, error);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "%s : %d !! \n", USB_SYS_LOG_PR_INIT_RESET_FAIL, error);
#else
				//sprintf_SetSystemLog("[USBLPT] Error : Initial Cannot Reset Printer : %d !! \n", error);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] Error : Initial Cannot Reset Printer : %d !! \n", error);
#endif // USB_SYS_LOG_PR_INIT_RESET_FAIL
			}
		#endif // _SYSLOG
	   }
	  else {
		wait_ms(USBLPT_INIT_WAIT_MS);
		iprintf("[USBLPT] Printer Initial Reset OK.\n");
		#ifdef _USB_SYSLOG
		    //sprintf(usbsyslog_buf,"[USBLPT] Printer Initial Reset OK. ");
						#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_PR_INIT_RESET_OK
				//sprintf_SetSystemLog(USB_SYS_LOG_PR_INIT_RESET_OK);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_PR_INIT_RESET_OK);
#else
				//sprintf_SetSystemLog("[USBLPT] Printer Initial Reset OK. ");
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] Printer Initial Reset OK. ");
#endif // USB_SYS_LOG_PR_INIT_RESET_OK
			}
		#endif // _SYSLOG
	   }
  #else
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] usblp_probe() : Not Call usblp_reset() !! \n"));
	#endif
  #else
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] usblp_probe() : Not Call usblp_reset() !! \n"));
 #endif // USBLPT_PROBE_RESET

 local_usblp->opened = 1 ; // 2003/90/27 B. S. Liaw

 USBLPT_DBG_INIT(iprintf(
	"[USBLPT INIT] usblp_probe() : Set Value : local_usblp->opened = %d \n",
	local_usblp->opened )) ;

 iprintf("[USB] Found Printer , V/P = %04XH/%04XH \n" , local_usblp->usb_dev->descriptor.idVendor,
					local_usblp->usb_dev->descriptor.idProduct);
                    
 #ifdef _USB_SYSLOG
    //sprintf(usbsyslog_buf,"[USB] Found Printer , V/P = %04XH/%04XH " , local_usblp->usb_dev->descriptor.idVendor,
	//				local_usblp->usb_dev->descriptor.idProduct);
				#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_PR_FOUND
		//sprintf_SetSystemLog("%s , V/P = %04XH/%04XH ", USB_SYS_LOG_PR_FOUND, local_usblp->usb_dev->descriptor.idVendor, local_usblp->usb_dev->descriptor.idProduct);
		sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "%s , V/P = %04XH/%04XH ", USB_SYS_LOG_PR_FOUND, local_usblp->usb_dev->descriptor.idVendor, local_usblp->usb_dev->descriptor.idProduct);
#else
		//sprintf_SetSystemLog("[USB] Found Printer , V/P = %04XH/%04XH " , local_usblp->usb_dev->descriptor.idVendor, local_usblp->usb_dev->descriptor.idProduct);
		sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USB] Found Printer , V/P = %04XH/%04XH " , local_usblp->usb_dev->descriptor.idVendor, local_usblp->usb_dev->descriptor.idProduct);
#endif // USB_SYS_LOG_PR_FOUND
	}
 #endif // _SYSLOG

 sprintf(usb_info_str, "Printer , V/P = %04XH/%04XH ", local_usblp->usb_dev->descriptor.idVendor,
					local_usblp->usb_dev->descriptor.idProduct);

 // usb_device_count ++ ; // >= 1 : enable ISR

 iprintf("[USB] Total External Device = %d \n",usb_device_count) ;
 #ifdef _USB_SYSLOG
    //sprintf(usbsyslog_buf,"[USB] Total External Device = %d ",usb_device_count);
				#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_DEV_NUM
		//sprintf_SetSystemLog(USB_SYS_LOG_DEV_NUM, usb_device_count);
		sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_DEV_NUM, usb_device_count);
#else
		//sprintf_SetSystemLog("[USB] Total External Device = %d ",usb_device_count);
		sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USB] Total External Device = %d ",usb_device_count);
#endif // USB_SYS_LOG_DEV_NUM
	}
 #endif // _SYSLOG

#if defined(_NB_SHARE)
	updateShareConfPlug(STYPE_PRINTQ, 0);
#endif // defined(_NB_SHARE)
	led_usb_on();
   
 // #if defined(USB_DBG_MEM_C_ON) || defined(USB_DBG_MEM_U_ON)
 // #ifdef USBLPT_INIT_ON
 #ifdef USB_DBG_MEM_COUNT_ON
	iprintf("[USB MEM_COUNT] usblp_probe(2) ====== show USB Mem Count .. \n"
    	) ;
	showUsbMemC() ;
	showUsbMemU() ;
 #endif
 
 #ifdef LPT_AUTO_DIAG
	ParallelTest() ;
 #endif // LPT_AUTO_DIAG
 
 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_probe() End.\n"));
 return 0;
 
 // -----------------------
 abort_minor:
 // -----------------------

 /* 
 #ifndef _ETCPIP
	usb_deregister_dev (1, local_usblp->minor);
 #endif // ndef _ETCPIP
 */

 // -----------------------
 abort:
 // -----------------------
 if (local_usblp)
   {
	if (local_usblp->writebuf)
       {
		usb_buffer_free(local_usblp->usb_dev, USBLPT_BUF_SIZE,
				local_usblp->writebuf,
				#ifndef USB_NO_VMM // no vitrtul memory management
	                (dma_addr_t) local_usblp->writeurb->transfer_dma
			     #else
    	            (dma_addr_t) local_usblp->writeurb->transfer_buffer
			    #endif
                );
       }
	#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	if (local_usblp->readbuf)
       {
		usb_buffer_free(local_usblp->usb_dev, USBLPT_BUF_SIZE,
				local_usblp->readbuf,
				#ifndef USB_NO_VMM // no vitrtul memory management
	                (dma_addr_t) local_usblp->readurb->transfer_dma
			     #else
    	            (dma_addr_t) local_usblp->readurb->transfer_buffer
			    #endif
                );
       }         
 	#endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw
               
	if (local_usblp->statusbuf)
       {
        /*
        kfree(local_usblp->statusbuf);
        */
		UsbFreeMemU(USB_MEM_U_S, (unsigned char *) // 8
			 		local_usblp->statusbuf
					 #if defined(USB_DBG_MEM_U_ON)
	    		    	,  STATUS_BUF_SIZE
    	    	     #endif
                   ) ;
       }
	if (local_usblp->device_id_string)
       {
        /*
        kfree(local_usblp->usb_device_id_string);
        */
		UsbFreeMemU(USB_MEM_U_L, (unsigned char *) // 1024
			 		local_usblp->device_id_string
					 #if defined(USB_DBG_MEM_U_ON)
	    		    	,  USBLPT_DEVICE_ID_SIZE
    	    	     #endif
			        ) ;
       }
	usb_free_urb(local_usblp->writeurb);

	#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
		usb_free_urb(local_usblp->readurb);
	#endif // USBLPT_BIDIR

	kfree(local_usblp);
   }
      
 // #ifdef _ETCPIP
 #ifdef USBLPT_PRINTERS
 if (global_usblp == local_usblp)
 	global_usblp = 0 ;
 usb_lpt_ok--;
 #else
 local_usblp = 0 ;
 global_usblp = 0 ;
 USBLPT_DBG_INFO(iprintf(
	"[USBLPT INFO] usblp_probe() : global_usblp=%08XH \n",
	global_usblp));
 usb_lpt_access = 0 ;
 #endif
 // #endif
 
 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_probe() 2 End with Fail !!\n" )) ;
 return -EIO;
}

/*
 // Interface descriptors advertise a "Class/Subclass/Protocol" code,
 // such as "7/1/3"
 
 * We are a "new" style driver with usb_device_id table,
 * but our requirements are too intricate for simple match to handle.
 *
 * The "proto_bias" option may be used to specify the preferred protocol
 * for all USB printers (1=7/1/1, 2=7/1/2, 3=7/1/3).  If the device
 * supports the preferred protocol, then we bind to it.
 *
 * The best interface for us is 7/1/2, because it is compatible
 * with a stream of characters. If we find it, we bind to it.
 *
 * Note that the people from hpoj.sourceforge.net need to be able to
 * bind to 7/1/3 (MLC/1284.4), so we provide them ioctls for this purpose.
 *
 * Failing 7/1/2, we look for 7/1/3, even though it's probably not
 * stream-compatible, because this matches the behaviour of the old code.
 *
 * If nothing else, we bind to 7/1/1 - the unidirectional interface.
 */
static int usblp_select_alts(struct usblp *usblp)
{
 struct usb_interface *if_alt;
 struct usb_host_interface *ifd;
 struct usb_endpoint_descriptor *epd, *epwrite ;
 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	struct usb_endpoint_descriptor *epread;
 #endif // USBLPT_BIDIR
 int p, i, e;
 /*
 #ifdef _USB_SYSLOG
	char usbsyslog_buf[100];
 #endif // _SYSLOG
 */


 /*
 for (i = 0; i < config->desc.bNumInterfaces; i++)
   {
	ifp = config->interface + i;

	if (!ifp)
       {
		break;
       }

	dprintf("\n  Interface: %d\n", i);
	for(j = 0; j < ifp->num_altsetting; j++)
        {
		usb_show_interface(ifp->altsetting + j);
       }
  }
 */

 USBLPT_DBG_INIT(iprintf(
	"[USBLPT INIT] usblp_select_alts() : usblp->ifnum=%d \n",
	usblp->ifnum));

 /*
 usblp->ifnum = 1 ;

 dprintf("[USB LPT] Set Value : usblp->ifnum=%d \n", usblp->ifnum);
 */
 // if_alt = &usblp->usb_dev->actconfig->interface[usblp->ifnum];
 if_alt = &usblp->usb_dev->actconfig->interface[usblp->ifnum];
 
 for (p = 0; p < USBLPT_MAX_PROTOCOLS; p++)
   {
	usblp->protocol[p].alt_setting = -1;
   }

 /* Find out what we have. */

 for (i = 0; i < if_alt->num_altsetting; i++)
   {
	ifd = &if_alt->altsetting[i];

	if (	ifd->desc.bInterfaceClass != USB_CLASS_PRINTER // 7 
    	|| ifd->desc.bInterfaceSubClass != 1)
       {
		continue;
       }

	if (ifd->desc.bInterfaceProtocol < USBLPT_FIRST_PROTOCOL ||
		    ifd->desc.bInterfaceProtocol > USBLPT_LAST_PROTOCOL)
       {
		continue;
       }

	/* Look for bulk OUT and IN endpoints. */
	epwrite = 0;

	#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
		epread = 0;
	#endif // USBLPT_BIDIR 

	for(e = 0; e < ifd->desc.bNumEndpoints; e++) {
		epd = &ifd->endpoint[e].desc;

		if ((epd->bmAttributes&USB_ENDPOINT_XFERTYPE_MASK)!=
			    USB_ENDPOINT_XFER_BULK)
				continue;

		if (!(epd->bEndpointAddress & USB_ENDPOINT_DIR_MASK)) {
				if (!epwrite)
				     epwrite=epd;

			} else {
		#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
				if (!epread) 
				    epread=epd;
		#endif // USBLPT_BIDIR 
			}
	   } // for e

	/* Ignore buggy hardware without the right endpoints. */
	// if (!epwrite || (ifd->desc.bInterfaceProtocol > 1 && !epread))
	if (!epwrite) {
		continue;
	   }

	#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
		if (ifd->desc.bInterfaceProtocol > 1 && !epread) {
			continue;
		   }
	#endif // USBLPT_BIDIR 

	#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
		/* Turn off reads for 7/1/1 (unidirectional) interfaces
		 * and buggy bidirectional printers. */

		if (ifd->desc.bInterfaceProtocol == 1) {
			epread = NULL;
		} else if (usblp->quirks & USBLPT_QUIRK_BIDIR) 
           {
			USBLPT_DBG_INFO(iprintf("[USBLPT INFO] USBLPT_QUIRK_BIDIR ..\n" )) ;
			USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Disabling reads from problem bidirectional printer on UsbLpt-%d", usblp->minor )) ;
			//info("Disabling reads from problem bidirectional "
			//	"printer on UsbLpt-%d", usblp->minor);
			dprintf("[USBLPT] This Printer is Output Only \n");
			#ifdef _USB_SYSLOG
			    //sprintf(usbsyslog_buf,"[USBLPT] This Printer is Output Only ");
							#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_OUTPUT_ONLY
					//sprintf_SetSystemLog(USB_SYS_LOG_OUTPUT_ONLY);
					sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_OUTPUT_ONLY);
#else
					//sprintf_SetSystemLog("[USBLPT] This Printer is Output Only ");
					sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] This Printer is Output Only ");
#endif // USB_SYS_LOG_OUTPUT_ONLY
				}
			#endif // _SYSLOG
			epread = NULL;
		}
		  else
		   {
			USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Not USBLPT_QUIRK_BIDIR \n")) ;
		   }
	#endif // USBLPT_BIDIR

	usblp->protocol[ifd->desc.bInterfaceProtocol].alt_setting = i;
	usblp->protocol[ifd->desc.bInterfaceProtocol].epwrite = epwrite;
	#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
		usblp->protocol[ifd->desc.bInterfaceProtocol].epread = epread;
	#endif // USBLPT_BIDIR
   } // for i

 /* If our requested protocol is supported, then use it. */
 if (proto_bias >= USBLPT_FIRST_PROTOCOL &&
	    proto_bias <= USBLPT_LAST_PROTOCOL &&
	    usblp->protocol[proto_bias].alt_setting != -1)
		return proto_bias;

 /* Ordering is important here. */
	if (usblp->protocol[2].alt_setting != -1)
	    return 2;
	if (usblp->protocol[1].alt_setting != -1) 
	   return 1;
	if (usblp->protocol[3].alt_setting != -1)
	    return 3;

 /* If nothing is available, then don't bind to this device. */
 return -1;
}

static int usblp_set_protocol(struct usblp *usblp, int protocol)
{
 int r, alts;
 /*
 #ifdef USBLPT_DBG_MEM_ON
 	struct usblp *usblp_p ;
 #endif //USBLPT_DBG_MEM_ON
 */
 /*
 #ifdef _USB_SYSLOG
    char usbsyslog_buf[100];
 #endif // _SYSLOG
 */


 if (protocol < USBLPT_FIRST_PROTOCOL || protocol > USBLPT_LAST_PROTOCOL)
   {
	return -EINVAL;
   }

 alts = usblp->protocol[protocol].alt_setting;
 if (alts < 0) return -EINVAL;
 r = usb_set_interface(usblp->usb_dev, usblp->ifnum, alts);
 if (r < 0)
    {
	dprintf("[USB LPT] Error : can't set desired altsetting %d on interface %d",
			alts, usblp->ifnum);
		return r;
	}
 /*
 #ifdef _ETCPIP
	usblp->writeurb = usb_alloc_urb(0);
	if (!usblp->writeurb)
       {
	 	iprintf("[USB] couldn't allocate write urb \n");
		// kfree(usblp->descriptor);
		return -1;
	   }
 #endif // _ETCPIP
 */
 USBLPT_DBG_INIT(iprintf(
	   "[USBLPT INIT] usblp_set_protocol() : FILL_BULK_URB write ..\n"
	   ));
    
 usb_fill_bulk_urb(usblp->writeurb, usblp->usb_dev,
		usb_sndbulkpipe(usblp->usb_dev,
		  usblp->protocol[protocol].epwrite->bEndpointAddress),
		usblp->writebuf, 0,
		usblp_bulk_write, usblp);
     
 USBLPT_DBG_MEM(iprintf(
 	"[USBLPT MEM] usblp_set_protocol() : &usblp->writeurb = %08XH\n",
	&(usblp->writeurb)
	)) ;
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_set_protocol() : usblp->writeurb = %08XH\n",
	usblp->writeurb
	)) ;
 /*    
 #ifdef USBLPT_DBG_MEM_ON
 	// usblp_p = usblp->writeurb->context;
 	usblp_p = (usblp->writeurb).context;
 #endif //USBLPT_DBG_MEM_ON
 
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_set_protocol() : usblp_p = %08XH \n",
	usblp_p )) ;
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_set_protocol() : &usblp_p->dev  = %08XH , "
    "usblp_p->dev = %08XH \n",
	&usblp_p->dev,usblp_p->dev
    )) ;
 USBLPT_DBG_MEM(iprintf(
	"[USBLPT MEM] usblp_set_protocol() : &usblp_p->used = %08XH , "
    "usblp_p->used = %d \n",
	&usblp_p->used,usblp_p->used
    )) ;
 */   
 
 USBLPT_DBG_MEM(iprintf(
		"[USBLPT MEM] usblp_set_protocol() : &(usblp->writeurb->context"
    	") = %08XH\n" ,
		&(usblp->writeurb->context)
	    )) ;
 USBLPT_DBG_MEM(iprintf(
		"[USBLPT MEM] usblp_set_protocol() : usblp->writeurb->context "
	    "= %08XH\n" ,
		usblp->writeurb->context
	    )) ;
 
 USBLPT_DBG_CTLMSG(iprintf(
    "[USBLPT CTLMSG] usblp_set_protocol() : &(usblp->writeurb->hcpriv) = %08XH usblp->writeurb->hcpriv=%08XH\n",
    &(usblp->writeurb->hcpriv) , usblp->writeurb->hcpriv
    ));
 USBLPT_DBG_CTLMSG(iprintf(
    "[USBLPT CTLMSG] usblp_set_protocol() : &(usblp->writeurb->callback) = %08XH usblp->writeurb->callback=%08XH\n",
    &(usblp->writeurb->callback) , usblp->writeurb->callback
    ));
 USBLPT_DBG_CTLMSG(iprintf(
    "[USBLPT CTLMSG] usblp_set_protocol() : &(usblp->writeurb->pipe) = %08XH usblp->writeurb->pipe=%08XH\n",
    &(usblp->writeurb->pipe) , usblp->writeurb->pipe
    ));
    


 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	usblp->bidir = (usblp->protocol[protocol].epread != 0);

	if (usblp->bidir) {
		USBLPT_DBG_INIT(iprintf(
			"[USBLPT INIT] usblp_set_protocol() : Bi-Directional \n"
			));
	    /*
		#ifdef _ETCPIP
		usblp->readurb = usb_alloc_urb(0);
		if (!usblp->readurb)
       	   {
	 		iprintf("[USB] couldn't allocate read urb \n");
			// kfree(usblp->descriptor);
			return -1;
	       }
		#endif // _ETCPIP
	    */
    
		USBLPT_DBG_INIT(iprintf(
			"[USBLPT INIT] usblp_set_protocol() : FILL_BULK_URB read ..\n"
			));
        
		USBLPT_DBG_MEM(iprintf(
			"[USBLPT MEM] usblp_set_protocol() : &usblp->readbuf=%08XH "
        	"usblp->readbuf=%08XH\n",
		    &usblp->readbuf , usblp->readbuf
			)) ;
	    /*
		USBLPT_DBG_MEM(iprintf(
			"[USBLPT MEM] usblp_set_protocol() : "
	        "usblp->readbuf+USBLP_BUF_SIZE = %08XH\n" ,
		    usblp->readbuf+USBLP_BUF_SIZE
			)) ;
	    */  
		usb_fill_bulk_urb(usblp->readurb, usblp->usb_dev,
			usb_rcvbulkpipe(usblp->usb_dev,
			usblp->protocol[protocol].epread->bEndpointAddress),
			usblp->readbuf, USBLPT_BUF_SIZE,
			usblp_bulk_read, usblp);
    
		USBLPT_DBG_MEM(iprintf(
			"[USBLPT MEM] usblp_set_protocol() : &(usblp->readurb) = "
        	"%08XH\n",
		    &(usblp->readurb)
			)) ;
		USBLPT_DBG_MEM(iprintf(
			"[USBLPT MEM] usblp_set_protocol() : usblp->readurb = %08XH\n",
			usblp->readurb
			)) ;
    	/*
		#ifdef USBLPT_DBG_MEM_ON
 		// usblp_p = usblp->readurb->context;
 		usblp_p = (usblp->readurb).context;
		#endif //USBLPT_DBG_MEM_ON
 
		USBLPT_DBG_MEM(iprintf(
			"[USBLPT MEM] usblp_set_protocol() : usblp_p = %08XH \n",
			usblp_p )) ;
		USBLPT_DBG_MEM(iprintf(
			"[USBLPT MEM] usblp_set_protocol() : &usblp_p->dev  = %08XH , "
	    	"usblp_p->dev = %08XH \n",
			&usblp_p->dev,usblp_p->dev
	    	)) ;
		USBLPT_DBG_MEM(iprintf(
			"[USBLPT MEM] usblp_set_protocol() : &usblp_p->used = %08XH , "
	    	"usblp_p->used = %d \n",
			&usblp_p->used,usblp_p->used
    		)) ;
	    */
    
		USBLPT_DBG_MEM(iprintf(
			"[USBLPT MEM] usblp_set_protocol() : &(usblp->readurb->context"
        	") = %08XH\n" ,
			&(usblp->readurb->context)
    		)) ;
		USBLPT_DBG_MEM(iprintf(
			"[USBLPT MEM] usblp_set_protocol() : usblp->readurb->context "
        	"= %08XH\n" ,
			usblp->readurb->context
    		)) ;
	    /*
		USBLPT_DBG_MEM(iprintf(
			"[USBLPT MEM] usblp_set_protocol() : &(usblp->readurb->context"
        	") = %08XH\n" ,
			&(((struct usblp *)usblp->readurb)->context)
    		)) ;
		USBLPT_DBG_MEM(iprintf(
			"[USBLPT MEM] usblp_set_protocol() : usblp->readurb->context "
        	"= %08XH\n" ,
			((struct usblp *)usblp->readurb)->context
    		)) ;
	    */
		USBLPT_DBG_MEM(iprintf(
			"[USBLPT MEM] usblp_set_protocol() : usblp->readurb= %08XH , "
        	"&(usblp->readurb->count)= %08XH usblp->readurb->count= %08XH\n",
	    	usblp->readurb , &(usblp->readurb->count) , usblp->readurb->count
    		));
		USBLPT_DBG_MEM(iprintf(
			"[USBLPT MEM] usblp_set_protocol() : &usblp->readurb->context = %08XH "
        	"usblp->readurb->context = %08XH \n",
			&usblp->readurb->context , usblp->readurb->context
    		)) ;
        
    
		USBLPT_DBG_MEM(iprintf(
    		"[USBLPT MEM] usblp_set_protocol() : &(usblp->readurb->status) = %08XH \n" ,
	    	&(usblp->readurb->status)
    	    )) ;
		USBLPT_DBG_MEM(iprintf(
    		"[USBLPT MEM] usblp_set_protocol() : usblp->readurb->status = %08XH \n" ,
	    	usblp->readurb->status
    	    )) ;

		USBLPT_DBG_CTLMSG(iprintf(
    		"[USBLPT CTLMSG] usblp_set_protocol() : &(usblp->readurb->hcpriv) = %08XH usblp->readurb->hcpriv=%08XH\n",
		    &(usblp->readurb->hcpriv) , usblp->readurb->hcpriv
		    ));
		USBLPT_DBG_CTLMSG(iprintf(
    		"[USBLPT CTLMSG] usblp_set_protocol() : &(usblp->readurb->callback) = %08XH usblp->readurb->callback=%08XH\n",
		    &(usblp->readurb->callback) , usblp->readurb->callback
		    ));
		USBLPT_DBG_CTLMSG(iprintf(
		    "[USBLPT CTLMSG] usblp_set_protocol() : &(usblp->readurb->pipe) = %08XH usblp->readurb->pipe=%08XH\n",
		    &(usblp->readurb->pipe) , usblp->readurb->pipe
		    ));

	   }
 #endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw
   
 usblp->current_protocol = protocol;
 
 USBLPT_DBG_INFO(iprintf(
	"[USBLPT INFO] UsbLpt-%d set protocol 7/1/%d \n",
    usblp->minor, protocol
    ));
 
 iprintf("[USBLPT] set USB LPT Protocol %d \n", protocol );

 #ifdef _USB_SYSLOG
	// sprintf(usbsyslog_buf,"[USBLPT] set USB LPT Protocol %d \n", protocol);
				#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_PR_PROTOCOL
		//sprintf_SetSystemLog(USB_SYS_LOG_PR_PROTOCOL, protocol);
		sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_PR_PROTOCOL, protocol);
#else
		//sprintf_SetSystemLog("[USBLPT] set USB LPT Protocol %d \n", protocol);
		sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] set USB LPT Protocol %d \n", protocol);
#endif // USB_SYS_LOG_PR_PROTOCOL
	}
 #endif // _SYSLOG
    
 return 0;
} // function usblp_set_protocol()

// ====================================================================
 
/* Gets the device ID string; either from the device directly (for good
 * devices) or from the cache (for buggy printers which can lock up if some
 * bulk I/O is in progress while the device ID is fetched). */

/*
static int usblp_get_device_id_string(struct usblp *usblp)
{
	int length;

	if (!(usblp->quirks & USBLP_QUIRK_GET_ID))
		return usblp_cache_device_id_string(usblp);

	length = (usblp->device_id_string[0] << 8) + usblp->device_id_string[1];
	if (length < 2)
		length = 2;
	else if (length >= DEVICE_ID_SIZE)
		length = DEVICE_ID_SIZE - 1;

	return length;
}
*/

// ====================================================================

/* Retrieves and caches device ID string.
 * Returns length, including length bytes but not null terminator.
 * On error, returns a negative errno value. */
// static 
int usblp_cache_device_id_string(struct usblp *usblp)
{
 int err, length;


 // Get the device_id string if possible.
 //  FIXME: Could make this kmalloc(length). 
 
 USBLPT_DBG_INIT(iprintf(
	"[USBLPT INIT] usblp_cache_device_id_string() : Get the device_id string ..\n"	));
    
 err = usblp_get_id(usblp, 0, usblp->device_id_string,
 						USBLPT_DEVICE_ID_SIZE - 1);
 
 if (err < 0)
   {
	dprintf("[USBLPT] Error : UsbLpt-%d: error = %d reading IEEE-1284 "
    	"Device ID string \n",
		usblp->minor, err);
	#ifdef _USB_SYSLOG
	    // sprintf(usbsyslog_buf,"[USBLPT] Error : reading IEEE-1284 Device ID string Device ID !! ");
					#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_PR_ID_ERR
			//sprintf_SetSystemLog(USB_SYS_LOG_PR_ID_ERR);
			sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_PR_ID_ERR);
#else
			//sprintf_SetSystemLog("[USBLPT] Error : reading IEEE-1284 Device ID string Device ID !! ");
			sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] Error : reading IEEE-1284 Device ID string Device ID !! ");
#endif // USB_SYS_LOG_PR_ID_ERR
		}
	#endif // _USB_SYSLOG
	usblp->device_id_string[0] = usblp->device_id_string[1] = '\0';
	return -EIO;
   }

 /* First two bytes are length in big-endian.
  * They count themselves, and we copy them into
  * the user's buffer. */
  
 // big-endian
 length = (usblp->device_id_string[0] << 8) + usblp->device_id_string[1];
 if (length < 2)
		length = 2;
	else if (length >= USBLPT_DEVICE_ID_SIZE)
		length = USBLPT_DEVICE_ID_SIZE - 1;
        
 usblp->device_id_string[length] = '\0';

 USBLPT_DBG_INFO(iprintf(
	"[USBLPT INFO] UsbLpt-%d Device ID string [len=%d]=[%s] \n",
	usblp->minor, length, &usblp->device_id_string[2]
    ));

 iprintf("[USBLPT] Device ID : [%s] \n",
 	&usblp->device_id_string[2]
    );
 #ifdef _USB_SYSLOG
    //sprintf(usbsyslog_buf,"[USBLPT] Device ID : [%s] ",
	// 	&usblp->device_id_string[2]
	//    );
				#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_PR_ID
		//sprintf_SetSystemLog("%s : [%s] ", USB_SYS_LOG_PR_ID, &usblp->device_id_string[2]);
		sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "%s : [%s] ", USB_SYS_LOG_PR_ID, &usblp->device_id_string[2]);
#else
		//sprintf_SetSystemLog("[USBLPT] Device ID : [%s] ", &usblp->device_id_string[2]);
		sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] Device ID : [%s] ", &usblp->device_id_string[2]);
#endif // USB_SYS_LOG_PR_ID
	}
 #endif // _SYSLOG

 return length;
} // function usblp_cache_device_id_string()

// ====================================================================

// #ifndef _ETCPIP
static void usblp_disconnect(struct usb_interface *intf)
{
 /*
 // #ifndef NO_SYS_FS
 #ifndef _ETCPIP
	struct usblp *usblp = usb_get_intfdata (intf);
  #else
 */
	// struct usblp *usblp ;
    unsigned int next_step ;
 /*
 #endif
 // #endif // ndef NO_SYS_FS
 */

#if defined(_NB_SHARE)
	updateShareConfPlug(STYPE_PRINTQ, 1);
#endif // defined(_NB_SHARE)
	led_usb_off();

 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] usblp_disconnect() Begin ..\n")) ;
 // dprintf("[USBLPT] usblp_disconnect() Begin ..\n") ;
 /*
 #ifdef _ETCPIP
	#ifndef NO_SYS_FS
    	// FixMe : Why Be NULL ??
		usblp = usb_get_intfdata (intf);
     #else
*/
		/*
        USBLPT_DBG_INFO(iprintf("[USBLPT INFO] FixMe : usblp Already NULL ! \n")) ;
		BLOCK_PREEMPTION;
		usblp = global_usblp ;
		UNBLOCK_PREEMPTION;
		*/
 /*
    #endif
 #endif
 */

 /*
 // FixMe : Why Be NULL ??
 USBLPT_DBG_INIT(iprintf(
	"[USBLPT INIT] usblp_disconnect() : intf=%08XH usblp=%08XH \n" ,
    intf, usblp
    )) ;
 dprintf(
	"[USBLPT INIT] usblp_disconnect() : intf=%08XH usblp=%08XH \n" ,
    intf, usblp
    ) ;
 */   
 #ifdef USBLPT_BS_ASSERT
	if ( !( (unsigned long) global_usblp > (unsigned long) 0x10000L
          )
       )
       {
		iprintf("[ASSERT][USBLPT] usblp_disconnect() : " ) ;
        iprintf("Error : Invalid pointer global_usblp %08XH !! \n",global_usblp) ;
		iprintf("[ASSERT][USBLPT] Continue .. \n" ) ;
	   }   
 #endif
    
 /* remove device id to disable open() */
 intf->kdev = NODEV;

 if (!global_usblp || !global_usblp->usb_dev) {
		dprintf("[USBLPT] Error : bogus disconnect");
		// BUG ();
	}

 #ifndef NO_SEMPHORE
		down (&global_usblp->sem);
 #endif // ndef NO_SEMPHORE
 #ifndef _ETCPIP
		lock_kernel();
 #endif // ndef _ETCPIP
 
 global_usblp->usb_dev = NULL;
 
 USBLPT_DBG_INIT(iprintf(
	"[USBLPT INIT] usblp_disconnect() : intf=%08XH "
    "usb_set_intfdata NULL \n" ,
    intf
    )) ;
    
 /*
 dprintf(
	"[USBLPT] usblp_disconnect() : intf=%08XH  "
    "usb_set_intfdata NULL \n" ,
    intf
    ) ;
 */
 
 USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_disconnect() : intf=%08XH , Set NULL ..\n",intf )) ;
 
 usb_set_intfdata (intf, NULL);
 #ifdef USBLPT_READ_BY_URBTASK
 /*
 BLOCK_PREEMPTION;
 global_usblp->readurb->next_step = NEXT_STEP_PAUSE ;
 next_step = NEXT_STEP_PAUSE ;
 UNBLOCK_PREEMPTION;
 
 USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_disconnect() : wait urb_task pause .. \n" )) ;
 
 while(!( next_step & NEXT_STEP_WC) )
   {
	dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
    
	BLOCK_PREEMPTION;
	next_step = usblp->readurb->next_step ;
	UNBLOCK_PREEMPTION;
   }
 */
  
 BLOCK_PREEMPTION;
 global_usblp->readurb->next_step = NEXT_STEP_KILL ;
 next_step = NEXT_STEP_KILL ;
 UNBLOCK_PREEMPTION;
 
 USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_disconnect() : wait urb_task stop .. \n" )) ;
 
 while( next_step != NEXT_STEP_NONE) 
   {
	#ifdef USBLPT_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
	   ) ) {
			iprintf("[ASSERT][USB] usblp_write(2) : " ) ;
    	    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
		    iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
	#endif // USB_BS_ASSERT
	dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
    NULL ;
    
	BLOCK_PREEMPTION;
	next_step = global_usblp->readurb->next_step ;
	UNBLOCK_PREEMPTION;
    NULL ;
   }
   
 USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_disconnect() : wait urb_task stop ok. \n" )) ;
 #endif // USBLPT_READ_BY_URBTASK

 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
 if (global_usblp->bidir)
   {
    
    if (global_usblp->readurb->hcpriv)
	   {
        /*
		USB_DBG_INIT(iprintf("[USB INIT] usblp_unlink_urbs() : free priv .. \n" )) ;
		USB_DBG_FLOW(iprintf( "[USB FLOW] usblp_unlink_urbs() Call hcd->driver->urb_free_priv() ...\n" )) ;
	
		hcd->driver->urb_free_priv(hcd, usblp->readurb->hcpriv);
    
		USB_DBG_URB(iprintf("[USB URB] usblp_unlink_urbs() : Set Vaue : urb->hcpriv = 0 ..\n" )) ;
    
		usblp->readurb->hcpriv = NULL;
        */
	   }
   }
 #endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw

 usblp_unlink_urbs(global_usblp);

 #ifdef _ETCPIP
	usblp_plug_out = 1 ;
	global_usblp->opened = 0;
	// usb_lpt_ok = 0 ;
	/*
	// while( usblp->opened) 
	// while(usb_lpt_ok)
	while(usb_lpt_released == 0)
	   {
		#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
		 #else
	       	#error : Please Make Sure !!
		#endif
	    NULL ;
	   }
    */
 #endif // _ETCPIP

 // #ifndef _ETCPIP
	if (!global_usblp->opened) // Not been opened
 // #else
 //	// if (! usb_lpt_ok)
 //	// if(usb_lpt_released)
 // if ((usblp->opened==0) ) // has been opened
 // #endif
   {
	USBLPT_DBG_INIT(iprintf(
		"[USBLPT INIT] usblp_disconnect() : usblp_cleanup() ..\n"
        )) ;
    // dprintf("[USBLPT] usblp_disconnect() : usblp_cleanup() ..\n") ;
    
	usblp_cleanup (global_usblp);
   }
  else 	/* cleanup later, on close */
   {
        
	#ifdef _ETCPIP
		USBLPT_DBG_INIT(iprintf(
			"[USBLPT INIT] usblp_disconnect() : Wait for Released ..\n"
            )) ;
		// dprintf("[USBLPT] usblp_disconnect() : Wait for Released "
        //			"..\n") ;
        
		while(usb_lpt_released == 0)
		   {
               
			BLOCK_PREEMPTION;
	        if( (! usb_lpt_access) && (usb_lpt_released == 0))
               {
				USBLPT_DBG_INIT(iprintf(
					"[USBLPT INIT] usblp_disconnect() : force "
                	"usblp_cleanup() ..\n"
                    )) ;
                    
				//dprintf("[USBLPT] usblp_disconnect() : force "
            	//   		"usblp_cleanup() ..\n") ;
                        
				usblp_cleanup(global_usblp);
				UNBLOCK_PREEMPTION;
                break ; 
              }
			UNBLOCK_PREEMPTION;
            
	#ifdef USBLPT_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
	   ) ) {
			iprintf("[ASSERT][USB] usblp_write(2) : " ) ;
    	    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
		    iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
	#endif // USB_BS_ASSERT
			#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
			 #else
		       	#error : Please Make Sure !!
			#endif
		    NULL ;
		   }
        /*
		if (usb_lpt_released)
   		   {
			dprintf("[USBLPT] usblp_disconnect() : usblp_cleanup() ..\n") ;
			usblp_cleanup (usblp);
           }
        */
	#endif
 
	#ifndef NO_SEMPHORE
		up (&global_usblp->sem);
	#endif // ndef NO_SEMPHORE
   }
 
 // #ifdef _ETCPIP
 global_usblp = 0 ;
 USBLPT_DBG_INFO(iprintf("[USBLPT INFO] usblp_disconnect() : global_usblp=%08XH \n",global_usblp));
	usb_lpt_ok = 0 ;
 // #endif // _ETCPIP
 /*
 #ifndef _ETCPIP
	unlock_kernel();
 #endif // ndef _ETCPIP
 */

 // dprintf("[USBLPT] usblp_disconnect() End.\n") ;
 USBLPT_DBG_FLOW(iprintf(
	"[USBLPT FLOW] usblp_disconnect() End.\n"));
}
// #endif // ndef _ETCPIP

#ifndef _ETCPIP
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
		static struct usb_device_id usblp_ids [] = {
		{ USB_DEVICE_INFO(7, 1, 1) },
		{ USB_DEVICE_INFO(7, 1, 2) },
		{ USB_DEVICE_INFO(7, 1, 3) },
		{ USB_INTERFACE_INFO(7, 1, 1) },
		{ USB_INTERFACE_INFO(7, 1, 2) },
		{ USB_INTERFACE_INFO(7, 1, 3) },
		{ }						/* Terminating entry */
		};

		MODULE_DEVICE_TABLE (usb, usblp_ids);
	#endif
#endif // ndef _ETCPIP

static struct usb_driver usblp_driver =
{
 /*
 #ifndef _ETCPIP
	.owner =	THIS_MODULE,
 #endif // ndef _ETCPIP
 */

 #ifdef USB_DBGSTRING_ON
	.name =		"usblp",
 #endif // USB_DBGSTRING_ON
 .usbdev_probe =	usblp_probe,
 
 // #ifndef _ETCPIP
	.disconnect =	usblp_disconnect,
 // #endif // ndef _ETCPIP
 
 
 #ifndef _ETCPIP
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
	.id_table =	usblp_ids,
	#endif
 #endif // ndef _ETCPIP
};

#ifndef _ETCPIP
	static 
#endif // ndef _ETCPIP
int __init usblp_init(void)
{
 USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_init() Begin ..\n"));
 
 #ifdef _ETCPIP
	usb_lpt_ok = 0 ;
 #endif // _ETCPIP
 
 USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usb_register(&usblp_driver) ..\n"));
 if (usb_register(&usblp_driver))
   {
	USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usb_register(&usblp_driver) Fail !!\n"));
	return -1;
   }
 USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usb_register(&usblp_driver) Ok.\n"));
        
 #ifndef _ETCPIP
	info(DRIVER_VERSION ":" DRIVER_DESC);
 #endif // ndef _ETCPIP
 
 USBLPT_DBG_INIT(iprintf("[USBLPT INIT] usblp_init() End.\n"));
 return 0;
}

#ifndef _ETCPIP
   /*
   static void __exit usblp_exit(void)
   {
 	usb_deregister(&usblp_driver);
   }
   */
#endif // ndef _ETCPIP

/*
#ifndef _ETCPIP
	
module_init(usblp_init);
module_exit(usblp_exit);

MODULE_AUTHOR( DRIVER_AUTHOR );
MODULE_DESCRIPTION( DRIVER_DESC );
MODULE_PARM(proto_bias, "i");
MODULE_PARM_DESC(proto_bias, "Favourite protocol number");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,12)
		MODULE_LICENSE("GPL");
#endif
   
#endif // ndef _ETCPIP
*/
// ====================================================================
// End of file : UsbLptIo.c

